package de.aristaflow.adept2.core.checks.processmodel.timemodel;

import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_AGENT;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_ORGPOSITION;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_PASSWORD;

import java.net.URI;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.base.security.AuthenticationException;
import de.aristaflow.adept2.base.service.InternalServiceException;
import de.aristaflow.adept2.base.service.Registry;
import de.aristaflow.adept2.base.service.ServiceNotKnownException;
import de.aristaflow.adept2.base.sessionmanagement.SessionFactory;
import de.aristaflow.adept2.core.ComponentTypes;
import de.aristaflow.adept2.core.checks.processmodel.AbstractProcessTemplateCheck;
import de.aristaflow.adept2.core.checks.processmodel.ProcessCheckService;
import de.aristaflow.adept2.core.timemanager.TimeManager;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.StructuredConstraintEdge;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultTemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.ActivityDuration;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDate;
import de.aristaflow.adept2.model.processmodel.timemodel.ProcessDuration;
import de.aristaflow.adept2.model.processmodel.timemodel.ScheduleRestrictedElement;
import de.aristaflow.adept2.model.processmodel.timemodel.TimeLag;
import de.aristaflow.adept2.model.processmodel.timemodel.ValidityPeriod;
import de.aristaflow.adept2.model.processmodel.tools.NodeRelations;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.CheckReport.ResultType;
import de.aristaflow.adept2.util.DataSourceException;

public class BasicTemporalProcessCheck extends AbstractProcessTemplateCheck
{

  private static final String ID = "BasicTimeModelCheck";

  private static final boolean XOR_SPLIT_HAS_DURATION = false;

  private static final boolean ENDLOOP_HAS_DURATION = false;

  private static final String PROCESS_DURATION_MISSING = "ProcessDuration.Missing";
  private static final String PROCESS_DURATION_MIN_MAX_MISSING = "ProcessDuration.MinMaxMissing";
  
  private static final String ACTIVITY_DURATION_MISSING = "ActivityDuration.Missing";
  private static final String ACTIVITY_DURATION_ZERO = "ActivityDuration.Zero";

  private static final String STUCTURAL_NODE_HAS_DURATION = "ActivityDuration.StructuralNode";

  private static final String XOR_SPLIT_DURATION_MISSING = "ActivityDuration.XOR.Missing";

  private static final String ENDLOOP_DURATION_MISSING = "ActivityDuration.Loop.Missing";

  private static final String STUCTURAL_NODE_HAS_VALIDITY = "ValidityPeriod.StructuralNode";

  private static final String STUCTURAL_NODE_HAS_SCHEDULE = "ScheduleRestrictedElement.StructuralNode";

  private static final String STUCTURAL_NODE_HAS_FIXEDDATE = "FixedDate.StructuralNode";

  protected Map<String, ResultType> checkReportLevel;



  private TimeModelOperations timeModelOperations;

  private ProcessCheckService checkService;

  private Registry registry;

  private TimeManager timeManager;


  

  public BasicTemporalProcessCheck(Configuration configuration,
      ProcessCheckService checkService, Registry registry)
  {
    super(ID);
    this.checkService = checkService;
    this.registry = registry;
    
    initCheckLevels();
  }
  
  public BasicTemporalProcessCheck(TimeModelOperations timeModelOperations)
  {
    super(ID);
    this.timeModelOperations = timeModelOperations;
    
    initCheckLevels();
  }

  protected void initCheckLevels()
  {
    this.checkReportLevel = new HashMap<String, ResultType>();
    this.checkReportLevel.put(ACTIVITY_DURATION_MISSING, ResultType.WARNING);
    this.checkReportLevel.put(ACTIVITY_DURATION_ZERO, ResultType.FAILURE);
    this.checkReportLevel.put(STUCTURAL_NODE_HAS_DURATION, ResultType.WARNING);
    this.checkReportLevel.put(XOR_SPLIT_DURATION_MISSING, ResultType.WARNING);
    this.checkReportLevel.put(ENDLOOP_DURATION_MISSING, ResultType.WARNING);
    this.checkReportLevel.put(STUCTURAL_NODE_HAS_VALIDITY, ResultType.FAILURE);
    this.checkReportLevel.put(STUCTURAL_NODE_HAS_SCHEDULE, ResultType.FAILURE);
    this.checkReportLevel.put(STUCTURAL_NODE_HAS_FIXEDDATE, ResultType.FAILURE);
    this.checkReportLevel.put(PROCESS_DURATION_MISSING, ResultType.WARNING);
    this.checkReportLevel.put(PROCESS_DURATION_MIN_MAX_MISSING, ResultType.WARNING);
  }

  @SuppressWarnings("deprecation")
  protected TimeManager getTimeManager()
  {
    if (timeManager == null)
    {
      URI[] uris = checkService.getURIs();
      SessionFactory sessionFactory;
      try
      {
        sessionFactory = registry.getSecurityManager()
          .authenticatePrivileged(SYSTEM_AGENT, SYSTEM_ORGPOSITION, SYSTEM_PASSWORD);
      }
      catch (AuthenticationException ae)
      {
        throw new InternalServiceException("Failed to authenticate at the "
                                           + "security manager due to wrong"
                                           + "authentication data. Check the "
                                           + "configuration of this "
                                           + "SubprocessManager.", ae);
      }
      catch (DataSourceException dse)
      {
        throw new InternalServiceException("Failed to authenticate at the "
                                           + "security manager due to "
                                           + "unavailability of the data.", dse);
      }
      
      try
      {
        timeManager = registry.getServiceOfType(sessionFactory.getSessionToken(uris),  ComponentTypes.COMPONENT_TYPE_TIME_MANAGER, TimeManager.class);
      }
      catch (ServiceNotKnownException e)
      {
        throw new InternalServiceException("Failed to get Time Manager Service.", e);
      }
    }
    return timeManager;
  }
  
  public TimeModelOperations getTimeModelOperations()
  {
    if (this.timeModelOperations == null)
    {
      TimeManager timeManager = getTimeManager();

      TimeModelFactory factory = timeManager.getTimeModellingService().getTimeModelFactory();
      this.timeModelOperations = factory.getTimeModelOperations();
    }
    return timeModelOperations;
  }

  protected ResultType getReportType(String id)
  {
    ResultType type = this.checkReportLevel.get(id);
    if (type != null)
    {
      return type;
    }
    
    logger.warning(String.format("Not ReportType defined for key '%s'!", id));
    return ResultType.INFO;
  }

  @Override
  protected void performCheckSpi(Template template, NodeRelations relations,
      CheckReport checkReport)
  {

    
    TemporalTemplate temporalTemplate = new DefaultTemporalTemplate(template);
    performCheckTemplate(temporalTemplate, checkReport);
    performCheckNodes(temporalTemplate, checkReport);
    performCheckConstraintEdges(temporalTemplate, checkReport);

  }

  protected void performCheckTemplate(TemporalTemplate template, CheckReport checkReport)
  {   
    
    performCheckProcessDuration(template, checkReport);
    
  }

  protected void performCheckProcessDuration(TemporalTemplate template, CheckReport checkReport)
  {
    ProcessDuration processDuration = template.getProcessDuration();
    
    if (processDuration != null)
    {
      checkProcessDuration(processDuration, template, checkReport);
    }
    else
    {
        String message = String
            .format("The template should have a process duration assigned!");
        addReportEntry(getReportType(PROCESS_DURATION_MISSING), message, template);
    }
  }

  protected void performCheckConstraintEdges(TemporalTemplate template, CheckReport checkReport)
  {
    for (StructuredConstraintEdge edge : template.getConstraintEdgeStructure())
    {
      for (RelationType r : RelationType.values())
      {
        TimeLag timeLag = template.getTimeLagConstraint(edge.getSourceNodeID(), edge.getDestinationNodeID(), r);
        
        if (timeLag != null)
        {                
          checkTimeLag(template, timeLag, edge, checkReport);
        }
      }
    }
    
  }

  protected void performCheckNodes(TemporalTemplate template,
      CheckReport checkReport)
  {
    for (Node node : template.getNodes())
    {
      TemporalNode temporalNode = template.getNode(node.getID());
      performCheckActivityDuration(temporalNode, template, checkReport);
      performCheckFixedDate(temporalNode, template, checkReport);
      performCheckSchedule(temporalNode, template, checkReport);
      performCheckValidityPeriod(temporalNode, template, checkReport);
    }

  }

  protected void checkTimeLag(TemporalTemplate template, TimeLag timeLag, StructuredConstraintEdge edge,
      CheckReport checkReport)
  {
    boolean minimumTimeLagOK = true;
    boolean maximumTimeLagOK = true;
    

    
    if (timeLag.getLowerMinimumTimeLag() == null)
    {
      if (timeLag.getUpperMinimumTimeLag() != null)
      {
        String message = String
          .format("Upper Minimum TimeLag must be null in case Lower Minimum TimeLag is null!");
        addReportEntry(ResultType.WARNING, message, 
          template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
      }
      minimumTimeLagOK = false;
    }
    else
    {
      
      minimumTimeLagOK &= checkTimePoint(timeLag.getLowerMinimumTimeLag(), checkReport, 
        template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
      if (timeLag.getUpperMinimumTimeLag() != null)
      {
        minimumTimeLagOK &= checkTimePoint(timeLag.getUpperMinimumTimeLag(), checkReport, 
          template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
        if (minimumTimeLagOK && this.getTimeModelOperations().compare(timeLag.getLowerMinimumTimeLag(),
                                    timeLag.getUpperMinimumTimeLag()) > 0)
        {
          String message = String
            .format("Lower Minimum TimeLag may not be bigger than Upper Minimum TimeLag!");
          addReportEntry(ResultType.FAILURE, message, 
            template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
        }
      }
    }
    
    if (timeLag.getUpperMaximumTimeLag() == null)
    {
      if (timeLag.getLowerMaximumTimeLag() != null)
      {
        String message = String
          .format("Lower Maximum TimeLag must be null in case Upper Maximum TimeLag is null!");
        addReportEntry(ResultType.WARNING, message, 
          template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
      }
      maximumTimeLagOK = false;
    }
    else
    {        
      
      maximumTimeLagOK &= checkTimePoint(timeLag.getLowerMaximumTimeLag(), checkReport, 
        template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));

      if (timeLag.getLowerMaximumTimeLag() != null)
      {
        maximumTimeLagOK &= checkTimePoint(timeLag.getUpperMaximumTimeLag(), checkReport, 
          template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
        
        if (maximumTimeLagOK && this.getTimeModelOperations().compare(timeLag.getLowerMaximumTimeLag(),
                                    timeLag.getUpperMaximumTimeLag()) > 0)
        {
          String message = String
            .format("Lower Maxmimum TimeLag may not be bigger than Upper Maximum TimeLag!");
          addReportEntry(ResultType.FAILURE, message, 
            template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
        }
      }
    }
    
    if (timeLag.getLowerMinimumTimeLag() == null
        && timeLag.getUpperMaximumTimeLag() == null)
    {
      if (timeLag.getTimeLagUpdateHandler() == null)
      {
        String message = String
          .format("Either minimum TimeLag, maximum TimeLag or activity update procedure needs to be provided!");
        addReportEntry(ResultType.FAILURE, message, 
          template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
      }
    }
    else
    {
      if (minimumTimeLagOK && maximumTimeLagOK && 
          this.getTimeModelOperations().compare(timeLag.getLowerMinimumTimeLag(),
                                  timeLag.getUpperMaximumTimeLag()) > 0)
      {
        String message = String
          .format("Lower Minimum TimeLag may not be bigger than Upper Maximum TimeLag!");
        addReportEntry(ResultType.FAILURE, message, 
          template.getNode(edge.getSourceNodeID()), template.getNode(edge.getDestinationNodeID()));
      }
    }
  }

  protected void checkProcessDuration(ProcessDuration processDuration, TemporalTemplate template,
      CheckReport checkReport)
  {
    boolean guardMinimumDurationOK = true;
    boolean guardMaximumDurationOK = true;
    boolean flexibleMaximumDurationOK = true;
   
    
    if (processDuration.getMinimumDuration() == null)
    {
      if (processDuration.getGuardMaximumDuration() != null)
      {
        String message = String
            .format("The template has a maximum contingent process duration but no minimum contingent process duration!");
        addReportEntry(getReportType(PROCESS_DURATION_MIN_MAX_MISSING), message, template);
      }
      else
      {
        String message = String
            .format("The template is missing a minimum contingent process duration!");
        addReportEntry(getReportType(PROCESS_DURATION_MISSING), message, template);
      }
 
      guardMinimumDurationOK = false;
    }
    else
    {
      guardMinimumDurationOK &= checkTimePoint(processDuration.getMinimumDuration(), checkReport, template);
    }
    
    if (processDuration.getGuardMaximumDuration() == null)
    {
      if (processDuration.getMinimumDuration() != null)
      {
        String message = String
            .format("The template has a minimum contingent process duration but no maximum contingent process duration!");
        addReportEntry(getReportType(PROCESS_DURATION_MIN_MAX_MISSING), message, template);
      }
      else
      {
        String message = String
            .format("The template is missing a maximum contingent process duration!");
        addReportEntry(getReportType(PROCESS_DURATION_MISSING), message, template);
      }
 
      guardMaximumDurationOK = false;
    }
    else
    {      
      guardMaximumDurationOK &= checkTimePoint(processDuration.getGuardMaximumDuration(), checkReport, template);
    }
    
    
    if (processDuration.getMaximumDuration() != null)
    {        
      flexibleMaximumDurationOK &= checkTimePoint(processDuration.getMaximumDuration(), checkReport, template);
    }
    else
    {
      flexibleMaximumDurationOK = false;
    }

    

    
    
    if (guardMinimumDurationOK && guardMaximumDurationOK)
    {
      if (this.getTimeModelOperations().compare(processDuration.getMinimumDuration(),
                                           processDuration.getGuardMaximumDuration()) > 0)
       {
         String message = String
           .format("Contingent minimum process duration may not be bigger than contingent maximum process duration!");
         addReportEntry(ResultType.FAILURE, message, template);
       }
    }
    
    if (guardMaximumDurationOK && flexibleMaximumDurationOK)
    {
      if (this.getTimeModelOperations().compare(processDuration.getGuardMaximumDuration(),
                                           processDuration.getMaximumDuration()) > 0)
       {
         String message = String
           .format("Flexible minimum process duration may not be lower than contingent maximum process duration!");
         addReportEntry(ResultType.FAILURE, message, template);
       }
    }
  }
  
  protected void performCheckValidityPeriod(TemporalNode node, TemporalTemplate template,
      CheckReport checkReport)
  {

    
    NodeType type = template.getNodeType(node.getID());

    ValidityPeriod startValidity = node.getStartValidityPeriod();
    
    ValidityPeriod completionValidity = node.getCompletionValidityPeriod();

    switch (type)
    {
      case NT_AND_JOIN:
      case NT_AND_SPLIT:
      case NT_ENDFLOW:
      case NT_STARTFLOW:
      case NT_STARTLOOP:
      case NT_XOR_JOIN:
      case NT_ENDLOOP:
      case NT_XOR_SPLIT:
        if (startValidity != null || completionValidity != null)
        {
          String message = String
            .format(
                    "Node '%1$s' (%2$d) is a structural Node of type '%3$s' and therefore should not have a Validity Period assigned!",
                    node.getName(), Integer.valueOf(node.getID()), type);
          addReportEntry(getReportType(STUCTURAL_NODE_HAS_VALIDITY), message,
                         node);
        }
        break;

      case NT_NORMAL:
        if (startValidity != null)
        {
          checkValidityPeriod(startValidity, node, checkReport);
        }
        if (completionValidity != null)
        {
          checkValidityPeriod(completionValidity, node, checkReport);
        }
        
        break;
    }
    
  }

  protected void checkValidityPeriod(ValidityPeriod validityPeriod, TemporalNode node,
      CheckReport checkReport)
  {
    boolean periodOK = true;
    
    if (validityPeriod.getValidFrom() != null)
    {
      periodOK &= checkCalendarElement(validityPeriod.getValidFrom(), node, checkReport);
    }
    
    if (validityPeriod.getValidUntil() != null)
    {
      periodOK &= checkCalendarElement(validityPeriod.getValidUntil(), node, checkReport);
    }
    
    if (validityPeriod.getValidFrom() == null &&
        validityPeriod.getValidUntil() == null)
    {
      if (validityPeriod.getValidityPeriodSource() == null)
      {
        String message = String
          .format("Either valid from, valid until or a source for the validity period needs to be provided!");
        addReportEntry(ResultType.FAILURE, message, node);
      }
    }
    
    if (periodOK && 
        validityPeriod.getValidFrom() != null &&
        validityPeriod.getValidUntil() != null)
    {
      if (this.getTimeModelOperations().compare(validityPeriod.getValidFrom() ,
                                  validityPeriod.getValidUntil()) >= 0)
      {
        String message = String.format("The from value of the validity period may not be bigger than until value!");
        addReportEntry(ResultType.FAILURE, message, node);
      }
    }    
  }

  protected boolean checkCalendarElement(CalendarElement calendarElement, Object affectedElement,
      CheckReport checkReport)
  {
    if (calendarElement == null)
      return true;
    
    if (calendarElement.getTimeInMillis() < 0)
    {
      String message = String
        .format("Value of calendarElement may not be < 0!");
      addReportEntry(ResultType.FAILURE, message, affectedElement);
      return false;
    }
    
    return true;
  }

  protected void performCheckSchedule(TemporalNode node, TemporalTemplate template,
      CheckReport checkReport)
  {

    
    NodeType type = template.getNodeType(node.getID());
    
    ScheduleRestrictedElement startSchedule = node.getStartSchedule();
    ScheduleRestrictedElement completionSchedule = node.getCompletionSchedule();

    switch (type)
    {
      case NT_AND_JOIN:
      case NT_AND_SPLIT:
      case NT_ENDFLOW:
      case NT_STARTFLOW:
      case NT_STARTLOOP:
      case NT_XOR_JOIN:
      case NT_ENDLOOP:
      case NT_XOR_SPLIT:
        if (startSchedule != null || completionSchedule != null)
        {
          String message = String
            .format(
                    "Node '%1$s' (%2$d) is a structural Node of type '%3$s' and therefore should not have a Schedule Restricted Element assigned!",
                    node.getName(), Integer.valueOf(node.getID()), type);
          addReportEntry(getReportType(STUCTURAL_NODE_HAS_SCHEDULE), message,
                         node);
        }
        break;

      case NT_NORMAL:
        if (startSchedule != null)
        {
          checkSchedule(startSchedule, node, checkReport);
        }
        if (completionSchedule != null)
        {
          checkSchedule(completionSchedule, node, checkReport);
        }
        
        break;
    }
    
  }

  protected void checkSchedule(
      ScheduleRestrictedElement scheduleRestrictedElement, Node node,
      CheckReport checkReport)
  {
    // XXX Check Schedule as soon as format is defined

  }

  protected void performCheckFixedDate(TemporalNode node, TemporalTemplate template,
      CheckReport checkReport)
  {

    
    NodeType type = template.getNodeType(node.getID());

    FixedDate startFixedDate = node.getStartFixedDate();
    FixedDate completionFixedDate = node.getCompletionFixedDate();
    
    switch (type)
    {
      case NT_AND_JOIN:
      case NT_AND_SPLIT:
      case NT_ENDFLOW:
      case NT_STARTFLOW:
      case NT_STARTLOOP:
      case NT_XOR_JOIN:
      case NT_ENDLOOP:
      case NT_XOR_SPLIT:
        if (startFixedDate != null || completionFixedDate != null)
        {
          String message = String
            .format(
                    "Node '%1$s' (%2$d) is a structural Node of type '%3$s' and therefore should not have a Fixed Date assigned!",
                    node.getName(), Integer.valueOf(node.getID()), type);
          addReportEntry(getReportType(STUCTURAL_NODE_HAS_FIXEDDATE), message,
                         node);
        }
        break;

      case NT_NORMAL:
      {
        if (startFixedDate != null)
        {
          checkFixedDate(startFixedDate, node, checkReport);
        }
        if (completionFixedDate != null)
        {
          checkFixedDate(completionFixedDate, node, checkReport);
        }
      }
      break;
    }
    
  }

  protected void checkFixedDate(FixedDate fixedDate, Node node,
      CheckReport checkReport)
  {
    // XXX Check fixed date as soon as format is defined
    
  }

  protected void performCheckActivityDuration(TemporalNode node, TemporalTemplate template,
      CheckReport checkReport)
  {
        
    NodeType type = template.getNodeType(node.getID());

    ActivityDuration activityDuration = node.getActivityDuration();

    switch (type)
    {
      case NT_AND_JOIN:
      case NT_AND_SPLIT:
      case NT_ENDFLOW:
      case NT_STARTFLOW:
      case NT_STARTLOOP:
      case NT_XOR_JOIN:
        if (activityDuration != null)
        {
          String message = String
            .format(
                    "Node '%1$s' (%2$d) is a structural Node of type '%3$s' and therefore should not have an Activity Duration assigned!",
                    node.getName(), Integer.valueOf(node.getID()), type);
          addReportEntry(getReportType(STUCTURAL_NODE_HAS_DURATION), message,
                         node);
        }
        break;

      case NT_NORMAL:
        if (activityDuration == null)
        {
          String message = String
            .format(
                    "Node '%1$s' (%2$d) does not have an Activity Duration assigned!",
                    node.getName(), Integer.valueOf(node.getID()));
          addReportEntry(getReportType(ACTIVITY_DURATION_MISSING), message,
                         node);
        }
        else
        {
          checkActivityDuration(activityDuration, node, checkReport);
        }
        break;
      case NT_XOR_SPLIT:
        if (XOR_SPLIT_HAS_DURATION)
        {
          if (activityDuration == null)
          {
            String message = String
              .format(
                      "Node '%1$s' (%2$d) does not have an Activity Duration assigned!",
                      node.getName(), Integer.valueOf(node.getID()));
            addReportEntry(getReportType(XOR_SPLIT_DURATION_MISSING), message,
                           node);
          }
          else
          {
            checkActivityDuration(activityDuration, node, checkReport);
          }
        }
        else
        {
          if (activityDuration != null)
          {
            String message = String
              .format(
                      "Node '%1$s' (%2$d) is a structural Node of type '%3$s' and therefore should not have an Activity Duration assigned!",
                      node.getName(), Integer.valueOf(node.getID()), type);
            addReportEntry(getReportType(STUCTURAL_NODE_HAS_DURATION), message,
                           node);
          }
        }
        break;

      case NT_ENDLOOP:
        if (ENDLOOP_HAS_DURATION)
        {
          if (activityDuration == null)
          {
            String message = String
              .format(
                      "Node '%1$s' (%2$d) does not have an Activity Duration assigned!",
                      node.getName(), Integer.valueOf(node.getID()));
            addReportEntry(getReportType(ENDLOOP_DURATION_MISSING), message,
                           node);
          }
          else
          {
            checkActivityDuration(activityDuration, node, checkReport);
          }
        }
        else
        {
          if (activityDuration != null)
          {
            String message = String
              .format(
                      "Node '%1$s' (%2$d) is a structural Node of type '%3$s' and therefore should not have an Activity Duration assigned!",
                      node.getName(), Integer.valueOf(node.getID()), type);
            addReportEntry(getReportType(STUCTURAL_NODE_HAS_DURATION), message,
                           node);
          }
        }
        break;

    }
  }

  protected void checkActivityDuration(
      ActivityDuration activityDuration, TemporalNode node, CheckReport checkReport)
  {
    boolean flexibleMinimumDurationOK = true;
    boolean contingentMinimumDurationOK = true;
    boolean contingentMaximumDurationOK = true;
    boolean flexibleMaximumDurationOK = true;
    
    if (activityDuration.getMinimumDuration() != null)
    {
      flexibleMinimumDurationOK &= checkTimePoint(activityDuration.getMinimumDuration(), checkReport, node);
      
      TimeDistance distance = getTimeModelOperations().convertToTimeDistance(activityDuration.getMinimumDuration());
      
      if (distance.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) <= 0)
      {
        String message = String.format("Flexible minimum duration must be greater than zero! (Currently: %s)", distance);
        addReportEntry(getReportType(ACTIVITY_DURATION_ZERO), message, node);
      }
      
    }
    else
    {
      String message = "Flexible minimum duration must be provided!";
      addReportEntry(ResultType.FAILURE, message, node);
      flexibleMinimumDurationOK = false;
    }
    
    if (activityDuration.getGuardMinimumDuration() != null)
    {
      contingentMinimumDurationOK &= checkTimePoint(activityDuration.getGuardMinimumDuration(), checkReport, node);
      
      TimeDistance distance = getTimeModelOperations().convertToTimeDistance(activityDuration.getGuardMinimumDuration());
      
      if (distance.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) <= 0)
      {
        String message = String.format("Contingent minimum duration must be greater than zero! (Currently: %s)", distance);
        addReportEntry(getReportType(ACTIVITY_DURATION_ZERO), message, node);
      }
      
    } // TODO CSTNPSU
//    else
//    {
//      String message = "Contingent minimum duration must be provided!";
//      addReportEntry(ResultType.FAILURE, message, node);
//      contingentMinimumDurationOK = false;
//    }
    
    if (activityDuration.getGuardMaximumDuration() != null)
    {
      contingentMaximumDurationOK  &= checkTimePoint(activityDuration.getGuardMaximumDuration(), checkReport, node);
    }
    else
    {
      contingentMaximumDurationOK = false;
    }
    
    if (activityDuration.getMaximumDuration() != null)
    {
      flexibleMaximumDurationOK  &= checkTimePoint(activityDuration.getMaximumDuration(), checkReport, node);
    }
    else
    {
      flexibleMaximumDurationOK = false;
    }
    
    if (!(contingentMaximumDurationOK || flexibleMaximumDurationOK))
    {
      String message = String.format("Maximum duration must be provided!");
      addReportEntry(ResultType.FAILURE, message, node);
    }
    
    if (contingentMaximumDurationOK && flexibleMinimumDurationOK)
    {
      if (this.getTimeModelOperations().compare(activityDuration.getMinimumDuration(),
                                  activityDuration.getGuardMaximumDuration()) > 0)
      {
        String message = String
          .format("Contingent minimum duration may not be bigger than contingent maximum duration!");
        addReportEntry(ResultType.FAILURE, message, node);
      }
    }
    
    if (flexibleMaximumDurationOK && contingentMaximumDurationOK)
    {        
      if (this.getTimeModelOperations().compare(activityDuration.getGuardMaximumDuration(), 
                                                                activityDuration.getMaximumDuration()) > 0)
      {
        String message = String
          .format("Contingent maxmimum duration may not be bigger than flexible maximum duration!");
        addReportEntry(ResultType.FAILURE, message, node);
      }
    }
  }

  protected boolean checkTimePoint(TimePoint timePoint, Object affectedElement,
      CheckReport checkReport)
  {
    return checkTimePoint(timePoint, checkReport, affectedElement);
  }

  protected boolean checkTimePoint(TimePoint timePoint, CheckReport checkReport,
      Object... affectedElement)
  {
    if (timePoint == null)
      return true;
    
    if (timePoint.getMonths() == TimePoint.UNDEFINED_TIME_POINT
        && timePoint.getDays() == TimePoint.UNDEFINED_TIME_POINT
        && timePoint.getHours() == TimePoint.UNDEFINED_TIME_POINT
        && timePoint.getMinutes() == TimePoint.UNDEFINED_TIME_POINT
        && timePoint.getSeconds() == TimePoint.UNDEFINED_TIME_POINT)
    {
      String message = String
        .format("At least one value of time point needs to be specified!");
      addReportEntry(ResultType.FAILURE, message, affectedElement);
      return false;
    }
    
    return true;
  }
  
}
