package de.aristaflow.adept2.core.checks.timemodel;

import java.util.logging.Level;
import java.util.logging.Logger;

import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.tools.NodeRelations;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.CheckReport.ResultType;
import de.aristaflow.adept2.util.LoggerTools;

public class TimeModelCheckRunner implements TimeModelCheck
{

  protected final TimeModelCheck[] checks;
  protected final boolean stopOnError;
  protected Logger logger =  LoggerTools.getLogger(this);

  public TimeModelCheckRunner(boolean stopOnError, TimeModelCheck[] checks)
  {
    this.stopOnError = stopOnError;
    if (checks != null)
    {
      this.checks = checks.clone();
    }
    else
    {
      this.checks = null;
    }
   
  }

  public TimeModelCheckRunner(
      TimeModelCheck... checks)
  {
    this(true, checks);
  }

  public boolean performCheck(TimeModel timeModel, Template template,
      NodeRelations relations, CheckReport checkReport)
  {
    boolean allOK = true;
    
    // run the checks and adjust the overall result
    for (TimeModelCheck check : checks)
    {
      // result for the current check
      boolean checkOK;
      // catch any runtime exception, which might be thrown by checks 
      // or implementations (e.g. of the process model), because of
      // faulty / inconsistent input 
      logger.fine(String.format("Running check: %s", check.getClass().getSimpleName()));
      try
      {
        // check template itself
        checkOK = check.performCheck(timeModel, template, relations, checkReport);
      }
      catch (RuntimeException e)
      {
        allOK = false;
        String message = String.format("An exception occurred performing the check '%1$s', failing and skipping subsequent checks!", check.getClass().getSimpleName());
        logger.log(Level.SEVERE, message, e);
        // add a check report entry
        if (checkReport != null)
        {
          checkReport.addReportEntry(message, ResultType.FAILURE, message);
        }
        break;
      }
      
      if (!checkOK)
      {
        allOK = false;
        logger.info("Check '"+ check.getClass().getSimpleName() +"' failed, skipping following checks.");
        
        if (stopOnError)
        {
          return false;
        }
      }
    }

    return allOK;

  }

  public boolean performCheck(TimeModel timeModel, CheckReport checkReport)
  {
    return performCheck(timeModel, null, null, checkReport);
  }

}
