package de.aristaflow.adept2.core.checks.timemodel.tcnimplementation;

import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_AGENT;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_ORGPOSITION;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_PASSWORD;

import java.net.URI;
import java.util.logging.Logger;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.base.security.AuthenticationException;
import de.aristaflow.adept2.base.service.InternalServiceException;
import de.aristaflow.adept2.base.service.Registry;
import de.aristaflow.adept2.base.service.ServiceNotKnownException;
import de.aristaflow.adept2.base.sessionmanagement.SessionFactory;
import de.aristaflow.adept2.core.ComponentTypes;
import de.aristaflow.adept2.core.checks.timemodel.TimeModelCheckService;
import de.aristaflow.adept2.core.timemanager.TimeManager;
import de.aristaflow.adept2.model.processmodel.DataElement;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.tools.NodeRelations;
import de.aristaflow.adept2.model.processmodel.tools.ProcessElementIdentifierTools;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LoggerTools;

public abstract class AbstractTimeModelCheck
{

  protected final String id;
  protected final TimeModelFactory factory;
  /** Thread local for the result. */
  private final ThreadLocal<CheckReport.ResultType> result = new ThreadLocal<CheckReport.ResultType>();
  /** Thread local for the check report. */
  private final ThreadLocal<CheckReport> checkReport = new ThreadLocal<CheckReport>();
  private Logger logger = LoggerTools.getLogger(this);

  public AbstractTimeModelCheck(String id, Configuration configuration,
      TimeModelCheckService checkService, Registry registry)
  {
    super();
    this.id = id;

    URI[] uris = checkService.getURIs();
    SessionFactory sessionFactory;
    try
    {
     
      sessionFactory = registry.getSecurityManager()
        .authenticatePrivileged(SYSTEM_AGENT, SYSTEM_ORGPOSITION, SYSTEM_PASSWORD);
    }
    catch (AuthenticationException ae)
    {
      throw new InternalServiceException("Failed to authenticate at the "
                                         + "security manager due to wrong"
                                         + "authentication data. Check the "
                                         + "configuration of this "
                                         + "SubprocessManager.", ae);
    }
    catch (DataSourceException dse)
    {
      throw new InternalServiceException("Failed to authenticate at the "
                                         + "security manager due to "
                                         + "unavailability of the data.", dse);
    }
    
    TimeManager timeManager;
    try
    {
      timeManager = registry.getServiceOfType(sessionFactory.getSessionToken(uris),  ComponentTypes.COMPONENT_TYPE_TIME_MANAGER, TimeManager.class);
    }
    catch (ServiceNotKnownException e)
    {
      throw new InternalServiceException("Failed to get Time Manager Service.", e);
    }
    
    this.factory = timeManager.getTimeModellingService().getTimeModelFactory();
  }

  public boolean performCheck(TimeModel timeModel, Template template, NodeRelations relations, CheckReport checkReport)
  {
    try
    {
      this.result.set(CheckReport.ResultType.OK);
      this.checkReport.set(checkReport);
  
      performCheckSpi(timeModel, template, relations);
  
      return (this.result.get().ordinal() < CheckReport.ResultType.WARNING.ordinal());
    }
    finally
    {
      this.checkReport.remove();
      this.result.remove();
    }
  }

  protected abstract void performCheckSpi(TimeModel timeModel, Template template,
      NodeRelations relations);
  
  
  public boolean performCheck(TimeModel timeModel, CheckReport checkReport)
  {
    return performCheck(timeModel, null, null, checkReport);
  }

  /**
   * Adds an entry to the provided check report (if it isn't null) and updates
   * {@link #result}. Note, that the objects passed as <code>elements</code>
   * must be supported, cf. implementation of {@link #getURIsForElement(URI, Object[])},
   * 
   * @param resultType
   * @param message
   * @param elements
   */
  protected final void addReportEntry(CheckReport.ResultType resultType, String message, Object ... elements)
  {
    // log the message in any case
    logger.info(message);
    
    CheckReport checkReport = this.checkReport.get();
  
    if (checkReport != null)
    {
      URI[] uris = getURIsForElement(checkReport.getBase(), elements);
      checkReport.addReportEntry(id, resultType, message, uris);
    }
    if (result.get().ordinal() < resultType.ordinal())
    {
      result.set(resultType);
    }
  }

  /**
   * Returns the element URIs for the given affected elements. Note, that
   * <code>element</code> must only contain supported objects.
   * 
   * @param base The base URI for creating the element URIs.
   * @param element The element objects.
   * @return An array with an entry for each element
   */
  private URI[] getURIsForElement(URI base, Object[] element)
  {
    URI[] res = new URI[element.length];
    
    for (int i = 0; i < res.length; i++)
    {
      if (element[i] instanceof DataElement)
      {
        res[i] = ProcessElementIdentifierTools.getDataElementIdentifier(base, (DataElement) element[i]);
      }
      else if(element[i] instanceof Node)
      {
        res[i] = ProcessElementIdentifierTools.getNodeIdentifier(base, (Node) element[i]);
      } 
    }
    
    return res;
  }

}