package de.aristaflow.adept2.core.timemanager;

import java.util.UUID;

import de.aristaflow.adept2.base.communication.annotations.ProxyAllow;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.util.CheckReport;

/**
 * This sub-interface of the time manager handles all runtime issues of process
 * template time models. It allows to store time models for process templates
 * and retrieve specific time models. The status of a time model depends on the
 * status of it's respective template, for instance, if a template is outdate
 * the time model is outdate as well.
 * 
 * If the designated template ID on any method expecting one does not exist, an
 * IllegalArgumentException will be thrown.
 * 
 * Methods do not synchronise, that is the retrieved data may be slightly out
 * of date.
 */
@ProxyAllow
public interface TemplateTimeManager
{
  /* {src_lang=Java} */

  /**
   * Gets the designated time model for the process template identfied by the
   * template ID. The returned time model can not be changed.
   * 
   * @param session The session which is used to check for access rights on this
   *          method.
   * @param templateID The ID of the process template to retrieve the time model
   *          for.
   * @return An object representing the designated time model.
   */
  public TemplateTimeModel getTimeModel(SessionToken session, UUID templateID);

  /**
   * Gets the IDs of the respective process templates of all time models managed
   * by this template time manage.
   * 
   * @param session The session which is used to check for access rights on this
   *          method.
   * @return The IDs of all process templates for which time models are managed
   *         by this template time manager.
   */
  public UUID[] getAllTimeModelIDs(SessionToken session);

  /**
   * Stores the designated time model in the template time manager. It is
   * checked whether a time model for the respective template already exists. If
   * it does, it will not be stored again.
   * 
   * A check report will be provided. Please refer to it for determining whether
   * the storage was successful (CheckReport.getCheckResult() is either
   * CheckReport.ResultType.OK or CheckReport.ResultType.INFO).
   * 
   * @param session The session which is used to check for access rights on this
   *          method.
   * @param templateID The ID of a process template
   * @param timeModel The representation of the time model
   * @returns A check report containing the result of the checks. Please refer
   *          to the CheckReport.getCheckResult() for checking whether the
   *          storage was successful (CheckReport.ResultType.OK and
   *          CheckReport.ResultType.INFO).
   */
  public TemplateTimeModel createAndGetTemplateTimeModel(SessionToken parentSession,
      UUID templateID, CheckReport checkReport);

}