package de.aristaflow.adept2.core.timemanager;

import de.aristaflow.adept2.base.communication.annotations.DeregisterCallback;
import de.aristaflow.adept2.base.communication.annotations.ProxyAllow;
import de.aristaflow.adept2.base.communication.annotations.RegisterCallback;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.worklistmanager.WorklistNotification;
import de.aristaflow.adept2.core.worklistmanager.WorklistNotification2;

/**
 * This interface is for the communication of time managers with an
 * execution manager. It allows for registering and deregistering of time
 * managers. Together with <code>WorklistNotification</code> this interface
 * realises the communication between execution manager and time manager.
 * 
 * @author Ulrich Kreher
 * @see de.aristaflow.adept2.core.worklistmanager.WorklistNotification
 */
@ProxyAllow
public interface WorklistInteraction
{
  /**
   * Adds a worklist manager to this execution manager to listen for state
   * changes. The worklist manager is notified when state changes occur, for
   * instance, a process step may be executed, is executed or has finished
   * execution.
   * 
   * <p />
   * Before any actual state changes are sent, the worklist manager will receive
   * all active activities asynchronously via the {@link WorklistNotification}
   * interface. This means the method will not block until the activities are
   * delivered. As soon as
   * {@link #removeWorklistManager(SessionToken, WorklistNotification)} is
   * called with the same <code>WorklistNotification</code>-instance, the
   * delivering of active activities is cancelled and no further active
   * activities will be sent.
   * 
   * @param session The session which is used to check for access rights on this
   *          method.
   * @param worklistManager An object that wants to listen to state changes of
   *          the managed instances.
   */
  @RegisterCallback
  public void addWorklistManager(SessionToken session,
      WorklistNotification2 worklistManager);

  /**
   * Removes a listening worklist manager from this execution manager.
   * Pending updates may still occur. If the worklist manager has not been
   * registered before, the removal will be ignored.
   *
   * @param session The session which is used to check for access rights on this
   *          method.
   * @param worklistManager An object that does not want to listen to state
   *        changes of the managed instances.
   */
  @DeregisterCallback
  public void removeWorklistManager(SessionToken session,
      WorklistNotification2 worklistManager);
  
//  /**
//   * Checks whether the designated worklist manager is still registered for
//   * updates. This allows a worklist manager to recover after the execution
//   * manager has been abruptly restarted. To detect this a worklist manager has
//   * to poll the execution manager in certain time spans.
//   * 
//   * @param session The session which is used to check for access rights on this
//   *          method.
//   * @param worklistManager The worklist manager to check whether it is
//   *          registered.
//   * @return Whether the designated worklist manager is registered for updates.
//   */
//  public boolean isRegistered(SessionToken session,
//      WorklistNotification worklistManager);
}