package de.aristaflow.adept2.core.timemanager.defaultimplementation;

import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_AGENT;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_ORGPOSITION;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_PASSWORD;

import java.net.URI;
import java.util.UUID;
import java.util.logging.Level;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.service.AbstractADEPT2Service;
import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.service.Registry;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.ComponentTypes;
import de.aristaflow.adept2.core.checks.timemodel.TimeModelCheck;
import de.aristaflow.adept2.core.checks.timemodel.TimeModelCheckRunner;
import de.aristaflow.adept2.core.processmanager.ProcessManager;
import de.aristaflow.adept2.core.timemanager.InstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.TimePrediction;
import de.aristaflow.adept2.core.timemanager.storage.InstanceTimeModelStorage;
import de.aristaflow.adept2.model.processmodel.Instance;
import de.aristaflow.adept2.model.timemodel.ChangeableTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelUpdateStatus;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.CheckReport.ResultType;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;

/**
 * @author  lanz
 */
public class DefaultInstanceTimeManager extends AbstractSubService implements
    InstanceTimeManager
{
  
  /**
   * @uml.property  name="instanceTimeModelStorage"
   * @uml.associationEnd  
   */
  private final InstanceTimeModelStorage instanceTimeModelStorage;

  /**
   * @uml.property  name="timeManager"
   * @uml.associationEnd  
   */
  private final DefaultTimeManager timeManager;

  /**
   * @uml.property  name="templateTimeManager"
   * @uml.associationEnd  
   */
  private final DefaultTemplateTimeManager templateTimeManager;

  /**
   * @uml.property  name="timeModelFactory"
   * @uml.associationEnd  
   */
  private final TimeModelFactory timeModelFactory;

  /**
   * @uml.property  name="processManager"
   * @uml.associationEnd  
   */
  private ProcessManager processManager;

  public DefaultInstanceTimeManager(
      DefaultTimeManager timeManager, TimeModelFactory timeModelFactory,
      DefaultTemplateTimeManager templateTimeManager,
      InstanceTimeModelStorage instanceTimeModelStorage)
  {
    super(timeManager);

    this.timeManager = timeManager;
    this.instanceTimeModelStorage = instanceTimeModelStorage;
    this.timeModelFactory = timeModelFactory;

    this.templateTimeManager = templateTimeManager;

  }

  @SuppressWarnings("deprecation")
  @Override
  public void init() throws AbortServiceException
  {
    this.processManager = timeManager.getProcessManager();
  }

  public InstanceTimeContext getInstanceTimeContext(SessionToken sessionToken,
      UUID instanceID)
  {
    InstanceTimeContext instanceTimeData;
    try
    {
      instanceTimeData = instanceTimeModelStorage.getInstanceTimeContextReadonly(instanceID);
      return instanceTimeData;
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    return null;
  }

  public TimePrediction getTimePrediction(SessionToken sessionToken,
      UUID instanceID)
  {
    // TODO Auto-generated method stub
    return null;
  }

  private boolean checkTimeModel(final SessionToken session,
      final TimeModel timeModel, final CheckReport report)
  {
    // super.sessionActive(session);
    try
    { // sessionActive
      // check the template for any problem
      boolean ret = true;
      // set up check runner
      TimeModelCheck[] checks;
      // get the checks
      checks = this.timeManager.getTimeModelChecks(session);
      TimeModelCheck allChecks = new TimeModelCheckRunner(false, checks);

      // perform consistency checks
      ret &= allChecks.performCheck(timeModel, report);

      return ret;
    }
    finally
    {
      // super.sessionFinished(session);
    }
  }

  public InstanceTimeModelStorage getStorage()
  {
    return instanceTimeModelStorage;
  }

  public void updateTimeModelStatus(SessionToken session, TimeModelStatus status)
      throws LockException
  {

    if (status.getTimeModelUpdateStatus().ordinal() <= TimeModelUpdateStatus.NO_UPDATE
      .ordinal())
    {
      // Time Model Update may only be enabled in case there is a valid time
      // model available
      try
      {
        if (!instanceTimeModelStorage.hasTimeModelForInstance(status.getInstanceID()))
        {
          throw new IllegalStateException(String.format(
                                                    "Time Model Update for instance '%1$s' may only be enabled in case there is a valid time model available",
                                                    status.getInstanceID()));
        }
      }
      catch (DataSourceException e)
      {
        throw new IllegalStateException(String.format(
                                                  "Time Model Update for instance '%1$s' may only be enabled in case there is a valid time model available",
                                                  status.getInstanceID()));
      }
    }

    try
    {
      instanceTimeModelStorage.updateTimeModelStatus(session, status);
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
  }

  public ChangeableTimeModelStatus getAndLockTimeModelStatus(SessionToken session,
      UUID instanceID)
  {
    try
    {
      return instanceTimeModelStorage.getAndLockTimeModelStatus(session, instanceID);
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    return null;
  }

  public void unlockTimeModelStatus(SessionToken session, UUID instanceID)
      throws LockException
  {
    instanceTimeModelStorage.unlockTimeModelStatus(session, instanceID);
  }

  public TimeModelStatus getTimeModelStatusReadonly(UUID instanceID)
  {
    try
    {
      return instanceTimeModelStorage.getTimeModelStatusReadonly(instanceID);
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    return null;
  }

  public UpdateableInstanceTimeContext getAndLockInstanceTimeContext(SessionToken session,
      UUID instanceID) throws DataSourceException
  {
    return instanceTimeModelStorage.getAndLockInstanceTimeContext(session, instanceID);
  }

  public void updateInstanceTimeContext(SessionToken session,
      InstanceTimeContext timeData) throws LockException, DataSourceException
  {
    instanceTimeModelStorage.updateInstanceTimeContext(session, timeData);
  }

  public void unlockInstanceTimeContext(SessionToken session, UUID instanceID)
      throws LockException
  {
    instanceTimeModelStorage.unlockInstanceTimeContext(session, instanceID);
  }

  public UpdateableInstanceTimeContext createAndLockInstanceTimeContext(SessionToken session,
      UUID instanceID) throws DataSourceException
  {
    return instanceTimeModelStorage.createAndLockInstanceTimeContext(session, instanceID);
  }

  private Instance getInstance(SessionToken parentSession, UUID instanceID)
  {
    SessionToken session = subSession(parentSession);
    Instance instance = this.processManager.getInstanceManager()
      .getInstance(session, instanceID);
    return instance;
  }

  private SessionToken subSession(SessionToken parentSession)
  {
    return this.timeManager.subSession(parentSession);
  }

  /**
   * Note this will NOT change the Time Model Status!
   * @param session NOTE: This SessionToken will hold a lock on the Time Model
   * @param templateID
   * @param instanceID
   * @param checkReport
   * @return
   * @throws DataSourceException
   */
  public UpdateableInstanceTimeModel createAndLockInstanceTimeModel(SessionToken session,
      UUID templateID, UUID instanceID, 
      CheckReport checkReport) throws DataSourceException
  {
    UpdateableInstanceTimeModel timeModel = null;
    
    // Get the Template Time Model from the Storage
    TemplateTimeModel templateTimeModel = this.templateTimeManager.getTimeModel(session, templateID);
    
    // Make sure the Template Time Model is available
    if (templateTimeModel != null) // Time Model not available or invalid?
    {
      // Get the Instance
      Instance instance = getInstance(session, instanceID);
      // and the Instance Time Context
      InstanceTimeContext timeContext = this.instanceTimeModelStorage.getInstanceTimeContextReadonly(instanceID);

      // Create the instance Time Model
      // FIXME AL: creation and initializing should be two separate steps?
      timeModel = (UpdateableInstanceTimeModel) this.timeModelFactory
        .createTimeModelForInstance(instance, timeContext, templateTimeModel);

     
      checkTimeModel(session, timeModel, checkReport);
      
      // Check if the time model has been created successfully
      if (checkReport.getCheckResult().ordinal() <= ResultType.WARNING
        .ordinal())
      {
        this.instanceTimeModelStorage.lockInstanceTimeModel(session, instanceID);
        try
        {  
          // try
          // {
          this.instanceTimeModelStorage.updateInstanceTimeModel(session, instanceID, timeModel);
          // }
          // finally
          // {
          // this.storage.unlockInstanceTimeModel(session, instanceID);
          // }
          // TODO AL: Unlock instance time model in case of an exception
        }
        catch (DataSourceException e)
        {
          logger.log(Level.SEVERE, "DataSourceException", e);
        }
        catch (LockException e)
        {
          logger.log(Level.SEVERE, "LockException", e);
        }
      }
    }
    else
    {
      checkReport.addReportEntry("General", ResultType.FAILURE, "No template time model available!");
    }
    
    return timeModel;
  }

  public void unlockTimeModel(SessionToken session, UUID instanceID)
      throws LockException
  {
    this.instanceTimeModelStorage.unlockInstanceTimeModel(session, instanceID);
  }

  public UpdateableInstanceTimeModel getAndLockTimeModel(SessionToken session,
      UUID instanceID) throws DataSourceException
  {
    return instanceTimeModelStorage.getAndLockInstanceTimeModel(session, instanceID);
  }

  public InstanceTimeModel getInstanceTimeModelReadonly(SessionToken session,
      UUID instanceID) throws DataSourceException
  {
    return instanceTimeModelStorage.getInstanceTimeModelReadonly(instanceID);
  }
  
  public void updateTimeModel(SessionToken session,
      UpdateableInstanceTimeModel timeModel)
  {
    // TODO Auto-generated method stub

  }
}
