package de.aristaflow.adept2.core.timemanager.defaultimplementation;

import java.util.UUID;
import java.util.logging.Level;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.ComponentTypes;
import de.aristaflow.adept2.core.checks.processmodel.ProcessTemplateCheck;
import de.aristaflow.adept2.core.checks.processmodel.ProcessTemplateCheckRunner;
import de.aristaflow.adept2.core.checks.timemodel.TimeModelCheck;
import de.aristaflow.adept2.core.checks.timemodel.TimeModelCheckRunner;
import de.aristaflow.adept2.core.processmanager.ProcessManager;
import de.aristaflow.adept2.core.timemanager.TemplateTimeManager;
import de.aristaflow.adept2.core.timemanager.storage.TemplateTimeModelStorage;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultTemporalTemplate;
import de.aristaflow.adept2.model.processmodel.tools.NodeRelations;
import de.aristaflow.adept2.model.processmodel.tools.ProcessElementIdentifierTools;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.CheckReport.ResultType;

/**
 * @author  lanz
 */
public class DefaultTemplateTimeManager extends AbstractSubService implements
    TemplateTimeManager
{

  /**
   * @uml.property  name="storage"
   * @uml.associationEnd  
   */
  private TemplateTimeModelStorage storage;

  /**
   * @uml.property  name="timeManager"
   * @uml.associationEnd  
   */
  private final DefaultTimeManager timeManager;

  /**
   * @uml.property  name="processManager"
   * @uml.associationEnd  
   */
  private ProcessManager processManager;

  /**
   * @uml.property  name="timeModelFactory"
   * @uml.associationEnd  
   */
  private TimeModelFactory timeModelFactory;

  public DefaultTemplateTimeManager(
      DefaultTimeManager timeManager, TimeModelFactory timeModelFactory,
      TemplateTimeModelStorage templateTimeStorage)
  {
    super(new String[]{ComponentTypes.COMPONENT_TYPE_PROCESS_MANAGER}, new String[0], timeManager);
    this.timeManager = timeManager;
    this.timeModelFactory = timeModelFactory;
    this.storage = templateTimeStorage;
  }

  @Override
  public void init() throws AbortServiceException
  {
//    SessionToken initSession = createSessionToken();
    
    this.processManager = this.timeManager.getProcessManager();
  }

  public UUID[] getAllTimeModelIDs(SessionToken session)
  {
    return this.storage.getAllTimeModels();
  }

  /**
   * Gets the designated time model for the process template identfied by the template ID. The returned time model can not be changed.
   * 
   * <b>Note that this method will create the time model if it does not exist yet.</b>
   */
  public TemplateTimeModel getTimeModel(SessionToken session, UUID templateID)
  {
    TemplateTimeModel templateTimeModel = this.storage.getTimeModel(templateID);
   
    // XXX AL this is an intermediate solution. Time Models should actually be created when the template is deployed 
    // If it has not yet been created, create it
    if (templateTimeModel == null)
    {
      CheckReport checkReport = new CheckReport(getURIs()[0]);
      templateTimeModel = this.createAndGetTemplateTimeModel(session, templateID, checkReport);
      
      if (checkReport.getCheckResult().ordinal() >= CheckReport.ResultType.FAILURE.ordinal())
      {
        templateTimeModel = null;
        logger.log(Level.SEVERE, checkReport.getReportSummary());
      }
    }

    return templateTimeModel;
  }

  public TemplateTimeModel createAndGetTemplateTimeModel(SessionToken parentSession,
      UUID templateID, CheckReport checkReport)
  {
    // XXX AL throw an exception in case the template time model cannot be created
    SessionToken session = subSession(parentSession);

    TemplateTimeModel timeModel = null;
    TemporalTemplate template = getTemplate(session, templateID);

    if (checkTemplate(session, template, checkReport))
    {
      timeModel = this.timeModelFactory.createTimeModelForTemplate(template);

      if (checkTimeModel(session, timeModel, template, checkReport))
      {
        this.storage.storeTimeModel(template.getID(), timeModel);
      }
      else
      {
        checkReport.addReportEntry("TemplateTimeManager", ResultType.FAILURE, "TimeModel contains errors",
                                   ProcessElementIdentifierTools.getTemplateIdentifier(template, "TemplateTimeManager"));
      }
    }
    else
    {
      checkReport.addReportEntry("TemplateTimeManager", ResultType.FAILURE, "Template contains errors",
                                 ProcessElementIdentifierTools.getTemplateIdentifier(template, "TemplateTimeManager"));
    }
      
    return timeModel;
  }
  
  private SessionToken subSession(SessionToken parentSession)
  {
    return this.timeManager.subSession(parentSession);
  }

  public TemporalTemplate getTemplate(SessionToken parentSession, UUID templateID)
  {
    SessionToken session = subSession(parentSession);
    Template template = this.processManager.getTemplateManager()
      .getTemplate(session, templateID);
    return new DefaultTemporalTemplate(template);
  }

  private boolean checkTemplate(final SessionToken session,
      final Template template, final CheckReport report)
  {
    super.sessionActive(session);
    try
    { // sessionActive
      // check the template for any problem
      boolean ret = true;
      // set up check runner
      ProcessTemplateCheck[] checks;
      // get the checks
      checks = this.timeManager.getTemplateChecks(session);
      ProcessTemplateCheckRunner allChecks = new ProcessTemplateCheckRunner(checks);

      // perform structural checks
      ret &= allChecks.performCheck(template, new NodeRelations(template),
                                    report);

      return ret;
    }
    finally
    {
      super.sessionFinished(session);
    }
  }

  private boolean checkTimeModel(final SessionToken session,
      final TimeModel timeModel, final Template template, final CheckReport report)
  {
    super.sessionActive(session);
    try
    { // sessionActive
      // check the template for any problem
      boolean ret = true;
      // set up check runner
      TimeModelCheck[] checks;
      // get the checks
      checks = this.timeManager.getTimeModelChecks(session);
      TimeModelCheck allChecks = new TimeModelCheckRunner(false, checks);

      
      // perform consistency checks
      ret &= allChecks.performCheck(timeModel, template, new NodeRelations(template), report);

      return ret;
    }
    finally
    {
      super.sessionFinished(session);
    }
  }


}
