package de.aristaflow.adept2.core.timemanager.defaultimplementation;

import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_AGENT;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_ORGPOSITION;
import static de.aristaflow.adept2.model.globals.WorklistConstants.SYSTEM_PASSWORD;

import java.net.URI;
import java.util.logging.Level;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.configuration.ConfigurationDescription;
import de.aristaflow.adept2.base.configuration.Property;
import de.aristaflow.adept2.base.configuration.Property.Type;
import de.aristaflow.adept2.base.service.AbstractADEPT2Service;
import de.aristaflow.adept2.base.service.InternalServiceException;
import de.aristaflow.adept2.base.service.Registry;
import de.aristaflow.adept2.base.service.ServiceAccessControlException;
import de.aristaflow.adept2.base.service.ServiceNotKnownException;
import de.aristaflow.adept2.base.sessionmanagement.SecurityTokenIntegrityException;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.ComponentTypes;
import de.aristaflow.adept2.core.checks.processmodel.ProcessTemplateCheck;
import de.aristaflow.adept2.core.checks.processmodel.timemodel.DefaultTimedProcessModelCheckService;
import de.aristaflow.adept2.core.checks.processmodel.timemodel.TemporalProcessCheckService;
import de.aristaflow.adept2.core.checks.timemodel.TimeModelCheck;
import de.aristaflow.adept2.core.checks.timemodel.TimeModelCheckService;
import de.aristaflow.adept2.core.checks.timemodel.tcnimplementation.TCNTimModelCheckService;
import de.aristaflow.adept2.core.complexeventmanager.EventManager;
import de.aristaflow.adept2.core.executionmanager.ExecutionManager;
import de.aristaflow.adept2.core.processmanager.ProcessManager;
import de.aristaflow.adept2.core.timemanager.InstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.TemplateTimeManager;
import de.aristaflow.adept2.core.timemanager.TimeManager;
import de.aristaflow.adept2.core.timemanager.TimeModellingService;
import de.aristaflow.adept2.core.timemanager.WorklistInteraction;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.escalation.defaultimplementation.DefaultEscalationManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation.DefaultActivityStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation.DefaultActivityTermination;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation.DefaultInstanceStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation.DefaultInstanceTermination;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation.DefaultProcessExecutionManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timeControl.defaultimplementation.DefaultInstanceTimeControlManager;
import de.aristaflow.adept2.core.timemanager.storage.InstanceTimeModelStorage;
import de.aristaflow.adept2.core.timemanager.storage.TemplateTimeModelStorage;
import de.aristaflow.adept2.core.timemanager.storage.TimeManagerStorage;
import de.aristaflow.adept2.core.timemanager.storage.transientstorage.TransientInstanceTimeModelStorage;
import de.aristaflow.adept2.core.timemanager.storage.transientstorage.TransientTemplateTimeStorage;
import de.aristaflow.adept2.core.timemanager.storage.xmlstorage.XMLTimeManagerStorage;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.util.ArrayTools;

/**
 * @author  lanz
 */
@ConfigurationDescription(
  properties = {
     @Property(
        name = XMLTimeManagerStorage.CONFIGURATION_DATA_DIR, 
        type = Type.STRING, isRequired = true,
        description = "Save Path for Time Manager Persistence.")
  }
)
public class DefaultTimeManager extends AbstractADEPT2Service implements
    TimeManager
{

  private static final boolean TEST_MODE = false;

  /**
   * @uml.property  name="eventManager"
   * @uml.associationEnd  
   */
  private EventManager eventManager;

  
  
  /**
   * @uml.property  name="instanceTimeManager"
   * @uml.associationEnd  
   */
  private DefaultInstanceTimeManager instanceTimeManager;
  
  /**
   * @uml.property  name="templateTimeManager"
   * @uml.associationEnd  
   */
  private DefaultTemplateTimeManager templateTimeManager;

  /**
   * @uml.property  name="instanceTimeControlManager"
   * @uml.associationEnd  
   */
  private DefaultInstanceTimeControlManager instanceTimeControlManager;
  

  /**
   * @uml.property  name="processExecutionManager"
   * @uml.associationEnd  
   */
  private DefaultProcessExecutionManager processExecutionManager;
  

  /**
   * @uml.property  name="timeModelFactory"
   * @uml.associationEnd  
   */
  private TimeModelFactory timeModelFactory;
  
  /**
   * @uml.property  name="timeModelChecks"
   * @uml.associationEnd  
   */
  private TemporalProcessCheckService processModelChecks;
  
  private TimeModelCheckService timeModelChecks;

  
  /**
   * @uml.property  name="instanceStarting"
   * @uml.associationEnd  
   */
  private DefaultInstanceStarting instanceStarting;

  /**
   * @uml.property  name="instanceTermination"
   * @uml.associationEnd  
   */
  private DefaultInstanceTermination instanceTermination;

  /**
   * @uml.property  name="activityStarting"
   * @uml.associationEnd  
   */
  private DefaultActivityStarting activityStarting;

  /**
   * @uml.property  name="activityTermination"
   * @uml.associationEnd  
   */
  private DefaultActivityTermination activityTermination;

  private final Configuration configuration;

  private TimeModellingService timeModellingService;

  private ProcessManager processManager;

  private DefaultEscalationManager escalationManager;

  private ExecutionManager executionManager;

  private TimeAwareWorklistInteraction worklistInteraction;

  
  
  
  public DefaultTimeManager(Configuration configuration, Registry registry,
      String[] startupRequiredServices, String[] runtimeRequiredServices)
  {
    super(configuration, registry, ArrayTools
      .join(startupRequiredServices,
            new String[] { ComponentTypes.COMPONENT_TYPE_COMPLEX_EVENT_MANAGER,
                           ComponentTypes.COMPONENT_TYPE_PROCESS_MANAGER,
                           ComponentTypes.COMPONENT_TYPE_EXECUTION_MANAGER}),
          runtimeRequiredServices);
    this.configuration = configuration;
    
    logger.log(Level.INFO, "Time Manager initialized");
  }

  public DefaultTimeManager(Configuration configuration, Registry registry)
  {
    this(configuration, registry, new String[0], new String[0]);
  }

  @SuppressWarnings("deprecation")
  @Override
  public void init(URI[] myURIs) throws AbortServiceException
  {
    super.init(myURIs, SYSTEM_AGENT, SYSTEM_ORGPOSITION, SYSTEM_PASSWORD);

    SessionToken initSession = createSessionToken();
    try
    {
      this.eventManager = this.registry
        .getServiceOfType(initSession,
                          ComponentTypes.COMPONENT_TYPE_COMPLEX_EVENT_MANAGER,
                          EventManager.class);
    }
    catch (ServiceNotKnownException e)
    {
      throw new InternalServiceException(e.getMessage());
    }
    
    try
    {
      this.processManager = this.registry
        .getServiceOfType(initSession,
                          ComponentTypes.COMPONENT_TYPE_PROCESS_MANAGER,
                          ProcessManager.class);
    }
    catch (ServiceNotKnownException e)
    {
      throw new InternalServiceException(e.getMessage());
    }
    
    try
    {
      this.executionManager = this.registry
        .getServiceOfType(initSession,
                          ComponentTypes.COMPONENT_TYPE_EXECUTION_MANAGER,
                          ExecutionManager.class);
    }
    catch (ServiceNotKnownException e)
    {
      throw new InternalServiceException(e.getMessage());
    }
    
    // May one day become a service of it's own...
    this.escalationManager = new DefaultEscalationManager(configuration, registry);
    
    // Model Factory
    this.timeModelFactory = registry.getModelFactory("TimeModelFactory",
                                                     TimeModelFactory.class);
    
    // Storage Facilities
    InstanceTimeModelStorage instanceTimeModelStorage = new TransientInstanceTimeModelStorage();
    TemplateTimeModelStorage templateTimeStorage = new TransientTemplateTimeStorage();
    TimeManagerStorage timeManagerStorage = new XMLTimeManagerStorage(configuration, timeModelFactory);
    

    // Time Model Managers
    this.templateTimeManager = new DefaultTemplateTimeManager(this,
                                                              timeModelFactory,
                                                              templateTimeStorage);
    this.instanceTimeManager = new DefaultInstanceTimeManager(this,
                                                              timeModelFactory, 
                                                              templateTimeManager, 
                                                              instanceTimeModelStorage);
    
    

    this.timeModellingService = new DefaultTimeModellingService(this, timeModelFactory, templateTimeManager, timeManagerStorage);
    
    // Instance Control
    this.instanceStarting = new DefaultInstanceStarting(this,
                                                        timeModelFactory,
                                                        templateTimeManager,
                                                        instanceTimeManager,
                                                        escalationManager);
    this.instanceTermination = new DefaultInstanceTermination(this,
                                                              instanceTimeManager);

    // ActivityControl
    this.activityStarting = new DefaultActivityStarting(this, timeModelFactory,
                                                        instanceTimeManager);
    this.activityTermination = new DefaultActivityTermination(this, timeModelFactory,
                                                              instanceTimeManager);

    // Process Execution Manager
    this.processExecutionManager = new DefaultProcessExecutionManager(this,
                                                                      instanceTimeManager,
                                                                      eventManager,
                                                                      instanceStarting,
                                                                      instanceTermination,
                                                                      activityStarting,
                                                                      activityTermination);
    
    
    this.instanceTimeControlManager = new DefaultInstanceTimeControlManager(this);
    
    this.worklistInteraction = new TimeAwareWorklistInteraction(this);

    // get the check service, may one day become a full service of its own.
    this.processModelChecks = new DefaultTimedProcessModelCheckService(this.configuration, registry);
    this.timeModelChecks = new TCNTimModelCheckService(this.configuration, this.registry);
    
    // Init Sub-Services
    this.escalationManager.init(myURIs);
    this.templateTimeManager.init();
    this.instanceTimeManager.init();
    this.instanceStarting.init();
    this.instanceTermination.init();
    this.activityStarting.init();
    this.activityTermination.init();
    this.processExecutionManager.init();
    this.instanceTimeControlManager.init();
    this.processModelChecks.init(myURIs);
    this.timeModelChecks.init(myURIs);

  }
  
  @Override
  public void start() throws AbortServiceException
  {
    super.start();

    this.escalationManager.start();
    
    // Time Model Managers
    this.templateTimeManager.start();
    this.instanceTimeManager.start();

    // Instance Control
    this.instanceStarting.start();
    this.instanceTermination.start();

    // ActivityControl
    this.activityStarting.start();
    this.activityTermination.start();

    // Process Execution Manager
    this.processExecutionManager.start();
    
    this.instanceTimeControlManager.start();
  }

  @Override
  public void shutdown()
  {
    this.escalationManager.shutdown();
    
    // Time Model Managers
    this.templateTimeManager.shutdown();
    this.instanceTimeManager.shutdown();

    // Instance Control
    this.instanceStarting.shutdown();
    this.instanceTermination.shutdown();

    // ActivityControl
    this.activityStarting.shutdown();
    this.activityTermination.shutdown();

    // Process Execution Manager
    this.processExecutionManager.shutdown();
    
    this.instanceTimeControlManager.shutdown();
    
    super.shutdown();
  }

  @Override
  public void emergencyShutdown()
  {
    this.escalationManager.emergencyShutdown();
    
    // Time Model Managers
    this.templateTimeManager.emergencyShutdown();
    this.instanceTimeManager.emergencyShutdown();

    // Instance Control
    this.instanceStarting.emergencyShutdown();
    this.instanceTermination.emergencyShutdown();

    // ActivityControl
    this.activityStarting.emergencyShutdown();
    this.activityTermination.emergencyShutdown();

    // Process Execution Manager
    this.processExecutionManager.emergencyShutdown();
    
    this.instanceTimeControlManager.emergencyShutdown();
      
    
    super.emergencyShutdown();
  }




  /**
   * @return
   * @uml.property  name="instanceTimeManager"
   */
  public InstanceTimeManager getInstanceTimeManager()
  {
    return this.instanceTimeManager;
  }

  /**
   * @return
   * @uml.property  name="templateTimeManager"
   */
  public TemplateTimeManager getTemplateTimeManager()
  {
    return this.templateTimeManager;
  }
  
//  @Override
//  protected SessionToken createSessionToken()
//  {
//    return super.createSessionToken();
//  }

  public TimeModelCheck[] getTimeModelChecks(SessionToken session)
  {    

    TimeModelCheck checks;
    if (TEST_MODE)
    {
      checks = this.timeModelChecks.getExecutionTestClientChecks();
    }
    else
    {
      checks = this.timeModelChecks.getExecutionProductionChecks();
    }
    
    return new TimeModelCheck[]{checks};
  }
  
  public ProcessTemplateCheck[] getTemplateChecks(SessionToken session)
  {
      ProcessTemplateCheck checks;
      if (TEST_MODE)
      {
        checks = this.processModelChecks.getExecutionTestClientChecks();
      }
      else
      {
        checks = this.processModelChecks.getExecutionProductionChecks();
      }
      
      return new ProcessTemplateCheck[] { this.processModelChecks.getProcessTemplateStructureChecks(),
                                          checks};
  }
//
//  public InstanceTimeStorage getInstanceTimeStorage()
//  {
//    return instanceTimeStorage;
//  }

  public SessionToken subSession(SessionToken parentSession)
  {
    checkAccessRights(parentSession);
    return getSessionFactory().getPrivilegedChildSession(parentSession, getURIs());
  }
  
  public Registry getDependencyRegistry()
  {
    return super.getDependencyRegistry();
  }
  

  /**
   * Checks if the caller has the authorisations to access this execution
   * manager.
   *
   * @param sessionToken The session token identifying the caller.
   */
  private final void checkAccessRights(SessionToken sessionToken)
  {
    try
    {
      sessionFactory.checkIntegrity(sessionToken);
    }
    catch (SecurityTokenIntegrityException stie)
    {
      throw new ServiceAccessControlException("The designated session token "
                                              + "failed the verification!", stie);
    }
  }

  @Override
  public TimeModellingService getTimeModellingService()
  {
    return timeModellingService;
  }

  public ProcessManager getProcessManager()
  {
    return processManager;
  }
  
  public ExecutionManager getExecutionManager()
  {
    return this.executionManager;
  }
  
  @Override
  public SessionToken createSessionToken()
  {
    return super.createSessionToken();
  }

  public WorklistInteraction getWorklistInteraction()
  {
    return this.worklistInteraction;
  }

  public WorklistAccess getWorklistAccess()
  {
     return this.worklistInteraction;
  }

}
