package de.aristaflow.adept2.core.timemanager.defaultimplementation;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import de.aristaflow.adept2.util.Adept2ThreadFactory;
import de.aristaflow.adept2.util.LoggerTools;

public class DefaultTriggerManager implements TriggerManager
{
  
  /**
   * @author  lanz
   */
  public abstract class DefaultTrigger implements Trigger
  {
    /**
     * @uml.property  name="runnable"
     */
    private final Runnable runnable;
    /**
     * @uml.property  name="triggerTime"
     */
    private final long triggerTime;
    private final boolean interruptOnCancel;

    public DefaultTrigger(long triggerTime, Runnable runnable, boolean interruptOnCancel)
    {
      this.triggerTime = triggerTime;
      this.runnable = runnable;
      this.interruptOnCancel = interruptOnCancel;
    }
    
    public DefaultTrigger(long triggerTime, Runnable runnable)
    {
      this(triggerTime, runnable, false);
    }

    /**
     * @return
     * @uml.property  name="runnable"
     */
    public Runnable getRunnable()
    {
      return runnable;
    }

    /**
     * @return
     * @uml.property  name="triggerTime"
     */
    public long getTriggerTime()
    {
      return triggerTime;
    }

    public boolean interruptOnCancel()
    {
      return interruptOnCancel;
    }
    
    @Override
    public boolean equals(Object other)
    {
      if (this == other)
        return true;
      
      if (other instanceof Trigger)
      {
        return equals((Trigger)other);
      }
      
      return false;
    }
    
    public abstract boolean equals(Trigger other);

    @Override
    public abstract int hashCode();
    
  }
  
  protected final Logger logger;
  
  /** Scheduled Executor used for Execution of the Trigger Routines */
  private final ScheduledExecutorService executor;
  
  private final Map<Trigger, ScheduledFuture<?>> triggers;
  
  
  public DefaultTriggerManager()
  {
    this.logger = LoggerTools.getLogger(this);
    this.executor = Executors
      .newScheduledThreadPool(2, new Adept2ThreadFactory("TriggerManager", logger));
    this.triggers = new HashMap<Trigger, ScheduledFuture<?>>();
  }
  
  public void setTrigger(Trigger trigger)
  {

    if (convertTime(trigger.getTriggerTime()) <= 200)
    {
      logger.info("Trigger date has already passed! Triggering now!");
      executor.execute(trigger.getRunnable());
      return;
    }
    
    synchronized (this.triggers)
    {
      ScheduledFuture<?> schedule = this.executor
        .schedule(trigger.getRunnable(), convertTime(trigger.getTriggerTime()),
                  TimeUnit.MILLISECONDS);
      this.triggers.put(trigger, schedule);
    }
    
  }

  public void cancelTrigger(Trigger trigger)
  {
    synchronized (this.triggers)
    {
      // Get the Scheduled trigger for the particular trigger
      ScheduledFuture<?> scheduled = this.triggers.remove(trigger);
      
      // if there was a trigger scheduled abort it.
      if (scheduled != null)
      {
        scheduled.cancel(trigger.interruptOnCancel());
      }
    }
  }

  private long convertTime(long time)
  {
    return ( ( time - System.currentTimeMillis() + 50) / 100 ) * 100;
    
  }
  
  /**
   * Requests this Trigger Manager to be shut down.
   */
  public void shutdown()
  {
    // Instruct the executor service to terminate ...
    this.executor.shutdownNow();
    try
    {
      // ... and wait for it's termination
      if (!this.executor.awaitTermination(10000l, TimeUnit.MILLISECONDS))
        logger.warning("Scheduler didn't terminate in time.");
    }
    catch (InterruptedException ie)
    {
      // ignore interruptions since we are shutting down
    }
  }
  
}
