package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import java.util.UUID;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;

import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultInstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ActivityStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ActivityStateUpdateManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ActivityTermination;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeContext;
import de.aristaflow.adept2.util.Adept2ThreadFactory;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;


public class DefaultActivityStateUpdateManager extends AbstractSubService
    implements ActivityStateUpdateManager
{
  
  /**
   * @uml.property  name="activityStarting"
   * @uml.associationEnd  
   */
  private ActivityStarting activityStarting;

  /**
   * @uml.property  name="activityTermination"
   * @uml.associationEnd  
   */
  private ActivityTermination activityTermination;

  /**
   * @uml.property  name="instanceTimeManager"
   * @uml.associationEnd  
   */
  private DefaultInstanceTimeManager instanceTimeManager;

  private ExecutorService executorService;

  public DefaultActivityStateUpdateManager(DefaultTimeManager timeManager,
      DefaultInstanceTimeManager instanceTimeManager, ActivityStarting activityStarting,
      ActivityTermination activityTermination)
  {
    super(timeManager);
    this.instanceTimeManager = instanceTimeManager;
    this.activityStarting = activityStarting;
    this.activityTermination = activityTermination;
  }

  @Override
  public void init()
  {
    executorService = Executors
      .newCachedThreadPool(new Adept2ThreadFactory("ActivityStateUpdate", logger)); //$NON-NLS-1$
  }

  /** Awaits the termination of the executor service. */
  private void awaitExecutorServiceShutdown()
  {
    boolean terminated = false;
    while (!terminated)
    {
      try
      {
        terminated = executorService.awaitTermination(30, TimeUnit.SECONDS);
      }
      // ignore interruptions since we are shutting down
      catch (InterruptedException ie)
      { 
        //
      }
    }
  }

  @Override
  public void emergencyShutdown()
  {
    executorService.shutdown();
    awaitExecutorServiceShutdown();
    super.emergencyShutdown();
  }

  @Override
  public void shutdown()
  {
    executorService.shutdownNow();
    awaitExecutorServiceShutdown();
    super.shutdown();
  }
  

  public void activityStarted(final SessionToken session, final UUID instanceID,
      final int nodeID, final int iteration, final long timestamp)
  {
    try
    {
      // Read the Time Context
      UpdateableInstanceTimeContext timeContext = this.instanceTimeManager.getAndLockInstanceTimeContext(session, instanceID);
      try
      {
        // Update the time context
        timeContext.setActivityStarted(nodeID, iteration, timestamp);
        // Store the updated time context
        this.instanceTimeManager.updateInstanceTimeContext(session, timeContext);
      }
      finally
      {
        // Unlock the Time Context
        this.instanceTimeManager.unlockInstanceTimeContext(session, instanceID);
      }
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "LockException", e);
    }
    
    
    executorService.execute(new Runnable(){

      public void run()
      {
        try
        {
          // Call ActivityStarting to handle the time model etc.
          activityStarting.activityStarted(session, instanceID, nodeID,
                                           iteration, timestamp);
        }
        catch (Exception e)
        {
          logger.log(Level.SEVERE, "An error occurred",e);
        }
      }
    });
  }

  public void activityFinished(final 
      SessionToken session, final UUID instanceID,
      final int nodeID, final int iteration, final long timestamp)
  {
    try
    {
      // Read the Time Context
      UpdateableInstanceTimeContext timeContext = this.instanceTimeManager.getAndLockInstanceTimeContext(session, instanceID);
      try
      {
        // Update the time context
        timeContext.setActivityFinished(nodeID, iteration, timestamp);
        // Store the updated time context
        this.instanceTimeManager.updateInstanceTimeContext(session, timeContext);
      }
      finally
      {
        // Unlock the Time Context
        this.instanceTimeManager.unlockInstanceTimeContext(session, instanceID);
      }
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "LockException", e);
    }
    
    
    executorService.execute(new Runnable(){

      public void run()
      {
        // Call ActivityTermination to handle the time model etc.
        activityTermination.activityFinished(session, instanceID, nodeID, iteration, timestamp);
      }
    });
  }



}
