package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import java.util.UUID;
import java.util.logging.Level;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultInstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTemplateTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.WorklistAccess;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.escalation.EscalationManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.InstanceStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timemodelcontrol.TemporalExceptionHandling;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timemodelcontrol.defaultimplementation.SysErrTemporalExceptionHandling;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.ProcessDuration;
import de.aristaflow.adept2.model.processmodel.timemodel.ValidityPeriod;
import de.aristaflow.adept2.model.timemodel.ChangeableTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.InconsistencyException;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelState;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelUpdateStatus;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.CheckReport.ResultType;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;

/**
 * @author  lanz
 */
public class DefaultInstanceStarting extends AbstractSubService implements
    InstanceStarting
{

  /**
   * @uml.property  name="templateTimeManager"
   * @uml.associationEnd  
   */
  protected final DefaultTemplateTimeManager templateTimeManager;

  /**
   * @uml.property  name="instanceTimeManager"
   * @uml.associationEnd  
   */
  protected final DefaultInstanceTimeManager instanceTimeManager;

  /**
   * @uml.property  name="timeModelTools"
   * @uml.associationEnd  
   */
  protected TimeModelOperations timeModelTools;
  
  /**
   * @uml.property  name="escalationManager"
   * @uml.associationEnd  
   */
  protected EscalationManager escalationManager;

  /**
   * @uml.property  name="timeModelFactory"
   * @uml.associationEnd  
   */
  protected final TimeModelFactory timeModelFactory;

  /**
   * @uml.property  name="temporalExceptionHandling"
   * @uml.associationEnd  
   */
  protected TemporalExceptionHandling temporalExceptionHandling;

  protected final DefaultTimeManager timeManager;

  public DefaultInstanceStarting(
      DefaultTimeManager timeManager,
      TimeModelFactory timeModelFactory,
      DefaultTemplateTimeManager templateTimeManager,
      DefaultInstanceTimeManager instanceTimeManager,
      EscalationManager escalationManager)
  {
    super(timeManager);
    this.logger.setLevel(Level.ALL);
    this.timeManager = timeManager;
    this.timeModelFactory = timeModelFactory;
    
    this.templateTimeManager = templateTimeManager;
    this.instanceTimeManager = instanceTimeManager;
    this.escalationManager = escalationManager;
    
  }
  
  @Override
  public void init() throws AbortServiceException
  {
    super.init();
    this.timeModelTools = timeModelFactory.getTimeModelOperations();
    this.temporalExceptionHandling = new SysErrTemporalExceptionHandling();
  }

  /////////////////////////////////////////////
  // BEGIN Instance Created
  /////////////////////////////////////////////
  
  public void instanceCreated(SessionToken parentSession, UUID templateID,
      UUID instanceID, long timestamp)
  {
    logger.entering(logger.getName(), "instanceCreated(SessionToken,UUID,UUID,long)");
   
    System.out.println("instanceCreated("+instanceID+")");
    SessionToken session = subSession(parentSession);
    
    // Get the Time Model status which is create along with the InstanceTimeContext (will be set to DISABLED by default)
    ChangeableTimeModelStatus status = this.instanceTimeManager
      .getAndLockTimeModelStatus(session, instanceID);
    
    try
    {
    
      SessionToken childSession = subSession(session);
      // First the time model is instantiated and the time stamp of the process
      // instantiation event is set to the current time point
      instantiateTimeModel(childSession, templateID, instanceID, timestamp, status);
      
      // If the time model is valid and may be updated ...
      if (status.getTimeModelUpdateStatus().ordinal() 
          < TimeModelUpdateStatus.NO_UPDATE.ordinal())
      {
        // the execution component of the time manager is initialized
        initializeInstanceExecution(childSession, instanceID, status);
      }
    }
    catch (RuntimeException e)
    {
      e.printStackTrace();
      throw e;
    }
    finally
    {
      // Finally the new time model is stored and the lock is released 
      try
      {
        try
        {
            this.instanceTimeManager.updateTimeModelStatus(session, status);
        }
        finally
        {
          this.instanceTimeManager.unlockTimeModelStatus(session, instanceID);
        }
      }
      catch (LockException e)
      {
        logger.log(Level.SEVERE, "LockException", e);
      }
    }

    
    //logger.exiting(logger.getName(), "instanceCreated(SessionToken,UUID,UUID,long)");
  }
  
  private SessionToken subSession(SessionToken parentSession)
  {
    return this.timeManager.subSession(parentSession);
  }

  /**
   * <ul>
   * <li>Zunaechst wird ueberprueft, ob die <u>Validity Period</u> auf dem
   * <u>Earliest</u> oder <u>Latest Start Date</u> des Prozesses erfuellt ist
   * (Wert mglw. aus Konfiguration lesen)<span> </span>und der Prozess somit
   * ueberhaupt gestartet werden darf.</li>
   * <li>Abschaetzen der Erfuellbarkeit der <u>Validity Period</u> auf dem
   * <u>Earliest</u> oder <u>Latest Completion Date</u> des Prozesses</span>
   * (Wert mglw. aus Konfiguration lesen) mithilfe des
   * Instanziierungszeitpunkts und der zur Build-time vorberechneten
   * Prozessdauer.
   * <ul>
   * <li>Eventuell kann eine <u>Validity Period</u> auf <u>Earliest</u> oder
   * <u>Latest Completion Date</u> waehrend der build-time bzw. waehrend dem
   * deploy mithilfe der berechneten Prozessdauer schon auf eine entsprechende
   * <u>Validity Period</u> auf dem Start abgebildet werden.</li>
   * </ul>
   * </li>
   * <li>Nun wird das Zeitmodell instanziiert und das
   * ProzessInstantiation-Event auf den aktuellen Zeitpunkt gesetzt.
   * <ul>
   * <li>Verwendung von zur Build-Time vorberechnetem Zeitmodell als Basis</li>
   * <li>XOR-Entscheidungen die lediglich von Prozess-Eingabeparametern
   * abhaengen werden ausgewertet und die abgewaehlten Pfade im Zeitmodell
   * deaktiviert bzw. aus diesem entfernt.</li>
   * <li><u>Time-dependent Variability</u> so weit moeglich auswerten</li>
   * <li>Es werden alle Constraints aktualisiert, die das
   * ProzessInstantiation-Event als Update-Trigger haben (d.h. ihre Daten aus
   * den Prozesseingabe-Parametern bzw. einer Konfiguration beziehen)<br/>
   * <ul>
   * <li>Die Constraints und Ausfuehrungsintervalle werden im Zeitmodell
   * entsprechend angepasst</li>
   * </ul>
   * </li>
   * <li>Auf Basis der aktuellen Daten des Zeitmodell wird fuer jedes Schedule
   * Restricted Element der erste und letzte im entsprechenden
   * Ausfuehrungs-Intervall des Events moegliche Slot ermittelt und falls noetig
   * das Ausfuehrungs-Intervall entsprechend eingeschraenkt
   * <ul>
   * <li>Falls kein Slot verfuegbar ist wir der Start des Prozesses abgebrochen
   * </li>
   * </ul>
   * </li>
   * <li>Nun findet eine erste Zeitrechnung auf dem vollstaendigen Zeitmodell
   * statt. Dadurch werden die Werte an die aktuelle Instanz angepasst.
   * <ul>
   * <li>Relativ effizient aufgrund der zur Build-Time vorberechneten Werte.
   * (nur Schedule Restricted Elements sind problematisch)
   * <ul>
   * <li>-&gt;Time-dependent Variability wird falls moeglich ausgewertet (-&gt;
   * wichtig fuer Validity Period von Aktivitaeten)</li>
   * </ul>
   * </li>
   * <li>Inkonsistenzen sollten dabei hauptsaechlich im Zusammenhang mit
   * folgenden Pattern auftreten (vorausgesetzt die Werte der Konfiguration
   * sind korrekt und es handelt sich um keinen Prozess mit Zeitdaten als
   * Eingabeparameter):
   * <ul>
   * <li><u>Schedule Restricted Elements</u>: Es ist kein gueltiger Slot
   * verfuegbar</li>
   * <li><u>Validity Period</u>: Eine der Aktivitaeten kann nicht innerhalb
   * ihrer <u>Validity Period</u> ausgefuehrt werden bzw. der Prozess kann
   * nicht innerhalb seiner <u>Validity Period</u> auf dem <u>Earliest</u>
   * bzw. <u>Latest Completion Date</u> beendet werden.</li>
   * </ul>
   * </li>
   * <li>Inkonsistenzen in Zusammenhang mit Schedule Restricted Elements und
   * Validity Period sowie Constraints, welche ihre Werte aus der
   * Konfiguration beziehen, sind generelle Probleme, die nicht behoben werden
   * koennen. Treten solche Inkonsistenzen auf wird der Start der
   * Prozessinstanz abgebrochen</li>
   * <li>Inkonsistenzen im Zusammenhang mit Prozess-Eingabeparameter koennen
   * moeglicherweise durch Veraendern der Eingabeparameter behoben werden.
   * <ul>
   * <li>-&gt; Exception und Einbezug des Benutzer der den Prozess gestartet
   * hat.</li>
   * <li>Betroffen: TP1, TP2, TP3, TP4, TP6, TP9, TP10</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>Damit ist, falls kein Fehler auftritt, die Initialisierung des
   * Zeitmodells abgeschlossen</li>
   * </ul>
   * </li>
   * 
   * @param parentSession
   * @param templateID
   * @param instanceID
   * @param timestamp
   * @param status 
   * @return
   */
  private void instantiateTimeModel(
      SessionToken parentSession, UUID templateID, UUID instanceID,
      long timestamp, ChangeableTimeModelStatus status)
  {
    System.out.println("instantiateTimeModel("+instanceID+")");
    SessionToken session = subSession(parentSession);
    UpdateableInstanceTimeModel timeModel = null;
    
    
    // First set the status to enabled and see what happens next (do not update the status in the storage just yet)
    status.setTimeModelUpdateStatus(TimeModelUpdateStatus.ENABLED, null);
    status.setTimeModelStatus(TimeModelState.UNKOWN);
    
    // Check any validity period for process start and process end (estimation)
    // XXX: This should actually be done in ExecutionManager to be able to
    // prevent the instance from beeing started at all
    handleValidityPeriodAtProcessCreate(subSession(parentSession), templateID, timestamp, status);
    
    
    
    if (status.getTimeModelUpdateStatus().ordinal() 
        < TimeModelUpdateStatus.DISABLED.ordinal())
    {
      // Now the time model is instantiated ...
      // NOTE: If successfull, this Method will keep a lock on the time model!
      timeModel = createTimeModel(session, templateID, instanceID, status);
    }
    
    if (timeModel != null)
    {
      System.out.println("Time Model for instance "+instanceID+" successfully created.");
      try
      {        
        try
        {
          // Check if the time model is disable because of an invalid time model
          if (status.getTimeModelUpdateStatus().ordinal() 
                < TimeModelUpdateStatus.DISABLED.ordinal())
          {
      
            // ... and the ProcessInstantiation-Event is set to the current time point.
            // Set Instance Creation Time
            timeModel.setInstanceCreationTime(timestamp-1000); // LOW for now 1sec before start...
            this.instanceTimeManager.updateTimeModel(session, timeModel);
          
            // Check if calculations are disable because of an invalid time model                
            if (status.getTimeModelUpdateStatus().ordinal() 
                < TimeModelUpdateStatus.NO_CALCULATION.ordinal())
            {
              
              // Update any constraint that depends on the Process Instantiation Event,
              // i.e. that read it's values from process input parameters or a configuration
              updateConstraints(session, timeModel, InstanceTimeModel.PROCESS_INSTANTIATION_EVENT);
              
              boolean stateChanged = TimeManagerTools.updateTimeModel(session, timeModelFactory, timeModel, status, this.temporalExceptionHandling);
              if (stateChanged)
              {
                this.instanceTimeManager.updateTimeModelStatus(session, status);
              //TODO AL: determine changes and update triggers!
              }
              
              /**
               * <li>XOR-Entscheidungen die lediglich von
               * Prozess-Eingabeparametern abhaengen werden ausgewertet und die
               * abgewaehlten Pfade im Zeitmodell deaktiviert bzw. aus diesem
               * entfernt.</li> 
               */
//              // Pre-Evaluate XOR Decisions if possible  (-> this may simplify the process/time model)
//              predictXORDecisions(session, timeModel);
//              
//              // Preevaluate Time Dependent Variability (-> this may simplify the process/time model)
//              evaluateTimeDependentVariability(session, timeModel);
//              
//              // Init Schedule Restricted Elements, i.e. update the schedule bounds
//              // according to the execution time
//              initScheduleRestrictedElements(session, timeModel);
//
//              // Update the time model
//              updateTimeModel(session, timeModel, status, this.temporalExceptionHandling);
              
              // Initialization Complete. Store the Time Model
              this.instanceTimeManager.updateTimeModel(session, timeModel);
              
              System.out.println("Time Model for instance "+instanceID+" successfully initialized.");
            }
          }
        }
        finally
        {
          // Release the lock
          this.instanceTimeManager.unlockTimeModel(session, instanceID);
        }
      
      }
      catch (LockException e)
      {
        logger.log(Level.SEVERE, "Lock Exception", e);
      }
    } //if (timeModel != null)
    else
    {
      assert status.getTimeModelStatus().equals(TimeModelState.INVALID);
      assert status.getTimeModelUpdateStatus().equals(TimeModelUpdateStatus.DISABLED);
    }
  }
  
  /**
   * <ul>
   * <li>First, it is checked whether the <em>Validity Period</em> on the
   * <em>Earliest</em> and <em>Latest Start Date</em> of the process is
   * fulfilled (read value from configuration if necessary) and the process
   * may therefore be started at all.</li>
   * <li>Estimate the satisfiability of the <em>Validity Period</em> on the
   * <em>Earliest</em> and <em>Latest Completion Date</em> of the process (read
   * value from configuration if necessary) using the instantiation time point
   * and the process duration pre-calculated at build-time.
   * </li>
   * </ul>
   * 
   * @param session
   * @param templateID
   * @param creationTimeStamp
   * @param currentStatus
   * @return
   */
  private void handleValidityPeriodAtProcessCreate(SessionToken session,
      UUID templateID, long creationTimeStamp, ChangeableTimeModelStatus currentStatus)
  {
    /*
     * First, it is checked whether the <em>Validity Period</em> on the
     * <em>Earliest</em> and <em>Latest Start Date</em> of the process is
     * fulfilled (read value from configuration if necessary) and the process
     * may therefore be started at all.
     */
    TemporalTemplate template = this.templateTimeManager.getTemplate(session, templateID);
    // Check if the start of the template has a validity period
    ValidityPeriod validityPeriod = template.getStartValidityPeriod();
    if (validityPeriod != null)
    {
      // Check if the current time stamp is after the start of the validity period
      if (timeModelTools.compare(validityPeriod.getValidFrom(), creationTimeStamp) > 0)
      {
        String msg = String.format("Template '%1$s' not valid until %2$s!",
                                   templateID, validityPeriod
                                     .getValidFrom());
        logger.severe(msg);
        
        // TODO AL: Check if the start of the validity period is in the near future. If so block instance until then. 
        // TODO AL: handle error
        
        // Fast: disable time calculations for instance
        TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
      }

      // Check if the current time stamp is before the end of the validity period
      if (timeModelTools.compare(validityPeriod.getValidUntil(), creationTimeStamp) < 0)
      {
        String msg = String
          .format("Template '%1$s' only valid until %2$s!", templateID,
                  validityPeriod.getValidUntil());
        logger.severe(msg);
        // TODO AL: handle error
        
        // Fast: disable time calculations for instance
        TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
      }
    }

    /*
     * Estimate the satisfiability of the <em>Validity Period</em> on the
     * <em>Earliest</em> and <em>Latest Completion Date</em> of the process
     * (read value from configuration if necessary) using the instantiation time
     * point and the process duration pre-calculated at build-time.
     */
    
    // Check if the completion of the template has a validity period
    validityPeriod = template.getCompletionValidityPeriod();
    if (validityPeriod != null)
    {
      
      ProcessDuration processDuration = template.getProcessDuration();
      //Check if the instance has a duration constraint or duration prediction
      if (processDuration != null)
      {
        CalendarElement earliestProcessEnd = timeModelTools.add(creationTimeStamp, processDuration.getMinimumDuration());
        // Check if the predicted earliest process end is after the start of the validity period
        if (timeModelTools.compare(validityPeriod.getValidFrom(), earliestProcessEnd) > 0)
        {
          String msg = String.format("Template '%1$s' may not be completed before %2$s!",
                                     templateID, validityPeriod
                                       .getValidFrom());
          logger.severe(msg);
          // TODO AL: handle error
          
          // Fast: disable time calculations for instance
          TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
        }
        
        CalendarElement latestProcessEnd = timeModelTools.add(creationTimeStamp, processDuration.getMaximumDuration());
        
        // Check if the predicted latest completion is before the end of the validity period
        if (timeModelTools.compare(validityPeriod.getValidUntil(), latestProcessEnd) < 0)
        {
          String msg = String
            .format("Template '%1$s' may not be completed after %2$s!", templateID,
                    validityPeriod.getValidUntil());
          logger.severe(msg);
          // TODO AL: handle error
          // Fast: disable time calculations for instance
          TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
        }
      }
    }
  }

  /**
   * Create and instantiate the time model
   * 
   * 
   * 
   * @param session: NOTE this session will hold the lock on the time model!
   * @param templateID
   * @param instanceID
   * @param status
   * @return
   * @throws LockException
   */
  private UpdateableInstanceTimeModel createTimeModel(SessionToken session, UUID templateID,
      UUID instanceID, ChangeableTimeModelStatus status)
  {
    UpdateableInstanceTimeModel timeModel = null;
    
    CheckReport checkReport = new CheckReport(getURIs()[0]);
    // Create the time model for the instance
    try
    {
      timeModel = this.instanceTimeManager.createAndLockInstanceTimeModel(session, templateID, instanceID, 
                              checkReport);
    }
    catch (DataSourceException e)
    {
      // Log the error
      this.logger.log(Level.SEVERE, "DataSourceException while trying to create Instance Time Model",e);
      // Completely disable time management
      TimeManagerTools.updateStatus(status, TimeModelState.INVALID, TimeModelUpdateStatus.DISABLED, 
        "Unable to create Instance Time Model due to a DataSourceException!");
    }
    
    if (checkReport.getCheckResult().ordinal() > ResultType.WARNING
        .ordinal())
    {
      // Time Model could not be created...
      String msg = String.format("Invalid Time Specification of Process Template %1$s! Reason:\n %2$s",
                templateID, checkReport.getReportSummary());
      logger.warning(msg);
      // Completely disable time management as the time data is invalid
      TimeManagerTools.updateStatus(status, TimeModelState.INVALID, TimeModelUpdateStatus.DISABLED, msg);
    }
    
    return timeModel;
  }
  
  /**
   * <li>Nun wird die Ausfuehrungskomponente des Zeitmanagers initialisiert
   * <ul>
   * <li>Zunaechst wird ueberprueft, ob der Prozess sofort gestartet werden kann,
   * d.h. das ProcessStart-Event sofort auftreten darf, oder aufgrund eines
   * Constraints noch gewartet werden muss
   * <ul>
   * <li>Falls der Prozess sofort gestartet werden kann wird der Prozess im
   * ExecutionManager zur Ausfuehrung freigegeben</li>
   * <li>Andernfalls wird entweder ein entsprechender Timer initialisiert, der
   * bei Ablauf die Ausfuehrung im Execution Manager freigibt (siehe unten),
   * oder auf das Auftreten eines entsprechenden (externen) Events gewartet.
   * (Trigger auf Event setzen)</li>
   * <li>Moegliche Ursachen fuer Verzoegerungen:
   * <ul>
   * <li><u>TP3</u> vom ProcessInstantiation-Event (sinnvoll?) oder von einem
   * externen Event zum ProcessStart-Event</li>
   * <li><u>Fixed Date</u> auf ProcessStart (Wert aus Prozess-Eingabeparameter
   * oder externen Quelle (--&gt; zusaetzlich vorgeschaltetes Update-Event))</li>
   * <li><u>Schedule Restricted Element</u> auf ProcessStart</li>
   * </ul>
   * </li>
   * <li>Nach Freigabe des Prozesses wird dieser damit als "Running"
   * betrachtet und kann vom ExecutionManager ausgefuehrt werden.</li>
   * <li>-&gt; Abstand zwischen ProcessInstantiation und ProcessStart ~0,
   * falls dies von keinen Constraint direkt beeinflusst wird</li>
   * </ul>
   * </li>
   * <li>Nun wird ueberprueft, ob eine der <u>Validity Periods</u> eine Gefahr
   * fuer die Prozessausfuehrung darstellt, d.h. ob ihr Wert an den berechneten
   * Wert des entsprechenden Events angrenzt bzw. diesen sogar beschraenkt.
   * <ul>
   * <li>Ist dies nicht der Fall wird die Validity Period im weiteren Verlauf
   * ignoriert und koennte aus dem Zeitmodell geloescht werden, falls dies die
   * Zeitrechnung beschleunigt (Achtung muesste ggf. beim
   * auftreten/"korrigieren" von Zeitfehlern ueberprueft / neu eingefuegt werden)
   * </li>
   * </ul>
   * </li>
   * <li>Parallel findet die Initialisierung des Laufzeitmodells statt
   * <ul>
   * <li>Das Laufzeitmodell besteht aus dem im ersten Schritt berechneten
   * Zeitmodell und vier Event-Mengen:
   * <ul>
   * <li><i>RecentHistory</i>: Die entlang des Kontrollflusses /
   * Abhaengigkeitsgraphen zuletzt aufgetretenen Events. Diese beschreiben den
   * aktuellen zeitlichen Zustand der Prozess-Instanz.</li>
   * <li><i>Pending</i>: Die Events deren Kontrollfluss-Vorbedingungen erfuellt
   * sind, d.h. die Events die nach Kontrollfluss als naechstes auftreten
   * koennten. (-&gt; Unabhaengig von Zeit-Constraints)</li>
   * <li><i>ActiveEvents</i>: Die Menge der Events, auf die sich ein aktives
   * Constraint bezieht. Das Ausfuehrungsintervall dieser Events ist meist
   * bereits stark eingeschraenkt und besitzt weniger aus dem Kontrollfluss
   * resultierende Varianz.
   * <ul>
   * <li>Dient zur Abschaetzung der Konsistenz, da diese Events grossen Einfluss
   * hierauf haben.</li>
   * <li>Nicht notwendigerweise disjunkt zu den anderen Mengen</li>
   * </ul>
   * </li>
   * <li><i>Horizon</i>: In "naher" Zukunft anstehende Events, die aufgrund
   * des aktuellen Vorschau-Horizonts bestimmt werden und fuer die Vorhersage
   * von Zeitdaten relevant sind.</li>
   * </ul>
   * </li>
   * <li>Zunaechst wird das ProcessInstantiation-Event zusammen mit dem
   * Zeitstempel in die Event-Menge <i>RecentHistory</i> eingefuegt</li>
   * <li>Dann werden zur Ausfuehrung anstehende Events in der Menge
   * <i>Pending</i> gespeichert (d.h. ProcessStart-Event und eventuelle
   * externe ProcessIntermediate-Events die kein eingehendes Constraint haben
   * bzw. nur eines zum ProzessInstantiation-Event)</li>
   * <li>Saemtliche <u>Fixed Date Elements</u> (TP4) deren Wert bereits bekannt
   * ist werden der Menge <i>ActiveEvents</i> hinzugefuegt.
   * 
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor anhand des
   * Constraints angepasst)</li>
   * <li>Moeglich: Exception-Timer starten, da Zeitmodell nicht immer
   * aktualisiert wird und damit Konsistenz nicht vollstaendig garantiert
   * werden kann <span> </span>(siehe auch weiter unten) <span
   * style="font-size: 11px;">&lt;-- Nur Events in Pending koennen direkt
   * beeinflusst werden. Alle anderen Events koennen nur indirekt ueber Pending
   * beeinflusst werden.</span></li>
   * </ul>
   * </li>
   * <li>Fuer saemtliche vom ProcessInstantiation-Event ausgehenden Constraints
   * (<u>TP3</u>) wird das Ziel-Event der <i>ActiveEvents</i> Menge
   * hinzugefuegt.
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor angepasst)</li>
   * <li>Moeglich: Exception-Timer starten</li>
   * </ul>
   * </li>
   * <li>Fuer jedes <u>Schedule Restricted Element</u> wird der naechste
   * Zeitslot berechnet, der innerhalb des "Ausfuehrungs"-Intervalls des Events
   * liegt (Wert bereits aus Instanziierung des Zeitmodells bekannt / kann
   * gecached werden). Fuer das Ende des Slots wird ein Update-Timer
   * registriert. Entsprechend wird das Event ebenfalls der
   * <i>ActiveEvents</i> Menge hinzugefuegt.
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor angepasst)</li>
   * <li>Hier sind Abstufungen fuer den Update-Timer moeglich, je nachdem wie
   * gross der Zwischenraum zum darauffolgenden Slot im Verhaeltnis zur Varianz
   * des Events ist (d.h. zur Laenge des "Ausfuehrungs"-Intervalls). Ist der
   * Zwischenraum klein im Verhaeltnis zur Varianz koennen weitere Slots
   * hinzugenommen werden bis ein grosser Zwischenraum auftritt oder eine
   * bestimmte (?) Zeitspanne ueberschritten ist und dann erst fuer diesen
   * Zeitpunkt ein Update Timer gesetzt werden.</li>
   * <li>Ebenfalls sind Abstufung bei der Entscheidung, ob das Event den
   * <i>ActiveEvents</i> hinzugefuegt wird, moeglich. Je nachdem wie viel
   * Varianz das Element hat und wie sich Slots und Zwischenraeume verhalten,
   * kann das Event zunaechst aus der <i>ActiveEvents</i> menge weggelassen
   * werden.
   * <ul>
   * <li>Was triggert dann schliesslich das hinzufuegen des Events zu
   * <i>ActiveEvents</i>? (-&gt; Update-Trigger?)</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>[Moegliche Erweiterung: Auch alle Constraints, die von einem Pending
   * Event ausgehen werden ausgewertet, d.h. <i>ActiveEvents</i> hinzugefuegt.]
   * </li>
   * <li>Auf Basis von <i>Pending</i>, <i>ActiveEvents</i> und des aktuellen
   * Horizonts wird die Event-Menge <i>Horizon</i> bestimmt (siehe Notizen zu
   * Sliding-Horizon`).
   * <ul>
   * <li>Verschiedene Strategien fuer den Horizont denkbar: Zeit, Anzahl,
   * Kombination (min/max)</li>
   * <li>Konsequenz: Desto mehr Constraint bereits betrachtet werden muessen
   * desto weniger Vorausschau ist moeglich
   * <ul>
   * <li>Zeitrechnung wird komplexer und ungenauer</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>Alle Events aus <i>ActiveEvents</i>, die in keiner der anderen Mengen
   * sind, werden mit ihren direkten Vorgaengern (nach Kontrollfluss) aus der
   * Vereinigungsmenge von <i>Pending</i>, <i>Horizon</i> und
   * <i>ActiveEvents</i> verbunden, in dem "neue" Constraints gebildet werden,
   * die den entsprechenden Abstand aus dem Zeitmodell entlang des
   * Kontrollflusses widerspiegeln.</li>
   * <li>Auf Basis der Event-Mengen <i>Horizon</i>, <i>ActiveEvents</i>,
   * <i>Pending</i> und <i>RecentHistory</i> wird nun eine erneute
   * Zeitrechnung durchgefuehrt.
   * 
   * <ul>
   * <li>d.h. es wird der von diesen Events bzw. Event-Instanzen aufgespannte
   * Sub-Graph des Constraint-Networks betrachtet.</li>
   * <li>Hierbei wird zusaetzlich ein Free/Bounded-Schedule berechnet
   * (Bettini02) der die Werte fuer die Prediction der Zeitdaten liefert</li>
   * </ul>
   * </li>
   * <li>Fuer alle Events aus <i>Pending</i> werden entsprechende Timer
   * gestartet:
   * <ul>
   * <li>Falls das ProcessStart-Event verzoegert werden muss wird es im
   * ExecutionManager blockiert und ein entsprechender Timer gestartet, der
   * das Event zum entsprechenden Zeitpunkt freigibt (siehe oben)</li>
   * <li>Falls das ProcessStart-Event bis zu einem gewissen Zeitpunkt
   * aufgetreten sein muss wird ein entsprechender Exception-Timer gestartet</li>
   * <li>Der frueheste Zeitpunkt fuer ProcessIntermediate-Events kann nicht
   * kontrolliert werden, es kann nur auf einen Fehler reagiert werden.</li>
   * <li>Falls ein ProcessIntermediate-Event bis zu einem gewissen Zeitpunkt
   * aufgetreten sein muss wird ein entsprechender Exception-Timer gestartet</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * @param parentSession 
   * @param instanceID  
   * @param status
   */
  private void initializeInstanceExecution(SessionToken parentSession, UUID instanceID, ChangeableTimeModelStatus status)
  {
    SessionToken session = subSession(parentSession);
    
    
    // Check if Updates are still enabled
    if (status.getTimeModelUpdateStatus().ordinal() 
        < TimeModelUpdateStatus.NO_UPDATE.ordinal())
    {
      
      // Get the instance time Model as readonly
      InstanceTimeModel timeModel;
      try
      {
        timeModel = this.instanceTimeManager.getInstanceTimeModelReadonly(session, instanceID);
      }
      catch (DataSourceException e)
      {
        this.logger.log(Level.SEVERE, "DataSourceException", e);
        return;
      }
      
      // Handle an earliest start time which is in the future, i.e., block the
      // execution of the process instance
      handleInstanceStartTime(timeModel);
      
      // Check all Validity Periods (Activity and Process) if there may become 
      // a problem. If so schedule a respective timer for escalation handling
      evaluateAndHandleValidityPeriod(timeModel);
      
      // Initialize the Runtime Model
      initializeRuntimeModel(timeModel);
      // TODO AL todo Initilize Escalation Management
      
    }
  }

  /**
   * Parallel findet die Initialisierung des Laufzeitmodells statt
   * <ul>
   * <li>Das Laufzeitmodell besteht aus dem im ersten Schritt berechneten
   * Zeitmodell und vier Event-Mengen:
   * <ul>
   * <li><i>RecentHistory</i>: Die entlang des Kontrollflusses /
   * Abhaengigkeitsgraphen zuletzt aufgetretenen Events. Diese beschreiben den
   * aktuellen zeitlichen Zustand der Prozess-Instanz.</li>
   * <li><i>Pending</i>: Die Events deren Kontrollfluss-Vorbedingungen erfuellt
   * sind, d.h. die Events die nach Kontrollfluss als naechstes auftreten
   * koennten. (-&gt; Unabhaengig von Zeit-Constraints)</li>
   * <li><i>ActiveEvents</i>: Die Menge der Events, auf die sich ein aktives
   * Constraint bezieht. Das Ausfuehrungsintervall dieser Events ist meist
   * bereits stark eingeschraenkt und besitzt weniger aus dem Kontrollfluss
   * resultierende Varianz.
   * <ul>
   * <li>Dient zur Abschaetzung der Konsistenz, da diese Events grossen Einfluss
   * hierauf haben.</li>
   * <li>Nicht notwendigerweise disjunkt zu den anderen Mengen</li>
   * </ul>
   * </li>
   * <li><i>Horizon</i>: In "naher" Zukunft anstehende Events, die aufgrund des
   * aktuellen Vorschau-Horizonts bestimmt werden und fuer die Vorhersage von
   * Zeitdaten relevant sind.</li>
   * </ul>
   * </li>
   * <li>Zunaechst wird das ProcessInstantiation-Event zusammen mit dem
   * Zeitstempel in die Event-Menge <i>RecentHistory</i> eingefuegt</li>
   * <li>Dann werden zur Ausfuehrung anstehende Events in der Menge
   * <i>Pending</i> gespeichert (d.h. ProcessStart-Event und eventuelle externe
   * ProcessIntermediate-Events die kein eingehendes Constraint haben bzw. nur
   * eines zum ProzessInstantiation-Event)</li>
   * <li>Saemtliche <u>Fixed Date Elements</u> (TP4) deren Wert bereits bekannt
   * ist werden der Menge <i>ActiveEvents</i> hinzugefuegt.
   * 
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor anhand des
   * Constraints angepasst)</li>
   * <li>Moeglich: Exception-Timer starten, da Zeitmodell nicht immer
   * aktualisiert wird und damit Konsistenz nicht vollstaendig garantiert werden
   * kann <span> </span>(siehe auch weiter unten) <span
   * style="font-size: 11px;">&lt;-- Nur Events in Pending koennen direkt
   * beeinflusst werden. Alle anderen Events koennen nur indirekt ueber Pending
   * beeinflusst werden.</span></li>
   * </ul>
   * </li>
   * <li>Fuer saemtliche vom ProcessInstantiation-Event ausgehenden Constraints
   * (<u>TP3</u>) wird das Ziel-Event der <i>ActiveEvents</i> Menge hinzugefuegt.
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor angepasst)</li>
   * <li>Moeglich: Exception-Timer starten</li>
   * </ul>
   * </li>
   * <li>Fuer jedes <u>Schedule Restricted Element</u> wird der naechste Zeitslot
   * berechnet, der innerhalb des "Ausfuehrungs"-Intervalls des Events liegt
   * (Wert bereits aus Instanziierung des Zeitmodells bekannt / kann gecached
   * werden). Fuer das Ende des Slots wird ein Update-Timer registriert.
   * Entsprechend wird das Event ebenfalls der <i>ActiveEvents</i> Menge
   * hinzugefuegt.
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor angepasst)</li>
   * <li>Hier sind Abstufungen fuer den Update-Timer moeglich, je nachdem wie gross
   * der Zwischenraum zum darauffolgenden Slot im Verhaeltnis zur Varianz des
   * Events ist (d.h. zur Laenge des "Ausfuehrungs"-Intervalls). Ist der
   * Zwischenraum klein im Verhaeltnis zur Varianz koennen weitere Slots
   * hinzugenommen werden bis ein grosser Zwischenraum auftritt oder eine
   * bestimmte (?) Zeitspanne ueberschritten ist und dann erst fuer diesen
   * Zeitpunkt ein Update Timer gesetzt werden.</li>
   * <li>Ebenfalls sind Abstufung bei der Entscheidung, ob das Event den
   * <i>ActiveEvents</i> hinzugefuegt wird, moeglich. Je nachdem wie viel Varianz
   * das Element hat und wie sich Slots und Zwischenraeume verhalten, kann das
   * Event zunaechst aus der <i>ActiveEvents</i> menge weggelassen werden.
   * <ul>
   * <li>Was triggert dann schliesslich das hinzufuegen des Events zu
   * <i>ActiveEvents</i>? (-&gt; Update-Trigger?)</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>[Moegliche Erweiterung: Auch alle Constraints, die von einem Pending
   * Event ausgehen werden ausgewertet, d.h. <i>ActiveEvents</i> hinzugefuegt.]</li>
   * <li>Auf Basis von <i>Pending</i>, <i>ActiveEvents</i> und des aktuellen
   * Horizonts wird die Event-Menge <i>Horizon</i> bestimmt (siehe Notizen zu
   * Sliding-Horizon`).
   * <ul>
   * <li>Verschiedene Strategien fuer den Horizont denkbar: Zeit, Anzahl,
   * Kombination (min/max)</li>
   * <li>Konsequenz: Desto mehr Constraint bereits betrachtet werden muessen
   * desto weniger Vorausschau ist moeglich
   * <ul>
   * <li>Zeitrechnung wird komplexer und ungenauer</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>Alle Events aus <i>ActiveEvents</i>, die in keiner der anderen Mengen
   * sind, werden mit ihren direkten Vorgaengern (nach Kontrollfluss) aus der
   * Vereinigungsmenge von <i>Pending</i>, <i>Horizon</i> und
   * <i>ActiveEvents</i> verbunden, in dem "neue" Constraints gebildet werden,
   * die den entsprechenden Abstand aus dem Zeitmodell entlang des
   * Kontrollflusses widerspiegeln.</li>
   * <li>Auf Basis der Event-Mengen <i>Horizon</i>, <i>ActiveEvents</i>,
   * <i>Pending</i> und <i>RecentHistory</i> wird nun eine erneute Zeitrechnung
   * durchgefuehrt.
   * 
   * <ul>
   * <li>d.h. es wird der von diesen Events bzw. Event-Instanzen aufgespannte
   * Sub-Graph des Constraint-Networks betrachtet.</li>
   * <li>Hierbei wird zusaetzlich ein Free/Bounded-Schedule berechnet (Bettini02)
   * der die Werte fuer die Prediction der Zeitdaten liefert</li>
   * </ul>
   * </li>
   * <li>Fuer alle Events aus <i>Pending</i> werden entsprechende Timer
   * gestartet:
   * <ul>
   * <li>Falls das ProcessStart-Event verzoegert werden muss wird es im
   * ExecutionManager blockiert und ein entsprechender Timer gestartet, der das
   * Event zum entsprechenden Zeitpunkt freigibt (siehe oben)</li>
   * <li>Falls das ProcessStart-Event bis zu einem gewissen Zeitpunkt
   * aufgetreten sein muss wird ein entsprechender Exception-Timer gestartet</li>
   * <li>Der frueheste Zeitpunkt fuer ProcessIntermediate-Events kann nicht
   * kontrolliert werden, es kann nur auf einen Fehler reagiert werden.</li>
   * <li>Falls ein ProcessIntermediate-Event bis zu einem gewissen Zeitpunkt
   * aufgetreten sein muss wird ein entsprechender Exception-Timer gestartet</li>
   * </ul>
   * </li>
   * </ul>
   */
  private void initializeRuntimeModel(InstanceTimeModel timeModel)
  {

  }

  /**
   * Nun wird ueberprueft, ob eine der <u>Validity Periods</u> eine Gefahr fuer die
   * Prozessausfuehrung darstellt, d.h. ob ihr Wert an den berechneten Wert des
   * entsprechenden Events angrenzt bzw. diesen sogar beschraenkt.
   * <ul>
   * <li>Ist dies nicht der Fall wird die Validity Period im weiteren Verlauf
   * ignoriert und koennte aus dem Zeitmodell geloescht werden, falls dies die
   * Zeitrechnung beschleunigt (Achtung muesste ggf. beim auftreten/"korrigieren"
   * von Zeitfehlern ueberprueft / neu eingefuegt werden)</li>
   * </ul>
   * 
   * @param timeModel
   */
  private void evaluateAndHandleValidityPeriod(InstanceTimeModel timeModel)
  {
    // TODO AL for now schedule each and every Validity Period
  }



//  /** 
//   * <li><u>Time-dependent Variability</u> so weit
//   * moeglich auswerten</li>
//   */
//  private void evaluateTimeDependentVariability(SessionToken session,
//      UpdateableInstanceTimeModel timeModel)
//  {
//    // TODO Auto-generated method stub
//    
//  }

//  /**
//   * 
//   * <li>Auf Basis der aktuellen Daten des Zeitmodell wird fuer jedes Schedule
//   * Restricted Element der erste und letzte im entsprechenden
//   * Ausfuehrungs-Intervall des Events moegliche Slot ermittelt und falls noetig
//   * das Ausfuehrungs-Intervall entsprechend eingeschraenkt
//   * <ul>
//   * <li>Falls kein Slot verfuegbar ist wir der Start des Prozesses abgebrochen</li>
//   * </ul>
//   * </li>
//   * 
//   * @param session
//   * @param timeModel
//   */
//  private void initScheduleRestrictedElements(SessionToken session,
//      UpdateableInstanceTimeModel timeModel)
//  {
//    // TODO Auto-generated method stub
//    
//  }

  /**
   * <li>Es werden alle Constraints aktualisiert, die das
   * ProzessInstantiation-Event als Update-Trigger haben (d.h. ihre Daten aus
   * den Prozesseingabe-Parametern bzw. einer Konfiguration beziehen)<br/>
   * <ul>
   * <li>Die Constraints und Ausfuehrungsintervalle werden im Zeitmodell
   * entsprechend angepasst</li>
   * </ul>
   * </li>
   * 
   * @param session
   * @param timeModel
   * @param processInstantiationEvent
   */
  private void updateConstraints(SessionToken session,
      UpdateableInstanceTimeModel timeModel, int processInstantiationEvent)
  {
    // XXX Future use
    
  }


  /////////////////////////////////////////////
  // END Instance Created
  /////////////////////////////////////////////
  
  /////////////////////////////////////////////
  // BEGIN Instance Started
  /////////////////////////////////////////////
  
  public void instanceStarted(SessionToken session, UUID instanceID,
      long timestamp)
  {
    System.out.println(String.format("Instance (%1$s) Started @ %2$d -> Updating Time Model", instanceID, Long.valueOf(timestamp)));
    //logger.entering(logger.getName(), "instanceStarted(SessionToken,UUID,long)");

    try
    {
      ChangeableTimeModelStatus status = this.instanceTimeManager
        .getAndLockTimeModelStatus(session, instanceID);
      System.out.println(String.format("Instance locked (%1$s)", instanceID, Long.valueOf(timestamp)));
      try
      {
    
        // Check if the time model is disable because of an invalid time model
        if (status.getTimeModelUpdateStatus().ordinal() 
            < TimeModelUpdateStatus.DISABLED.ordinal())
        {
          SessionToken childSession = subSession(session);
          try
          {
            UpdateableInstanceTimeModel timeModel = this.instanceTimeManager
              .getAndLockTimeModel(childSession, instanceID);
      
            try
            {
              // Set Instance Start Time
              timeModel.setInstanceStartTime(timestamp);
              this.instanceTimeManager.updateTimeModel(childSession, timeModel);
              
              // Check if calculations are disable because of an invalid time model      
              if (status.getTimeModelUpdateStatus().ordinal() 
                  < TimeModelUpdateStatus.NO_CALCULATION.ordinal())
              {
                System.out.println(String.format("Updating Time Model of Instance (%1$s) @ %2$d", instanceID, Long.valueOf(timestamp)));
                  boolean stateChanged = TimeManagerTools.updateTimeModel(session, timeModelFactory, timeModel, status, this.temporalExceptionHandling);
                  if (stateChanged)
                  {
                    this.instanceTimeManager.updateTimeModelStatus(session, status);
                  //TODO AL: determine changes and update triggers!
                  }
                  this.instanceTimeManager.updateTimeModel(childSession, timeModel);
              }
              handleValidityPeriodAtProcessStart(childSession, instanceID, timestamp, status);
            }
            finally
            {
              this.instanceTimeManager.unlockTimeModel(childSession, instanceID);
            }
            
          }
          catch (LockException e)
          {
            logger.log(Level.SEVERE, "Lock Exception", e);
          }
          catch (DataSourceException e)
          {
            logger.log(Level.SEVERE, "Data Source Exception", e);
          }
        }
      }
      finally 
      {
        this.instanceTimeManager.unlockTimeModelStatus(session, instanceID);
      }

      if (status.getTimeModelUpdateStatus().ordinal() <= TimeModelUpdateStatus.NO_UPDATE.ordinal())
      {
        SessionToken childSession = subSession(session);
        handleProcessDuration(childSession, instanceID, timestamp);
        
      }
      
      WorklistAccess notification = timeManager.getWorklistAccess();
      
      notification.timeModelUpdated(session, instanceID);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "Lock Exception", e);
      throw new RuntimeException(e);
    }
   
    //logger.exiting(logger.getName(), "instanceStarted(SessionToken,UUID,long)");
  }

  
  private void handleProcessDuration(SessionToken session, UUID instanceID,
      long timestamp)
  {
    // TODO AL create trigger for process duration if necessary
//    ProcessDuration duration = getProcessTemplateDuration(session, instanceID);
//    if (duration != null)
//    {
//      escalationManager.scheduleProcessDuration(session, instanceID, timestamp, duration);
//    }
  }
 
  /**
   * Zunaechst wird ueberprueft, ob der Prozess sofort gestartet werden kann, d.h.
   * das ProcessStart-Event sofort auftreten darf, oder aufgrund eines
   * Constraints noch gewartet werden muss
   * <ul>
   * <li>Falls der Prozess sofort gestartet werden kann wird der Prozess im
   * ExecutionManager zur Ausfuehrung freigegeben</li>
   * <li>Andernfalls wird entweder ein entsprechender Timer initialisiert, der
   * bei Ablauf die Ausfuehrung im Execution Manager freigibt (siehe unten), oder
   * auf das Auftreten eines entsprechenden (externen) Events gewartet. (Trigger
   * auf Event setzen)</li>
   * <li>Moegliche Ursachen fuer Verzoegerungen:
   * <ul>
   * <li><u>TP3</u> vom ProcessInstantiation-Event (sinnvoll?) oder von einem
   * externen Event zum ProcessStart-Event</li>
   * <li><u>Fixed Date</u> auf ProcessStart (Wert aus Prozess-Eingabeparameter
   * oder externen Quelle (--&gt; zusaetzlich vorgeschaltetes Update-Event))</li>
   * <li><u>Schedule Restricted Element</u> auf ProcessStart</li>
   * </ul>
   * </li>
   * <li>Nach Freigabe des Prozesses wird dieser damit als "Running" betrachtet
   * und kann vom ExecutionManager ausgefuehrt werden.</li>
   * <li>-&gt; Abstand zwischen ProcessInstantiation und ProcessStart ~0, falls
   * dies von keinen Constraint direkt beeinflusst wird</li>
   * </ul>
   * 
   * @param timeModel
   */
  private void handleInstanceStartTime(InstanceTimeModel timeModel)
  {
    // FIXME
    Interval<CalendarElement> startTime = timeModel.getInstanceStartTimeframe();
    
  }

  private void handleValidityPeriodAtProcessStart(SessionToken session,
      UUID instanceID, long creationTimeStamp, ChangeableTimeModelStatus currentStatus)
      throws LockException, DataSourceException
  {
    // Check if the start of the template has a validity period
    InstanceTimeModel instanceTimeModel = this.instanceTimeManager.getInstanceTimeModelReadonly(session, instanceID);
    UUID templateID = instanceTimeModel.getBaseTemplateID();  
    TemporalTemplate template = this.templateTimeManager.getTemplate(session, templateID);
    // Check if the start of the template has a validity period
    ValidityPeriod validityPeriod = template.getStartValidityPeriod();
    if (validityPeriod != null)
    {
      // Check if the current time stamp is after the start of the validity period
      if (timeModelTools.compare(validityPeriod.getValidFrom(), creationTimeStamp) > 0)
      {
        String msg = String.format("Instance '%1$s' not valid until %2$s!",
                                   instanceID, validityPeriod
                                     .getValidFrom());
        logger.warning(msg);
        // TODO AL: handle error
        
        // Fast: disable time calculations for instance
        TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
        this.instanceTimeManager.updateTimeModelStatus(session, currentStatus);
      }

      // Check if the current time stamp is before the end of the validity period
      if (timeModelTools.compare(validityPeriod.getValidUntil(), creationTimeStamp) < 0)
      {
        String msg = String
          .format("Instance '%1$s' only valid until %2$s!", instanceID,
                  validityPeriod.getValidUntil());
        logger.warning(msg);
        // TODO AL: handle error
        // Fast: disable time calculations for instance
        TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
        this.instanceTimeManager.updateTimeModelStatus(session, currentStatus);
      }
    }
    
    // Check if the completion of the template has a validity period
    validityPeriod =  template.getCompletionValidityPeriod();
    if (validityPeriod != null)
    {
      Interval<TimeDistance> processDuration = getProcessInstanceDuration(session, instanceID);
      //Check if the instance has a duration constraint or duration prediction
      if (processDuration != null)
      {
        CalendarElement earliestProcessEnd = timeModelTools.add(creationTimeStamp, processDuration.getMinValue());
        // Check if the predicted earliest process end is after the start of the validity period
        if (timeModelTools.compare(validityPeriod.getValidFrom(), earliestProcessEnd) > 0)
        {
          String msg = String.format("Template '%1$s' may not be completed before %2$s!",
                                     instanceID, validityPeriod
                                       .getValidFrom());
          logger.warning(msg);
          // TODO AL: handle error
          
          // Fast: disable time calculations for instance
          TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
          this.instanceTimeManager.updateTimeModelStatus(session, currentStatus);
        }
        
        CalendarElement latestProcessEnd = timeModelTools.add(creationTimeStamp, processDuration.getMaxValue());
        
        // Check if the predicted latest completion is before the end of the validity period
        if (timeModelTools.compare(validityPeriod.getValidUntil(), latestProcessEnd) < 0)
        {
          String msg = String
            .format("Template '%1$s' may not be completed after %2$s!", instanceID,
                    validityPeriod.getValidUntil());
          logger.warning(msg);
          // TODO AL: handle error
          // Fast: disable time calculations for instance
          TimeManagerTools.downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
          this.instanceTimeManager.updateTimeModelStatus(session, currentStatus);
        }
      }
    }
  }
  
  /////////////////////////////////////////////
  // END Instance Started
  /////////////////////////////////////////////
  
  /////////////////////////////////////////////
  // BEGIN Helper Methods
  /////////////////////////////////////////////

  
  private Interval<TimeDistance> getProcessInstanceDuration(SessionToken session, UUID instanceID) throws DataSourceException
  {
    InstanceTimeModel timeModel = this.instanceTimeManager.getInstanceTimeModelReadonly(session, instanceID);
    return timeModel.getInstanceDuration();
  }

//  private ValidityPeriod getInstanceValidityPeriod(SessionToken session,
//      UUID instanceID, ReferencePoint referencePoint)
//  {
//    InstanceTimeModel instanceTimeModel = this.instanceTimeManager.getInstanceTimeModelReadonly(session, instanceID);
//    UUID templateID = instanceTimeModel.getBaseTemplateID();    
//    TemplateTimeModel timeModel = this.templateTimeManager.getTimeModel(session, templateID);
//    
//    return this.instanceTimeManager.getValidityPeriod(session, instanceID,
//                                                      referencePoint);
//  }





  
}
