package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import java.util.UUID;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;

import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultInstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.InstanceStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.InstanceStateUpdateManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.InstanceTermination;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeContext;
import de.aristaflow.adept2.util.Adept2ThreadFactory;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;

/**
 * @author  lanz
 */
public class DefaultInstanceStateUpdateManager extends AbstractSubService
    implements InstanceStateUpdateManager
{

  /**
   * @uml.property  name="instanceStarting"
   * @uml.associationEnd  
   */
  private InstanceStarting instanceStarting;

  /**
   * @uml.property  name="instanceTermination"
   * @uml.associationEnd  
   */
  private InstanceTermination instanceTermination;

  /**
   * @uml.property  name="instanceTimeManager"
   * @uml.associationEnd  
   */
  private DefaultInstanceTimeManager instanceTimeManager;
  
  private ExecutorService executorService;

  public DefaultInstanceStateUpdateManager(DefaultTimeManager timeManager,
      DefaultInstanceTimeManager instanceTimeManager, InstanceStarting instanceStarting,
      InstanceTermination instanceTermination)
  {
    super(timeManager);
    this.instanceTimeManager = instanceTimeManager;
    this.instanceStarting = instanceStarting;
    this.instanceTermination = instanceTermination;
  }
  
  @Override
  public void init()
  {
    executorService = Executors
      .newCachedThreadPool(new Adept2ThreadFactory("InstanceStateUpdate", logger)); //$NON-NLS-1$
  }

  /** Awaits the termination of the executor service. */
  private void awaitExecutorServiceShutdown()
  {
    boolean terminated = false;
    while (!terminated)
    {
      try
      {
        terminated = executorService.awaitTermination(30, TimeUnit.SECONDS);
      }
      // ignore interruptions since we are shutting down
      catch (InterruptedException ie)
      { 
        //
      }
    }
  }

  @Override
  public void emergencyShutdown()
  {
    executorService.shutdown();
    awaitExecutorServiceShutdown();
    super.emergencyShutdown();
  }

  @Override
  public void shutdown()
  {
    executorService.shutdownNow();
    awaitExecutorServiceShutdown();
    super.shutdown();
  }
  


  public void instanceCreated(final SessionToken session, final UUID templateID,
      final UUID instanceID, final long timestamp)
  {
    try
    {
      // Create the time context
      UpdateableInstanceTimeContext timeContext = this.instanceTimeManager
        .createAndLockInstanceTimeContext(session, instanceID);
      try
      {
        // Update the time context
        timeContext.setInstanceCreated(timestamp);
        // Store the updated time context
        this.instanceTimeManager.updateInstanceTimeContext(session, timeContext);
      }
      finally
      {
        // Unlock the Time Context
        this.instanceTimeManager.unlockInstanceTimeContext(session, instanceID);
      }
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "LockException", e);
    }
    
    executorService.execute(new Runnable(){

      public void run()
      {
        // Call InstanceStarting to handle the time model etc.
        try
        {
          instanceStarting.instanceCreated(session, templateID, instanceID, timestamp);
        }
        catch (Exception e)
        {
          e.printStackTrace();
          logger.log(Level.SEVERE, "An exception occurred in instanceCreate.", e);
          
        }
      }
      
    });

  }

  public void instanceStarted(final SessionToken session, final UUID instanceID,
      final long timestamp)
  {
    try
    {
      // Read the Time Context
      UpdateableInstanceTimeContext timeContext = this.instanceTimeManager.getAndLockInstanceTimeContext(session, instanceID);
      try
      {
        // Update the time context
        timeContext.setInstanceStarted(timestamp);
        // Store the update time context
        this.instanceTimeManager.updateInstanceTimeContext(session, timeContext);
      }
      finally
      {
        // Unlock the Time Context
        this.instanceTimeManager.unlockInstanceTimeContext(session, instanceID);
      }
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "LockException", e);
    }
    
    
    executorService.execute(new Runnable(){

      public void run()
      {
        try
        {
          // Call InstanceStarting to handle the time model etc.
          instanceStarting.instanceStarted(session, instanceID, timestamp);
        }
        catch (Throwable e)
        {
          logger.log(Level.SEVERE, "An exception occurred in instanceStarted.", e);
        }
      }
    });
  }

  public void instanceFinished(final SessionToken session, final UUID instanceID,
      final long timestamp)
  {
    try
    {
      // Read the Time Context
      UpdateableInstanceTimeContext timeContext = this.instanceTimeManager.getAndLockInstanceTimeContext(session, instanceID);
      try
      {
        // Update the time context
        timeContext.setInstanceFinished(timestamp);
        // Store the update time context
        this.instanceTimeManager.updateInstanceTimeContext(session, timeContext);
      }
      finally
      {
        // Unlock the Time Context
        this.instanceTimeManager.unlockInstanceTimeContext(session, instanceID);
      }
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "LockException", e);
    }
    
    
    executorService.execute(new Runnable(){

      public void run()
      {
        try
        {
          // Call InstanceTermination to handle the time model etc.
          instanceTermination.instanceFinished(session, instanceID, timestamp);
        }
        catch (Throwable e)
        {
          logger.log(Level.SEVERE, "An exception occurred in instanceFinished.", e);
        }
      }
    });
  }

}
