package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import java.util.UUID;
import java.util.logging.Level;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.complexeventmanager.EventManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultInstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeEventNotification;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ActivityStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ActivityTermination;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.InstanceStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.InstanceTermination;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ProcessExecutionManager;
import de.aristaflow.adept2.model.eventmodel.EventReference;
import de.aristaflow.adept2.model.globals.ExecutionEventConstants;
import de.aristaflow.adept2.model.globals.ExecutionEventConstants.ActivityEventType;
import de.aristaflow.adept2.model.globals.ExecutionEventConstants.ProcessEventType;

/**
 * @author  lanz
 */
public class DefaultProcessExecutionManager extends AbstractSubService
    implements ProcessExecutionManager
{

  /**
   * @uml.property  name="processInstanceUpdateManager"
   * @uml.associationEnd  
   */
  private final DefaultInstanceStateUpdateManager processInstanceUpdateManager;

  /**
   * @uml.property  name="activityInstanceUpdateManager"
   * @uml.associationEnd  
   */
  private final DefaultActivityStateUpdateManager activityInstanceUpdateManager;

  /**
   * @uml.property  name="eventManager"
   * @uml.associationEnd  
   */
  private final EventManager eventManager;

  /**
   * @uml.property  name="eventNotification"
   * @uml.associationEnd  
   */
  private DefaultTimeEventNotification eventNotification;

  private DefaultTimeManager timeManager;

  public DefaultProcessExecutionManager(DefaultTimeManager timeManager,
      DefaultInstanceTimeManager instanceTimeManager,
      EventManager eventManager, InstanceStarting instanceStarting,
      InstanceTermination instanceTermination,
      ActivityStarting activityStarting, ActivityTermination activityTermination)
  {
    super(timeManager);
    this.timeManager = timeManager;
    this.eventManager = eventManager;

    this.processInstanceUpdateManager = new DefaultInstanceStateUpdateManager(timeManager,
                                                                              instanceTimeManager,
                                                                              instanceStarting,
                                                                              instanceTermination);
    this.activityInstanceUpdateManager = new DefaultActivityStateUpdateManager(timeManager,
                                                                               instanceTimeManager,
                                                                               activityStarting,
                                                                               activityTermination);
  }

  @Override
  public void init() throws AbortServiceException
  {
    this.processInstanceUpdateManager.init();
    this.activityInstanceUpdateManager.init();
    
    this.eventNotification = new DefaultTimeEventNotification(
                                                              this,
                                                              eventManager,
                                                              ExecutionEventConstants.PROCESS_INSTANCE_EVENT_CLASS_ID,
                                                              ExecutionEventConstants.ACTIVITY_INSTANCE_EVENT_CLASS_ID);
    

  }

  @Override
  public void emergencyShutdown()
  {
    this.eventManager.getEventSinkManager()
      .deregisterEventListener(this.eventNotification);
    
    this.processInstanceUpdateManager.emergencyShutdown();
    this.activityInstanceUpdateManager.emergencyShutdown();
    super.emergencyShutdown();
  }

  @Override
  public void shutdown()
  {
    this.eventManager.getEventSinkManager()
      .deregisterEventListener(this.eventNotification);
    this.processInstanceUpdateManager.shutdown();
    this.activityInstanceUpdateManager.shutdown();
    super.shutdown();
  }

  @Override
  public void start() throws AbortServiceException
  {
    super.start();
    
    this.processInstanceUpdateManager.start();
    this.activityInstanceUpdateManager.start();
    
    // Go Live by registering with the Event Manager
    this.eventManager.getEventSinkManager()
      .registerEventListener(this.eventNotification);
  }

  public void processEvent(EventReference event)
  {
//    logger.entering(logger.getName(), "processEvent(EventReference)", event);
    try
    {
      // Decide what to do based on event class.
      if (eventManager.getEventClassManager()
        .isSubClassOf(event.getEventClassID(),
                      ExecutionEventConstants.ACTIVITY_INSTANCE_EVENT_CLASS_ID))
      {
        processActivityEvent(event);
      }
      if (eventManager.getEventClassManager()
        .isSubClassOf(event.getEventClassID(),
                      ExecutionEventConstants.PROCESS_INSTANCE_EVENT_CLASS_ID))
      {
        processProcessEvent(event);
      }
      if (eventManager.getEventClassManager()
        .isSubClassOf(event.getEventClassID(),
                      ExecutionEventConstants.SUBPROCESS_INSTANCE_EVENT_CLASS_ID))
      {
        processSubProcessEvent(event);
      }
    }
    catch (Throwable t)
    {
      logger.log(Level.SEVERE, "Exception while processing event!", t);
    }
//    logger.exiting(logger.getName(), "processEvent(EventReference)", event);
  }



  protected void processActivityEvent(EventReference event)
  {
    SessionToken session = createSessionToken();
    ActivityEventType eventType = (ActivityEventType) event
      .getAttribute(ExecutionEventConstants.ATTR_ACTIVITY_EVENT_TYPE);
    UUID instanceID = (UUID) event
      .getAttribute(ExecutionEventConstants.ATTR_INSTANCE_ID);
    int nodeID = ((Integer) event
      .getAttribute(ExecutionEventConstants.ATTR_NODE_ID)).intValue();
    int nodeIteration = ((Integer) event
      .getAttribute(ExecutionEventConstants.ATTR_NODE_ITERATION)).intValue();

    switch (eventType)
    {
      case STARTED:
        this.activityInstanceUpdateManager.activityStarted(childSession(session), instanceID, nodeID, nodeIteration,
                                                           event.getTimeStamp());
        break;
      case FINISHED:
        this.activityInstanceUpdateManager.activityFinished(childSession(session), instanceID, nodeID, nodeIteration,
                                                            event.getTimeStamp());
        break;
      case ACTIVATED:
      case FAILED:
      case RESET:
      case RESUMED:
      case SUSPENDED:
        logger.fine(String
          .format("Received Event of Type '%1$s'. Ignoring it!", eventType));
        break;
      default:
        this.logger.severe(String.format("Unkown Process Event Type: %1$s",
                                         eventType));
    }
  }



  protected void processProcessEvent(EventReference event)
  {
//    logger.entering(logger.getName(), "processProcessEvent(EventReference)",
//                    event);
    SessionToken session = createSessionToken();
    ProcessEventType eventType = (ProcessEventType) event
      .getAttribute(ExecutionEventConstants.ATTR_PROCESS_EVENT_TYPE);
    UUID instanceID = (UUID) event
      .getAttribute(ExecutionEventConstants.ATTR_INSTANCE_ID);
    UUID templateID = (UUID) event
      .getAttribute(ExecutionEventConstants.ATTR_TEMPLATE_ID);

    switch (eventType)
    {
      case STARTED:
        this.processInstanceUpdateManager.instanceCreated(childSession(session), 
                                templateID, instanceID, event.getTimeStamp()-1000); // FIXME
        this.processInstanceUpdateManager.instanceStarted(childSession(session), 
                                instanceID, event.getTimeStamp());
        break;
      case FINISHED:
        this.processInstanceUpdateManager.instanceFinished(childSession(session), 
                                instanceID, event.getTimeStamp());
        break;
      default:
        this.logger.severe("Unkown Process Event Type: " + eventType);
    }
//    logger.exiting(logger.getName(), "processProcessEvent(EventReference)",
//                   event);
  }

  private SessionToken childSession(SessionToken session)
  {
    return timeManager.subSession(session);
  }
  
  private SessionToken createSessionToken()
  {
    
    return timeManager.createSessionToken();
  }
  
  protected void processSubProcessEvent(EventReference event)
  {
    // TODO Auto-generated method stub
    
  }
}
