package de.aristaflow.adept2.core.timemanager.defaultimplementation.schedules;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.aristaflow.adept2.base.service.InternalServiceException;
import de.aristaflow.adept2.core.timemanager.ScheduleEvaluation;
import de.aristaflow.adept2.core.timemanager.ScheduleManager;
import de.aristaflow.adept2.core.timemanager.ScheduleResolutionException;
import de.aristaflow.adept2.core.timemanager.storage.TimeManagerStorage;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultCalendarElement;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation.ComplexIntervalSet;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation.ContinuousIntervalSet;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation.Evaluator;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation.Interval;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation.IntervalSet;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation.TimeFrame;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.ScheduleParser;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.CalendarExpression;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.Ident;
import de.aristaflow.adept2.model.timemodel.Schedule;
import de.aristaflow.adept2.model.timemodel.ScheduleInterval;
import de.aristaflow.adept2.model.timemodel.ScheduleIntervalSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultSchedule;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultScheduleInterval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultScheduleIntervalSet;
import de.aristaflow.adept2.util.DataSourceException;

public class DefaultScheduleEvaluation implements ScheduleEvaluation, ScheduleManager
{
  
  protected Evaluator evaluator;
  private TimeManagerStorage storage;
  private Logger logger;

  public DefaultScheduleEvaluation(TimeManagerStorage timeManagerStorage)
  {
    this.storage = timeManagerStorage;
    Map<String, Schedule> allCustomCalendars = null;
    try
    {
      
      allCustomCalendars = this.storage.getAllCustomCalendars();
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "Could not load custom calendars", e);
      throw new InternalServiceException("Could not load custom calendars", e);
    }
    
    Map<String, CalendarExpression> customeCal = new HashMap<String, CalendarExpression>();
    
    for (Entry<String, Schedule> cal : allCustomCalendars.entrySet())
    {
      CalendarExpression calendarExpression = getCalendarExpression(cal.getValue());
      if (calendarExpression != null)
      {
        customeCal.put(cal.getKey(), calendarExpression);
      }
    }
    
    evaluator = new Evaluator(customeCal);
  }

  @Override
  public Schedule parseSchedule(String schedule) throws ScheduleResolutionException
  {
    CalendarExpression calendarExpression = ScheduleParser.parse(schedule);
//    ScheduleParser.parseScript(schedule);
    return new DefaultSchedule(schedule, null, calendarExpression);
  }

  @Override
  public ScheduleIntervalSet evaluateSchedule(Schedule schedule, CalendarElement from, CalendarElement until) throws ScheduleResolutionException
  {
    DefaultSchedule s = (DefaultSchedule) schedule;
    CalendarExpression expression = s.getExpression();
    TimeFrame timeFrame;
    try
    {
      timeFrame = evaluator.evaluate(expression, toCalendar(from), toCalendar(until), true);
    }
    catch (Exception e)
    {
      throw new ScheduleResolutionException(e.getMessage(), e);
    }
    timeFrame = evaluator.flatten(timeFrame);
    
    return toScheduleIntervalSet(schedule, timeFrame);
  }

  private ScheduleIntervalSet toScheduleIntervalSet(Schedule schedule, TimeFrame timeFrame)
  {
    List<ScheduleInterval> list = new ArrayList<ScheduleInterval>();
    
    if (timeFrame instanceof Interval)
    {
      list.add(new DefaultScheduleInterval(toCalendarElement(timeFrame.getStart()), toCalendarElement(timeFrame.getEnd())));
    }
    else if (timeFrame instanceof ContinuousIntervalSet)
    {
      list.add(new DefaultScheduleInterval(toCalendarElement(timeFrame.getStart()), toCalendarElement(timeFrame.getEnd())));
    }
    else if (timeFrame instanceof ComplexIntervalSet)
    {
      IntervalSet intervalSet = (IntervalSet)timeFrame;
      
      for (TimeFrame tf : intervalSet.getTimeFrames())
      {
        if (tf instanceof Interval || tf instanceof ContinuousIntervalSet)
        {
          list.add(new DefaultScheduleInterval(toCalendarElement(tf.getStart()), toCalendarElement(tf.getEnd())));
        }
        else
        {
          throw new IllegalStateException();
        }
      }
    }
    else
    {
      throw new IllegalStateException();
    }
     
    
    
    return new DefaultScheduleIntervalSet(schedule, list);
  }
  

  @Override
  public ScheduleIntervalSet evaluateSchedule(Schedule schedule, CalendarElement from, int intervalCount)
  {
    // TODO Auto-generated method stub
    return null;
  }

//  @Override
//  public ScheduleIntervalSet evaluateSchedule(Schedule schedule, BaseGranularity granularity, CalendarElement from,
//      CalendarElement until)
//  {
//    // TODO Auto-generated method stub
//    return null;
//  }
//
//  @Override
//  public ScheduleIntervalSet evaluateSchedule(Schedule schedule, BaseGranularity granularity, CalendarElement from,
//      int intervalCount)
//  {
//    // TODO Auto-generated method stub
//    return null;
//  }

  @Override
  public ScheduleInterval getNextInterval(Schedule schedule, CalendarElement from)
  {
    DefaultSchedule s = (DefaultSchedule) schedule;
    CalendarExpression expression = s.getExpression();
    TimeFrame timeFrame = evaluator.nextInterval(expression, toCalendar(from), true);
    timeFrame = evaluator.flatten(timeFrame);
    
    if (timeFrame instanceof Interval)
    {
      return new DefaultScheduleInterval(toCalendarElement(timeFrame.getStart()), toCalendarElement(timeFrame.getEnd()));
    }
    else
    {
      throw new IllegalStateException();
    }
  }

  @Override
  public boolean hasNextInterval(Schedule schedule, CalendarElement from)
  {
    DefaultSchedule s = (DefaultSchedule) schedule;
    CalendarExpression expression = s.getExpression();
    TimeFrame timeFrame = evaluator.nextInterval(expression, toCalendar(from), true);
    return timeFrame != null;
  }
  
  @Override
  public boolean hasNextInterval(Schedule schedule, CalendarElement from, CalendarElement until)
  {
    DefaultSchedule s = (DefaultSchedule) schedule;
    CalendarExpression expression = s.getExpression();
    TimeFrame timeFrame = evaluator.evaluate(expression, toCalendar(from), toCalendar(until), true);
    
    if (timeFrame instanceof Interval)
    {
      return true;
    } else if (timeFrame instanceof IntervalSet)
    {
      IntervalSet set = (IntervalSet) timeFrame;
      return set.size() > 0;
    }
    else
    {
      throw new IllegalStateException();
    }
  }

  private CalendarElement toCalendarElement(Calendar start)
  {
    return new DefaultCalendarElement(start.getTimeInMillis());
  }

  private Calendar toCalendar(CalendarElement from)
  {
    Calendar fromC = Calendar.getInstance();
    fromC.setTimeInMillis(from.getTimeInMillis());
    return fromC;
  }

  public Map<String, Schedule> getKnownCustomCalendars() throws DataSourceException
  {
    return storage.getAllCustomCalendars();
  }

  public void addCustomCalendar(String id, Schedule calendar) throws DataSourceException
  {
    CalendarExpression calendarExpression = getCalendarExpression(calendar);
    
    if (calendarExpression != null)
    {
      storage.storeCustomCalendar(id, calendar);    
      
      
      this.evaluator.registerIdent(new Ident(id), calendarExpression);
    }
  }

  protected CalendarExpression getCalendarExpression(Schedule calendar)
  {
    CalendarExpression calendarExpression;
    DefaultSchedule s;
    if (calendar instanceof DefaultSchedule)
    {      
      s = (DefaultSchedule) calendar;
    }
    else
    {
      try
      {
        s = (DefaultSchedule) parseSchedule(calendar.getRule());
      }
      catch (ScheduleResolutionException e)
      {
        // TODO Auto-generated catch block
        e.printStackTrace();
        return null;
      }
    }
    
    calendarExpression = s.getExpression();
    return calendarExpression;
  }

  public void removeCustomCalendar(String id) throws DataSourceException
  {
      storage.removeCustomCalendar(id);    

      this.evaluator.unregisterIdent(new Ident(id));
    
  }

}
