package de.aristaflow.adept2.core.timemanager.storage.transientstorage;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.aristaflow.adept2.base.service.InternalServiceException;
import de.aristaflow.adept2.base.service.InvalidServiceStateException;
import de.aristaflow.adept2.base.sessionmanagement.SessionLock;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.storage.InstanceTimeModelStorage;
import de.aristaflow.adept2.model.timemodel.ChangeableTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelState;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelUpdateStatus;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultInstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTimeModelFactory;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;
import de.aristaflow.adept2.util.LoggerTools;
import de.aristaflow.adept2.util.locking.AcquireLockException;
import de.aristaflow.adept2.util.locking.ObjectLockManager;

/**
 * @author  lanz
 */
public class TransientInstanceTimeModelStorage implements InstanceTimeModelStorage
{

  /**
   * @author  lanz
   */
  protected class ActivityTimeData
  {

    /**
     * @uml.property  name="nodeID"
     */
    private final int nodeID;
    private long startTime = -1;
    private long finishTime = -1;

    public ActivityTimeData(int nodeID)
    {
      this.nodeID = nodeID;
    }

    public void setStartTime(int iteration, long timestamp)
    {
      this.startTime = timestamp;
      // TODO AL Iteration
    }

    public void setFinishTime(int iteration, long timestamp)
    {
      this.finishTime = timestamp;
      // TODO AL Iteration
    }

    public long getStartTime(int iteration)
    {
      return startTime;
    }

    public long getFinishTime(int iteration)
    {
      return finishTime;
    }

    /**
     * @return
     * @uml.property  name="nodeID"
     */
    public int getNodeID()
    {
      return nodeID;
    }
    
    

  }

  
  protected final Logger logger = LoggerTools.getLogger(this);
  
  /** Lock for accessing the <code>timeData</code>-Map */
  protected final ReentrantReadWriteLock timeDataLock ;
  
  /**
   * Lock Manager Responsible for Locking Instance Time Data. It is only used with exclusive (write) locks.
   * @uml.property  name="lockManager"
   * @uml.associationEnd  
   */
  protected final ObjectLockManager<UUID, SessionToken> lockManager;
  
  
  private Map<UUID, UpdateableInstanceTimeContext> timeData = new HashMap<UUID, UpdateableInstanceTimeContext>();
  
  private Map<UUID, InstanceTimeModel> timeModels = new HashMap<UUID, InstanceTimeModel>();
  private Map<UUID, ChangeableTimeModelStatus> timeModelStati = new HashMap<UUID, ChangeableTimeModelStatus>();
  
  private long timeOut = 0;

  
  public TransientInstanceTimeModelStorage()
  {
    timeDataLock = new ReentrantReadWriteLock();
    try
    {
      lockManager = new ObjectLockManager<UUID, SessionToken>(SessionLock.SubsessionLockCount.class,
          SessionLock.class, "time manager instance time storage");
    }
    catch (InvocationTargetException ite)
    {
      String msg = "Can not instantiate object lock manager for synchronising "
                   + "storage of worklists. There were problems instantiating "
                   + "the lock count manager and/or the session locks.";
      logger.log(Level.SEVERE, msg, ite);
      throw new InternalServiceException(msg, ite);
    }
  }
  
  
  ///////////////////////////////////////////
  // Instance Time Data Handling
  ///////////////////////////////////////////
  
  public  void lockInstanceTimeContext(SessionToken session, UUID instanceID)
  {
    logger.fine(String.format("Session '%s' aquiring lock for time data '%s'",
                              session, instanceID));
    
    try
    {
      lockManager.getLockForObject(true, session, instanceID, timeOut);
    }
    catch (InterruptedException ie)
    {
      Thread.currentThread().interrupt();
    }
    catch (TimeoutException te)
    {
      String msg =
        String.format("Got a timeout while waiting to lock the time data"
                      + " '%s' for session '%s'. This is possibly a deadlock.",
                      instanceID, session);
      throw new InvalidServiceStateException(msg, te);
    }
  }

  
  protected UpdateableInstanceTimeContext createAndLockInstanceTimeContext(
      UUID instanceID)
  {
    timeDataLock.writeLock().lock();
    try
    {
      if (timeData.containsKey(instanceID))
        return timeData.get(instanceID);
      
      UpdateableInstanceTimeContext instanceTimeData = new DefaultInstanceTimeContext(instanceID);
      ChangeableTimeModelStatus timeModelStatus = new DefaultTimeModelStatus(instanceID, TimeModelState.UNKOWN, TimeModelUpdateStatus.DISABLED);
      
      timeData.put(instanceID, instanceTimeData);
      timeModelStati.put(instanceID, timeModelStatus);
      return instanceTimeData;
    }
    finally
    {
      timeDataLock.writeLock().unlock();
    }
  }

  public UpdateableInstanceTimeContext createAndLockInstanceTimeContext(SessionToken session,
      UUID instanceID) throws DataSourceException
  {
    lockInstanceTimeContext(session, instanceID);
    return createAndLockInstanceTimeContext(instanceID);
  }
  
  public InstanceTimeContext getInstanceTimeContextReadonly(
      UUID instanceID) throws DataSourceException
  {
    timeDataLock.readLock().lock();
    try
    {
      InstanceTimeContext instanceTimeData = timeData.get(instanceID);
      return instanceTimeData;
    }
    finally
    {
      timeDataLock.readLock().unlock();
    }
  }
  
  public UpdateableInstanceTimeContext getAndLockInstanceTimeContext(SessionToken session,
      UUID instanceID) throws DataSourceException
  {
    lockInstanceTimeContext(session, instanceID);
    return (UpdateableInstanceTimeContext)getInstanceTimeContextReadonly(instanceID);
  }

  public void updateInstanceTimeContext(SessionToken session,
      InstanceTimeContext instanceTimeData)
      throws DataSourceException, LockException
  {
    if (!lockManager.hasLock(true, session, instanceTimeData.getInstanceID()))
    {
      String message =  String.format("Session '%s' is not holiding any lock on " +
                                      "time data '%s'!", session, instanceTimeData.getInstanceID());
      logger.log(Level.SEVERE, message);
      throw new AcquireLockException(instanceTimeData.getInstanceID(), session, "updateInstanceTimeContext");
      
    }

    // Nothing to do because this is a transient implementation and we are
    // working directly on the items
  }
  
  public void unlockInstanceTimeContext(SessionToken session, UUID instanceID) throws LockException
  {
    logger.fine(String.format("Session '%s' releasing lock for time data '%s'",
                              session, instanceID));
    
    lockManager.unlockObject(true, session, instanceID);
  }

  //////////////////////////////////////////
  //
  //////////////////////////////////////////
  

  public InstanceTimeModel getInstanceTimeModelReadonly(UUID instanceID) throws DataSourceException
  {
    timeDataLock.readLock().lock();
    try
    {
      InstanceTimeModel timeModel = timeModels.get(instanceID);
      return timeModel;
    }
    finally
    {
      timeDataLock.readLock().unlock();
    }
  }
  
  public UpdateableInstanceTimeModel getAndLockInstanceTimeModel(SessionToken session, UUID instanceID) throws DataSourceException
  {
    lockInstanceTimeModel(session, instanceID);
    
    // TODO AL make sure lock is released in case of exception
    TimeModelFactory tmf;
//    try
//    {
      //tmf = registry.getModelFactory(MODEL_FACTORY_PROCESS, ProcessModelFactory.class);
      //TODO AL: use registry
      tmf = new CSTNUTimeModelFactory();
      tmf.setCreateChangeable(true);
//    }
//    catch (ServiceNotKnownException snke)
//    {
//      String msg = "Could not retrieve time model factory for creating the "
//                  + "updateable time model '%s'. Aborting retrieval of template.";
//      throw new InternalServiceException(String.format(msg, instanceID), snke);
//    }
    
    InstanceTimeModel ret = getInstanceTimeModelReadonly(instanceID);
    return (UpdateableInstanceTimeModel) ret; // FIXME  AL//de.aristaflow.adept2.model.timemodel.TimeModelTools.cloneInstanceTimeModel(ret, tmf);
  }
  
  public void lockInstanceTimeModel(SessionToken session, UUID instanceID)
  {
    lockInstanceTimeContext(session, instanceID);
  }

  public void updateInstanceTimeModel(SessionToken session, UUID instanceID,
      InstanceTimeModel timeModel)
      throws DataSourceException, LockException
  {
    if (!lockManager.hasLock(true, session, instanceID))
    {
      String message = String.format(
                                     "Session '%s' is not holiding any lock on "
                                         + "time data '%s'!", session, instanceID);
      logger.log(Level.SEVERE, message);
      throw new AcquireLockException(instanceID, session,
                                          "updateInstanceTimeModel");
    }
    
    timeDataLock.writeLock().lock();
    try
    {
      InstanceTimeContext data = timeData.get(instanceID);
      if (data != null)
      {
        timeModels.put(instanceID, timeModel);
      }
      else
      {
        throw new DataSourceException();
      }
    }
    finally
    {
      timeDataLock.writeLock().unlock();
    }
  }
  
  public void unlockInstanceTimeModel(SessionToken session, UUID instanceID) throws LockException
  {
    unlockInstanceTimeContext(session, instanceID);
  }
  
  //////////////////////////////////////////
  // Time Model Status
  //////////////////////////////////////////

  public TimeModelStatus getTimeModelStatusReadonly(UUID instanceID)
  {
    timeDataLock.readLock().lock();
    try
    {
      return copyTimeModelStatus(timeModelStati.get(instanceID));
    }
    finally
    {
      timeDataLock.readLock().unlock();
    }
  }
  
  public ChangeableTimeModelStatus getAndLockTimeModelStatus(SessionToken session, UUID instanceID)
  {
    lockInstanceTimeContext(session, instanceID);
    return (ChangeableTimeModelStatus)getTimeModelStatusReadonly(instanceID);
  }

  public void updateTimeModelStatus(SessionToken session, TimeModelStatus status)
      throws LockException
  {
    if (!lockManager.hasLock(true, session, status.getInstanceID()))
    {
      String message = String.format(
                                     "Session '%s' is not holiding any lock on "
                                         + "time data '%s'!", session, status
                                       .getInstanceID());
      logger.log(Level.SEVERE, message);
      throw new AcquireLockException(status.getInstanceID(), session,
                                          "updateTimeModelStatus");
    }

    timeDataLock.writeLock().lock();
    try
    {
      timeModelStati.put(status.getInstanceID(), copyTimeModelStatus(status));
    }
    finally
    {
      timeDataLock.writeLock().unlock();
    }
  }
  

  
  public void unlockTimeModelStatus(SessionToken session, UUID instanceID) throws LockException
  {
    unlockInstanceTimeContext(session, instanceID);
  }

  private ChangeableTimeModelStatus copyTimeModelStatus(TimeModelStatus status)
  {
    ChangeableTimeModelStatus timeModelStatus = new DefaultTimeModelStatus(
      status.getInstanceID(), status.getTimeModelStatus(),
      status.getTimeModelUpdateStatus(), 
      status.getTimeModelUpdateStatusComment());
    return timeModelStatus;
  }


  public boolean hasTimeModelForInstance(UUID instanceID)
      throws DataSourceException
  {
    return getInstanceTimeModelReadonly(instanceID) != null;
  }



  
}
