package de.aristaflow.adept2.core.timemanager.storage.xmlstorage;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.configuration.Configuration;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import de.aristaflow.adept2.base.configuration.ConfigurationException;
import de.aristaflow.adept2.core.timemanager.ScheduleResolutionException;
import de.aristaflow.adept2.core.timemanager.storage.TimeManagerStorage;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.ScheduleParser;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.CalendarExpression;
import de.aristaflow.adept2.model.timemodel.Schedule;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultSchedule;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LoggerTools;
import de.aristaflow.adept2.util.xml.XMLFormatException;
import de.aristaflow.adept2.util.xml.XMLHelperTools;

public class XMLTimeManagerStorage implements TimeManagerStorage
{

  public static final String CONFIGURATION_DATA_DIR = "StorageDataDir";
  
  public static final String CONFIGURATION_FILE = "customCalendars.config";
  
  private String dataDirName;
  private Logger logger = LoggerTools.getLogger(this);
  private File dataDir;

  private TimeModelFactory factory;

  private Map<String, Schedule> customCalendars;
  
  public XMLTimeManagerStorage(Configuration configuration, TimeModelFactory factory) throws ConfigurationException
  {
    this.factory = factory;
    // get the data dir from the configuration
    dataDirName = configuration.getString(CONFIGURATION_DATA_DIR);

    // configuration value set?
    if (dataDirName == null)
    {
      String message = String
        .format("The required configuration value '%s' is not set.",
                CONFIGURATION_DATA_DIR);
      logger.severe(message);
      throw new ConfigurationException(message);
    }

    // valid path? need read / write access, and it must be a directory.
    dataDir = new File(dataDirName);
    
    if (!dataDir.exists())
    {
      logger.warning(String.format("Data dir '%s' does not exist. Creating it.", 
                                   dataDir));
      boolean success = dataDir.mkdir();
      if (!success)
      {
        String msg = String.format("The directory %s could not be created!",
                                   dataDir.getAbsolutePath());
        logger.severe(msg);
      }
    }
  }
  
  public Map<String, Schedule> getAllCustomCalendars() throws DataSourceException
  {
    if (customCalendars == null)
    {
      customCalendars = internalGetReadCustomCalendars();
    }    
    return Collections.unmodifiableMap(customCalendars);
  }

  private Map<String, Schedule> internalGetReadCustomCalendars() throws DataSourceException
  {
    try
    {
      File customCalendarFile = getCustomCalendarFile();
      if (customCalendarFile.exists())
        return readCustomCalendarsFromFile(customCalendarFile);
      else
        return Collections.emptyMap();
    }
    catch (FileNotFoundException e)
    {
      String message = String
          .format("Exception while reading custom calendar file.");
        logger.log(Level.SEVERE, message, e);
        throw new DataSourceException(message, e);
    }
    catch (XMLFormatException e)
    {
      String message = String
          .format("Exception while reading custom calendar file.");
        logger.log(Level.SEVERE, message, e);
        throw new DataSourceException(message, e);
    }
    catch (IOException e)
    {
      String message = String
          .format("Exception while reading custom calendar file.");
        logger.log(Level.SEVERE, message, e);
        throw new DataSourceException(message, e);
    }
  }

  private Map<String, Schedule> readCustomCalendarsFromFile(File file) throws FileNotFoundException, XMLFormatException, IOException
  {
 // get the document from the file (without validation)
    Document document = XMLHelperTools.getDocumentFromFile(file, null, null);
    
    // check the version attribute
//    validateCustomCalendarFormatVersion(document);
    
    // now validate the document
//    XMLHelperTools.validateDoc(document, TimeModelXMLHelperTools
//      .getSchema(SchemaType.CUSTOM_CALENDAR));
    
    return getCustomCalendarsFromFile(document);
  }

  private Map<String, Schedule> getCustomCalendarsFromFile(Document document)
  {
    Element root = document.getDocumentElement();
    return getCustomCalendarsFromFile(this.factory, root);
  }

  private Map<String, Schedule> getCustomCalendarsFromFile(TimeModelFactory factory, Element root)
  {
    NodeList nodeList = root.getElementsByTagName("CustomCalendar");
    Map<String, Schedule> result = new HashMap<String, Schedule>();
    for (int i = 0; i < nodeList.getLength(); i++)
    {
      Node item = nodeList.item(i);
      Element element = XMLHelperTools.getElement(item);
      
      String id = element.getAttribute("id");
      Schedule schedule = getCustomCalendar(factory, element);
      result.put(id, schedule);
    }
    
    return result;
  }

  private Schedule getCustomCalendar(TimeModelFactory factory, Element element)
  {
    String schedule = element.getTextContent();
    
    CalendarExpression calendarExpression;
    try
    {
      calendarExpression = ScheduleParser.parse(schedule);
      return new DefaultSchedule(schedule, null, calendarExpression);
    }
    catch (ScheduleResolutionException e)
    {
      logger.log(Level.SEVERE, "An error occurred", e);
      return null;
    }
    
  }

  public void storeCustomCalendar(String id, Schedule calendar) throws DataSourceException
  {
    Map<String, Schedule> allCustomCalendars = new HashMap<String, Schedule>(getAllCustomCalendars());
    allCustomCalendars.put(id, calendar);
    updateCustomCalendars(allCustomCalendars);
  }

  private void updateCustomCalendars(Map<String, Schedule> allCustomCalendars) throws DataSourceException
  {
    try
    {
      File file = getCustomCalendarFile();
      
      if (!file.exists())
        file.createNewFile();
      
      writeCustomCalendarsToFile(file, allCustomCalendars);
    }
    catch (IOException e)
    {
      String message = String
          .format("Exception while reading custom calendar file.");
        logger.log(Level.SEVERE, message, e);
        throw new DataSourceException(message, e);
    }
    catch (XMLFormatException e)
    {
      String message = String
          .format("Exception while reading custom calendar file.");
        logger.log(Level.SEVERE, message, e);
        throw new DataSourceException(message, e);
    }
    this.customCalendars = allCustomCalendars;
  }
  
  private void writeCustomCalendarsToFile(File file, Map<String, Schedule> allCustomCalendars) throws IOException, XMLFormatException
  {
    // get the document from the file (without validation)
    Document document = XMLHelperTools.createDocument();
    
    writeCustomCalendarsFromFile(document, allCustomCalendars);
    
    XMLHelperTools.writeDocumentToFile(document, file, null);
  }

  private void writeCustomCalendarsFromFile(Document document, Map<String, Schedule> allCustomCalendars)
  {
    Element root = document.createElement("CustomCalendars");
    writeCustomCalendarsFromFile(root, allCustomCalendars);
    document.appendChild(root);
  }

  private void writeCustomCalendarsFromFile(Element root, Map<String, Schedule> allCustomCalendars)
  {
    for (Entry<String, Schedule> s : allCustomCalendars.entrySet())
    {
      Element element = root.getOwnerDocument().createElement("CustomCalendar");
      element.setAttribute("id", s.getKey());
      writeCustomCalendar(element, s.getValue());
      root.appendChild(element);
    }
    
  }

  private void writeCustomCalendar(Element element, Schedule value)
  {
     element.setTextContent(value.getRule());
  }

  private File getCustomCalendarFile()
  {
    return new File(
                    getDataDir(),
                    CONFIGURATION_FILE);
  }

  private File getDataDir()
  {
    return dataDir;
  }

  public void removeCustomCalendar(String id) throws DataSourceException
  {
    Map<String, Schedule> allCustomCalendars = new HashMap<String, Schedule>(getAllCustomCalendars());
    allCustomCalendars.remove(id);
    updateCustomCalendars(allCustomCalendars);
  }
  
}
