package de.aristaflow.adept2.core.worklistmanager.timeaware;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.InstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.TimeManager;
import de.aristaflow.adept2.core.worklistmanager.WorklistNotification2;
import de.aristaflow.adept2.core.worklistmanager.defaultimplementation.AbstractWorklistManager;
import de.aristaflow.adept2.core.worklistmanager.defaultimplementation.DefaultWorklistNotification;
import de.aristaflow.adept2.model.common.defaultimplementation.DefaultPluginData;
import de.aristaflow.adept2.model.common.defaultimplementation.DefaultPluginDataContainer;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.WorklistItemTimeContext;
import de.aristaflow.adept2.model.common.viewer.ModelViewer;
import de.aristaflow.adept2.model.processmodel.EBPInstanceReference;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelState;
import de.aristaflow.adept2.model.worklistmodel.InternalWorklistItem;
import de.aristaflow.adept2.util.DataSourceException;

public class TimeAwareWorklistNotification extends DefaultWorklistNotification implements WorklistNotification2
{

  public TimeAwareWorklistNotification(Configuration configuration, AbstractWorklistManager worklistManager)
  {
    super(configuration, worklistManager);
  }
  
  Map<UUID, Set<InternalWorklistItem>> instanceItems = new HashMap<UUID, Set<InternalWorklistItem>>();
  
  
  @Override
  protected void postAssignAndUpdateActivityHook(SessionToken session, InternalWorklistItem worklistItem) throws DataSourceException
  {
    // 
    EBPInstanceReference ebpInstanceReference = (EBPInstanceReference) worklistItem.getActivityReference();
    injectWorklistItemTimeContext(session, ebpInstanceReference, worklistItem);
  }
  
  @Override
  public void postUpdateActivityStateHook(SessionToken session, InternalWorklistItem worklistItem) throws DataSourceException
  {
    //
    EBPInstanceReference ebpInstanceReference = (EBPInstanceReference) worklistItem.getActivityReference();
    injectWorklistItemTimeContext(session, ebpInstanceReference, worklistItem);
  }
  
  @Override
  public void postAddActivityHook(SessionToken session, InternalWorklistItem item) throws DataSourceException
  {
    // 
    EBPInstanceReference ebpInstanceReference = (EBPInstanceReference) item.getActivityReference();
    injectWorklistItemTimeContext(session, ebpInstanceReference, item);
    
    synchronized (instanceItems)
    {
      Set<InternalWorklistItem> list = instanceItems.get(ebpInstanceReference.getInstanceID());
      if (list == null)
      {
        list = new HashSet<InternalWorklistItem>();
        instanceItems.put(ebpInstanceReference.getInstanceID(), list);
      }
      list.add(item);
    }
  }
  
  @Override
  protected void postRemoveWorklistItemHook(SessionToken session, InternalWorklistItem worklistItem)
  {
    EBPInstanceReference ebpInstanceReference = (EBPInstanceReference) worklistItem.getActivityReference();
    synchronized (instanceItems)
    {
      Set<InternalWorklistItem> list = instanceItems.get(ebpInstanceReference.getInstanceID());
      if (list != null)
      {
        list.remove(worklistItem);
        if (list.isEmpty())
        {
          instanceItems.remove(ebpInstanceReference.getInstanceID());
        }
      }
      
    }
  }


  private SessionToken getSessionToken()
  {
    return worklistManager.getSessionFactory().getSessionToken(worklistManager.getURIs());
  }
  
  protected void injectWorklistItemTimeContext(SessionToken sessionToken, EBPInstanceReference ebpInstanceReference,
      InternalWorklistItem item) throws DataSourceException
  {
    if (!(item instanceof DefaultPluginDataContainer))
    {
      logger.severe(String.format("EBPInstance '%s' does not implement DefaultPluginDataContainer! Unable to inject temporal information!", item));
    }
    else
    {
      
      
      DefaultPluginDataContainer pluginDataContainer = (DefaultPluginDataContainer) item;
      Map<String, String> attributes = new HashMap<String, String>();
      WorklistItemTimeContext context = ModelViewer.createView(attributes, WorklistItemTimeContext.class);
      InstanceTimeManager instanceTimeManager = getInstanceTimeManager();
      
      
      
      TimeManager timeManager = getWorklistManager().getTimeManager();
      TimeModelOperations operations = timeManager.getTimeModellingService().getTimeModelFactory().getTimeModelOperations();
      
      InstanceTimeContext timeContext = instanceTimeManager.getInstanceTimeContext(sessionToken, ebpInstanceReference.getInstanceID());
      InstanceTimeModel timeModel = instanceTimeManager.getInstanceTimeModelReadonly(sessionToken, ebpInstanceReference.getInstanceID());
      TimeModelStatus timeModelStatus = instanceTimeManager.getTimeModelStatusReadonly(ebpInstanceReference.getInstanceID());
      
      // 
      if (timeModelStatus.getTimeModelStatus() == TimeModelState.VALID)
      {
        Interval<CalendarElement> startTimeframe = timeModel.getActivityInstanceStartTimeframe(ebpInstanceReference.getNodeID(), ebpInstanceReference.getNodeIteration());
        Interval<TimeDistance> duration = timeModel.getActivityInstanceDuration(ebpInstanceReference.getNodeID(), ebpInstanceReference.getNodeIteration());
        Interval<CalendarElement> completionTimeframe = timeModel.getActivityInstanceCompletionTimeframe(ebpInstanceReference.getNodeID(), ebpInstanceReference.getNodeIteration());
        
        context.setEarliestStartTime(startTimeframe.getMinValue());
        context.setLatestStartTime(startTimeframe.getMaxValue());
        context.setEarliestCompletionTime(completionTimeframe.getMinValue());
        context.setLatestCompletionTime(completionTimeframe.getMaxValue());
        context.setMinimumDurationRestriction(operations.convertToTimePoint(duration.getMinValue()));
        context.setMaximumDurationRestriction(operations.convertToTimePoint(duration.getMaxValue()));
        context.setTimeModelStatus(TimeModelState.VALID);
        
        System.out.println("Injecting temporal data... ("+startTimeframe.getMinValue()+"; "+startTimeframe.getMaxValue()+")");
      }
      else
      {
        context.setTimeModelStatus(timeModelStatus.getTimeModelStatus());
        System.out.println("No temporal data available to be injected... ("+timeModelStatus.getTimeModelStatus()+")");
      }
      
      // Inject WorklistItemTimeContext... This is not the best way to do this but the most practical one
      DefaultPluginData pluginData = new DefaultPluginData( WorklistItemTimeContext.ID, attributes);
      pluginDataContainer.addSupportedPlugin(WorklistItemTimeContext.ID, WorklistItemTimeContext.ID, pluginData);
    }
  }



  public InstanceTimeManager getInstanceTimeManager()
  {
    return getWorklistManager().getInstanceTimeManager();
  }



  private TimeAwareWorklistManager getWorklistManager()
  {
    return (TimeAwareWorklistManager)this.worklistManager;
  }




  public void updateInstance(UUID instance)
  {
    System.out.println("Asynchronously updating Items for (" + instance + ") to ");
    List<InternalWorklistItem> items;
    synchronized (instanceItems)
    {
      Set<InternalWorklistItem> set = instanceItems.get(instance);
      if (set != null)
      {
        items = new ArrayList<InternalWorklistItem>(set);
      }
      else
      {
        items = Collections.emptyList();
      }
    }
    
    
    
    SessionToken sessionToken = getSessionToken();

    for (InternalWorklistItem item : items)
    {
      EBPInstanceReference ebpInstanceReference = (EBPInstanceReference) item.getActivityReference();
      
      try
      {
        injectWorklistItemTimeContext(sessionToken, ebpInstanceReference, item);
        this.worklistManager.notifyWorklistsAboutUpdate(item);
      }
      catch (DataSourceException e)
      {
        logger.log(Level.SEVERE, "DataSourceException occurred!", e);
      }
    }
    
    
  }
  

}
