package de.aristaflow.adept2.model.common.timedata.defaultimplementation;

import java.util.Map;

import org.apache.commons.collections.map.LRUMap;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;

public class AdvancedTimeDistance implements TimeDistance, TimePoint, Comparable<TimeDistance>
{
  
  private class IdentityTuple<T1, T2>
  {

    private final T1 t1;
    private final T2 t2;
    private final int hashCode;

    public IdentityTuple(T1 t1, T2 t2)
    {
      this.t1 = t1;
      this.t2 = t2;
      this.hashCode = 31 * t1.hashCode() * t2.hashCode();
    }
    
    @Override
    public int hashCode()
    {
      return this.hashCode;
    }
    
    @Override
    public boolean equals(Object o)
    {
      if (o == this)
        return true;
      
      if (o.hashCode() != this.hashCode())
        return false;
      
      if (o instanceof IdentityTuple<?, ?>)
      {
        IdentityTuple<?, ?> other = (IdentityTuple<?, ?>) o;
        return this.t1.equals(other.t1) && this.t2.equals(other.t2) ||
            this.t1.equals(other.t2) && this.t2.equals(other.t1);
      }
      
      return false;
    }
    
    @Override
    public String toString()
    {
      return "<"+t1+", "+t2+">";
    }
    

  }
  
  private long miliseconds;
  private long seconds;
  private long minutes;
  private long hours;
  private long days;
  private long months;
  
  private AdvancedTimeDistance negative = null;
  private int cachedHashCode = 0;

  public static final TimeDistance NEGATIVE_INFINITE_TIME_DISTANCE = new AdvancedTimeDistance(
    TimeDistance.NEGATIVE_INFINITY, TimeDistance.NEGATIVE_INFINITY, TimeDistance.NEGATIVE_INFINITY,
    TimeDistance.NEGATIVE_INFINITY, TimeDistance.NEGATIVE_INFINITY);

  public static final TimeDistance POSITIVE_INFINITE_TIME_DISTANCE = new AdvancedTimeDistance(
    TimeDistance.POSITIVE_INFINITY, TimeDistance.POSITIVE_INFINITY, TimeDistance.POSITIVE_INFINITY,
    TimeDistance.POSITIVE_INFINITY, TimeDistance.POSITIVE_INFINITY);
  

  public static final TimeDistance ONE_TIME_DISTANCE = new AdvancedTimeDistance(1, 0, 0, 0, 0);
  
  /** 
   * Zero time distance 
   */
  public static final TimeDistance ZERO_TIME_DISTANCE = new AdvancedTimeDistance(0, 0, 0, 0, 0);
  
  /** 
   * The smallest possible time distance that is greater than zero, currently 1ms
   */
  public static final TimeDistance EPSILON_TIME_DISTANCE = new AdvancedTimeDistance(1, 0, 0, 0, 0, 0);

  public AdvancedTimeDistance(long seconds, long minutes, long hours, long days, long months)
  {
    this(0, seconds, minutes, hours, days, months);
  }
  
  public AdvancedTimeDistance(long miliseconds, long seconds, long minutes, long hours, long days, long months)
  {
    // If one value is INFINITY all others are too with the same sign
    
    assert miliseconds != TimeDistance.POSITIVE_INFINITY
           || seconds != TimeDistance.POSITIVE_INFINITY
           || minutes != TimeDistance.POSITIVE_INFINITY
           || hours != TimeDistance.POSITIVE_INFINITY
           || days != TimeDistance.POSITIVE_INFINITY
           || months != TimeDistance.POSITIVE_INFINITY
           || (seconds == minutes && seconds == hours && seconds == days && seconds == months);
    
    assert miliseconds != TimeDistance.NEGATIVE_INFINITY
        || seconds != TimeDistance.NEGATIVE_INFINITY
        || minutes != TimeDistance.NEGATIVE_INFINITY
        || hours != TimeDistance.NEGATIVE_INFINITY
        || days != TimeDistance.NEGATIVE_INFINITY
        || months != TimeDistance.NEGATIVE_INFINITY
        || (seconds == minutes && seconds == hours && seconds == days && seconds == months);
    
    // Not all values are undefined
    assert miliseconds != TimeDistance.UNDEFINED_TIME_DISTANCE
        || seconds != TimeDistance.UNDEFINED_TIME_DISTANCE
        || minutes != TimeDistance.UNDEFINED_TIME_DISTANCE
        || hours != TimeDistance.UNDEFINED_TIME_DISTANCE
        || days != TimeDistance.UNDEFINED_TIME_DISTANCE
        || months != TimeDistance.UNDEFINED_TIME_DISTANCE;
    
    this.miliseconds = miliseconds;
    this.seconds = seconds;
    this.minutes = minutes;
    this.hours = hours;
    this.days = days;
    this.months = months;    
  }

//  /**
//   * Copy constructor
//   * @param timeDistance
//   */
//  public DefaultTimeDistance(TimeDistance timeDistance)
//  {
//    this(timeDistance.getSeconds(), timeDistance.getMinutes(), 
//         timeDistance.getHours(), timeDistance.getDays(), 
//         timeDistance.getMonths());    
//  }

  public AdvancedTimeDistance(TimePoint timePoint)
  {
    this(timePoint.getSeconds(), timePoint.getMinutes(), timePoint.getHours(),
         timePoint.getDays(), timePoint.getMonths()); 
  }

  public long getSeconds()
  {
    return seconds;
  }

  public long getMilliseconds()
  {
    return miliseconds;
  }
  
  public void setSeconds(long seconds)
  {
    this.seconds = seconds;
    resetCache();
  }

  private void resetCache()
  {
    if (negative != null)
    {
      negative.negative = null;
      this.negative = null;
    }
    cachedHashCode = 0;
  }

  public long getMinutes()
  {
    return minutes;
  }

  public void setMinutes(long minutes)
  {
    this.minutes = minutes;
    resetCache();
  }

  public long getHours()
  {
    return hours;
  }

  public void setHours(long hours)
  {
    this.hours = hours;
    resetCache();
  }

  public long getDays()
  {
    return days;
  }

  public void setDays(long days)
  {
    this.days = days;
    resetCache();
  }

  public long getMonths()
  {
    return months;
  }

  public void setMonths(long months)
  {
    this.months = months;
    resetCache();
  }
  
  @Override
  public String toString()
  {
    
    if (this.getMonths() == TimeDistance.POSITIVE_INFINITY)
    {
      return "infty";
    }
    
    if (this.getMonths() == TimeDistance.NEGATIVE_INFINITY)
    {
      return "-infty";
    }

    StringBuilder s = new StringBuilder();
    boolean notZero = false;
    
    
    if (this.getMonths() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getMonths() != 0)
      {
        s.append(this.getMonths());
        s.append("M");
        notZero = true;
      }
      
    }

    if (this.getDays() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getDays() != 0 || notZero)
      {
        s.append(this.getDays());
        s.append("d");
        notZero = true;
      }
    }

    if (this.getHours() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getHours() != 0 || notZero)
      {
        s.append(this.getHours());
        s.append("h");
        notZero = true;
      }
    }

    if (this.getMinutes() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getMinutes() != 0 || notZero)
      {
        s.append(this.getMinutes());
        s.append("m");
        notZero = true;
      }
    }

    if (this.getSeconds() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      s.append(this.getSeconds());
      s.append("s");
      notZero = true;
    }
    
    if (this.getMilliseconds() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getMilliseconds() != 0)
      {
        //s.append(".");
        s.append(this.getMilliseconds());
        s.append("ms");
        notZero = true;
      }
    }
    
    if (!notZero)
    {
      s.append("0s");
    }

    return s.toString();

  }

  public TimeDistance neg()
  {
    if (negative == null)
    {
      negative = new AdvancedTimeDistance(miliseconds != TimeDistance.UNDEFINED_TIME_DISTANCE ? -miliseconds : TimeDistance.UNDEFINED_TIME_DISTANCE,
                                   seconds != TimeDistance.UNDEFINED_TIME_DISTANCE ? -seconds : TimeDistance.UNDEFINED_TIME_DISTANCE, 
                                   minutes != TimeDistance.UNDEFINED_TIME_DISTANCE ? -minutes : TimeDistance.UNDEFINED_TIME_DISTANCE, 
                                   hours != TimeDistance.UNDEFINED_TIME_DISTANCE ? -hours : TimeDistance.UNDEFINED_TIME_DISTANCE, 
                                   days != TimeDistance.UNDEFINED_TIME_DISTANCE ? -days : TimeDistance.UNDEFINED_TIME_DISTANCE, 
                                   months != TimeDistance.UNDEFINED_TIME_DISTANCE ? -months : TimeDistance.UNDEFINED_TIME_DISTANCE);
      negative.negative = this;
    }
    return negative;
  }

  @Override
  public int hashCode()
  {
    if (cachedHashCode == 0)
    {
      final int prime = 31;
      cachedHashCode = 1;
      cachedHashCode = prime * cachedHashCode + (days != TimeDistance.UNDEFINED_TIME_DISTANCE ? (int) (days ^ (days >>> 32)) : 0);
      cachedHashCode = prime * cachedHashCode + (hours != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (hours ^ (hours >>> 32)) : 0);
      cachedHashCode = prime * cachedHashCode + (minutes != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (minutes ^ (minutes >>> 32)) : 0);
      cachedHashCode = prime * cachedHashCode + (months != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (months ^ (months >>> 32)) : 0);
      cachedHashCode = prime * cachedHashCode + (seconds != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (seconds ^ (seconds >>> 32)) : 0);
      cachedHashCode = prime * cachedHashCode + (miliseconds != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (miliseconds ^ (miliseconds >>> 32)) : 0);
    }
    return cachedHashCode;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;
    
    if (obj instanceof TimeDistance)
    {
      TimeDistance other = (TimeDistance) obj;
      // TODO reconsider whether to use convertToSeconds or compareValue...
//      return this.convertToSeconds(this) == convertToSeconds(other);
//      return compareValue(this.seconds, other.getSeconds())
//             && compareValue(this.minutes, other.getMinutes())
//             && compareValue(this.hours, other.getHours()) 
//             && compareValue(this.days, other.getDays())
//             && compareValue(this.months, other.getMonths());
      return compareTo(other) == 0;
    }

    return super.equals(obj);
  }

  private boolean compareValue(long v1, long v2)
  {
    if (v1 == v2)
      return true;
    if ((v1 == 0 && v2 == TimeDistance.UNDEFINED_TIME_DISTANCE) || (v1 == TimeDistance.UNDEFINED_TIME_DISTANCE && v2 == 0))
      return true;
    return false;
  }

  public int compareTo(TimeDistance o)
  {
    
    TimeDistance v2 = o;
    
    if (this.isInfinity())
    {
      // Both are infinity with the same signum
      int s1 = this.sign();
      if (v2.isInfinity() && s1 == v2.sign())
        return 0;
      // If v1 is positive infinity it will always be greater, otherwise it will always be less
      return s1 > 0 ? 1 : -1;
    }
    
    if (v2.isInfinity())
    {
      // If v2 is positive infinity it will always be greater, otherwise it will always be less
      return v2.sign() > 0 ? -1 : 1;
    }
    
    long diff = this.toMilliseconds() - v2.toMilliseconds();
    int sgn = diff > 0 ? 1 : diff < 0 ? -1 : 0;
    return sgn ;
  }

  public boolean isInfinity()
  {
    if (this.getMonths() == TimeDistance.POSITIVE_INFINITY)
      return true;
    
    if (this.getMonths() == TimeDistance.NEGATIVE_INFINITY)
      return true;
    
    return false;
  }
  
  public int sign()
  {
    if (this.isInfinity())
      return this.getMonths() > 0 ? 1 : -1;
    long milis = this.toMilliseconds();
    return milis > 0 ? 1 : milis < 0 ? -1 : 0;
  }
  
  public long toMilliseconds()
  {
    long milis = toSeconds();
    milis *= 1000;
    if (miliseconds != TimePoint.UNDEFINED_TIME_POINT)
    {
      milis += miliseconds;
    }
    return milis;
  }
  
  @Override
  public long toSeconds()
  {
    if (this.isInfinity())
      return this.getMonths() > 0 ? TimeDistance.POSITIVE_INFINITY : TimeDistance.NEGATIVE_INFINITY;
    
    long seconds = 0;
    
    if (this.getMonths() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += this.getMonths();
    }
    seconds *= 30;
    
    if (this.getDays() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += this.getDays();
    }
    seconds *= 24;
    
    if (this.getHours() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += this.getHours();
    }
    seconds *= 60;
    
    if (this.getMinutes() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += this.getMinutes();
    }
    seconds *= 60;
    
    if (this.getSeconds() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += this.getSeconds();
    }

    return seconds;
  }
  
  @SuppressWarnings("unchecked")
  private static final Map<IdentityTuple<TimeDistance, TimeDistance>, TimeDistance> cacheAddTimeDistance = new LRUMap(100000);
  
  public TimeDistance add(TimeDistance timeDistance)
  {
    IdentityTuple<TimeDistance, TimeDistance> cacheKey = new IdentityTuple<TimeDistance, TimeDistance>(this, timeDistance);
    
    TimeDistance sum = cacheAddTimeDistance.get(cacheKey);
    
    if (sum != null)
    {
      return sum;
    }
    
    sum = internalAdd(timeDistance);
    
    cacheAddTimeDistance.put(cacheKey, sum);
    
    return sum;
  }

  protected TimeDistance internalAdd(TimeDistance summand)
  {
    // infinity - infinity
    if (this.isInfinity() && summand.isInfinity() && this.sign() != summand.sign())
      throw new IllegalArgumentException("infinity + (- infinity)");
    
    // If one argument is infinity return it
    if (this.isInfinity())
      return this;
    if (summand.isInfinity())
      return summand;
    
    long months = this.getMonths() == TimeDistance.UNDEFINED_TIME_DISTANCE ? summand.getMonths() : summand.getMonths() == TimeDistance.UNDEFINED_TIME_DISTANCE ? this.getMonths() : this.getMonths() + summand.getMonths();
    long days = this.getDays() == TimeDistance.UNDEFINED_TIME_DISTANCE ? summand.getDays() : summand.getDays() == TimeDistance.UNDEFINED_TIME_DISTANCE ? this.getDays() : this.getDays() + summand.getDays();
    long hours = this.getHours() == TimeDistance.UNDEFINED_TIME_DISTANCE ? summand.getHours() : summand.getHours() == TimeDistance.UNDEFINED_TIME_DISTANCE ? this.getHours() : this.getHours() + summand.getHours();
    long minutes = this.getMinutes() == TimeDistance.UNDEFINED_TIME_DISTANCE ? summand.getMinutes() : summand.getMinutes() == TimeDistance.UNDEFINED_TIME_DISTANCE ? this.getMinutes() : this.getMinutes() + summand.getMinutes();
    long seconds = this.getSeconds() == TimeDistance.UNDEFINED_TIME_DISTANCE ? summand.getSeconds() : summand.getSeconds() == TimeDistance.UNDEFINED_TIME_DISTANCE ? this.getSeconds() : this.getSeconds() + summand.getSeconds();
    long milliseconds = this.getMilliseconds() == TimeDistance.UNDEFINED_TIME_DISTANCE ? summand.getMilliseconds() : summand.getMilliseconds() == TimeDistance.UNDEFINED_TIME_DISTANCE ? this.getMilliseconds() : this.getMilliseconds() + summand.getMilliseconds();
        
    if (seconds != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (Math.abs(seconds) >= 60)
      {
        //System.out.println("a) Seconds: "+seconds + " Minutes: "+minutes);
        if (minutes == TimeDistance.UNDEFINED_TIME_DISTANCE)
        {
          minutes = seconds / 60;
        }
        else
        {
          minutes += seconds / 60;
        }
        seconds = seconds % 60;
        //System.out.println("b) Seconds: "+seconds + " Minutes: "+minutes);
      }
    }
    
    if (minutes != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (sign(seconds) != 0 && sign(minutes) != 0 && sign(seconds) != sign(minutes))
      {
        if (minutes > 0)
        {
          minutes -= 1;
          seconds += 60;
        }
        else
        {
          minutes += 1;
          seconds -= 60;
        }
      }
          
      
      if (Math.abs(minutes) >= 60)
      {
        if (hours == TimeDistance.UNDEFINED_TIME_DISTANCE)
        {
          hours = minutes / 60;
        }
        else
        {
          hours += minutes / 60;
        }
        minutes = minutes % 60;
      }
    }
    
    if (hours != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (sign(minutes) != 0 && sign(hours) != 0 && sign(minutes) != sign(hours))
      {
        if (hours > 0)
        {
          hours -= 1;
          minutes += 60;
        }
        else
        {
          hours += 1;
          minutes -= 60;
        }
        
        if (sign(seconds) != 0 && sign(minutes) != 0 && sign(seconds) != sign(minutes))
        {
          if (minutes > 0)
          {
            minutes -= 1;
            seconds += 60;
          }
          else
          {
            minutes += 1;
            seconds -= 60;
          }
        }
      }
      
      if (Math.abs(hours) >= 24)
      {
        if (days == TimeDistance.UNDEFINED_TIME_DISTANCE)
        {
          days = hours / 24;
        }
        else
        {
          days += hours / 24;
        }
        hours = hours % 24;
      }
    }
    
    if (sign(days) != 0 && sign(hours) != 0 && sign(hours) != sign(days))
    {
      if (days > 0)
      {
        days -= 1;
        hours += 24;
      }
      else
      {
        days += 1;
        hours -= 24;
      }
      if (sign(minutes) != 0 && sign(hours) != 0 && sign(minutes) != sign(hours))
      {
        if (hours > 0)
        {
          hours -= 1;
          minutes += 60;
        }
        else
        {
          hours += 1;
          minutes -= 60;
        }
        
        if (sign(seconds) != 0 && sign(minutes) != 0 && sign(seconds) != sign(minutes))
        {
          if (minutes > 0)
          {
            minutes -= 1;
            seconds += 60;
          }
          else
          {
            minutes += 1;
            seconds -= 60;
          }
        }
      }
    }
    return new AdvancedTimeDistance(seconds, minutes, hours, days, months);
  }

  public TimeDistance sub(
      TimeDistance timeDistance)
  {
    return this.add(timeDistance.neg());
  }
  
  private int sign(long value)
  {
    if (value == TimeDistance.UNDEFINED_TIME_DISTANCE)
      return 0;
    return value == 0 ? 0 : value > 0 ? 1 : -1;
  }
}
