package de.aristaflow.adept2.model.common.timedata.defaultimplementation;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;

public class DefaultTimeDistance implements TimeDistance, Comparable<TimeDistance>
{
    
  private long miliseconds;

  
  private DefaultTimeDistance negative = null;
  private int cachedHashCode = 0;

  public static final TimeDistance NEGATIVE_INFINITE_TIME_DISTANCE = new DefaultTimeDistance(
    TimeDistance.NEGATIVE_INFINITY, TimeDistance.NEGATIVE_INFINITY, TimeDistance.NEGATIVE_INFINITY,
    TimeDistance.NEGATIVE_INFINITY, TimeDistance.NEGATIVE_INFINITY);

  public static final TimeDistance POSITIVE_INFINITE_TIME_DISTANCE = new DefaultTimeDistance(
    TimeDistance.POSITIVE_INFINITY, TimeDistance.POSITIVE_INFINITY, TimeDistance.POSITIVE_INFINITY,
    TimeDistance.POSITIVE_INFINITY, TimeDistance.POSITIVE_INFINITY);
  

  public static final TimeDistance ONE_TIME_DISTANCE = new DefaultTimeDistance(1, 0, 0, 0, 0);
  
  /** 
   * Zero time distance 
   */
  public static final TimeDistance ZERO_TIME_DISTANCE = new DefaultTimeDistance(0, 0, 0, 0, 0);
  
  /** 
   * The smallest possible time distance that is greater than zero, currently 1ms
   */
  public static TimeDistance EPSILON_TIME_DISTANCE = new DefaultTimeDistance(1, 0, 0, 0, 0, 0);

  public DefaultTimeDistance(long seconds, long minutes, long hours, long days, long months)
  {
    this(0, seconds, minutes, hours, days, months);
  }
  
  public DefaultTimeDistance(long miliseconds, long seconds, long minutes, long hours, long days, long months)
  {
    // Not all values are undefined
    assert miliseconds != TimeDistance.UNDEFINED_TIME_DISTANCE
        || seconds != TimeDistance.UNDEFINED_TIME_DISTANCE
        || minutes != TimeDistance.UNDEFINED_TIME_DISTANCE
        || hours != TimeDistance.UNDEFINED_TIME_DISTANCE
        || days != TimeDistance.UNDEFINED_TIME_DISTANCE
        || months != TimeDistance.UNDEFINED_TIME_DISTANCE;
    
    long value;
    
    if (miliseconds == TimeDistance.POSITIVE_INFINITY
           || seconds == TimeDistance.POSITIVE_INFINITY
           || minutes == TimeDistance.POSITIVE_INFINITY
           || hours == TimeDistance.POSITIVE_INFINITY
           || days == TimeDistance.POSITIVE_INFINITY
           || months == TimeDistance.POSITIVE_INFINITY)
    {
      value = TimeDistance.POSITIVE_INFINITY;
    }
    else if (miliseconds == TimeDistance.NEGATIVE_INFINITY
        || seconds == TimeDistance.NEGATIVE_INFINITY
        || minutes == TimeDistance.NEGATIVE_INFINITY
        || hours == TimeDistance.NEGATIVE_INFINITY
        || days == TimeDistance.NEGATIVE_INFINITY
        || months == TimeDistance.NEGATIVE_INFINITY)
    {
      value = TimeDistance.NEGATIVE_INFINITY;
    }
    else
    {
      value = months != TimeDistance.UNDEFINED_TIME_DISTANCE ? months : 0;
      value *= 30;
      value += days != TimeDistance.UNDEFINED_TIME_DISTANCE ? days : 0;
      value *= 24;
      value += hours != TimeDistance.UNDEFINED_TIME_DISTANCE ? hours : 0;
      value *= 60;
      value += minutes != TimeDistance.UNDEFINED_TIME_DISTANCE ? minutes : 0;
      value *= 60;
      value += seconds != TimeDistance.UNDEFINED_TIME_DISTANCE ? seconds : 0;
      value *= 1000;
      value += miliseconds;
    }
    
    this.miliseconds = value; 
  }
  
  private DefaultTimeDistance(long miliseconds)
  {
    this.miliseconds = miliseconds;
  }

  public DefaultTimeDistance(TimePoint timePoint)
  {
    this(timePoint.getSeconds(), timePoint.getMinutes(), timePoint.getHours(),
         timePoint.getDays(), timePoint.getMonths()); 
  }

  public long getSeconds()
  {
    if (this.isInfinity())
      return miliseconds;
    return (miliseconds / 1000) % 60;
  }

  public long getMilliseconds()
  {
    if (this.isInfinity())
      return miliseconds;
    return miliseconds % 1000;
  }

  public long getMinutes()
  {
    if (this.isInfinity())
      return miliseconds;
    return (miliseconds / 1000 / 60) % 60;
  }

  public long getHours()
  {
    if (this.isInfinity())
      return miliseconds;
    return (miliseconds / 1000 / 60 / 60) % 24;
  }

  public long getDays()
  {
    if (this.isInfinity())
      return miliseconds;
    return (miliseconds / 1000 / 60 / 60 / 24);
  }

  public long getMonths()
  {
    if (this.isInfinity())
      return miliseconds;
    return 0;
  }
  
  @Override
  public String toString()
  {
    
    if (this.miliseconds == TimeDistance.POSITIVE_INFINITY)
    {
      return "infty";
    }
    
    if (this.miliseconds == TimeDistance.NEGATIVE_INFINITY)
    {
      return "-infty";
    }

    StringBuilder s = new StringBuilder();
    boolean notZero = false;
    
    

    if (this.getMonths() != 0)
    {
      s.append(this.getMonths());
      s.append("M");
      notZero = true;
    }
    


    if (this.getDays() != 0 || notZero)
    {
      s.append(this.getDays());
      s.append("d");
      notZero = true;
    }
  


    if (this.getHours() != 0 || notZero)
    {
      s.append(this.getHours());
      s.append("h");
      notZero = true;
    }
  


    if (this.getMinutes() != 0 || notZero)
    {
      s.append(this.getMinutes());
      s.append("m");
      notZero = true;
    }
  

 
    if (this.getSeconds() != 0 || notZero)
    {
      s.append(this.getSeconds());
      s.append("s");
      notZero = true;
    }
  
  
  
    if (this.getMilliseconds() != 0)
    {
      //s.append(".");
      s.append(this.getMilliseconds());
      s.append("ms");
      notZero = true;
    }
    
    
    if (!notZero)
    {
      s.append("0s");
    }

    return s.toString();

  }

  public TimeDistance neg()
  {
    if (negative == null)
    {
      negative = new DefaultTimeDistance(-miliseconds);
      negative.negative = this;
    }
    return negative;
  }

  @Override
  public int hashCode()
  {
    if (cachedHashCode == 0)
    {
      final int prime = 31;
      cachedHashCode = 1;
      cachedHashCode = prime * cachedHashCode + (int) (miliseconds ^ (miliseconds >>> 32));
    }
    return cachedHashCode;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;
    
    if (obj instanceof TimeDistance)
    {
      TimeDistance other = (TimeDistance) obj;
      // TODO reconsider whether to use convertToSeconds or compareValue...
//      return this.convertToSeconds(this) == convertToSeconds(other);
//      return compareValue(this.seconds, other.getSeconds())
//             && compareValue(this.minutes, other.getMinutes())
//             && compareValue(this.hours, other.getHours()) 
//             && compareValue(this.days, other.getDays())
//             && compareValue(this.months, other.getMonths());
      return compareTo(other) == 0;
    }

    return super.equals(obj);
  }
  public int compareTo(TimeDistance o)
  {
    
    TimeDistance v2 = o;
    
    if (this.isInfinity())
    {
      // Both are infinity with the same signum
      int s1 = this.sign();
      if (v2.isInfinity() && s1 == v2.sign())
        return 0;
      // If v1 is positive infinity it will always be greater, otherwise it will always be less
      return s1 > 0 ? 1 : -1;
    }
    
    if (v2.isInfinity())
    {
      // If v2 is positive infinity it will always be greater, otherwise it will always be less
      return v2.sign() > 0 ? -1 : 1;
    }
    
    long diff = this.toMilliseconds() - v2.toMilliseconds();
    int sgn = diff > 0 ? 1 : diff < 0 ? -1 : 0;
    return sgn ;
  }

  public boolean isInfinity()
  {
    if (this.miliseconds == TimeDistance.POSITIVE_INFINITY)
      return true;
    
    if (this.miliseconds == TimeDistance.NEGATIVE_INFINITY)
      return true;
    
    return false;
  }
  
  public int sign()
  {
    if (this.isInfinity())
      return this.miliseconds > 0 ? 1 : -1;
    long milis = this.toMilliseconds();
    return milis > 0 ? 1 : milis < 0 ? -1 : 0;
  }
  
  public long toMilliseconds()
  {
    if (this.isInfinity())
      return this.miliseconds > 0 ? TimeDistance.POSITIVE_INFINITY : TimeDistance.NEGATIVE_INFINITY;
      
    return miliseconds;
  }
  
  @Override
  public long toSeconds()
  {
    if (this.isInfinity())
      return this.miliseconds > 0 ? TimeDistance.POSITIVE_INFINITY : TimeDistance.NEGATIVE_INFINITY;
    
    return miliseconds / 1000;
  }
  
  public TimeDistance add(TimeDistance timeDistance)
  {
    TimeDistance sum;

    sum = internalAdd(timeDistance);
    

    return sum;
  }

  protected TimeDistance internalAdd(TimeDistance summand)
  {
    // infinity - infinity
    if (this.isInfinity() && summand.isInfinity() && this.sign() != summand.sign())
      throw new IllegalArgumentException("infinity + (- infinity)");
    
    // If one argument is infinity return it
    if (this.isInfinity())
      return this;
    if (summand.isInfinity())
      return summand;
    
    long milliseconds = this.miliseconds + summand.toMilliseconds();
    
    return new DefaultTimeDistance(milliseconds);
  }

  public TimeDistance sub(
      TimeDistance timeDistance)
  {
    return this.add(timeDistance.neg());
  }
  
}
