package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import de.aristaflow.adept2.model.timemodel.tools.AbstractGraphVisitor;
import de.aristaflow.adept2.model.timemodel.tools.DepthFirstSearch;
import de.aristaflow.adept2.model.timemodel.tools.UndirectedGraph;

/**
 * Helper class to compute the cycle equivalence classes of an undirected graph.
 * 
 * See <i>Johnson et al. 1994 - The Program Structure Tree: Computing Control
 * Regions in Linear Time</i> as well as <i>Ananian 1997 - The Static Single
 * Information Form</i> for details.
 * 
 * Implemented with some help from an implementation by Scott Ananian under GNU GPL Licence (see
 * <link>
 * http://bpmn2bpel.googlecode.com/svn/trunk/BPMN2BPEL/src/translation/analysis
 * /sese/CycleEq.java</link>)
 * 
 * @author lanz
 * 
 */
public class CycleEquivalence
{


  public CycleEquivalence()
  {
    
  }

  public Map<CQClass, List<EdgeStruct>> transform(UndirectedGraph<NodeStruct, EdgeStruct> graph, NodeStruct sourceNode)
  {
    
    final Queue<NodeStruct> reverseDepthFirst = constructDepthFirstSpanningTree(graph, sourceNode);
    
    for (NodeStruct n : reverseDepthFirst)
    {
      computeCycleEquivalency(graph, n);
    }
    
 // DFS search to order elements in cycle-equivalency classes.
    Map<CQClass, List<EdgeStruct>> equivalencyClasses = createEquivalencyClasses(graph, sourceNode);
    
    return equivalencyClasses;
  }




  private Map<CQClass, List<EdgeStruct>> createEquivalencyClasses(UndirectedGraph<NodeStruct, EdgeStruct> graph, NodeStruct sourceNode)
  {
    final Map<CQClass, List<EdgeStruct>> equiv = new HashMap<CQClass, List<EdgeStruct>>();
    DepthFirstSearch<NodeStruct,EdgeStruct> depthFirst = new DepthFirstSearch<NodeStruct,EdgeStruct>(graph);
    depthFirst.performDepthFirstSearch(new AbstractGraphVisitor<NodeStruct, EdgeStruct>()
    {

      @Override
      public void discoverEdge(NodeStruct startVertex, NodeStruct endVertex, EdgeStruct edge)
      {
        // 
        CQClass cq = edge.CQclass;
        List<EdgeStruct> l = equiv.get(cq);
        if (l == null)
        {
          l = new LinkedList<EdgeStruct>();
          equiv.put(cq, l);
        }
        if (!l.contains(edge))
          l.add(edge); // append edge to end of list.
      }


    }, sourceNode);
    
    return equiv;
  }

  private void computeCycleEquivalency(UndirectedGraph<NodeStruct, EdgeStruct> graph, NodeStruct n)
  {
  //    /* Compute Hi(n) */
  //    /* hi0 is highest using backedges only */
  //    hi0 := min{ Number(t) | (t,n) is a backedge };
      NodeStruct hi0 = null;
      for (EdgeStruct e : n.backedges)
      {
        NodeStruct t = otherEnd(e, n);
        if (hi0 == null || t.number < hi0.number)
          hi0 = t;
      }
      
      //    /* hi1 is highest through children */
      //    hi1 := min{ Hi(c) | c is a child of n };
      //    /* hi2 is lowest through children */
      //    hi2 := max{ Hi(c) | c is a child of n };
      NodeStruct hi1 = null;
      NodeStruct hi2 = null;
      for (EdgeStruct e : n.children)
      {
        NodeStruct t = otherEnd(e, n);
        if (hi1 == null || t.hi.number < hi1.number)
        {
          if (hi2 == null || (hi1 != null && hi1.number < hi2.number))
            hi2 = hi1;
          hi1 = t.hi;
        }
        else if (hi2 == null || t.hi.number < hi2.number)
        {
          hi2 = t.hi; // second-highest.
        }
//        if (hi2 == null || t.hi.number > hi2.number)
//        {
//          hi2 = t.hi;
//        }
      }
      
    //    Hi(n) := min{ hi0, hi1 };
      n.hi = (hi0 == null) ? hi1 : (hi1 == null) ? hi0 : (hi0.number < hi1.number) ? hi0 : hi1;


  //    /* Compute BList(n) */
  //    BList(n) := create ();
      n.blist = new LinkedList<EdgeStruct>();
      
  //    for each child c of n, do
  //      BList(n) := concat (BList(n), BList(c));
      for (EdgeStruct e: n.children)
      {
        NodeStruct t = otherEnd(e, n);
        List<EdgeStruct> blist = t.blist;
        if (blist != null)
        {
          n.blist.addAll(blist);
        }
      }
    //    for each capping backedge <d, n> of n, do
    //      BList(n) := delete (BList(n), <d, n>);
      for (EdgeStruct e: n.capping)
      {
         n.blist.remove(e);
      }
      
  //    for each backedge <d, n> from a descendant d of n to n, do
  //      BList(n) := delete (BList(n), <d, n>);

      for (EdgeStruct e : n.backedges)
      {
        if (otherEnd(e,n).number > n.number)
        { /* descendant of n */
          n.blist.remove(e);
          if (e.CQclass == null)
            e.CQclass = new CQClass(e, n.number); /* new CQ class */
        }
      }

  //    for each backedge <n, a> from n to an ancestor a of n, do {
  //      BList(n) := push (BList(n), <n, a>)
  //      RecentSize(<n, a>) := -1; /* not a representative node */
  //    }
      
      for (EdgeStruct e : n.backedges)
      {
        if (otherEnd(e, n).number <= n.number) /* ancestor or itself */
        {
          n.blist.push(e);
          e.recentSize = -1;
        }
      }
      
  //    if n has more than one child, then {
  //      BList(n) := push (BList(n), <n, hi2>); /* capping backedge */
  //      RecentSize(<n, hi2>) := -1;
  //      add <n, hi2> to capping backedges list of hi2;
  //    }
      if (hi2 != null && (hi0 == null || hi0.number > hi2.number))
//      if (n.children.size() > 1)
      {
        assert n.children.size() > 1;
        /* create capping backedge */
        EdgeStruct d = new EdgeStruct(n, hi2);
        hi2.capping.add(d);
//        hi2.capping.addAll(hi2.capping);
        n.blist.push(d);
      }
      else
      {
        assert n.children.size() <= 1 : hi0.number + ">" + hi2.number;
      }

  //    /* Compute NQClass (n) */
  //    if n is a representative node, then {
  //      if RecentSize (top (BList(n))) != size (BList(n)), then {
  //        /* start a new equivalence class */
  //        RecentSize (top (BList(n))) := size (BList(n));
  //        RecentClass (top (BList(n))) := new-class-name();
  //      }
  //      NQClass (n) := RecentClass (top (Blist(n)));
  //    }
      /* determine class for edge from parent(n) to n */
      EdgeStruct e = n.parent;
      if (e != null)
      {
        EdgeStruct b = n.blist.get(0);
        if (b.recentSize != n.blist.size())
        {
          b.recentSize = n.blist.size();
          b.recentClass = new CQClass(b, n.blist.size()); /* new CQ class */
        }
        e.CQclass = b.recentClass;

        /* handle one tree one backedge case */
        if (b.recentSize == 1)
          b.CQclass = e.CQclass;
      }
    
  }

  private NodeStruct otherEnd(EdgeStruct e, NodeStruct n)
  {
    return e.n1 == n ? e.n2 : e.n1;
  }


  private Queue<NodeStruct> constructDepthFirstSpanningTree(UndirectedGraph<NodeStruct,EdgeStruct> graph, NodeStruct sourceNode)
  {
    DepthFirstSearch<NodeStruct,EdgeStruct> depthFirst = new DepthFirstSearch<NodeStruct,EdgeStruct>(graph);
    
    final Queue<NodeStruct> reverseDepthFirst = new LinkedList<NodeStruct>();
    depthFirst.performDepthFirstSearch(new AbstractGraphVisitor<NodeStruct,EdgeStruct>()
    {
      int dfsnum = 0;

      @Override
      public void discoverVertex(NodeStruct vertex)
      {
        // Processing vertex set dfs num
        vertex.number = dfsnum++;
      }
      
      @Override
      public void finishVertex(NodeStruct vertex)
      {
        reverseDepthFirst.offer(vertex);
      }

      @Override
      public void discoverEdge(NodeStruct startVertex, NodeStruct endVertex, EdgeStruct edge)
      {
        // Has the endVertex already been processed?
        if (endVertex.number < 0)
        { // No, so add edge to list of children and set this as parent
          startVertex.children.add(edge);
          endVertex.parent = edge;
        }
        else if (edge != startVertex.parent)
        {
          // EndVertex has been process -> this is a backedge
          startVertex.backedges.add(edge);
        }
      }

    }, sourceNode);
    return reverseDepthFirst;
  }


  
}
