package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;

import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.StructuredConstraintEdge;
import de.aristaflow.adept2.model.processmodel.StructuredEdge;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tools.DefaultUndirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.DepthFirstSearch;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.Graph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;
import de.aristaflow.adept2.model.timemodel.tools.UndirectedGraph;

/**
 * Helper class to transform an Directed Graph or Template to an undirected
 * graph which can be used as input for {@link CycleEquivalence}.
 * 
 * <b>Node</b> that any node of the input graph will be represented by three
 * different nodes in the undirected graph, of which only one is a
 * "representative" of the original node and the others are
 * "non-representatives" (see {@link NodeStruct#representative}).
 * 
 * @author lanz
 * 
 * @param <V>
 * @param <E>
 */
public class UndirectedGraphTransformation<V, E>
{

  public interface TopoVisitor
  {

    void visit(Node node);

  }
  
  public interface UndirectedGraphFactory<V, E>
  {
    public V createVertex(Object vertex, boolean representative);
    
    public E createEdge(V v1, V v2, Object edge);
    
    public UndirectedGraph<V, E> createGraph(List<V> vertices);
    
    
  }

  private UndirectedGraphFactory<V, E> factory;
  public V sourceNode;
  public V sinkNode;
  
  public UndirectedGraphTransformation(UndirectedGraphFactory<V, E> factory)
  {
    this.factory = factory;
    
  }
  
  @SuppressWarnings("unchecked")
  public <N, C> UndirectedGraph<V, E> transform(DirectedGraph<N, C> directedGraph, N startNode)
  {
    DepthFirstSearch<N, C> depthFirstSearch = new DepthFirstSearch<N, C>(directedGraph);
    final List<V> vertices = new ArrayList<V>();
    final Map<N, Object[]> representatives = new HashMap<N, Object[]>();
    final Object[] turnaroundpoint = new Object[1];
    depthFirstSearch.performDepthFirstSearch(new GraphVisitor<N, C>()
    {
      public void discoverGraph(Graph<N, C> graph)
      {
        // 
      }

      public void finishGraph(Graph<N, C> graph)
      {
        // 
      }

      public void discoverVertex(N vertex)
      {
          V ni = factory.createVertex(vertex, false);
          V v = factory.createVertex(vertex, true);
          V no = factory.createVertex(vertex, false);
          
          vertices.add(ni);
          vertices.add(v);
          vertices.add(no);
          
          representatives.put(vertex, new Object[]{ni, v, no});
      }

      public void finishVertex(N vertex)
      {
        // 
        if (turnaroundpoint[0] == null)
          turnaroundpoint[0] = representatives.get(vertex)[2];
      }

      public void discoverEdge(N startVertex, N endVertex, C edge)
      {
        // 
      }

      public void finishEdge(N startVertex, N endVertex, C edge)
      {
        // 
      }
    }, startNode);
  
    DefaultUndirectedGraph<V, E> undirectedGraph = new DefaultUndirectedGraph<V, E>(vertices);
    
    for (Object[] rep : representatives.values())
    {
      V ni = (V) rep[0];
      V v = (V) rep[1];
      V no = (V) rep[2];

      undirectedGraph.addEdge(ni, v, factory.createEdge(ni, v, null));
      undirectedGraph.addEdge(v, no, factory.createEdge(v, no, null));
    }
    
    // Graph Structure
    for (N sourceNode : directedGraph.getVertices())
    {
      for (N destNode : directedGraph.getVertices())
      {        
        
        C edge = directedGraph.getEdge(sourceNode, destNode);
        if (edge != null)
        {
        
          Object[] sourceRep = representatives.get(sourceNode);
          Object[] destRep = representatives.get(destNode);
          
          V source = (V) sourceRep[2];
          V dest = (V) destRep[0];
          
          undirectedGraph.addEdge(source, dest, factory.createEdge(source, dest, edge));
        }
      }
    }
    
    V v1 = vertices.get(0);
    V v2 = (V) turnaroundpoint[0];
//    undirectedGraph.addEdge(v1, v2, factory.createEdge(v1, v2, null));
    
    Object[] source = representatives.get(startNode);
    this.sourceNode = (V)source[0];
    this.sinkNode = (V) turnaroundpoint[0];
    
    return undirectedGraph;
  }
  
  @SuppressWarnings("unchecked")
  public UndirectedGraph<V, E> transform(TemporalTemplate template)
  {
    Set<Node> nodes = template.getNodes();
    
    final List<V> vertices = new ArrayList<V>(nodes.size()*3);
    final Map<Integer, Object[]> representatives = new HashMap<Integer, Object[]>(nodes.size()*2);
    
    traverseTopological(template, new TopoVisitor()
    {
      public void visit(Node node)
      {
        V ni = factory.createVertex(node, false);
        V v = factory.createVertex(node, true);
        V no = factory.createVertex(node, false);
        
        vertices.add(ni);
        vertices.add(v);
        vertices.add(no);
        
        representatives.put(node.getID(), new Object[]{ni, v, no});
      }
    });
    
    
    DefaultUndirectedGraph<V, E> graph = new DefaultUndirectedGraph<V, E>(vertices);
    
    for (Object[] rep : representatives.values())
    {
      V ni = (V) rep[0];
      V v = (V) rep[1];
      V no = (V) rep[2];

      graph.addEdge(ni, v, factory.createEdge(ni, v, null));
      graph.addEdge(v, no, factory.createEdge(v, no, null));
    }
    
    // Process Structure
    Set<StructuredEdge> edgeStructure = template.getEdgeStructure(EdgeType.ET_CONTROL, EdgeType.ET_LOOP, EdgeType.ET_SYNC);
    
    for (StructuredEdge edge : edgeStructure)
    {
      int sourceNodeID = edge.getSourceNodeID();
      int destinationNodeID = edge.getDestinationNodeID();
      
      Object[] sourceRep = representatives.get(sourceNodeID);
      Object[] destRep = representatives.get(destinationNodeID);
      
      V source = (V) sourceRep[2];
      V dest = (V) destRep[0];
      

      graph.addEdge(source, dest, factory.createEdge(source, dest, null));
    }
    
    // Time Lags
    Set<StructuredConstraintEdge> constraints = template.getConstraintEdgeStructure();
    
    for (StructuredConstraintEdge edge : constraints)
    {
      // IMPROVE check whether or not edge is really a time edge 
      int sourceNodeID = edge.getSourceNodeID();
      int destinationNodeID = edge.getDestinationNodeID();
      
      Object[] sourceRep = representatives.get(sourceNodeID);
      Object[] destRep = representatives.get(destinationNodeID);
      
      V source = (V) sourceRep[2];
      V dest = (V) destRep[0];
      
      graph.addEdge(source, dest, factory.createEdge(source, dest, edge));
      
    }
    
    // Process Duration
    {
      int startNodeID = template.getStartNode().getID();
      int endNodeID = template.getEndNode().getID();

      Object[] sourceRep = representatives.get(startNodeID);
      Object[] destRep = representatives.get(endNodeID);
      
      // Duration
      {
        V source = (V) sourceRep[2];
        V dest = (V) destRep[0];
  
        graph.addEdge(source, dest, factory.createEdge(source, dest, null));
      }

      // Backedge
      {
        V source = (V) sourceRep[0];
        V dest = (V) destRep[2];
        graph.addEdge(source, dest, factory.createEdge(source, dest, null));
      }
    }
    
    Object[] source = representatives.get(template.getStartNode().getID());
    this.sourceNode = (V)source[0];
    Object[] end = representatives.get(template.getEndNode().getID());
    this.sinkNode = (V)end[2];
    
    return graph;
  }

  
  public void traverseTopological(Template template, TopoVisitor visitor)
  {
    Queue<Integer> queue = new LinkedList<Integer>();
    queue.offer(template.getStartNode().getID());

    Set<Integer> processed = new HashSet<Integer>();
    
    while (!queue.isEmpty())
    {
      int node = queue.poll();
      
      int[] pred = template.getPredByEdgeType(node, EdgeType.ET_CONTROL, EdgeType.ET_SYNC);
      
      boolean complete = true;
      
      for (int p : pred)
      {
        if (!processed.contains(p))
        {
          complete = false;
          break;
        }
      }
      
      if (complete)
      {
        visitor.visit(template.getNode(node));
        processed.add(node);
        
        int[] succs = template.getSuccByEdgeType(node, EdgeType.ET_CONTROL, EdgeType.ET_SYNC);
        
        for (int s : succs)
        {
          if (!queue.contains(s))
            queue.offer(s);
        }
      }
      else
      {
        queue.offer(node);
      }
      
    }

  }

  public Map<Node, SESERegion> transform(Map<NodeStruct, RegionStruct> regionMapping)
  {
    Map<Node, SESERegion> result = new HashMap<Node, SESERegion>();
    for (Entry<NodeStruct, RegionStruct> entry : regionMapping.entrySet())
    {
      if (entry.getKey().representative)
      {
        RegionStruct regionStruct = entry.getValue();
        EdgeStruct entryEdge = regionStruct.entry;
        EdgeStruct exitEdge = regionStruct.exit;
        Node startNode = (Node) entryEdge.n2.node;
        Node endNode = (Node) exitEdge.n1.node;
        SESERegion region = new SESERegion(startNode, endNode);
        result.put((Node)entry.getKey().node, region);
      }
    }
    return result;
  }

  public SESERegion transform(RegionStruct regionStruct)
  {
    EdgeStruct entryEdge = regionStruct.entry;
    EdgeStruct exitEdge = regionStruct.exit;
    Node startNode = null;
    Node endNode = null;
    if (entryEdge != null)
      startNode = (Node) entryEdge.n2.node;
    if (exitEdge != null)
      endNode = (Node) exitEdge.n1.node;
    SESERegion region = new SESERegion(startNode, endNode);
    
    boolean hasNodes = false;
    for (NodeStruct node : regionStruct.nodes)
    {
      if (node.representative)
      {
        region.addNode((Node)node.node);
        hasNodes = true;
      }
    }
    

    
    for (RegionStruct child : regionStruct.children)
    {
      SESERegion subRegion = transform(child);
      if (subRegion != null)
      {
        region.addChildRegion(subRegion);
        hasNodes = true;
      }
    }
    
    if (!hasNodes)
      return null;
    
    return region;
  }
  

}
