package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;

/*
 * TP2
 * 
 * Checks for ActivityDuration
 */
/**
 * <p>
 * An activity duration represents the duration constraints of an activity in a
 * process as described by time pattern 2.<br />
 * The respective activity has a certain duration restriction. Durations result
 * from both waiting and processing times. Durations are often determined by
 * external benchmarks (e.g., regulations, policies, QoS agreements).
 * </p>
 * <p>
 * Activity Durations may vary within certain boundaries. More precisely the
 * minimum duration is by default the lower bound (i.e., LowerMinimumDuration)
 * but may be restricted up to the upper bound (i.e., UpperMinimumDuration). For
 * the maximum duration the default value is the upper bound (i.e.,
 * UpperMaximumDuration) and it may be restricted down to the lower bound (i.e.,
 * LowerMaximumDuration). Ultimately the minimum duration must still be lower
 * then the respective maximum duration although UpperMinimumDuration may be
 * higher than LowerMaximumDuration. <br />
 * LowerMinimumDuration and UpperMaximumDuration may be null. If
 * LowerMinimumDuration is null UpperMinimumDuration also has to be null.
 * Likewise if UpperMaximumDuration is null LowerMaximumDuration also has to be
 * null.
 * </p>
 * <p>
 * Activity durations may be set or updated during runtime, e.g. they may depend
 * on a path choosen during the execution or the context of the executions. Late
 * binding is another example for a case where the activity duration can only be
 * determined during runtime. Therefore the ActivityDurationUpdateProcedure
 * provides means to determine and update the respective activity duration
 * constraint during runtime.
 * </p>
 * <p>
 * Use <blockquote>
 * 
 * <pre>
 * PluginData pluginData = node.getPluginData(ActivityDuration.ID);
 * ActivityDuration activityDuration = null;
 * if (pluginData != null)
 * {
 *   activityDuration = ModelViewer.createView(pluginData.getPluginData(),
 *                                             ActivityDuration.class);
 * }
 * </pre>
 * 
 * </blockquote> or
 * {@link TimeModelTools#getActivityDuration(de.aristaflow.adept2.model.processmodel.Node)}
 * to get an instance of this interface from a
 * {@link de.aristaflow.adept2.model.processmodel.Node} .
 * </p>
 */
public interface ActivityDuration {
  /* {src_lang=Java, author=Andreas Lanz, java_import=de.aristaflow.adept2.model.common.viewer.Attribute, java_import=de.aristaflow.adept2.model.common.viewer.AttributeSetter}*/


    
  //static final String ID = "TimeData.ActivityDuration";

  /**
   * <p> The lower bound of the minimum duration of the respective activity. This is the inital value used when determining the validity of the time model. In other words this is the duration the execution of the activity will take at least. </p>
   * @see #getGuardMinimumDuration()
   * @returns   the lower bound of the minimum duration of the respective  activity.
   * @uml.property  name="lowerMinimumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="LowerMinimumDuration")
  public TimePoint getMinimumDuration();

  /**
   * <p> The upper bound of the minimum duration of the respective activity. That is, the minimum duration restriction of the activity may be extended to this value when determining the validity of the time model. In other words this is the maximum time which the execution of the respective activity may have to take at least. </p>
   * @see #getLowerMinimumDuration()
   * @returns   the upper bound of the minimum duration of the respective  activity.
   * @uml.property  name="upperMinimumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="UpperMinimumDuration")
  public TimePoint getGuardMinimumDuration();

  /**
   * <p> The lower bound of the maximum duration of the respective activity. That is, the maximum duration restriction of the activity may be reduced to this value when determining the validity of the time model. In other words this is the minimal time which needs to available to the user for executing the respective activity. </p>
   * @see   getUpperMaximumDuration()
   * @returns   the lower bound of the maximum duration of the respective  activity.
   * @uml.property  name="lowerMaximumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="LowerMaximumDuration")
  public TimePoint getGuardMaximumDuration();

  /**
   * <p> The upper bound of the maximum duration of the respective activity. This is the inital value used when determining the validity of the time model. In other words this is the duration the execution of the activity may take at most. </p>
   * @see #getGuardMaximumDuration()
   * @returns   the upper bound of the maximum duration of the respective  activity.
   * @uml.property  name="upperMaximumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="UpperMaximumDuration")
  public TimePoint getMaximumDuration();
  
  @Attribute(name="Contingency", defaultValue="0")
  public TimePoint getContingency();

  /**
   * <p> The expected duration of the respective activity. This may be used during process execution to calculate the expected execution time of this and other activities for a more precise prediction of execution times. </p>
   * @returns   the expected execution duration of this activity
   * @uml.property  name="expectedDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="ExpectedDuration")
  public TimePoint getExpectedDuration();

  /**
   * <p> The Configuration of the ActivityDurationUpdateProcedure which allows to update the constraint of this ActivityDuration during process execution. </p>
   * @returns   the Configuration of the ActivityDurationUpdateProcedure which  will be used to update this duration constraint during runtime.
   * @uml.property  name="activityDurationUpdateProcedure"
   * @uml.associationEnd  
   */
  @Attribute(name="ActivityDurationUpdateProcedureData")
  public ActivityDurationUpdateProcedureData getActivityDurationUpdateProcedure();

  /**
   * <p> The Configuration of the TemporalExceptionHandlingProcedure. The reference temporal exception handling procedure will be called when, during runtime, an exception in the temporal dimension of the process is detected which is caused by this activity duration. Exceptions in the temporal dimension should be avoided as much as possible as their fixing requires ad-hoc changes to the process instance. Therefore it is possible to specify escalation strategies to prevent the time used by this activity from getting out of hand (see #getEscalationData). </p>
   * @returns   the configuration of the temporal exception handling procedure for  this activity duration restriction.
   * @uml.property  name="temporalExceptionHandlingProcedure"
   * @uml.associationEnd  
   */
  @Attribute(name="TemporalExceptionHandlingProcedureData")
  public PluginData getTemporalExceptionHandlingProcedure();

  /**
   * Plug-in data for escalation of process steps duration.
   * @returns   The Plug-in data for escalation of process steps duration.
   * @uml.property  name="escalationData"
   * @uml.associationEnd  
   */
  @Attribute(name="EcalationData")
  public PluginData getEscalationData();

  /**
   * @param  duration
   * @uml.property  name="lowerMinimumDuration"
   */
  @AttributeSetter("LowerMinimumDuration")
  public void setFlexibleMinimumDuration(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="upperMinimumDuration"
   */
  @AttributeSetter("UpperMinimumDuration")
  public void setGuardMinimumDuration(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="lowerMaximumDuration"
   */
  @AttributeSetter("LowerMaximumDuration")
  public void setGuardMaximumDuration(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="upperMaximumDuration"
   */
  @AttributeSetter("UpperMaximumDuration")
  public void setFlexibleMaximumDuration(TimePoint duration);

  @AttributeSetter("Contingency")
  public void setContingency(TimePoint contingency);
  
  /**
   * @param  duration
   * @uml.property  name="expectedDuration"
   */
  @AttributeSetter("ExpectedDuration")
  public void setExpectedDuration(TimePoint duration);

  /**
   * @param  data
   * @uml.property  name="activityDurationUpdateProcedure"
   */
  @AttributeSetter("ActivityDurationUpdateProcedureData")
  public void setActivityDurationUpdateProcedure(ActivityDurationUpdateProcedureData data);

  /**
   * @param  data
   * @uml.property  name="temporalExceptionHandlingProcedure"
   */
  @AttributeSetter("TemporalExceptionHandlingProcedureData")
  public void setTemporalExceptionHandlingProcedure(PluginData data);

  /**
   * @param  data
   * @uml.property  name="escalationData"
   */
  @AttributeSetter("EcalationData")
  public void setEscalationData(PluginData data);

}