package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.eventmodel.EventDescription;

/** 
 *  <p><code>ActivityDurationUpdateMeasures</code> provides the means to update the duration restriction of an activity (and prevents accessing arbitrary functions of the time manager). It allows to update the minimum, maximum and expected duration as well as the list of triggers responsible or triggering the respective {@link ActivityDurationUpdateHandler}.</p>
 */
public interface ActivityDurationUpdateMeasures {
  /* {src_lang=Java, author=Andreas Lanz}*/


  /** 
   *  Changes the minimum duration restriction of the specified activity instance.
   *  
   *  @param session The session token to be used for updating the activtiy duration. It is the same token as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param activity  The activity to set the minimum duration for. It is the same activity as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param minimumDuration The new minimum duration restriction of the activity
   *  @param ignoreInconsistency This indicates whether inconistencies in the time model resulting from this update should be accepted or if the update should be rejected in case of a resulting inconsistency.
   *  
   *  @returns Whether the update was successfull or would lead the time model of the process in an inconsistent state.
   *  @throws IllegalArgumentException if the specified minimum duration value is invalid
   */
  public boolean updateMinimumDuration(SessionToken session, ActivityInstanceReference activity, TimePoint minimumDuration, boolean ignoreInconsistency);

  /** 
   *  Changes the maximum duration restriction of the specified activity instance.
   *  
   *  @param session The session token to be used for updating the activtiy duration. It is the same token as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param activtiy  The activity to set the maximum duration for. It is the same activity as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param maximumDuration The new maximum duration restriction of the activity
   *  @param ignoreInconsistency This indicates whether inconistencies in the time model resulting from this update should be accepted or if the update should be rejected in case of a resulting inconsistency.
   *  
   *  @returns Whether the update was successfull or would lead the time model of the process in an inconsistent state.
   *  @throws IllegalArgumentException if the specified maximum duration value is invalid
   */
  public boolean updateMaximumDuration(SessionToken session, ActivityInstanceReference activity, TimePoint maximumDuration, boolean ignoreInconsistency);

  /** 
   *  Changes the expected duration of the specified activity instance.
   *  
   *  @param session The session token to be used for updating the activtiy duration. It is the same token as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param activtiy  The activity to set the expected duration for. It is the same activity as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param expectedDuration The new expected duration of the activity
   *  
   *  @throws IllegalArgumentException if the specified expected duration value is invalid
   */
  public void updateExpectedDuration(SessionToken session, ActivityInstanceReference activity, TimePoint expectedDuration);

  /** 
   *  <p>Removes the specified trigger from the activity's list of update triggers.</p>
   *  
   *  <p>
   *  Note:<br />
   *  If the update trigger list doesn't contain the given item, it should be ignored and this incident be properly logged, so it doesn't go unnoticed.</p>
   *  
   *  @param session The session token to be used for updating the activtiy's trigger list. It is the same token as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param activtiy  The activity to update the trigger list for. It is the same activity as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param trigger The process event reference to be removed from the list of update triggers
   */
  public void removeUpdateTrigger(SessionToken session, ActivityInstanceReference activity, EventDescription trigger);

  /** 
   *  <p>Adds the specified trigger to the activity's list of update triggers.</p>
   *  
   *  <p>
   *  Note:<br />
   *  If the update trigger list already contain the given item, it should be ignored and this incident be properly logged, so it doesn't go unnoticed.</p>
   *  
   *  
   *  @param session The session token to be used for updating the activtiy's trigger list. It is the same token as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param activity  The activity to update the trigger list for. It is the same activity as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param trigger The process event reference to be added from the list of update triggers
   */
  public void addUpdateTrigger(SessionToken session, ActivityInstanceReference activity, EventDescription trigger);

  /** 
   *  <p>Clears the activity's list of update triggers.</p>
   *  
   *  
   *  @param session The session token to be used for updating the activtiy's trigger list. It is the same token as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   *  @param activtiy  The activity to clear the trigger list for. It is the same activity as provided to {@link ActivityDurationUpdateHandler#updateDuration()}
   */
  public void clearUpdateTriggerList(SessionToken session, ActivityInstanceReference activity);

}