package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.eventmodel.EventDescription;

public interface FixedDateUpdateMeasures
{
  /* {src_lang=Java} */

  /**
   * Changes the ealiest execution time restriction of the specified activity
   * instance event.
   * 
   * @param session The session token to be used for updating the activtiy's
   *          earliest execution time. It is the same token as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param activtiy The activity to set the earliest execution time for. It is
   *          the same activity as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param date The new earlist execution time restriction of the activity
   * @param ignoreInconsistency This indicates whether inconistencies in the
   *          time model resulting from this update should be accepted or if the
   *          update should be rejected in case of a resulting inconsistency.
   * 
   * @returns Whether the update was successfull or would lead the time model of
   *          the process in an inconsistent state.
   * @throws IllegalArgumentException if the specified execution time value is
   *           invalid
   */
  public boolean updateEarliestDate(SessionToken session, int nodeID, int iteration, CalendarElement date,
      boolean ignoreInconsistency);

  /**
   * Changes the latest execution time restriction of the specified activity
   * instance event.
   * 
   * @param session The session token to be used for updating the activtiy
   *          latest execution time. It is the same token as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param activtiy The activity to set the latest execution time for. It is
   *          the same activity as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param date The new latest execution time restriction of the activity
   * @param ignoreInconsistency This indicates whether inconistencies in the
   *          time model resulting from this update should be accepted or if the
   *          update should be rejected in case of a resulting inconsistency.
   * 
   * @returns Whether the update was successfull or would lead the time model of
   *          the process in an inconsistent state.
   * @throws IllegalArgumentException if the specified execution time value is
   *           invalid
   */
  public boolean updateLatestDate(SessionToken session, int nodeID, int iteration, CalendarElement date,
      boolean ignoreInconsistency);

  /**
   * <p>
   * Adds the specified trigger to the activity's list of update triggers.
   * </p>
   * 
   * <p>
   * Note:<br />
   * If the update trigger list already contain the given item, it should be
   * ignored and this incident be properly logged, so it doesn't go unnoticed.
   * </p>
   * 
   * 
   * @param session The session token to be used for updating the activtiy's
   *          trigger list. It is the same token as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param activtiy The activity to update the trigger list for. It is the same
   *          activity as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param trigger The process event reference to be added from the list of
   *          update triggers
   */
  public void addUpdateTrigger(SessionToken session, EventDescription activity, EventDescription trigger);

  /**
   * <p>
   * Removes the specified trigger from the activity's list of update triggers.
   * </p>
   * 
   * <p>
   * Note:<br />
   * If the update trigger list doesn't contain the given item, it should be
   * ignored and this incident be properly logged, so it doesn't go unnoticed.
   * </p>
   * 
   * @param session The session token to be used for updating the activtiy's
   *          trigger list. It is the same token as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param activtiy The activity to update the trigger list for. It is the same
   *          activity as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param trigger The process event reference to be removed from the list of
   *          update triggers
   */
  public void removeUpdateTrigger(SessionToken session, EventDescription activity, EventDescription trigger);

  /**
   * <p>
   * Clears the activity's list of update triggers.
   * </p>
   * 
   * 
   * @param session The session token to be used for updating the activtiy's
   *          trigger list. It is the same token as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   * @param activtiy The activity to clear the trigger list for. It is the same
   *          activity as provided to
   *          {@link FixedDateUpdateHandler#updateFixedDate()}
   */
  public void clearUpdateTriggerList(SessionToken session, EventDescription activity);

  public String resolveParameterString(SessionToken session, int targetNodeID, String parameterString);

}