package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;
import de.aristaflow.adept2.model.globals.TimeModelConstants.ReferencePoint;

/*
 * Checks for ScheduleRestrictedElement
 * 
 */
/**
 * <p>
 * A schedule restricted element represents a restriction on the possible
 * execution time of an activity in a process to a schedule as described by time
 * pattern 5.<br />
 * The structure of this schedule is known at process type level, while the
 * concrete date is determined at instance level. The schedule provides
 * restrictions on when the respective element can be executed. In particular,
 * for rather restricted schedules even small delays in process execution can
 * become critical (if schedule restricted elements being on a critical path are
 * affected by the delay or the path becomes critical due to the delays).
 * Schedules may contain exceptions (e.g., every year except leap years)./p>
 * <p>
 * Whether the Schedule restricts the execution of the start or end event of the
 * activity is decided by the ReferencePoint of the ScheduleRestrictedElement.
 * </p>
 * <p>
 * The schedule is represented by a string representation, which, during process
 * execution, is piece-wiese evaluation using the {@link de.aristaflow.adept2.core.timemanager.ScheduleEvaluation} of the
 * TimeManager.
 * </p>
 * <p>
 * The Schedule of this schedule restricted element may be provided by an
 * external source, e.g. a data element or the configuration, as referenced by
 * #getScheduleSource().
 * </p>
 * <p>
 * Either #getSchedule() or #getScheduleSource() needs to be provided but not
 * both at the same time
 * </p>
 * <p>
 * Use <blockquote>
 * 
 * <pre>
 * PluginData pluginData = node.getPluginData(ScheduleRestrictedElement.ID);
 * ScheduleRestrictedElement scheduleRestrictedElement = null;
 * if (pluginData != null)
 * {
 *   scheduleRestrictedElement = ModelViewer.createView(pluginData.getPluginData(), ScheduleRestrictedElement.class);
 * }
 * </pre>
 * 
 * </blockquote> or
 * {@link TimeModelTools#getScheduleRestrictedElement(de.aristaflow.adept2.model.processmodel.Node)}
 * to get an instance of this interface from a
 * {@link de.aristaflow.adept2.model.processmodel.Node} .
 * </p>
 * 
 * @see de.aristaflow.adept2.core.timemanager.ScheduleEvaluation
 */
@SuppressWarnings("deprecation")
public interface ScheduleRestrictedElement {
  /* {src_lang=Java, java_import=de.aristaflow.adept2.model.common.viewer.Attribute, java_import=de.aristaflow.adept2.model.common.viewer.AttributeSetter}*/


  
  //public static final String ID = "TimeData.ScheduleRestrictedElement";

  /**
   * <p> The schedule of this schedule restricted element respresented as string. The concrete structure of the string is implementation dependent (see e.g. [Leban])</p> <p> During process execution the schedule, is piece-wiese evaluation using the ScheduleEvaluation of the TimeManager.</p> <p>Either #getSchedule() or #getScheduleSource() needs to be provided but not both at the same time</p>
   * @see  de.aristaflow.adept2.core.timemanager.ScheduleEvaluation
   * @uml.property  name="schedule"
   */
  @Attribute(name="Schedule")
  public String getSchedule();

  /**
   * The reference point determines whether the schedule applies to the start or end of the respective activity.
   * @returns   whether this schedule restricted element applies to the start or end of the respective activity.
   * @uml.property  name="referencePoint"
   * @uml.associationEnd  
   * @deprecated
   */
  @Attribute(name="ReferencePoint")
  public ReferencePoint getReferencePoint();

  /**
   * <p> TODO AL</p> <p>Either #getSchedule() or #getScheduleSource() needs to be provided but not both at the same time</p>
   * @see  de.aristaflow.adept2.core.timemanager.ScheduleEvaluation
   * @uml.property  name="scheduleSource"
   */
  @Attribute(name="ScheduleSource")
  public String getScheduleSource();

  /**
   * @param  schedule
   * @uml.property  name="schedule"
   */
  @AttributeSetter("Schedule")
  public void setSchedule(String schedule);

  /**
   * @param  referencePoint
   * @uml.property  name="referencePoint"
   */
  @AttributeSetter("ReferencePoint")
  public void setReferencePoint(ReferencePoint referencePoint);

  /**
   * @param  source
   * @uml.property  name="scheduleSource"
   */
  @AttributeSetter("ScheduleSource")
  public void setScheduleSource(String source);

}