package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.common.plugindata.EscalationData;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;

/*
 * TP1
 * 
 * Checks for TimeLag
 */
/**
 * <p>
 * The interface for time edges, used to model time lags between activity
 * events. There may be only one time edge connecting a pair of activity events,
 * where each event of the pair needs to belong to a different activity. This
 * implies, that the two activities and the relation type identify a time edge
 * unambiguously.
 * </p>
 * <p>
 * A time lag means, that there is a given time lag between two activities which
 * needs to be respected. Time Lags may not only exist between succeeding
 * activities, but also between arbitrary ones. Time lags are often required to
 * comply with existing rules and regulations. The time lag may or may not have
 * binding character.
 * </p>
 * <p>
 * Use <blockquote>
 * <pre>
 * PluginData pluginData = constraintEdge.getPluginData(TimeLag.ID);
 * 
 * TimeLag timeLag = null;
 * if (pluginData != null)
 * {
 *   timeLag = ModelViewer.createView(pluginData.getPluginData(), TimeLag.class);
 * }
 * </pre>
 * </blockquote> or
 * {@link TimeModelTools#getTimeLag(de.aristaflow.adept2.model.processmodel.ConstraintEdge)}
 * to get an instance of this interface from a
 * {@link de.aristaflow.adept2.model.processmodel.ConstraintEdge}.
 * </p>
 */
public interface TimeLag
{
  /* {src_lang=Java, author=Andreas Lanz} */

  //public static final String ID = "TimeData.TimeLag";

  /**
   * <p>
   * The relation type of a time edge describes whether the time lag describes a
   * start-to-start, start-to-end, end-to-start, or end-to-end relation between
   * the two process steps.
   * </p>
   * 
   * @returns the relation type of this time edge
   */
  @Attribute(name="RelationType")
  public RelationType getRelationType();

  /**
   * <p>
   * The lower bound of the minimum time lag of the respective time edge. This
   * is the inital value used when determining the validity of the time model.
   * In other words this is the minimal time distance between the two events.
   * </p>
   * 
   * @see #getUpperMinimumTimeLag()
   * @returns the lower bound of the minimum time lag of the respective time
   *          edge.
   */
  @Attribute(name="LowerMinimumTimeLag")
  public TimePoint getLowerMinimumTimeLag();

  /**
   * <p>
   * The upper bound of the minimum time lag of the respective time edge. That
   * is, the minimum time lag restriction of the time edge may be extended to
   * this value when determining the validity of the time model. In other words
   * this is the maximum time distance which needs to be at least between the
   * two events.
   * </p>
   * 
   * @see #getLowerMinimumTimeLag()
   * @returns the upper bound of the minimum time lag of the respective time
   *          edge.
   */
  @Attribute(name="UpperMinimumTimeLag")
  public TimePoint getUpperMinimumTimeLag();

  /**
   * <p>
   * The lower bound of the maximum time lag of the respective time edge. That
   * is, the maximum time lag restriction of the time edge may be reduced to
   * this value when determining the validity of the time model. In other words
   * this is the minimal time which needs to available between the two
   * respective events.
   * </p>
   * 
   * @see getUpperMaximumTimeLag()
   * @returns the lower bound of the maximum time lag of the respective time
   *          edge.
   */
  @Attribute(name="LowerMaximumTimeLag")
  public TimePoint getLowerMaximumTimeLag();

  /**
   * <p>
   * The upper bound of the maximum time lag of the respective time edge. This
   * is the inital value used when determining the validity of the time model.
   * In other words this is the maximal time distance between the two events.
   * </p>
   * 
   * @see #getLowerMaximumTimeLag()
   * @returns the upper bound of the maximum time lag of the respective time
   *          edge.
   */
  @Attribute(name="UpperMaximumTimeLag")
  public TimePoint getUpperMaximumTimeLag();

  /**
   * <p>
   * The expected time lag of the respective time deg. This may be used during
   * process execution to calculate the expected execution time activities for a
   * more precise prediction of execution times.
   * </p>
   * 
   * @returns the expected time lag of this time edge
   */
  @Attribute(name="ExpectedTimeLag")
  public TimePoint getExpectedTimeLag();

  /**
   * <p>
   * The Configuration of the TimeLagUpdateProcedure which allows to update the
   * constraint of this <code>TimeEdge</code> during process execution.
   * </p>
   * 
   * @returns the Configuration of the TimeLagUpdateProcedure which will be used
   *          to update this time lag constraint during runtime.
   */
  @Attribute(name="TimeLagUpdateHandler")
  public TimeLagUpdateProcedureData getTimeLagUpdateHandler();

  /**
   * <p>
   * The Configuration of the TemporalExceptionHandlingProcedure. The reference
   * temporal exception handling procedure will be called when, during runtime,
   * an exception in the temporal dimension of the process is detected which is
   * caused by this time edge. Exceptions in the temporal dimension should be
   * avoided as much as possible as their fixing requires ad-hoc changes to the
   * process instance. Therefore it is possible to specify escalation strategies
   * to prevent the time "used by this time edge" from getting out of hand (see
   * #getEscalationData).
   * </p>
   * 
   * @returns the configuration of the temporal exception handling procedure for
   *          this time edge.
   */
  @Attribute(name="TemporalExceptionHandlingProcedure")
  public PluginData getTemporalExceptionHandlingProcedure();

  /**
   * Plug-in data for escalation of time lags.
   * 
   * @returns The Plug-in data for escalation of time lags.
   */
  @Attribute(name="EscalationData")
  public EscalationData getEscalationData();
  
  
  @AttributeSetter("RelationType")
  public void setRelationType(RelationType relationType);

  /**
   * @param  duration
   * @uml.property  name="lowerMinimumDuration"
   */
  @AttributeSetter("LowerMinimumTimeLag")
  public void setLowerMinimumTimeLag(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="upperMinimumDuration"
   */
  @AttributeSetter("UpperMinimumTimeLag")
  public void setUpperMinimumTimeLag(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="lowerMaximumDuration"
   */
  @AttributeSetter("LowerMaximumTimeLag")
  public void setLowerMaximumTimeLag(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="upperMaximumDuration"
   */
  @AttributeSetter("UpperMaximumTimeLag")
  public void setUpperMaximumTimeLag(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="expectedDuration"
   */
  @AttributeSetter("ExpectedTimeLag")
  public void setExpectedTimeLag(TimePoint duration);

}