package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.eventmodel.EventDescription;
import de.aristaflow.adept2.model.eventmodel.EventReference;

/** 
 *  <p>This interface supports user-defined update procedures for time lag constraints. This allows to update time lag constraints during runtime.<br />
 *  An implementing class is notified when one out of a predefined set of triggers is raised. Based on the activity instance reference, the trigger, and the access to  <code>TimeLagUpdateMeasures</code> arbitrary source can be used to determine the new time lag constraint.</p>
 *  <p>
 *  Each time lag update handler is identified by a (unique) ID. This ID is also found in the time edge.</p>
 */
public interface TimeLagUpdateHandler {
  /* {src_lang=Java}*/


  /** 
   *  <p>Notifies that on of the triggers specified by {@link TimeLagUpdateHandlerData#getUpdateTriggerList()} has been raised. The <code>TimeLagUpdateMeasures</code> allow to access the time manager, for instance, to determine the current time lag restriction of the respective activity, or to update the time lag restriction with a new value and to set a new set of trigger events.</p>
   *  
   *  @param start A reference to the activity from which the respective time lag originates from.
   *  @param end A reference to the activity to which the respective time lag leads to.
   *  @param trigger The trigger that has been raised; The trigger corresponds to an event within the process
   *  @param updateMeasures Update measures that can be used to update the time lag restrictions, e.g. set a new maximum time lag restriction.
   *  @param session The session token that must be used for accessing the given {@link TimeLagUpdateMeasures} object and only for it. When e.g. other services are accessed through the registry, a new session token must be used.
   *  @param timeManager The Time Manager
   */
  public void updateTimeLag(EventDescription start, EventDescription end, TimeLagUpdateProcedureData configuration, EventReference trigger, TimeLagUpdateMeasures measures, SessionToken session);

  /** 
   *  <p>Gets the (unique) ID of this time lag update handling procedure to identify it as specified by the time edge.</p>
   *  
   *  @returns The (unique) ID of this time lag update handler procedure.
   */
  public String getID();

}