package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.eventmodel.EventDescription;

/** 
 *  <p><code>TimeLagUpdateMeasures</code> provides the means to update the time lag restriction betwenn two activities (and prevents accessing arbitrary functions of the time manager). It allows to update the minimum, maximum and expected time lag as well as the list of triggers responsible or triggering the respective {@link TimeLagUpdateHandler}.</p>
 */
public interface TimeLagUpdateMeasures {
  /* {src_lang=Java}*/


  /** 
   *  Changes the minimum time lag restriction of the specified time edge instance.
   *  
   *  @param session The session token to be used for updating the time edge. It is the same token as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param start  The activity event the time edge to set the minimum time lag for starts in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param end  The activity event the time edge to set the minimum time lag for ery in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param timeLag The new minimum time lag restriction of the time edge
   *  @param ignoreInconsistency This indicates whether inconistencies in the time model resulting from this update should be accepted or if the update should be rejected in case of a resulting inconsistency.
   *  
   *  @returns Whether the update was successfull or would lead the time model of the process in an inconsistent state.
   *  @throws IllegalArgumentException if the specified minimum time lag value is invalid
   */
  public boolean updateMinimumTimeLag(SessionToken session, EventDescription start, EventDescription end, TimePoint timeLag, boolean ignoreInconsistency);

  /** 
   *  Changes the maximum time lag restriction of the specified time edge instance.
   *  
   *  @param session The session token to be used for updating the time edge. It is the same token as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param start  The activity event the time edge to set the maximum time lag for starts in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param end  The activity event the time edge to set the maximum time lag for ery in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param timeLag The new maximum time lag restriction of the time edge
   *  @param ignoreInconsistency This indicates whether inconistencies in the time model resulting from this update should be accepted or if the update should be rejected in case of a resulting inconsistency.
   *  
   *  @returns Whether the update was successfull or would lead the time model of the process in an inconsistent state.
   *  @throws IllegalArgumentException if the specified maximum time lag value is invalid
   */
  public boolean updateMaximumTimeLag(SessionToken session, EventDescription start, EventDescription end, TimePoint timeLag, boolean ignoreInconsistency);

  /** 
   *  Changes the expected time lag of the specified time edge instance.
   *  
   *  @param session The session token to be used for updating the time edge. It is the same token as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param start  The activity event the time edge to set the expected time lag for starts in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param end  The activity event the time edge to set the expected time lag for ery in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param timeLag The new expected time lag of the time edge
   *  @param ignoreInconsistency This indicates whether inconistencies in the time model resulting from this update should be accepted or if the update should be rejected in case of a resulting inconsistency.
   *  
   *  @returns Whether the update was successfull or would lead the time model of the process in an inconsistent state.
   *  @throws IllegalArgumentException if the specified expected time lag value is invalid
   */
  public boolean updateExpectedTimeLag(SessionToken session, EventDescription start, EventDescription end, TimePoint timeLag);

  /** 
   *  <p>Adds the specified trigger to the time edge's list of update triggers.</p>
   *  
   *  <p>
   *  Note:<br />
   *  If the update trigger list already contain the given item, it should be ignored and this incident be properly logged, so it doesn't go unnoticed.</p>
   *  
   *  
   *  @param session The session token to be used for updating the time edge's trigger list. It is the same token as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param start  The activity event the time edge  to update the trigger list for starts in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param end  The activity event the time edge  to update the trigger list for ends in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param trigger The process event reference to be added from the list of update triggers
   */
  public void addUpdateTrigger(SessionToken session, EventDescription start, EventDescription end, EventDescription trigger);

  /** 
   *  <p>Removes the specified trigger to the time edge's list of update triggers.</p>
   *  
   *  <p>
   *  Note:<br />
   *  If the update trigger list doesn't contain the given item, it should be ignored and this incident be properly logged, so it doesn't go unnoticed.</p>
   *  
   *  
   *  @param session The session token to be used for updating the time edge's trigger list. It is the same token as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param start  The activity event the time edge  to update the trigger list for starts in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param end  The activity event the time edge  to update the trigger list for ends in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param trigger The process event reference to be removed from the list of update triggers
   */
  public void removeUpdateTrigger(SessionToken session, EventDescription start, EventDescription end, EventDescription trigger);

  /** 
   *  <p>Clears the time edges's list of update triggers.</p>
   *  
   *  
   *  @param session The session token to be used for updating the time edge's trigger list. It is the same token as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param start  The activity event the time edge  to update the trigger list for starts in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   *  @param end  The activity event the time edge  to update the trigger list for ends in. It is the same activity event as provided to {@link TimeLagUpdateHandler#updateTimeLag()}
   */
  public void clearUpdateTriggerList(SessionToken session, EventDescription start, EventDescription end);

}