package de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.BasicCalendar;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.BasicCalendarExpression;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.CalendarExpression;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.Dicing;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.Ident;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.IntervalSelection;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.MultiSelection;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.SetSelection;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.SingleSelection;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.Slicing;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.Union;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.AssignementStmt;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.CalendarScript;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.CalendarStatement;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.ReturnExpression;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.ReturnStmt;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.ReturnString;

public class Evaluator
{
  public static final SimpleDateFormat format = new SimpleDateFormat("dd.MM.yy hh:mm:ss.S");

//  private static Evaluator instance;
  
  private Map<Ident, CalendarCache> variabes = new HashMap<Ident, CalendarCache>();
  
  public Evaluator(Map<String, CalendarExpression> knownCalendars)
  {
    if (knownCalendars != null)
    {
      for (Entry<String, CalendarExpression> cal : knownCalendars.entrySet())
      {
        registerIdent(new Ident(cal.getKey()), cal.getValue());
      }
    }
  }

//  public static Evaluator getInstance()
//  {
//    if (instance == null)
//    {
//      instance = new Evaluator(null);
//    }
//    return instance;
//  }

  public IntervalSet evaluate(CalendarScript script, Calendar from, Calendar until, boolean inclusive)
  {
//    IntervalSet set = null;
    for (CalendarStatement stmt : script.getStmtList())
    {
      if (stmt instanceof ReturnStmt)
      {
        return evaluateReturnStmt((ReturnStmt) stmt, from, until, inclusive);
      }
      else
      {
        IntervalSet set = evaluate(stmt, from, until, inclusive);
        if (set != null) // We have reached a return statement somewhere
          return set;
      }
    }
    
    return null;
  }
  
  public void registerIdent(Ident ident, CalendarExpression expression)
  {
    variabes.put(ident, new CalendarCache(expression));
  }
  
  public void unregisterIdent(Ident ident)
  {
    variabes.remove(ident);
  }
  
  
  protected IntervalSet evaluate(CalendarStatement stmt, Calendar from, Calendar until, boolean inclusive)
  {
    if (stmt instanceof AssignementStmt)
    {
      evaluateAssignment((AssignementStmt)stmt, from, until, inclusive);
      return null;
    } 
    else if (stmt instanceof ReturnStmt)
    {
      return evaluateReturnStmt((ReturnStmt)stmt, from, until, inclusive);
    }
    
    throw new IllegalArgumentException(stmt.getClass().getSimpleName() + " not supported!");
  }

  
  protected IntervalSet evaluateReturnStmt(ReturnStmt stmt, Calendar from, Calendar until, boolean inclusive)
  {
    if (stmt instanceof ReturnString)
    {
      System.out.println(((ReturnString) stmt).getIdentString());
      return new ContinuousIntervalSet();
    }
    else if (stmt instanceof ReturnExpression)
    {
      ReturnExpression expression = (ReturnExpression) stmt;
      CalendarExpression exp = expression.getCalendarExpression();
      return evaluate(exp, from, until, inclusive);
    }

    throw new IllegalArgumentException(stmt.getClass().getSimpleName() + "not supported!");
  }


  protected void evaluateAssignment(AssignementStmt stmt, Calendar from, Calendar until, boolean inclusive)
  {
//    IntervalSet evaluate = evaluate(stmt.getCalendarExpression(), from, until, inclusive);

    CalendarExpression expression = stmt.getCalendarExpression();
    Ident ident = stmt.getIdent();
    registerIdent(ident, expression);
  }




  public TimeFrame nextInterval(CalendarExpression expression, Calendar from, boolean inclusive)
  {
    Calendar start = from;
    while (true)
    {
      IntervalSet intervalSet = evaluate(expression, start, start, true);
      
      for (TimeFrame tf : intervalSet.getTimeFrames())
      {
        if (inclusive)
        {
          if (tf.getEnd().compareTo(from) >= 0)
          {
            return tf;
          }
  
        }
        else
        {
          if (tf.getStart().compareTo(from) >= 0 && tf.getEnd().compareTo(from) >= 0)
          {
            return tf;
          }
        }
      }
      start = (Calendar) intervalSet.getEnd().clone();
      start.add(intervalSet.getGranule().getCalendarConstant(), 1);
    }
    
    
//    return null;
  }

  public IntervalSet evaluate(CalendarExpression expression, Calendar from, Calendar until, boolean inclusive)
  {
    
    Ident ident = new Ident(expression.toString());
    CalendarCache cache = variabes.get(ident);
    if (cache == null)
    {
      cache = new CalendarCache(expression);
      variabes.put(ident, cache);
    }
    
    return cache.evaluate(this, from, until, inclusive);
  }
  

  public TimeFrame flatten(TimeFrame intervalSet)
  {
    if (intervalSet instanceof Interval)
      return intervalSet;
    if (intervalSet instanceof ContinuousIntervalSet)
      return intervalSet;
    else if (intervalSet instanceof ComplexIntervalSet)
    {
      IntervalSet set = (IntervalSet) intervalSet;
      ComplexIntervalSet result = new ComplexIntervalSet(set.getGranule());

      TimeFrame last = null;
      for (TimeFrame tf : set.getTimeFrames())
      {
        TimeFrame flatten = flatten(tf);
        
        if (flatten instanceof ComplexIntervalSet)
        {
          if (last != null)
          {
            add(result, last);
            last = null;
          }
          add(result, flatten);
        }
        else
        {
          if (last == null)
          {
            last = flatten;
          }
          else
          {
            Calendar clone = (Calendar) last.getEnd().clone();
            clone.add(Calendar.MILLISECOND, 1);
            if (clone.compareTo(flatten.getStart()) == 0)
            {
              last = new ContinuousIntervalSet(last.getGranule(), last.getStart(), flatten.getEnd());
            }
            else
            {
              add(result, last);
              last = flatten;
            }
          }
        }
      }
      
      if (last != null)
      {
        add(result, last);
      }
      
      return result;
    }

    return intervalSet; 
  }

  protected IntervalSet evaluteNoCache(CalendarExpression expression, Calendar from, Calendar until, boolean inclusive)
  {
    if (expression instanceof Dicing)
    {
      return evaluateDicing((Dicing)expression, from, until, inclusive);
    } 
    else if (expression instanceof BasicCalendarExpression)
    {
      return evaluateBasicCalendar((BasicCalendarExpression)expression, from, until, inclusive);
    } 
    else if (expression instanceof Ident)
    {
      return evaluateIdent((Ident)expression, from, until, inclusive);
    }
    else if (expression instanceof Slicing)
    {
      return evaluateSlicing((Slicing)expression, from, until, inclusive);
    }
    else if (expression instanceof Union)
    {
      return evaluateUnion(expression, from, until, inclusive);
    }
    throw new IllegalArgumentException("Unkown expression type "+expression);
  }

  private IntervalSet evaluateUnion(CalendarExpression expression, Calendar from, Calendar until, boolean inclusive)
  {
    Union union = (Union) expression;
    IntervalSet part1 = evaluate(union.getFirst(), from, until, inclusive);
    IntervalSet part2 = evaluate(union.getSecond(), from, until, inclusive);
    
    if (part1.getGranule() != part2.getGranule())
      throw new IllegalArgumentException(part1.getGranule() + "!=" + part2.getGranule());
    
    ComplexIntervalSet result = new ComplexIntervalSet(part1.getGranule());
    int i = 0;
    int j = 0;
    while (i < part1.size() && j < part2.size())
    {
//      ComplexIntervalSet s = new ComplexIntervalSet(part1.getGranule());
      TimeFrame t1 = part1.get(i);
      TimeFrame t2 = part2.get(j);
      
      if (t1.getEnd().before(t2.getStart())) // t1 ends before t2 starts
      {
        add(result, t1);
        i++;
        continue;
      }
      else if (t2.getEnd().before(t1.getStart())) // t2 ends before t2 starts
      {
        add(result, t2);
        j++;
        continue;
      }
      else // Overlapping
      {
        if ((t1 instanceof Interval || t1 instanceof ContinuousIntervalSet) 
            && (t2 instanceof Interval || t2 instanceof ContinuousIntervalSet))
        {
          Calendar start = t1.getStart().compareTo(t2.getStart()) <= 0 ? t1.getStart() : t2.getStart();
          Calendar end = t1.getEnd().compareTo(t2.getEnd()) >= 0 ? t1.getEnd() : t2.getEnd();
          result.add(new ContinuousIntervalSet(t1.getGranule(), start, end));
        }
        else // one is ComplexIntervalSet
        {
          if (t1 instanceof ComplexIntervalSet)
          {
            throw new UnsupportedOperationException("Not yet implemented");
          }
          else if (t2 instanceof ComplexIntervalSet)
          {
            throw new UnsupportedOperationException("Not yet implemented");
          }
          else
          {
            throw new IllegalStateException("Illegal combination of intervals detected.");
          }
        }
        i++;
        j++;
      }
    }
    
    for (;i < part1.size(); i++)
    {
      add(result, part1.get(i));
    }
    
    for (;j < part2.size(); j++)
    {
      add(result, part2.get(j));
    }
    
    return result;
  }
  
  private void add(ComplexIntervalSet s, TimeFrame t1)
  {
    if (t1 instanceof ComplexIntervalSet)
    {
      addAll(s, (IntervalSet) t1);
    }
    else
    {
     s.add(t1);
    }
  }

  protected IntervalSet evaluateSlicing(Slicing expression, Calendar from, Calendar until, boolean inclusive)
  {
    IntervalSet result = evaluate(expression.getCalendarExpression(), from, until, inclusive);
    
    result = performSlicing(expression, result);

    return result;
  }

  protected IntervalSet evaluateIdent(Ident ident, Calendar from, Calendar until, boolean inclusive)
  {
    CalendarCache calendarCache = variabes.get(ident);
    if (calendarCache == null)
      throw new IllegalArgumentException("Unkown variable '"+ident.getId()+"'");

    return calendarCache.evaluate(this, from, until, inclusive);
  }
  
  protected IntervalSet evaluateDicing(Dicing expression, Calendar from, Calendar until, boolean inclusive)
  {
    
    IntervalSet secondSet = evaluate(expression.getSecondOperand(), from, until, inclusive); // This is the more coarse grain calendar
    
    
    
    IntervalSet result = null;
    switch (expression.getSetOperator())
    {
      case contains:
        throw new UnsupportedOperationException("Operation "+ expression.getSetOperator()+ " not supported");
        
      case during:
      {
        
        IntervalSet firstSet = evaluate(expression.getFirstOperand(), secondSet.getStart(), secondSet.getEnd(), true);
        result = evaluateDuring(firstSet, secondSet); // During has the same result for strict and relaxed
      }
        break;
      case lessThan:
        throw new UnsupportedOperationException("Operation "+ expression.getSetOperator()+ " not supported");
      case lessThanEqual:
        throw new UnsupportedOperationException("Operation "+ expression.getSetOperator()+ " not supported");
      case meets:
        throw new UnsupportedOperationException("Operation "+ expression.getSetOperator()+ " not supported");
      case moreThan:
        throw new UnsupportedOperationException("Operation "+ expression.getSetOperator()+ " not supported");
      case moreThanEqual:
        throw new UnsupportedOperationException("Operation "+ expression.getSetOperator()+ " not supported");
      case overlaps:
        IntervalSet firstSet = evaluate(expression.getFirstOperand(), secondSet.getStart(), secondSet.getEnd(), true);
        result = evaluateOverlaps(firstSet, secondSet, expression.isStrict()); 
        break;
      default:
        throw new UnsupportedOperationException("Operation "+ expression.getSetOperator()+ " not supported");
      
    }
    
    
    
    return result;
  }
  
  protected IntervalSet performSlicing(Slicing slicing, IntervalSet result)
  {
    if (slicing instanceof SingleSelection)
    {
      return performSingleSlicing((SingleSelection)slicing, result);
    }
    else if (slicing instanceof IntervalSelection)
    {
      return performIntervalSelection((IntervalSelection)slicing, result);
    }
    else if (slicing instanceof SetSelection)
    {
      return performSetSlicing((SetSelection)slicing, result);
    }
    else if (slicing instanceof MultiSelection)
    {
      return performMultiSelection((MultiSelection)slicing, result);
    }
      
    
    throw new IllegalArgumentException(slicing.getClass().getSimpleName() + " not supported!");
  }
  
  protected IntervalSet performIntervalSelection(IntervalSelection slicing, IntervalSet set)
  {
    int startIndex = slicing.getStartIndext();
    int endIndex = slicing.getEndIndex();
    ComplexIntervalSet result = new ComplexIntervalSet(set.getGranule());
    for (TimeFrame frame : set.getTimeFrames())
    {
      if (frame instanceof IntervalSet)
      {
        IntervalSet range = ((IntervalSet) frame).range(startIndex-1, endIndex-1);
        
        result.add(range);        
      }
      else
      {
        throw new IllegalArgumentException("Cannot apply interval selection to "+frame.getClass().getSimpleName());
       
      }
    }
    return result;
  }

  protected IntervalSet performMultiSelection(MultiSelection slicing, IntervalSet result)
  {
    
    // TODO Auto-generated method stub
    throw new UnsupportedOperationException("Unsupported Operation!");
  }

  protected IntervalSet performSetSlicing(SetSelection slicing, IntervalSet set)
  {
    List<Integer> selections = slicing.getSetSelection();
    
    ComplexIntervalSet result = new ComplexIntervalSet(set.getGranule());
    
    for (Integer s : selections)
    {

      int index = s.intValue();
//      ComplexIntervalSet result = new ComplexIntervalSet(set.getGranule());
      for (TimeFrame frame : set.getTimeFrames())
      {
        if (frame instanceof IntervalSet)
        {
          TimeFrame timeFrame;
          if (index > 0)
          {
            timeFrame = ((IntervalSet) frame).get(index-1);
          }
          else
          {
            timeFrame = ((IntervalSet) frame).get(((IntervalSet) frame).size()+index); // == -index
          }
          result.add(timeFrame);
        }
        else
        {
          if (result.size() > 0)
            throw new IllegalArgumentException("Set slicing to possible for "+frame.getClass().getSimpleName()+" with size > 0");
          result.add(set.get(index));
          break;
        }
      }
      
//    IntervalSet res = performSlicing(s, set);
//         
//      if (result == null)
//      {
//        result = new ComplexIntervalSet(res.getGranule());
//        for (TimeFrame t : res.getTimeFrames())
//        {
//          ComplexIntervalSet is = new ComplexIntervalSet(t.getGranule());
//          is.add(t);
//          result.add(is);
//        }
//        
//      }
//      else
//        result = merge(result, res);
    }
    
    return result;
  }
  
  protected ComplexIntervalSet merge(IntervalSet set1, IntervalSet set2)
  {
    if (set1.size() != set2.size())
      throw new IllegalArgumentException("Merge not possible: set1.size() != set2.size()");
    
    if (set1.getGranule() != set2.getGranule())
      throw new IllegalArgumentException(set1.getGranule() + "!=" + set2.getGranule());
    
    ComplexIntervalSet result = new ComplexIntervalSet(set1.getGranule());
    
    for (int i = 0; i < set1.size(); i++)
    {
      ComplexIntervalSet s = new ComplexIntervalSet(set1.getGranule());
      TimeFrame t1 = set1.get(i);
      if (t1 instanceof IntervalSet)
      {
        addAll(s, (IntervalSet) t1);
      }
      else
      {
       s.add(t1);
      }
      TimeFrame t2 = set2.get(i);
      if (t2 instanceof IntervalSet)
      {
        addAll(s, (IntervalSet) t2);
      }
      else
      {
       s.add(t2);
      }
      result.add(s);
    }
    
    return result;
  }

  protected void addAll(ComplexIntervalSet s, IntervalSet intervalSet)
  {
    for (TimeFrame t : intervalSet.getTimeFrames())
    {
      s.add(t);
    }
  }

  protected IntervalSet performSingleSlicing(SingleSelection slicing, IntervalSet set)
  {
   
    switch (slicing.getSelectionFunction())
    {
      case any:
      {
        ComplexIntervalSet result = null;
        for (TimeFrame frame : set.getTimeFrames()) 
        {
          if (result == null)
          {
            result = new ComplexIntervalSet(frame.getGranule());
          }
          
          if (frame instanceof IntervalSet)
          {
            for (TimeFrame fr : ((IntervalSet)frame).getTimeFrames())
            {
              result.add(fr);
            }
          }
          else
          {
            throw new IllegalArgumentException("Any selection not supported for "+frame.getClass().getSimpleName());
          }
        }
        return result;
      }
      case n:
      {
        int index = slicing.getIntSelectionFunction();
        ComplexIntervalSet result = null;
        for (TimeFrame frame : set.getTimeFrames())
        {
          if (result == null)
          {
            result = new ComplexIntervalSet(frame.getGranule());
          }
          
          if (frame instanceof IntervalSet)
          {
            TimeFrame timeFrame;
            if (index > 0)
            {
              timeFrame = ((IntervalSet) frame).get(index-1);
            }
            else
            {
              timeFrame = ((IntervalSet) frame).get(((IntervalSet) frame).size()+index); // == -index
            }
            result.add(timeFrame);
          }
          else
          {
            if (result.size() > 0)
              throw new IllegalArgumentException("n-th selection not supported for "+frame.getClass().getSimpleName()+" with size > 0");
            result.add(set.get(index-1));
            break;
          }
        }
        return result;
      }
      case the:
      {
        if (set.size() == 1)
        {
          TimeFrame frame = set.get(0);
          if (frame instanceof IntervalSet)
          {
            return (IntervalSet)frame;
          }
          else
          {
            IntervalSet result = new ContinuousIntervalSet(frame.getGranule(), frame.getStart(), frame.getEnd());
            return result;
          }
        }
        else
        {
          throw new IllegalArgumentException("the selection not supported for size != 1");
        }
      }
      default:
        throw new IllegalArgumentException(slicing.getSelectionFunction() + " not supported");
      
    }
    
//    return null;
  }

  protected IntervalSet evaluateDuring(IntervalSet firstSet, IntervalSet secondSet)
  {
    ComplexIntervalSet result = new ComplexIntervalSet(secondSet.getGranule());
    for (TimeFrame tf2 : secondSet.getTimeFrames())
    {
      if (tf2 instanceof Interval)
      {
        Interval if2 = (Interval) tf2;

        IntervalSet set = firstSet.range(if2.getStart(), if2.getEnd());
        
        result.add(set);
      }
      else
      {
        IntervalSet set = evaluateDuring(firstSet, (IntervalSet) tf2);
//        result.add(set);
        addAll(result, set);
      }
    }
    return result;
  }
  
  protected IntervalSet evaluateOverlaps(IntervalSet firstSet, IntervalSet secondSet, boolean strict)
  {
    ComplexIntervalSet result = new ComplexIntervalSet(secondSet.getGranule());
    for (TimeFrame tf2 : secondSet.getTimeFrames())
    {
      ComplexIntervalSet set = new ComplexIntervalSet(firstSet.getGranule());
      for (TimeFrame tf1 : firstSet.getTimeFrames())
      {
        if (strict && tf1.getStart().compareTo(tf2.getStart()) >= 0&& tf1.getEnd().compareTo(tf2.getEnd()) <= 0)
        {
          set.add(tf1);
        } 
        else if (!strict && tf1.getStart().compareTo(tf2.getEnd()) <= 0 && tf1.getEnd().compareTo(tf2.getStart()) >= 0)
        {
          set.add(tf1);
        }
      }
      result.add(set);
    }
    return result;
  }

  protected IntervalSet evaluateBasicCalendar(BasicCalendarExpression expression, Calendar from, Calendar until, boolean inclusive)
  {
//    System.out.println(expression.getBasicCalendar() + " from "+format.format(from.getTime()) + " until "+format.format(until.getTime()));
    Calendar start = (Calendar) from.clone();
    Calendar end = (Calendar) until.clone();
    
    start.getTimeInMillis(); // This is to fix a bug in calendar implementation regarding set(Calender.DAY_OF_WEEK, int)
    
    int increment = 0;
//    int offset;
    BasicCalendar basicCalendar = expression.getBasicCalendar();
    switch (basicCalendar)
    {
      case YEARS:
        start.set(Calendar.DAY_OF_YEAR, 1);
        start.set(Calendar.HOUR_OF_DAY, 0);
        start.set(Calendar.MINUTE, 0);
        start.set(Calendar.SECOND, 0);
        start.set(Calendar.MILLISECOND, 0);
        
        end.set(Calendar.DAY_OF_YEAR, 1);
        end.set(Calendar.HOUR_OF_DAY, 0);
        end.set(Calendar.MINUTE, 0);
        end.set(Calendar.SECOND, 0);
        end.set(Calendar.MILLISECOND, 0);
        
        increment = Calendar.YEAR;
        
        break;
      case MONTHS:
        start.set(Calendar.DAY_OF_MONTH, 1);
        start.set(Calendar.HOUR_OF_DAY, 0);
        start.set(Calendar.MINUTE, 0);
        start.set(Calendar.SECOND, 0);
        start.set(Calendar.MILLISECOND, 0);
        
        end.set(Calendar.DAY_OF_MONTH, 1);
        end.set(Calendar.HOUR_OF_DAY, 0);
        end.set(Calendar.MINUTE, 0);
        end.set(Calendar.SECOND, 0);
        end.set(Calendar.MILLISECOND, 0);
        
        increment = Calendar.MONTH;
        break;
      case WEEKS: 
        start.set(Calendar.DAY_OF_WEEK, start.getFirstDayOfWeek());
        start.set(Calendar.HOUR_OF_DAY, 0);
        start.set(Calendar.MINUTE, 0);
        start.set(Calendar.SECOND, 0);
        start.set(Calendar.MILLISECOND, 0);
        
        end.set(Calendar.DAY_OF_WEEK, start.getFirstDayOfWeek());
        end.set(Calendar.HOUR_OF_DAY, 0);
        end.set(Calendar.MINUTE, 0);
        end.set(Calendar.SECOND, 0);
        end.set(Calendar.MILLISECOND, 0);
        
        increment = Calendar.WEEK_OF_YEAR;
        break;
      case DAYS:
        start.set(Calendar.HOUR_OF_DAY, 0);
        start.set(Calendar.MINUTE, 0);
        start.set(Calendar.SECOND, 0);
        start.set(Calendar.MILLISECOND, 0);
        
        end.set(Calendar.HOUR_OF_DAY, 0);
        end.set(Calendar.MINUTE, 0);
        end.set(Calendar.SECOND, 0);
        end.set(Calendar.MILLISECOND, 0);
        
        increment = Calendar.DAY_OF_YEAR;
        break;
      case HOURS:
        start.set(Calendar.MINUTE, 0);
        start.set(Calendar.SECOND, 0);
        start.set(Calendar.MILLISECOND, 0);
        
        end.set(Calendar.MINUTE, 0);
        end.set(Calendar.SECOND, 0);
        end.set(Calendar.MILLISECOND, 0); 
        
        increment = Calendar.HOUR_OF_DAY;
        break;
      case MINUTES:
        start.set(Calendar.SECOND, 0);
        start.set(Calendar.MILLISECOND, 0);
        
        end.set(Calendar.SECOND, 0);
        end.set(Calendar.MILLISECOND, 0);   
        
        increment = Calendar.MINUTE;
        break;
      case SECONDS:
        start.set(Calendar.MILLISECOND, 0);        
        
        end.set(Calendar.MILLISECOND, 0);
        
        increment = Calendar.SECOND;
        break;
      default:
        break;
    }
    
    if (!inclusive && start.compareTo(from) < 0)
    {
      start.add(increment, 1);
    }
    
    if (inclusive && end.compareTo(until) < 0)
    {
      end.add(increment, 1);
    }
    
    end.add(Calendar.MILLISECOND, -1);
    
//    offset = start.get(increment);
    
    
    return new ContinuousIntervalSet(basicCalendar, start, end);
  }



  

}
