package de.aristaflow.adept2.model.timemodel.defaultimplementation;

import java.util.Calendar;

import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultCalendarElement;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimePoint;
import de.aristaflow.adept2.model.timemodel.GuardedInterval;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNAlgorithms;
import de.aristaflow.adept2.util.ArgChecks;

public class DefaultTimeModelOperations implements TimeModelOperations
{

  public int compare(CalendarElement date, long timestamp)
  {
    return date.getTimeInMillis() < timestamp ? -1 :
      (date.getTimeInMillis() == timestamp ? 0 : 1);
  }

  public int compare(TimePoint timePoint1, TimePoint timePoint2)
  {
    long t1 = convertToSeconds(timePoint1);
    long t2 = convertToSeconds(timePoint2);
    return t1 < t2 ? -1 : (t1 == t2 ? 0 : 1);
  }

  public CalendarElement add(long timeStamp, TimePoint duration)
  {
    long seconds = convertToSeconds(duration);
    long timeInMillis = timeStamp + (seconds * 1000l);
    
    return new DefaultCalendarElement(timeInMillis);
  }

  public int compare(CalendarElement date1, CalendarElement date2)
  {
    return date1.getTimeInMillis() < date2.getTimeInMillis() ? -1 :
      (date1.getTimeInMillis() == date2.getTimeInMillis() ? 0 : 1);
  }
    
  public long convertToSeconds(TimePoint timePoint)
  {
    long seconds = 0;
    
    if (timePoint.getMonths() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += timePoint.getMonths();
    }
    seconds *= 30;
    
    if (timePoint.getDays() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += timePoint.getDays();
    }
    seconds *= 24;
    
    if (timePoint.getHours() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += timePoint.getHours();
    }
    seconds *= 60;
    
    if (timePoint.getMinutes() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += timePoint.getMinutes();
    }
    seconds *= 60;
    
    if (timePoint.getSeconds() != TimePoint.UNDEFINED_TIME_POINT)
    {
      seconds += timePoint.getSeconds();
    }

    return seconds;
  }

  
  @Override
  public TimePoint convertToTimePoint(TimeDistance value)
  {
    ArgChecks.checkForNull(value, "value");
    
    if (value.isInfinity())
      return null;
    
    long months = TimePoint.UNDEFINED_TIME_POINT;
    long days = TimePoint.UNDEFINED_TIME_POINT;
    long hours = TimePoint.UNDEFINED_TIME_POINT;
    long minutes = TimePoint.UNDEFINED_TIME_POINT;
    long seconds = 0;
    
    boolean hasNotZero = false;
    if (value.getMonths() != 0)
    {
      months = value.getMonths();
      hasNotZero = true;
    }
    
    if (value.getDays() != 0 || hasNotZero)
    {
      days = value.getDays();
      hasNotZero = true;
    }
    
    if (value.getHours() != 0 || hasNotZero)
    {
      hours = value.getHours();
      hasNotZero = true;
    }
   
    if (value.getMinutes() != 0 || hasNotZero)
    {
      minutes = value.getMinutes();
      hasNotZero = true;
    }
    
    if (value.getSeconds() != 0)
    {
      seconds = value.getSeconds();
    }    
    
    return new DefaultTimePoint(seconds, minutes,
                                hours, days,
                                months);
  }

  @Override
  public CalendarElement convertToCalendarElement(TimeDistance value)
  {
    if (value.isInfinity())
      return null;
    
    return new DefaultCalendarElement(value.toMilliseconds());
  }
 
  
  public Calendar convertToCalendar(TimeDistance value)
  {
    if (value.isInfinity())
      return null;
    
    long milliseconds = value.toMilliseconds();
    
    return convertToCalendar(milliseconds);
  }

  private Calendar convertToCalendar(long milliseconds)
  {
    Calendar calendar = Calendar.getInstance();
    calendar.setTimeInMillis(milliseconds);
    return calendar;
  }
  
  public Calendar convertToCalendar(CalendarElement date)
  {
    if (date == null)
      return null;
    
    return convertToCalendar(date.getTimeInMillis());
  }
  
  public TimeDistance secondsToTimeDistance(long v)
  {
    return new DefaultTimeDistance(v % 60, (v / 60) % 60, (v / 3600) % 24, (v / 86400), 0);
  }

  public TimeDistance convertToTimeDistance(TimePoint timePoint)
  {
    return new DefaultTimeDistance(timePoint);
  }

  public CalendarElement add(long timeStamp, TimeDistance duration)
  {
    long timeInMillis = timeStamp + duration.toMilliseconds();
    
    return new DefaultCalendarElement(timeInMillis);
  }

  public TimeDistance millisecondsToTimeDistance(long milliseconds)
  {
//    long v = milliseconds;
//    long miliseconds = v % 1000;
//    v /= 1000;
//    long seconds = v % 60;
//    v /= 60;
//    long minutes = v % 60;
//    v /= 60;
//    long hours = v % 24;
//    v /= 24;
//    long days = v;
//    return new DefaultTimeDistance(miliseconds, seconds, minutes, hours, days, 0);
    return new DefaultTimeDistance(milliseconds, 0, 0, 0, 0, 0);
  }

  public GuardedInterval<TimeDistance> getRestrictableProcessDuration(TimeModel timeModel)
  {
    if (timeModel instanceof CSTNUTemplateTimeModel)
    {
//      TCN<LabeledTimePoint,LabeledEdge> baseTCN = ((CSTNUTemplateTimeModel) timeModel).getBaseTimeModel();
      return TCNAlgorithms.getRestrictableDuration((CSTNUTemplateTimeModel) timeModel);
    }
    else if (timeModel instanceof TCNInstanceTimeModel<?, ?>)
    {
//      TCN<LabeledTimePoint,LabeledEdge> baseTCN = ((CSTNUInstanceTimeModel) timeModel).getBaseTimeModel();
      return TCNAlgorithms.getRestrictableDuration((CSTNUInstanceTimeModel) timeModel);
    }
    throw new IllegalArgumentException("Time Model not supported");
  }

  @Override
  public CalendarElement add(CalendarElement timeStamp, TimePoint duration)
  {
    long seconds = convertToSeconds(duration);
    long timeInMillis = timeStamp.getTimeInMillis() + (seconds * 1000l);
    
    return new DefaultCalendarElement(timeInMillis);
  }




  
}
