package de.aristaflow.adept2.model.timemodel.tcn;





/**
 * A label denotes a particular (potentially partial) execution path of a process.
 * 
 * 
 * see e.g., Conrad et al. (2011)
 * 
 * <pre>
 * Definition 3.7 (Environment) An environment is a partial assignment of the choice 
 * variables in X, written e = {x_i = d_{ij}, ...}. An environment may have at most one assignment
 * to each variable to be consistent. A complete environment contains an assignment to every
 * choice variable in X. An empty environment provides no assignments and is written {}.
 * We denote the set of possible environments as E and the set of complete environments as
 * E_c. The length of an environment is the number of assigned variables, denoted |e|.
 * </pre>
 * 
 * @author lanz
 *
 */
public interface Label
{
  
  /**
   * 
   * see e.g., Conrad et al. (2011)
   * 
   * <pre>
   * Definition 3.5 (Choice Variables) Each choice is associated with a finite domain variable x_i. 
   * Each of these variables has a domain with size equal to the number of options
   * of that choice. An assignment is a selection of a single option for a choice, 
   * represented as an assignment to the choice's associated choice variable.
   * </pre>
   * 
   * @author lanz
   *
   */
  public interface ChoiceVariable
  {
    /**
     * 
     * @return the domain size of the {@link ChoiceVariable}
     */
    int getOptions();
    
    ObservationTimePoint getDecisionTimePoint();
  }
  
  /**
   * 
   * see e.g., Conrad et al. (2011)
   * 
   * <pre>
   * Definition 3.5 (Choice Variables) Each choice is associated with a finite domain variable x_i. 
   * Each of these variables has a domain with size equal to the number of options
   * of that choice. An assignment is a selection of a single option for a choice, 
   * represented as an assignment to the choice's associated choice variable.
   * </pre>
   * 
   * @author lanz
   *
   */
  public interface ChoiceVariableBinding
  {

    /**
     * 
     * @return the {@link ChoiceVariable} of this choice variable binding
     */
    public ChoiceVariable getVariable();
    
  }

  /**
   * 
   * A label l1 subsumes a label l2 if and only if |= (l1 => l2).
   * That is, in any case l1 is true, l2 must be true as well.
   * 
   * @param other the reference object with which to compare.
   * @return true if <code>this</code> subsumes <code>other</code>.
   */
  public boolean subsumes(Label other);
  
  /**
   * Determines whether it is possible to satisfy the labels <code>this</code> and <code>other</code>
   * at the same time.
   * 
   * @param other
   * @return true if a configuration <code>conf()</code> for <code>this</code> and <code>other</code> exists such that
   *  <pre>conf(this AND other) == true</pre>
   */
  public boolean satisfiable(Label other);
  
  /**
   * Merges the two labels into one.
   * 
   * @param other The other Label
   * @return A label consisting of the conjunction of <code>this</code> and <code>other</code>
   * @throws IllegalArgumentException in case <code>this</code> AND <code>other</code> is not satisfiable. (see {@link #satisfiable(Label)).
   */
  public Label merge(Label other) throws IllegalArgumentException;
  
  /**
   * 
   * @return The negation of this label. The negation of a label is the set of
   *         all possible labels which may be satisfied in case <code>this</code> cannot be satisfied.
   * @throws IllegalStateException in case this label is empty and cannot be
   *           negated (see {@link #isEmptyLabel()}).
   */
  public Label[] negate() throws IllegalStateException;

  /**
   * <p>
   * Removes <code>labelVariable</code> from <code>this</code>.
   * </p>
   * 
   * <p>
   * <b>Note</b> that any binding of <code>labelVariable</code> will be ignored,
   * i.e., only the variable counts. 
   * </p>
   * 
   * @param label The variable to be removed from <code>this</code>.
   * @return A new label where <code>labelVariable</code> has been removed from
   *         <code>this</code>.
   * @see #removeLabel(Label)
   */
  public Label removeVariable(ChoiceVariable labelVariable);

  /**
   * <p>
   * Removes any part of <code>label</code> from <code>this</code>.
   * </p>
   * 
   * <p>
   * <b>Note</b> that this operation is sensitive regarding the value binding of
   * the variables in both <code>this</code> and <code>label</code>. In
   * particular, if <code>label</code> has a variable <code>p</code> bound to
   * <code>1</code> and <code>this</code> contains variable <code>p</code>
   * bound to <code>2</code>, <code>p</code> will not be removed from the
   * label.<br />
   * In particular, it should be ensured that
   * <code>this.satisfiable(other)</code> holds before calling this operation.
   * Otherwise unexpected results may occur.
   * </p>
   * 
   * @param label The label to be removed from <code>this</code>.
   * @return A new label where any part of <code>label</code> has been removed
   *         from <code>this</code>.
   * @see #removeVariable(ChoiceVariable)
   */
  public Label removeLabel(Label label);
  
  /**
   * <p>
   * <b>Note</b> that any binding of <code>labelVariable</code> will be ignored,
   * i.e., only the variable counts. 
   * </p>
   * 
   * @param labelVariable
   * @return <code>true</code> if <code>this</code> contains <code>labelVariable</code>, <code>false</code> otherwise.
   */
  public boolean containsVariable(ChoiceVariable labelVariable);
  
  /**
   * 
   * @return <code>true</code> if this is the empty label, <code>false</code> otherwise.
   */
  public boolean isEmptyLabel();

  public ChoiceVariableBinding getBinding(ChoiceVariable variable);

  public Label addVariableBinding(ChoiceVariableBinding observation);

}