package de.aristaflow.adept2.model.timemodel.tcn;

import java.util.Collection;

import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;

/**
 * A TCN is a directed graph consisting of nodes and edges, where each edge may be annotated with a particular temporal
 * constraint. 
 * 
 * Note that a TCN is considered to be a simple graph. That is, between any pair of nodes at most one edge may exist.
 * In particular, if more than one edge shall exist between two nodes this has to be encoded in the value of a single 
 * edge (i.e., the value of an edge may be complex and contain the annotation of different edges). 
 * 
 * @author lanz
 *
 * @param <N> The type of nodes this TCN consists of.
 * @param <E> The type of edges this TCN consists of.
 */
public interface TCN<N extends TCNTimePoint, E extends TCNEdge<N, ?>> extends Cloneable, DirectedGraph<N, E>
{

 
  /**
   * 
   * @return The time base event of this TCN. The time base event represents the origin of time (i.e., time point 
   *  00.00.0001) and is used as reference for any absolute constraint.
   */
  public N getTimeBaseEvent();
  

  /**
   * 
   * @param from The source node of the edge in question.
   * @param to The target node of the edge in question.
   * @return Returns the edge between nodes <code>from</code> and </code>to</code> or <code>null</code> if there is no 
   *    such edge.
   */
  public E getEdge(N from, N to);

  /**
   * 
   * @return The set of all edges of this TCN.
   */
  public Collection<E> getEdges();
  
  /**
   * 
   * @return The set of all nodes of this TCN.
   */
  public Collection<N> getTimePoints();
  
  public boolean contains(N timePoint);


  
  /**
   * Determines whether consistency of this TCN has been assured (cf. {@link TCNSolver}). 
   * 
   * Note that any change to the TCN will destroy consistency until it has been re-assured.
   * 
   * @return <code>true</code> if this TCN currently is consistent, <code>false</code> otherwise.
   */
  public boolean isConsistent();


  
  TCN<N, E> clone();

  Collection<E> getIncomingEdges(N to);

  Collection<E> getOutgoingEdges(N from);
  
}
