package de.aristaflow.adept2.model.timemodel.tcn;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultInterval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractTCN;

public abstract class TCNInstanceTimeModel<N extends TCNTimePoint, E extends TCNEdge<N, ?>> 
  implements InstanceTimeModel, UpdateableInstanceTimeModel
{

  protected STN<N, E> stn;
  protected UUID templateID;
  public IterationChain iterationChain;
  
  protected TimeModelOperations timeModelOperations = new DefaultTimeModelOperations();
  private UUID instanceID;
  
  protected List<NodeMapping<N>> nodeMappings;
  
  private N processCreationNode;
  private N processStartNode;
  private N processEndNode;
  protected Map<Integer, List<URI>> updateTriggers;

  public TCNInstanceTimeModel(UUID instanceID, UUID templateID, STN<N, E> tcn, Collection<NodeMapping<N>> nodeMapping,
      N creationTimePoint, N startTimePoint, N endTimePoint, Map<Integer, List<URI>> updateTriggers)
  {
    this.instanceID = instanceID;
    this.templateID = templateID;
    this.stn = tcn;
    if (updateTriggers != null)
    {
      this.updateTriggers = new HashMap<Integer, List<URI>>(updateTriggers);
    }
    else
    {
      this.updateTriggers = Collections.emptyMap();
    }
    
    this.nodeMappings = new ArrayList<NodeMapping<N>>(nodeMapping);
    this.processCreationNode = creationTimePoint;
    this.processStartNode = startTimePoint;
    this.processEndNode = endTimePoint;
    
    
    iterationChain = getProcessCreationTimePoint().getNodeInstance().getIteration();
  }

  public UUID getBaseTemplateID()
  {
    return templateID;
  }

  public UUID getInstanceID()
  {
    return instanceID;
  }

  public Interval<TimeDistance> getInstanceDuration()
  {
    
    return getTimeFrame(getProcessStartTimePoint(), getProcessEndTimePoint());
  }

  protected Interval<TimeDistance> getTimeFrame(N from, N to)
  {
    TimeDistance minValue;
    TimeDistance maxValue;
    if (from != to)
    {
      E maxEdge = stn.getEdge(from, to);
      E minEdge = stn.getEdge(to, from);
      
      if (maxEdge != null)
      {
        maxValue = getValue(maxEdge);
      }
      else
      { // No edge no restriction...
        maxValue = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
      }
      if (minEdge != null)
      { // If there is an edge there is a value, so calling .neg() is fine.
        minValue = getValue(minEdge).neg();
      }
      else
      { // No edge no restriction...
        minValue = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
      }
    }
    else
    {
      minValue = maxValue = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    }
    
    return toTimeFrame(minValue, maxValue);
  }

  protected Interval<TimeDistance> toTimeFrame(TimeDistance minValue, TimeDistance maxValue)
  {
//    TimePoint minTimePoint = timeModelOperations.convertToTimePoint(minValue);
//    TimePoint maxTimePoint = timeModelOperations.convertToTimePoint(maxValue);
    return new DefaultInterval<TimeDistance>(minValue, maxValue);
  }
  
  protected Interval<CalendarElement> toTimePoint(TimeDistance minValue, TimeDistance maxValue)
  {
    return new DefaultInterval<CalendarElement>(timeModelOperations.convertToCalendarElement(minValue), 
        timeModelOperations.convertToCalendarElement(maxValue));
  }
  


  protected abstract TimeDistance getValue(E edge);

  public Interval<CalendarElement> getInstanceStartTimeframe()
  {   
//    NodeMapping<N> startNodeMapping = stn.getNodeMapping(getBaseTemplate().getStartNode().getID(), iterationChain);
    
    return getTimePoint(getProcessStartTimePoint());
  }

  protected N getTimeBaseEvent()
  {
    return stn.getTimeBaseEvent();
  }

  protected Interval<CalendarElement> getTimePoint(
                                           N of)
  {
    N baseEvent = getTimeBaseEvent();
    
    TimeDistance minValue;
    TimeDistance maxValue;
    if (baseEvent != of)
    {
      E maxEdge = stn.getEdge(baseEvent, of);
      E minEdge = stn.getEdge(of, baseEvent);
      
      if (maxEdge != null)
      {
        maxValue = getValue(maxEdge);
      }
      else
      { // No edge, means no restriction...
        maxValue = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
      }
      
      if (minEdge != null)
      {
        // If we have got an edge we've got a value, so calling .neg() is fine.
        minValue = getValue(minEdge).neg();
      }
      else
      { // No edge, means no restriction...
        minValue = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
      }
    }
    else
    { // Start and Target are the same event, so there is no distance...
      minValue = maxValue = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    }
    
    return toTimePoint(minValue, maxValue);
  }



  public Interval<CalendarElement> getInstanceCompletionTimeframe()
  {
//    Node endNode = template.getEndNode();
//    
//    NodeMapping<N> endNodeMapping = stn.getNodeMapping(endNode.getID(), iterationChain);
//    
//    
    return getTimePoint(getProcessEndTimePoint());
  }

  public Interval<TimeDistance> getActivityInstanceDuration(int nodeID, int iteration)
  {
   
    NodeMapping<N> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    
    return getTimeFrame(nodeMapping.getStartNode(), nodeMapping.getEndNode());
  }

  public Interval<CalendarElement> getActivityInstanceStartTimeframe(int nodeID, int iteration)
  {   
    NodeMapping<N> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    return getTimePoint(nodeMapping.getStartNode());
  }

  public Interval<CalendarElement> getActivityInstanceCompletionTimeframe(int nodeID, int iteration)
  {    
    NodeMapping<N> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    return getTimePoint(nodeMapping.getEndNode());
  }

  public Interval<TimeDistance> getTimeLag(int fromNodeID, int fromIteration, int toNodeID, int toIteration,
      RelationType relationType)
  {
    NodeMapping<N> fromNodeMapping = getNodeMapping(fromNodeID, iterationChain);
    NodeMapping<N> toNodeMapping = getNodeMapping(toNodeID, iterationChain);
    
    N fromNode;
    N toNode;
    
    switch (relationType)
    {
      case START_START:
        fromNode = fromNodeMapping.getStartNode();
        toNode = toNodeMapping.getStartNode();
        break;
      case START_END:
        fromNode = fromNodeMapping.getStartNode();
        toNode = toNodeMapping.getEndNode();
        break;
      case END_END:
        fromNode = fromNodeMapping.getEndNode();
        toNode = toNodeMapping.getEndNode();
        break;
      case END_START:
      default:
        fromNode = fromNodeMapping.getEndNode();
        toNode = toNodeMapping.getStartNode();
        break;
    }
    
    return getTimeFrame(fromNode, toNode);
  }

  /**
   * 
   * @return The node representing the process creation time point of the process represented by this TCN.
   */
   public N getProcessCreationTimePoint()
   {
     return processCreationNode;
   }
   
   /**
   * 
   * @return The node representing the process start time point of the process represented by this TCN.
   */
   public N getProcessStartTimePoint()
   {
     return processStartNode;
   }
   
   /**
   * 
   * @return The node representing the process end time point of the process represented by this TCN.
   */
   public N getProcessEndTimePoint()
   {
     return processEndNode;
   }
   
   /**
   * Gets the node mapping for a particular {@link NodeInstance} described by its <code>nodeID</code> and 
   * <code>iterationChain</code>.
   * 
   * @param nodeId The ID of the node.
   * @param iterationChain The {@link IterationChain} denoting the particular iteration of the node
   * @return The node mapping of the given {@link NodeInstance}.
   */
   public NodeMapping<N> getNodeMapping(int nodeId, IterationChain iterationChain)
   {
     for (NodeMapping<N> nm : this.getNodeMapping())
     {
       if (nm.getNodeInstance().getNodeID() == nodeId && nm.getNodeInstance().getIteration().equals(iterationChain))
         return nm;
     }
     return null;
   }
   
   /**
   * 
   * @return The mapping of process nodes (to be precisely, {@link NodeInstance}s) to tcn nodes.
   */
   public Collection<NodeMapping<N>> getNodeMapping()
   {
     return this.nodeMappings;
   }
   
   /**
   * Sets the nodes representing the bounds of the process represented by this TCN.
   * 
   * @param processCreationNode The node representing the process creation time point.
   * @param processStartNode The node representing the process start time point.
   * @param processEndNode The node representing the process end time point.
   */
   public void setProcessBounds(N processCreationNode, N processStartNode, N processEndNode)
   {
     this.processCreationNode = processCreationNode;
     this.processStartNode = processStartNode;
     this.processEndNode = processEndNode;
     
   }
   
   protected void insertNodeMapping(NodeMapping<N> nodeMapping)
   {
     assert nodeMapping.checkComplete();
     
     for (N n : nodeMapping.getNodes())
     {
       ((AbstractTCN) getBaseTimeModel()).insertTimePoint(n);
     }
     
     this.nodeMappings.add(nodeMapping);
   }
   
   public void removeNode(NodeMapping<N> node)
   {
     this.nodeMappings.remove(node);
     for (N n : node.getNodes())
     {
       ((AbstractTCN) getBaseTimeModel()).removeTimePoint(n);
     }
   }

   public void addNodeMapping(NodeMapping<N> nodeMapping)
   {
     
     for (N n : nodeMapping.getNodes())
     {
       
//       int key = getNodeKey(n);
       if (!stn.contains(n))
         throw new IllegalArgumentException("Node does not exist: "+n);
     }
     
     insertNodeMapping(nodeMapping);
     
   }
  
  public TCN<N, E> getBaseSTN()
  {
    return stn;
  }

  public abstract TCN<N, ?> getBaseTimeModel();
  
  public Map<Integer, List<URI>> getUpdateTriggers()
  {
    return Collections.unmodifiableMap(this.updateTriggers);
  }
  
  public List<URI> getUpdateTrigger(int nodeID)
  {
    List<URI> list = this.updateTriggers.get(Integer.valueOf(nodeID));
    if (list != null)
      return Collections.unmodifiableList(list);
    else
      return Collections.emptyList();
  }
}
