package de.aristaflow.adept2.model.timemodel.tcn;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultInterval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractTCN;

public class TCNTemplateTimeModel<N extends TCNTimePoint, E extends TCNEdge<N, ?>> implements
    TemplateTimeModel
{
  
  
  private final STN<N, E> tcn;
  protected Template template;
  protected IterationChain iterationChain;
  
  protected TimeModelOperations timeModelOperations = new DefaultTimeModelOperations();
  private int startNodeID;
  
  
  protected List<NodeMapping<N>> nodeMappings;
  
  private N processCreationNode;
  private N processStartNode;
  private N processEndNode;
  protected Map<Integer, List<URI>> updateTriggers;

  public TCNTemplateTimeModel(Template template, STN<N, E> tcn, Collection<NodeMapping<N>> nodeMapping,
      N creationTimePoint, N startTimePoint, N endTimePoint, Map<Integer, List<URI>> updateTriggers)
  {
    this.template = template;
    this.tcn = tcn;
    this.processCreationNode = creationTimePoint;
    this.processStartNode = startTimePoint;
    this.processEndNode = endTimePoint;
    if (updateTriggers != null)
    {
      this.updateTriggers = new HashMap<Integer, List<URI>>(updateTriggers);// TODO clone lists...
    }
    else
    {
      this.updateTriggers = Collections.emptyMap();
    }
    this.nodeMappings = new ArrayList<NodeMapping<N>>(nodeMapping);
    
    this.startNodeID = template.getStartNode().getID();
    iterationChain = new IterationChain(template.getStartNode().getID());
    
//    assert
//    for (NodeMapping<N> mapping : this.nodeMappings)
//    {
//      mapping.checkComplete();
//    }
  }

  public UUID getTemplateID()
  {
    return template.getID();
  }

  public Template getBaseTemplate()
  {
    return template;
  }
  
  /**
  * 
  * @return The node representing the process creation time point of the process represented by this TCN.
  */
  public N getProcessCreationTimePoint()
  {
    return processCreationNode;
  }
  
  /**
  * 
  * @return The node representing the process start time point of the process represented by this TCN.
  */
  public N getProcessStartTimePoint()
  {
    return processStartNode;
  }
  
  /**
  * 
  * @return The node representing the process end time point of the process represented by this TCN.
  */
  public N getProcessEndTimePoint()
  {
    return processEndNode;
  }
  
  /**
  * Gets the node mapping for a particular {@link NodeInstance} described by its <code>nodeID</code> and 
  * <code>iterationChain</code>.
  * 
  * @param nodeId The ID of the node.
  * @param iterationChain The {@link IterationChain} denoting the particular iteration of the node
  * @return The node mapping of the given {@link NodeInstance}.
  */
  public NodeMapping<N> getNodeMapping(int nodeId, IterationChain iterationChain)
  {
    for (NodeMapping<N> nm : this.getNodeMapping())
    {
      if (nm.getNodeInstance().getNodeID() == nodeId && nm.getNodeInstance().getIteration().equals(iterationChain))
        return nm;
    }
    return null;
  }
  
  public NodeMapping<N> getNodeMapping(N event)
  {
    for (NodeMapping<N> nm : this.getNodeMapping())
    {
      if (nm.getNodes().contains(event))
        return nm;
    }
    return null;
  }
  
  /**
  * 
  * @return The mapping of process nodes (to be precisely, {@link NodeInstance}s) to tcn nodes.
  */
  public Collection<NodeMapping<N>> getNodeMapping()
  {
    return this.nodeMappings;
  }
  
  /**
  * Sets the nodes representing the bounds of the process represented by this TCN.
  * 
  * @param processCreationNode The node representing the process creation time point.
  * @param processStartNode The node representing the process start time point.
  * @param processEndNode The node representing the process end time point.
  */
  public void setProcessBounds(N processCreationNode, N processStartNode, N processEndNode)
  {
    this.processCreationNode = processCreationNode;
    this.processStartNode = processStartNode;
    this.processEndNode = processEndNode;
    
  }
  
  protected void insertNodeMapping(NodeMapping<N> nodeMapping)
  {
    assert nodeMapping.checkComplete();
    
    for (N n : nodeMapping.getNodes())
    {
      ((AbstractTCN) getBaseTimeModel()).insertTimePoint(n);
    }
    
    this.nodeMappings.add(nodeMapping);
  }
  
  public void removeNode(NodeMapping<N> node)
  {
    this.nodeMappings.remove(node);
    for (N n : node.getNodes())
    {
      ((AbstractTCN) getBaseTimeModel()).removeTimePoint(n);
    }
  }

  public void addNodeMapping(NodeMapping<N> nodeMapping)
  {
    
    for (N n : nodeMapping.getNodes())
    {
      
//      int key = getNodeKey(n);
      if (!tcn.contains(n))
        throw new IllegalArgumentException("Node does not exist: "+n);
    }
    
    insertNodeMapping(nodeMapping);
    
  }


  public Interval<TimeDistance> getProcessDuration()
  {
    Node startNode = template.getStartNode();
    Node endNode = template.getEndNode();
    NodeMapping<N> startNodeMapping = getNodeMapping(startNode.getID(), iterationChain);
    NodeMapping<N> endNodeMapping = getNodeMapping(endNode.getID(), iterationChain);
    
    return getTimeFrame(startNodeMapping.getEndNode(), endNodeMapping.getStartNode());
  }

  protected TimeDistance getValue(E edge)
  {
    if (edge == null)
      return null;
    Object value = edge.getValue();
    if (value instanceof LabeledValueSet)
    {
      LabeledValueSet valueSet = (LabeledValueSet) value;
      assert valueSet.getLabeledValues().size() == 1;
      LabeledValue label = valueSet.getLabeledValues().iterator().next();
      return label.getValue();
    }
    else if (value instanceof SimpleConstraintValue)
    {
      return ((SimpleConstraintValue) value).getValue();
    }
      
    throw new IllegalArgumentException();
  }

  public Interval<TimePoint> getProcessStartTimeframe()
  {
    N timeBase = getTimeBaseEvent();
   
    NodeMapping<N> startNodeMapping = getNodeMapping(startNodeID, iterationChain);
    
    
    return getTimePoint(timeBase, startNodeMapping.getEndNode());
  }

  private N getTimeBaseEvent()
  {
    NodeMapping<N> startNodeMapping = getNodeMapping(startNodeID, iterationChain);
    return startNodeMapping.getStartNode();
  }



  public Interval<TimePoint> getProcessCompletionTimeframe()
  {
    Node endNode = template.getEndNode();
    
    N timeBase = getTimeBaseEvent();
    
    NodeMapping<N> endNodeMapping = getNodeMapping(endNode.getID(), iterationChain);
    
    
    return getTimePoint(timeBase, endNodeMapping.getStartNode());
  }

  public Interval<TimeDistance> getActivityDuration(int nodeID)
  {
   
    NodeMapping<N> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    
    return getTimeFrame(nodeMapping.getStartNode(), nodeMapping.getEndNode());
  }

  public Interval<TimePoint> getActivityStartTimeframe(int nodeID)
  {
    N timeBase = getTimeBaseEvent();
    
    NodeMapping<N> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    return getTimePoint(timeBase, nodeMapping.getStartNode());
  }

  public Interval<TimePoint> getActivityCompletionTimeframe(int nodeID)
  {
    N timeBase = getTimeBaseEvent();
    
    NodeMapping<N> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    return getTimePoint(timeBase, nodeMapping.getEndNode());
  }

  public Interval<TimeDistance> getTimeLag(int fromNodeID, int toNodeID,
      RelationType relationType)
  {
    NodeMapping<N> fromNodeMapping = getNodeMapping(fromNodeID, iterationChain);
    NodeMapping<N> toNodeMapping = getNodeMapping(toNodeID, iterationChain);
    
    N fromNode;
    N toNode;
    
    switch (relationType)
    {
      case START_START:
        fromNode = fromNodeMapping.getStartNode();
        toNode = toNodeMapping.getStartNode();
        break;
      case START_END:
        fromNode = fromNodeMapping.getStartNode();
        toNode = toNodeMapping.getEndNode();
        break;
      case END_END:
        fromNode = fromNodeMapping.getEndNode();
        toNode = toNodeMapping.getEndNode();
        break;
      case END_START:
      default:
        fromNode = fromNodeMapping.getEndNode();
        toNode = toNodeMapping.getStartNode();
        break;
    }
    
    return getTimeFrame(fromNode, toNode);
  }

  public STN<N, E> getBaseSTN()
  {
    return tcn;
  }
  
  protected Interval<TimePoint> toInterval(TimeDistance minValue, TimeDistance maxValue)
  {
    assert minValue != null;
    assert maxValue != null;
    TimePoint min = timeModelOperations.convertToTimePoint(minValue);
    TimePoint max = timeModelOperations.convertToTimePoint(maxValue);
    return new DefaultInterval<TimePoint>(min, max);
  }
  
  protected Interval<TimePoint> getTimePoint(N from, N to)
  {
    TimeDistance minValue;
    TimeDistance maxValue;
    if (from != to)
    {
      E maxEdge = tcn.getEdge(from, to);
      E minEdge = tcn.getEdge(to, from);

      maxValue = getValue(maxEdge);
      TimeDistance timeDistance = getValue(minEdge);
      minValue = timeDistance != null ? timeDistance.neg() : null;
    }
    else
    {
      minValue = maxValue = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    }

    return toInterval(minValue, maxValue);
  }
  
  protected Interval<TimeDistance> getTimeFrame(N from, N to)
  {
    TimeDistance minValue = null;
    TimeDistance maxValue = null;
    if (from != to)
    {
      E maxEdge = tcn.getEdge(from, to);
      E minEdge = tcn.getEdge(to, from);

      maxValue = getValue(maxEdge);
      
      TimeDistance timeDistance = getValue(minEdge);

      minValue = timeDistance != null ? timeDistance.neg() : null;
    }
    else
    {
      minValue = maxValue = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    }

    return new DefaultInterval<TimeDistance>(minValue, maxValue);
  }
  
  public TCN<N, ?> getBaseTimeModel()
  {
    return tcn;
  }

  public Map<Integer, List<URI>> getUpdateTriggers()
  {
    return Collections.unmodifiableMap(this.updateTriggers);
  }
  
  public List<URI> getUpdateTrigger(int nodeID)
  {
    List<URI> list = this.updateTriggers.get(Integer.valueOf(nodeID));
    if (list != null)
      return Collections.unmodifiableList(list);
    else
      return Collections.emptyList();
  }
  

}
