package de.aristaflow.adept2.model.timemodel.tcn;

import static de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.processmodel.StructuredConstraintEdge;
import de.aristaflow.adept2.model.processmodel.StructuredEdge;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.ActivityDuration;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDate;
import de.aristaflow.adept2.model.processmodel.timemodel.ProcessDuration;
import de.aristaflow.adept2.model.processmodel.timemodel.TimeLag;
import de.aristaflow.adept2.model.processmodel.tools.ProcessModelTools;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultInterval;
import de.aristaflow.adept2.util.Tuple;

public abstract class TCNTransformation<N extends TCNTimePoint, E extends TCNEdge<N, ?>>
{
  
  public static final String CONF_PREFIX = "TCNTransformation.";
  
  public static final  String CONF_NODE_MIN_DURATION = CONF_PREFIX + "NodeMinDuration";
  public static final  String CONF_NODE_MAX_DURATION = CONF_PREFIX + "NodeMaxDuration";
  
  public static final  String CONF_STRUCTURE_NODE_MIN_DURATION = CONF_PREFIX + "StructureNodeMinDuration";
  public static final  String CONF_STRUCTURE_NODE_MAX_DURATION = CONF_PREFIX + "StructureNodeMaxDuration";
  
  public static final  String CONF_STRUCTURE_NODE_MIN_TIME_LAG = CONF_PREFIX + "StructureNodeMinTimeLag";
  public static final  String CONF_STRUCTURE_NODE_MAX_TIME_LAG = CONF_PREFIX + "StructureNodeMaxTimeLag";
  
  public static final  String CONF_CONTROL_FLOW_MIN_TIME_LAG = CONF_PREFIX + "ControlFlowMinTimeLag";
  public static final  String CONF_CONTROL_FLOW_MAX_TIME_LAG = CONF_PREFIX + "ControlFlowMaxTimeLag";
  
  public static final  String CONF_NO_INSTANCE_CREATION_EVENT = CONF_PREFIX + "NoInstanceCreationEvent";
  public static final  String CONF_NO_DISTINCT_TIMEBASE_EVENT = CONF_PREFIX + "NoDistinctTimeBaseEvent";
  
  public static  TimeDistance DEFAULT_NODE_MIN_DURATION = DefaultTimeDistance.EPSILON_TIME_DISTANCE;
  public static  TimeDistance DEFAULT_NODE_MAX_DURATION = POSITIVE_INFINITE_TIME_DISTANCE;
  
  public static  TimeDistance DEFAULT_STRUCTURE_NODE_MIN_DURATION = DefaultTimeDistance.EPSILON_TIME_DISTANCE;
  public static  TimeDistance DEFAULT_STRUCTURE_NODE_MAX_DURATION = new DefaultTimeDistance(5, 0, 0, 0, 0);
  
  public static  TimeDistance DEFAULT_STRUCTURE_NODE_MIN_TIME_LAG = DefaultTimeDistance.EPSILON_TIME_DISTANCE;
  public static  TimeDistance DEFAULT_STRUCTURE_NODE_MAX_TIME_LAG = DefaultTimeDistance.ONE_TIME_DISTANCE;
  
  public static  TimeDistance DEFAULT_CONTROL_FLOW_MIN_TIME_LAG = DefaultTimeDistance.EPSILON_TIME_DISTANCE;
  public static  TimeDistance DEFAULT_CONTROL_FLOW_MAX_TIME_LAG = POSITIVE_INFINITE_TIME_DISTANCE;

  protected TimeDistance NODE_MIN_DURATION;
  protected TimeDistance NODE_MAX_DURATION;
  
  protected TimeDistance STRUCTURE_NODE_MIN_DURATION;
  protected TimeDistance STRUCTURE_NODE_MAX_DURATION;
  
  protected TimeDistance STRUCTURE_NODE_MIN_TIME_LAG;
  protected TimeDistance STRUCTURE_NODE_MAX_TIME_LAG;
  
  protected TimeDistance CONTROL_FLOW_MIN_TIME_LAG;
  protected TimeDistance CONTROL_FLOW_MAX_TIME_LAG;
    
  
  protected boolean NO_INSTANCE_CREATION_EVENT;

  protected boolean NO_DISTINCT_TIMEBASE_EVENT;

  
  protected TemporalTemplate template;
  private int lookAhead;

  private List<N> nodes;

  private List<E> edges;

  private Collection<NodeMapping<N>> nodeMapping;
  private N instanceCreationTimePoint;
  private N instanceStartTimePoint;
  private N instanceEndTimePoint;
  protected Configuration configuration;

  
 

  public TCNTransformation(Configuration configuration, TemporalTemplate template)
  {
    this(configuration, template, 0);
  }

  public TCNTransformation(Configuration configuration, TemporalTemplate template, int lookAhead)
  {
   
    this.template = template;
    this.lookAhead = lookAhead;
    
    setConfiguration(configuration);
  }
  
  protected void setConfiguration(Configuration configuration)
  {
    if (configuration == null)
    {
      this.configuration = new BaseConfiguration();
    }
    else
    {
      this.configuration = configuration;
    }
      
    readConfiguration();
  }

  protected void readConfiguration()
  {
    NODE_MIN_DURATION = longToTimeDistance(configuration.getLong(CONF_NODE_MIN_DURATION,
      DEFAULT_NODE_MIN_DURATION.toMilliseconds()));
    NODE_MAX_DURATION = longToTimeDistance(configuration.getLong(CONF_NODE_MAX_DURATION,
      DEFAULT_NODE_MAX_DURATION.toMilliseconds()));

    STRUCTURE_NODE_MIN_DURATION = longToTimeDistance(configuration.getLong(CONF_STRUCTURE_NODE_MIN_DURATION,
      DEFAULT_STRUCTURE_NODE_MIN_DURATION.toMilliseconds()));
    STRUCTURE_NODE_MAX_DURATION = longToTimeDistance(configuration.getLong(CONF_STRUCTURE_NODE_MAX_DURATION,
      DEFAULT_STRUCTURE_NODE_MAX_DURATION.toMilliseconds()));

    STRUCTURE_NODE_MIN_TIME_LAG = longToTimeDistance(configuration.getLong(CONF_STRUCTURE_NODE_MIN_TIME_LAG,
      DEFAULT_STRUCTURE_NODE_MIN_TIME_LAG.toMilliseconds()));
    STRUCTURE_NODE_MAX_TIME_LAG = longToTimeDistance(configuration.getLong(CONF_STRUCTURE_NODE_MAX_TIME_LAG,
      DEFAULT_STRUCTURE_NODE_MAX_TIME_LAG.toMilliseconds()));

    CONTROL_FLOW_MIN_TIME_LAG = longToTimeDistance(configuration.getLong(CONF_CONTROL_FLOW_MIN_TIME_LAG,
      DEFAULT_CONTROL_FLOW_MIN_TIME_LAG.toMilliseconds()));
    CONTROL_FLOW_MAX_TIME_LAG = longToTimeDistance(configuration.getLong(CONF_CONTROL_FLOW_MAX_TIME_LAG,
      DEFAULT_CONTROL_FLOW_MAX_TIME_LAG.toMilliseconds()));
    
    NO_INSTANCE_CREATION_EVENT = configuration.getBoolean(CONF_NO_INSTANCE_CREATION_EVENT, false);
    NO_DISTINCT_TIMEBASE_EVENT = configuration.getBoolean(CONF_NO_DISTINCT_TIMEBASE_EVENT, false);
  }

  protected TimeDistance longToTimeDistance(long miliseconds)
  {
    return new DefaultTimeDistance(miliseconds, 0, 0, 0, 0, 0);
  }

  public TCN<N, E> transform()
  {
    nodes = new ArrayList<N>();
    edges = new ArrayList<E>();
    nodeMapping = new ArrayList<NodeMapping<N>>();
    instanceCreationTimePoint = null;
    instanceEndTimePoint = null;
    instanceStartTimePoint = null;

    
    N timeBaseNode = getTimeBaseNode();
    
    addNode(timeBaseNode);
    
    // Determine start and end node of the template
    int startNodeID = template.getStartNode().getID();
    int endNodeID = template.getEndNode().getID();
    
    Label branches = createInitialBranch();
    IterationChain iterationChain = new IterationChain(startNodeID);
    
    N instanceCreationEvent;
    if (NO_INSTANCE_CREATION_EVENT)
    {
      instanceCreationEvent = getTimeBaseNode();
    }
    else
    {
      instanceCreationEvent = transformInstanceCreationEvent(iterationChain, branches);
    }
    
    
    
    // Initialize the process structure
    Tuple<NodeMapping<N>,NodeMapping<N>> boundary = transformSequenceBlock(startNodeID, endNodeID, iterationChain,
                                                    branches);
    
    
    if (instanceCreationEvent != timeBaseNode)
    {
      // Create a positive control constraint between base event and instance creation event
      createTCNEdge(timeBaseNode, instanceCreationEvent, branches, CONTROL_FLOW_MIN_TIME_LAG, CONTROL_FLOW_MAX_TIME_LAG, true);
    }
    
    // Create a positive control constraint between instance creation event and the start event of the start node
    N startNodeStartEvent = boundary.getFirst().getStartNode(); 
    if (startNodeStartEvent != instanceCreationEvent)
    {
      createTCNEdge( instanceCreationEvent, startNodeStartEvent, branches, CONTROL_FLOW_MIN_TIME_LAG, CONTROL_FLOW_MAX_TIME_LAG, true);
    }
    
    // Process sync edges
    processSyncEdges();
    
 // Process process duration constraint
    transformProcessDuration(boundary.getFirst(), boundary.getSecond());
       
    // Process any time lags
    transformTimeLags();
    
    // Process any validity periods
    transformProcessValidityPeriod(boundary.getFirst(), boundary.getSecond());
    
    ChangeableTCN<N, E> tcn = createTCN(nodes, edges, nodeMapping);
     
    instanceCreationTimePoint = instanceCreationEvent;
    instanceStartTimePoint = boundary.getFirst().getStartNode();
    instanceEndTimePoint = boundary.getSecond().getEndNode();
    
    return tcn;
  }
  
  /**
  * 
  * @return The node representing the process creation time point of the process represented by this TCN.
  */
  public N getProcessCreationTimePoint()
  {
    return instanceCreationTimePoint;
  }
  
  /**
  * 
  * @return The node representing the process start time point of the process represented by this TCN.
  */
  public N getProcessStartTimePoint()
  {
    return instanceStartTimePoint;
  }
  
  /**
  * 
  * @return The node representing the process end time point of the process represented by this TCN.
  */
  public N getProcessEndTimePoint()
  {
    return instanceEndTimePoint;
  }

  public Collection<NodeMapping<N>> getNodeMapping()
  {
    return nodeMapping;
  }
  
  protected abstract N getTimeBaseNode();
  
  protected abstract ChangeableTCN<N, E> createTCN(Collection<N> nodes, Collection<E> edges, Collection<NodeMapping<N>> nodeMapping);
  
  
  protected Tuple<NodeMapping<N>,NodeMapping<N>> transformSequenceBlock(int startNodeID,
      int endNodeID, IterationChain iterationChain, Label branches)
  {
    // Variable to hold the start (i.e., first) node of this block
    NodeMapping<N> startNode = null;
    // Variable to hold the latest node processed
    NodeMapping<N> latestNodeProcessed = null;

    // Start with the first node of the block
    int currentNodeID = startNodeID;
    Label currentBranches = branches;
    
    while (true) 
    {
      // Determine the type of the current node
      NodeType nodeType = template.getNodeType(currentNodeID);
      switch (nodeType)
      {
        // If the current node is a split node process the whole block
        case NT_AND_SPLIT:
        case NT_XOR_SPLIT:
        case NT_STARTLOOP:
        {
          // Determine the bounds of the block
          int blockStartNode = currentNodeID;
          int blockEndNode = template.getNodeCorrespondingBlockNodeID(blockStartNode);
          
          // Process the whole block at once
          Tuple<NodeMapping<N>,NodeMapping<N>> block = transformBlock(
                                                  blockStartNode, blockEndNode, iterationChain, currentBranches);
          
          // Just some check to verify the whole block has been processed
          assert block.getFirst().getNodeInstance().getNodeID() == blockStartNode;
          assert block.getSecond().getNodeInstance().getNodeID() == blockEndNode;
          
          // If start node is not yet set, the start node of the block is also the start node of this sequence
          if (startNode == null)
          {
            startNode = block.getFirst();
          }
          
          // If we have a previous node in our sequence connect the first node of the block to it
          if (latestNodeProcessed != null)
          {
            if (NodeType.NT_AND_SPLIT.equals(nodeType) || NodeType.NT_STARTLOOP.equals(nodeType))
            {
              transformControlEdge(latestNodeProcessed, block.getFirst(), currentBranches, STRUCTURE_NODE_MIN_TIME_LAG,
                            STRUCTURE_NODE_MAX_TIME_LAG);
            }
            else
            {
              transformControlEdge(latestNodeProcessed, block.getFirst(), null);
            }
          }
          
          // Remember the last processed node
          latestNodeProcessed = block.getSecond();
          
          // Update the new branches according to the result of the block
          currentBranches = latestNodeProcessed.getSuffixNode().getLabel();

          // Check if the end node of the block is also the last node of this sequence
          if (endNodeID != blockEndNode)
          {
            // If not proceed with the next node
            currentNodeID = template.getSuccByEdgeType(blockEndNode,
                                                       EdgeType.ET_CONTROL)[0];
          }
          else
          {
            // Otherwise return the boundaries of this sequence
            return new Tuple<NodeMapping<N>,NodeMapping<N>>( startNode, latestNodeProcessed );
          }
        }
        break;
        // If the current node is the start node of the process this is the process block itself
        case NT_STARTFLOW:
        {
          // Create the process start node
          NodeMapping<N> processStartNode = transformProcessStartNode( currentNodeID, 
                                       iterationChain, currentBranches);
          
          // The process start node is always the first node of a sequence (i.e., the process)...
          startNode = processStartNode;
          latestNodeProcessed = processStartNode;
          
          // ...and it's never the last node in a sequence so proceed with the next node
          currentNodeID = template.getSuccByEdgeType(currentNodeID,
                                                     EdgeType.ET_CONTROL)[0];
        }
        break;
        // If the current node is "normal" node process it
        case NT_NORMAL:
        {
          // Create the node
          NodeMapping<N> node = transformActivity( currentNodeID, iterationChain, currentBranches);
          
          // If start node of the block is not yet set, the current node is also this start node of this sequence
          if (startNode == null)
          {
            startNode = node;
          }
          
          // If we have a previous node in our sequence connect the node to it
          if (latestNodeProcessed != null)
          {
            transformControlEdge( latestNodeProcessed, node, null);
          }
          
          // Remember the last processed node
          latestNodeProcessed = node;
          
          // Check if the current node is also the end node of this sequence
          if (currentNodeID != endNodeID)
          {
            // If not proceed with the next node
            currentNodeID = template.getSuccByEdgeType(currentNodeID,
                                                     EdgeType.ET_CONTROL)[0];
          }
          else
          {
            // Otherwise return the boundaries of this sequence
            return new Tuple<NodeMapping<N>,NodeMapping<N>>( startNode, latestNodeProcessed );
          }
        }
        break;
        // If the current node is the end node of the process this is the process block itself
        case NT_ENDFLOW:
        {
          NodeMapping<N> processEndNode = transformProcessEndNode( currentNodeID, iterationChain,
                                       currentBranches);
          
          // We always have at least a start node ...
          if (latestNodeProcessed != null) // so this is just to silence the compiler
          {
            // Constraint from last Event of the process to process end event
            transformControlEdge(latestNodeProcessed, processEndNode, 
                          null, 
                          STRUCTURE_NODE_MIN_TIME_LAG, STRUCTURE_NODE_MAX_TIME_LAG);
          }
          
          // This is always the last node of a sequence so return the boundaries
          return new Tuple<NodeMapping<N>,NodeMapping<N>>( startNode, processEndNode );
        }
        // A Join always belongs to a block so this is an error
        case NT_XOR_JOIN:
        case NT_AND_JOIN:
        case NT_ENDLOOP:
        default:
          throw new IllegalArgumentException("" + currentNodeID + " "
                                             + nodeType);
      }
    } 
  }


  protected Tuple<NodeMapping<N>,NodeMapping<N>> transformBlock(int startNodeID,
      int endNodeID, IterationChain iterationChain, Label branches)
  {
    NodeType blockType = template.getNodeType(startNodeID);

    // Determine block type and initiate appropriate transformation
    switch (blockType)
    {
      case NT_AND_SPLIT:
        return transformANDBlock( startNodeID, endNodeID, iterationChain, branches);
      case NT_STARTLOOP:
        return transformLoopBlock( startNodeID, endNodeID, iterationChain, branches);
      case NT_XOR_SPLIT:
        return transformXORBlock( startNodeID, endNodeID, iterationChain, branches);
      // In any other case this is not a split block...
      default:
        throw new IllegalArgumentException();
    }

  }
  
  protected Tuple<NodeMapping<N>,NodeMapping<N>> transformANDBlock(int startNodeID,
      int endNodeID, IterationChain iteration, Label branches)
  {
    NodeType blockType = template.getNodeType(startNodeID);
    
    if (blockType != NodeType.NT_AND_SPLIT)
      throw new IllegalArgumentException();
    
    SplitNodeMapping<N> splitNode;
    JoinNodeMapping<N> joinNode;
    
    // Create the split Node

    Label currentBranches = branches;
    splitNode = transformANDSplit( startNodeID, iteration, currentBranches);
      
    int[] succs = template.getSuccByEdgeType(startNodeID, EdgeType.ET_CONTROL);
    
    @SuppressWarnings({ "cast", "unchecked" })
    Tuple<NodeMapping<N>,NodeMapping<N>>[] branchBlocks = (Tuple<NodeMapping<N>,NodeMapping<N>>[])(new Tuple[succs.length]);
    
    Label[] newBranches = new Label[succs.length];
    
    
    for (int i = 0; i < succs.length; i++)
    {
      N splitNodeEndEvent = splitNode.getSuffixNode(i);
      int branchStartNode = succs[i];
      // empty branch?
      if (branchStartNode == endNodeID)
      {
        branchBlocks[i] = null;
        newBranches[i] = splitNodeEndEvent.getLabel();// AND-Split so exactly one endEvent
      }
      else
      {
        int branchID1 = template.getNodeBranchID(branchStartNode);
        int branchEndNode = ProcessModelTools.getBranchEndNode(template, startNodeID, branchID1);
        branchBlocks[i] = transformSequenceBlock( branchStartNode, branchEndNode, iteration, currentBranches);
        newBranches[i] = branchBlocks[i].getSecond().getSuffixNode().getLabel(); // Join Node or Normal node so exactly one endEvent
      }
    }
    
    currentBranches = mergeANDBranches(currentBranches, newBranches);

    joinNode = transformANDJoin( endNodeID, iteration, currentBranches);
    
    
    for (int i = 0; i < branchBlocks.length; i++)
    {
      
      Tuple<NodeMapping<N>,NodeMapping<N>> b = branchBlocks[i];
      N splitNodeEndEvent = splitNode.getSuffixNode(i);
      N joinNodeStartEvent = joinNode.getPrefixNode(i);
      if (b != null)
      {

        // can only be one event, i.e., can not be a join
        N branchStartNodeStartEvent = b.getFirst().getPrefixNode();
        transformControlEdge(splitNode, splitNodeEndEvent, 
                          b.getFirst(), branchStartNodeStartEvent, 
                          splitNodeEndEvent.getLabel());
        

        // can only be one event, i.e., can not be a split
        N branchEndNodeEndEvent = b.getSecond().getSuffixNode();
        transformControlEdge(b.getSecond(), branchEndNodeEndEvent, joinNode, joinNodeStartEvent,
                          branchEndNodeEndEvent.getLabel());
      }
      else
      {
        // Empty branch -> connect split to join node
        transformControlEdge(splitNode, splitNodeEndEvent, joinNode, joinNodeStartEvent,
                          splitNodeEndEvent.getLabel());
      }
    }
    
    return new Tuple<NodeMapping<N>,NodeMapping<N>>(splitNode, joinNode);
  }
  

  protected Tuple<NodeMapping<N>, NodeMapping<N>> transformXORBlock(int startNodeID, int endNodeID, IterationChain iteration,
      Label branches)
  {
    if (template.getNodeType(startNodeID) != NodeType.NT_XOR_SPLIT)
      throw new IllegalArgumentException();
    
    SplitNodeMapping<N> xorSplitNode;
    JoinNodeMapping<N> xorJoinNode;
        
    Label currentBranches = branches;
    
    // Create the split Node
    xorSplitNode = transformXORSplitNode( startNodeID, iteration, currentBranches);
    
    // Create the branches
    int[] succs = template.getSuccByEdgeType(startNodeID, EdgeType.ET_CONTROL);
    
    // Store a block and a set of branch paths for each branches
    @SuppressWarnings({ "cast", "unchecked" })
    Tuple<NodeMapping<N>, NodeMapping<N>>[] branchBlocks = (Tuple<NodeMapping<N>, NodeMapping<N>>[])(new Tuple[succs.length]);
    Label[] newBranchPaths = new Label[succs.length];
    
    // Process the branches
    for (int i = 0; i < succs.length; i++)
    {
      int branchStartNode = succs[i];
      
      N branchDecisionEvent = xorSplitNode.getSuffixNode(i);
      
      Label newBranch = branchDecisionEvent.getLabel();
      
      // empty branch?
      if (branchStartNode == endNodeID)
      {
        //long edgeCode = getEdgeCode(startNodeID, branchStartNode);
        newBranchPaths[i] = newBranch;
        branchBlocks[i] = null;
      }
      else
      {
        
        int branchID = template.getNodeBranchID(branchStartNode);
        //long edgeCode = getEdgeCode(startNodeID, succs[i]);
        int branchEndNode = ProcessModelTools.getBranchEndNode(template, startNodeID, branchID);
        Tuple<NodeMapping<N>, NodeMapping<N>> b = transformSequenceBlock( branchStartNode, branchEndNode, iteration, newBranch);
        
        // Constraint from Decision Event to first event of branch
        N branchStartEvent = b.getFirst().getPrefixNode();
        transformControlEdge(xorSplitNode, branchDecisionEvent, b.getFirst(), branchStartEvent, 
                             newBranch);
        
        branchBlocks[i] = b;
        newBranchPaths[i] = b.getSecond().getSuffixNode().getLabel();
      }
    }
    
    // Merge the XOR Branches
    currentBranches = mergeXORBranches(currentBranches, newBranchPaths);
    
    // Create the join Node
    xorJoinNode = transformXORJoinNode( endNodeID, iteration, currentBranches);
    
    // Connect the branch to the xor join 
    
    
    for (int i = 0; i < branchBlocks.length; i++)
    {
      N xorJoinStartEvent = xorJoinNode.getPrefixNode(i);
      Tuple<NodeMapping<N>, NodeMapping<N>> b = branchBlocks[i];
      if (b != null)
      {
        // Constraint from last event of branch to Join Event
        N branchEndEvent = b.getSecond().getSuffixNode();
        transformControlEdge(b.getSecond(), branchEndEvent, xorJoinNode, xorJoinStartEvent, branchEndEvent.getLabel(), STRUCTURE_NODE_MIN_TIME_LAG,
          STRUCTURE_NODE_MAX_TIME_LAG);        
      }
      else
      {
        //long edgeCode = getEdgeCode(startNodeID, succs[i]);
        // Empty branch -> connect split to join node
        N branchDecisionEvent = xorSplitNode.getSuffixNode(i);
        transformControlEdge(xorSplitNode, branchDecisionEvent, xorJoinNode, xorJoinStartEvent, branchDecisionEvent.getLabel());
      }
    }
   
    
    return new Tuple<NodeMapping<N>, NodeMapping<N>>(xorSplitNode, xorJoinNode);
  }

  


  protected Tuple<NodeMapping<N>, NodeMapping<N>> transformLoopBlock(int startNodeID, int endNodeID,
      IterationChain iteration, Label branches)
  {
    NodeType blockType = template.getNodeType(startNodeID);
    
    if (blockType != NodeType.NT_STARTLOOP)
      throw new IllegalArgumentException();
    
    
    IterationChain newIteration = iteration.subIteration(startNodeID);

    Tuple<NodeMapping<N>, NodeMapping<N>> blockNodes = transformLoopContent(startNodeID, endNodeID, 
                                                                            newIteration, branches, 
                                                                            lookAhead);
    
    
    return blockNodes;
  }
  
  protected static final int REPEAT_BRANCH = 0;
  protected static final int EXIT_BRANCH = 1;
  
  protected Tuple<NodeMapping<N>, NodeMapping<N>> transformLoopContent(
      int startNodeID, int endNodeID, IterationChain iteration,
      Label branches, int lookAhead)
  {

    
    SplitNodeMapping<N> loopStartNode;
    N latestEvent;
    NodeMapping<N> latestNode;
    JoinNodeMapping<N> loopEndNode;
    
    loopStartNode = transformLoopStartNode(startNodeID, iteration, branches);
    
    int[] succs = template.getSuccByEdgeType(startNodeID, EdgeType.ET_CONTROL);
    int branchStartNode = succs[0];
    int branchEndNode = 0;
    
    Label exitBranch = loopStartNode.getSuffixNode(EXIT_BRANCH).getLabel();
    
    latestNode = loopStartNode;
    latestEvent = loopStartNode.getSuffixNode(REPEAT_BRANCH);
    Label repeatBranch = latestEvent.getLabel();
    
    Tuple<NodeMapping<N>,NodeMapping<N>> branchBlock;
    // empty branch?
    if (branchStartNode == endNodeID)
    {
      branchBlock = null;
    }
    else
    {
      int branchID = template.getNodeBranchID(branchStartNode);
      branchEndNode = ProcessModelTools.getBranchEndNode(template, startNodeID, branchID);
      
      branchBlock = transformSequenceBlock(branchStartNode, branchEndNode, iteration, repeatBranch);
      
      N loopIterationEvent = loopStartNode.getSuffixNode(REPEAT_BRANCH);
      N branchStartEvent = branchBlock.getFirst().getPrefixNode();
      transformControlEdge(loopStartNode, loopIterationEvent, branchBlock.getFirst(), branchStartEvent, repeatBranch);
      
      latestNode = branchBlock.getSecond();
      latestEvent = latestNode.getSuffixNode();
      repeatBranch = latestEvent.getLabel();
    }
       
    
    if (lookAhead > 0)
    {
      Tuple<NodeMapping<N>,NodeMapping<N>> loopContent = transformLoopContent(startNodeID, endNodeID, 
                                                                              iteration.incrementIteration(), 
                                                                              repeatBranch, lookAhead-1);
      
      NodeMapping<N> loopContentStart = loopContent.getFirst();
      transformControlEdge(latestNode, latestEvent, loopContentStart, loopContentStart.getPrefixNode(), repeatBranch);
      
      
      latestNode = loopContent.getSecond();
      latestEvent = latestNode.getSuffixNode();
    }
    
    // connect repeatbranch & exitbranch to loopendnode 
    
    loopEndNode = transformLoopEndNode(endNodeID, iteration, branches);
    transformControlEdge(loopStartNode, loopStartNode.getSuffixNode(EXIT_BRANCH ), loopEndNode, loopEndNode.getPrefixNode(EXIT_BRANCH ), exitBranch);
    transformControlEdge(latestNode, latestEvent, loopEndNode, loopEndNode.getPrefixNode(REPEAT_BRANCH), repeatBranch);
    
    return new Tuple<NodeMapping<N>, NodeMapping<N>>(loopStartNode, loopEndNode);
  }


  protected abstract JoinNodeMapping<N> transformLoopEndNode(int endNodeID,
      IterationChain iteration, Label branches);

  protected abstract SplitNodeMapping<N> transformLoopStartNode(int startNodeID,
      IterationChain iteration, Label branches);

  protected static class TCNActivityDuration
  {
    public final TimeDistance minDuration;
    public final TimeDistance maxDuration;
    
    public TCNActivityDuration(TimeDistance minDuration, TimeDistance maxDuration)
    {
      this.minDuration = minDuration;
      this.maxDuration = maxDuration;
    }
  }
  
  /**
   * 
   * @param nodeID
   * @return First element is Contingent Duration; second element is Flexible Duration
   */
  protected TCNActivityDuration getActivityDuration(int nodeID)
  {
    TemporalNode node = template.getNode(nodeID);
    ActivityDuration activityDuration = node.getActivityDuration();

    TimeDistance flexibleMinDuration = NODE_MIN_DURATION;
    TimeDistance flexibleMaxDuration = NODE_MAX_DURATION;
    
    if (activityDuration != null)
    {
      

      TimePoint lowerMinimumDuration = activityDuration
        .getMinimumDuration();
      if (lowerMinimumDuration != null)
      {
        flexibleMinDuration = new DefaultTimeDistance(lowerMinimumDuration);
      }
      
      TimePoint flexibleMaximumDuration = activityDuration
        .getMaximumDuration();
      if (flexibleMaximumDuration != null)
      {
        flexibleMaxDuration = new DefaultTimeDistance(flexibleMaximumDuration);
      }

    }

    return new TCNActivityDuration(flexibleMinDuration, flexibleMaxDuration);
  }
  
  protected void processSyncEdges()
  {
    for (StructuredEdge syncEdge : template.getEdgeStructure(EdgeType.ET_SYNC))
    {
      NodeMapping<N>[] nodes = findNodes(syncEdge.getSourceNodeID());

      for (NodeMapping<N> source : nodes)
      {
        // NodeMapping source = timeModel.findNode(ste.getSourceNodeID(), new
        // Iteration(0, new int[]{0}));
        NodeMapping<N> destination = findNode(syncEdge.getDestinationNodeID(),
                                              source.getNodeInstance().getIteration());

        if (destination != null)
        {
          N e1 = source.getEndNode();
          N e2 = destination.getStartNode();
          {
            Label branches = mergeBranches(e1, e2);

            transformSyncEdge(e1, e2, branches);
          }

        }
      }
    }
  }
  
  protected void transformProcessDuration(NodeMapping<N> startNode, NodeMapping<N> endNode)
  {
    Interval<TimeDistance> processDuration = getProcessDuration();
    N processStartNodeEndEvent = startNode.getEndNode();
    N processEndNodeStartEvent = endNode.getStartNode();
    
    Label branches = processEndNodeStartEvent.getLabel();
    
    createTCNEdge(processStartNodeEndEvent,
                         processEndNodeStartEvent, branches, processDuration.getMinValue(), processDuration.getMaxValue(), false);
  }
  
  protected void transformTimeLags()
  {
    for (StructuredConstraintEdge ste : template.getConstraintEdgeStructure())
    {
      transformTimeLag(ste);
    }
  }
  
  protected void transformTimeLag(StructuredConstraintEdge ste)
  {
    for (TimeLag timeLag : template
      .getTimeLagsConstraint(ste.getSourceNodeID(), ste.getDestinationNodeID()))
    {
      if (timeLag != null)
      {
        // TODO AL handle loop incoming and loop outgoing edges
        NodeMapping<N>[] nodes = findNodes(ste.getSourceNodeID());

        for (NodeMapping<N> source : nodes)
        {
          NodeMapping<N> destination = findNode(ste.getDestinationNodeID(),
                                                source.getNodeInstance().getIteration());

          if (destination == null)
            continue;

          N sourceEvents;
          N destinationEvents;

          switch (timeLag.getRelationType())
          {
            case END_END:
              sourceEvents = source.getEndNode();
              destinationEvents = destination.getEndNode();
              break;
            case END_START:
              sourceEvents = source.getEndNode();
              destinationEvents = destination.getStartNode();
              break;
            case START_END:
              sourceEvents = source.getStartNode();
              destinationEvents = destination.getEndNode();
              break;
            case START_START:
              sourceEvents = source.getStartNode();
              destinationEvents = destination.getStartNode();
              break;
            default:
              throw new RuntimeException();
          }

          transformTimeLag(sourceEvents, destinationEvents, timeLag);

        }
      }
    }
  }

  protected Interval<TimeDistance> transformTimeLag(N sourceEvent, N destinationEvent, TimeLag timeLag)
  {
    Interval<TimeDistance> timeSpan = toTimeSpan(timeLag);
    {
      Label branches = mergeBranches(sourceEvent, destinationEvent);
      createTCNEdge(sourceEvent, destinationEvent, branches, timeSpan.getMinValue(),
                    timeSpan.getMaxValue(), false);
    }
    return timeSpan;
  }

  private Interval<TimeDistance> toTimeSpan(TimeLag timeLag)
  {
    TimeDistance minDuration;
    TimeDistance maxDuration;

    TimePoint minimumTimeLag = timeLag.getLowerMinimumTimeLag();

    if (minimumTimeLag != null)
    {
      minDuration = new DefaultTimeDistance(minimumTimeLag);
    }
    else
    {
      minDuration = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
    }

    TimePoint maximumTimeLag = timeLag.getUpperMaximumTimeLag();

    if (maximumTimeLag != null)
    {
      maxDuration = new DefaultTimeDistance(maximumTimeLag);
    }
    else
    {
      maxDuration = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    }

    Interval<TimeDistance> timeSpan = new DefaultInterval<TimeDistance>(minDuration, maxDuration);
    return timeSpan;
  }

    
  protected void processActivityFixedDate(int nodeID, NodeMapping<N> node)
  {
    TemporalNode n = template.getNode(nodeID);

    FixedDate startFixedDate = n.getStartFixedDate();
    if (startFixedDate != null) // Do we have a fixed date on the start of the node?
    {
      N startNode = node.getStartNode();
      createFixedDateForNode(startNode);
    }
    
    FixedDate completionFixedDate = n.getCompletionFixedDate();
    if (completionFixedDate != null) // do we have a fixed date on the end of the node?
    {
      N endNode = node.getEndNode();
      createFixedDateForNode(endNode);
    }

  }

  protected void createFixedDateForNode(N startNode)
  {
    createTCNEdge(getTimeBaseNode(), startNode, startNode.getLabel(),
      DefaultTimeDistance.ZERO_TIME_DISTANCE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, false);
  }
  


  protected void processActivityValidityPeriod(int nodeID, NodeMapping<LabeledTimePoint> node)
  {
    // TemporalNode n = template.getNode(nodeID);
    //
    // ValidityPeriod startValidityPeriod = n.getStartValidityPeriod();
    //
    // if (startValidityPeriod != null)
    // {
    // TimeSpan timeSpan = toTimeSpan(startValidityPeriod);
    // timeModel.setConstraint(TCNTimeModel.TIMEBASE_EVENT, node.startEvents[0],
    // timeSpan,
    // ConstraintType.Time_C);
    // }
    //
    // ValidityPeriod completionValidityPeriod =
    // n.getCompletionValidityPeriod();
    //
    // if (completionValidityPeriod != null)
    // {
    // TimeSpan timeSpan = toTimeSpan(completionValidityPeriod);
    // timeModel.setConstraint(TCNTimeModel.TIMEBASE_EVENT, node.endEvents[0],
    // timeSpan,
    // ConstraintType.Time_C);
    // }
  }


  protected void transformProcessValidityPeriod(NodeMapping<N> startNode, NodeMapping<N> endNode)
  {
    // ValidityPeriod startValidityPeriod = template.getStartValidityPeriod();
    //
    // if (startValidityPeriod != null)
    // {
    // TCNEvent instanceCreationEvent = boundary.getFirstNode().endEvents[0];
    // TimeSpan timeSpan = toTimeSpan(startValidityPeriod);
    // timeModel.setConstraint(TCNTimeModel.TIMEBASE_EVENT,
    // instanceCreationEvent,
    // timeSpan,
    // ConstraintType.Time_C);
    // }
    //
    // ValidityPeriod completionValidityPeriod =
    // template.getCompletionValidityPeriod();
    // if (completionValidityPeriod != null)
    // {
    // TCNEvent instanceTerminationEvent = boundary.getLastNode().endEvents[0];
    // TimeSpan timeSpan = toTimeSpan(completionValidityPeriod);
    // timeModel.setConstraint(TCNTimeModel.TIMEBASE_EVENT,
    // instanceTerminationEvent,
    // timeSpan,
    // ConstraintType.Time_C);
    // }
  }

  protected Interval<TimeDistance> getProcessDuration()
  {
// create a constraint for overall process duration
    
    ProcessDuration processDuration = template.getProcessDuration();

    TimeDistance minDuration = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    TimeDistance maxDuration = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    if (processDuration != null)
    {

      TimePoint lowerMinimumDuration = processDuration
        .getMinimumDuration();
      if (lowerMinimumDuration != null)
      {
        minDuration = new DefaultTimeDistance(lowerMinimumDuration);
      }
      
      TimePoint upperMaximumDuration = processDuration
          .getMaximumDuration();
      if (upperMaximumDuration != null)
      {
        maxDuration = new DefaultTimeDistance(upperMaximumDuration);
      }
    }
    return new DefaultInterval<TimeDistance>(minDuration, maxDuration);
  }

  protected abstract E[] transformSyncEdge(N from, N to, Label branches);

  private NodeMapping<N> findNode(
      int nodeID, IterationChain iterationChain)
  {
    for (NodeMapping<N> m : nodeMapping)
    {
      if (m.getNodeInstance().getNodeID() == nodeID && m.getNodeInstance().getIteration().equals(iterationChain))
        return m;
    }
    return null;
  }
  
  @SuppressWarnings("unchecked")
  private NodeMapping<N>[] findNodes(
      int nodeID)
  {
    Collection<NodeMapping<N>> res = new ArrayList<NodeMapping<N>>();
    for (NodeMapping<N> m : nodeMapping)
    {
      if (m.getNodeInstance().getNodeID() == nodeID)
        res.add(m);
    }
    
    return res.toArray(new NodeMapping[res.size()]);
  }

//  protected abstract E[] createTCNEdge(N from,
//      N to, BranchSet branches);
  
  
  protected abstract E[] createTCNEdge(N from,
      N to, Label branches, TimeDistance minValue,
      TimeDistance maxValue, boolean controlEdge);
  
  
  protected abstract NodeMapping<N> transformProcessEndNode(
      int currentNodeID, IterationChain iterationChain, Label branches);

  protected abstract NodeMapping<N> transformActivity( int currentNodeID,
      IterationChain iterationChain, Label branches);

  protected abstract NodeMapping<N> transformProcessStartNode(
      int currentNodeID, IterationChain iterationChain, Label branches);

  protected abstract JoinNodeMapping<N> transformXORJoinNode( int endNodeID,
      IterationChain iteration, Label branches);

  protected abstract SplitNodeMapping<N> transformXORSplitNode(
      int startNodeID, IterationChain iteration, Label branches);

  protected abstract JoinNodeMapping<N> transformANDJoin(int endNodeID,
      IterationChain iteration, Label branches);

  protected abstract SplitNodeMapping<N> transformANDSplit(int startNodeID,
      IterationChain iteration, Label branches);

  protected abstract Label createInitialBranch();

  protected abstract N transformInstanceCreationEvent(
      IterationChain iterationChain, Label branches);

//  protected abstract E[] transformControlEdge(N from, N to, Label branches);
  
  protected void transformControlEdge(NodeMapping<N> pred, NodeMapping<N> succ, Label branches)
  {
    transformControlEdge(pred, succ, branches, CONTROL_FLOW_MIN_TIME_LAG, CONTROL_FLOW_MAX_TIME_LAG);
  }

  protected void transformControlEdge(NodeMapping<N> pred, NodeMapping<N> succ, Label branches, TimeDistance minDistance, TimeDistance maxDistance)
  {
    N from = pred.getSuffixNode(); 
    N to = succ.getPrefixNode();
    
    transformControlEdge(pred, from, succ, to, branches, minDistance, maxDistance);
  }
  
  protected void transformControlEdge(NodeMapping<N> pred, N from, NodeMapping<N> succ, N to, Label branches)
  {
    transformControlEdge(pred, from, succ, to, branches, CONTROL_FLOW_MIN_TIME_LAG, CONTROL_FLOW_MAX_TIME_LAG);
  }

  protected void transformControlEdge(NodeMapping<N> pred, N from, NodeMapping<N> succ, N to,
      Label branches, TimeDistance minDistance, TimeDistance maxDistance)
  {
    if (branches == null)
    {
      branches = to.getLabel().merge(from.getLabel());
    }
    
    createTCNEdge(from, to, branches, minDistance, maxDistance, true);
    pred.setBranch(from, succ);
    succ.setBranch(to, pred);
  }


  protected abstract Label mergeBranches(N node1, N node2);
  
  protected abstract Label mergeANDBranches(Label branches, Label[] newBranches);
  
  protected abstract Label mergeXORBranches(Label branches, Label[] newBranchPaths);
  
  protected <M extends NodeMapping<N>> M addNodeMapping(M mapping)
  {
    this.nodeMapping.add(mapping);
    
    return mapping;    
  }
  
  protected N addNode(N nodeTemplate)
  {
    if (this.nodes.contains(nodeTemplate))
    {
      int index = this.nodes.indexOf(nodeTemplate);
      throw new IllegalArgumentException("Node already exists: "+nodeTemplate+" as "+this.nodes.get(index));
    }
    
    this.nodes.add(nodeTemplate);
    
    return nodeTemplate;
  }


  protected E addEdge(E edgeTemplate)
  {
    if (!nodes.contains(edgeTemplate.getSourceNode()))
      throw new IllegalArgumentException();
    
    if (!nodes.contains(edgeTemplate.getTargetNode()))
      throw new IllegalArgumentException();
    
    for (E e : this.edges)
    {
      if (e.getSourceNode() == edgeTemplate.getSourceNode() && e.getTargetNode() == edgeTemplate.getTargetNode())
      {
        E edge = mergeEdges(e, edgeTemplate);

        if (edge != e)
        {
          this.edges.add(edge);
          this.edges.remove(e);
        }
        return edge;
      }
    }
    this.edges.add(edgeTemplate);
        
    
    return edgeTemplate;
  }

  protected abstract E mergeEdges(E e1, E e2);

  protected String getNodeName(int nodeID)
  {
    return template.getNode(nodeID).getName();
  }
}
