package de.aristaflow.adept2.model.timemodel.tcn.cstnImpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Path;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUReductionProjection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractSTNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ObservationBinding;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;
import de.aristaflow.adept2.util.ArgChecks;

public abstract class AbstractCSTNSolver extends AbstractSTNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge>
{
  
  public AbstractCSTNSolver()
  {
    super();
  }

  public AbstractCSTNSolver(Level debugLevel)
  {
    super(debugLevel);
  }
  
  @Override
  protected Label mergeLabels(Label branch1, Label branch2)
  {
    return branch1.merge(branch2);
  }
  
  /**
   * 
   * @param labeledValueSet
   * @return
   */
 
  public LabeledValueSet applySimplificationRules(LabeledValueSet labeledValueSet)
  {
    Map<Label, LabeledValue> labels = DefaultLabeledValueSet.toMap(labeledValueSet);
    
    applySimplificationRules(labels);
    
    if (labels.size() != labeledValueSet.getLabeledValues().size())
    {
    
      LabeledValueSet newLabeledValueSet = new DefaultLabeledValueSet(labels.values());
      
      assert !labeledValueSet.equals(newLabeledValueSet);
      
      assert debug(Level.FINER, "\tSimplifying %s to %s",labeledValueSet, newLabeledValueSet);
      
      return newLabeledValueSet;
    }
    return labeledValueSet;
  }

  public boolean applySimplificationRules(Map<Label, LabeledValue> labels)
  {
    boolean changes = false;
    boolean localChanges = true;
    while (localChanges)
    {
      localChanges = false;
      // Each of these has complexity O(n^2)
      // Each of these will remove at least either a ChoiceVariableBinding from one of the labels or a LabeledValue
      // since both sets are finite this method will terminate
      localChanges |= redundantLabelEliminationRule1(labels);
      localChanges |= redundantLabelEliminationRule2(labels);
      
      changes |= localChanges;
    }
    
    assert labels.size() > 0;
    
    return changes;
  }

  /**
   * Redundant Label Elimination Rule 2
   * 
   * <p>
   * This is a variant of the RLE2 of Combi et al. (2013) pushing values down
   * instead of up. That is, for two labels \alpha p and \alpha -p instead of
   * removing any \alpha labeled value we push the value of either \alpha p or
   * \alpha -p down to \alpha (depending on which one is greater) and then
   * remove the labeled value which has been pushed down (or both if they have
   * the same value).<br/>
   * This ensures that all labels are as short (i.e., generic) as possible. 
   * As as side product calling this method (multiple times) will yield the same result
   * as when applying RLE 3.
   * 
   * 
   * <br/>
   * See Conrad et al. 2011 for details on when a labeled value is dominated by
   * another labeled value and thus redundant.
   * </p>
   * 
   * Complexity of this is roughly O(n^2) in the number of entries of
   * <code>labels</code>
   * 
   * @param labels
   * @return
   */
  protected boolean redundantLabelEliminationRule2(Map<Label, LabeledValue> labels)
  {
    boolean changes = false;
    
    Set<LabeledValue> markForAdd = new HashSet<LabeledValue>();
    Set<Label> markForRemoval = new HashSet<Label>();
    
    List<LabeledValue> list = new ArrayList<LabeledValue>(labels.values());
    
    // First find all choices for each choice variable in each context
    Map<Label, Map<ChoiceVariable, Set<LabeledValue>>> map = new HashMap<Label, Map<ChoiceVariable, Set<LabeledValue>>>();
    for (ListIterator<LabeledValue> iter1 = list.listIterator(); iter1.hasNext(); )
    {
      LabeledValue v1 = iter1.next();
      if (!iter1.hasNext())
        break;
      
      // Each pair needs to be considered only once as singlediff(l1, l2) = singlediff(l2, l1)
      for (ListIterator<LabeledValue> iter2 = list.listIterator(iter1.nextIndex()); iter2.hasNext(); )
      {
        LabeledValue v2 = iter2.next();
        
        ChoiceVariable diff = singlediff(v1.getLabel(), v2.getLabel());
        if (diff != null)
        {
          Label context = v1.getLabel().removeVariable(diff);
          
          Map<ChoiceVariable, Set<LabeledValue>> baseLabels = map.get(context);
          if (baseLabels == null)
          {
            baseLabels = new HashMap<Label.ChoiceVariable, Set<LabeledValue>>();
            map.put(context, baseLabels);
          }
          
          Set<LabeledValue> choices = baseLabels.get(diff);
          if (choices == null)
          {
            choices = new HashSet<LabeledValue>(diff.getOptions()*2);
            baseLabels.put(diff, choices);
          }
          
          choices.add(v1);
          choices.add(v2);
        }
      }
    }
    
    
    for (Entry<Label, Map<ChoiceVariable, Set<LabeledValue>>> m : map.entrySet())
    {
      // Go over all contexts
      Label context = m.getKey();
      Map<ChoiceVariable, Set<LabeledValue>> choices = m.getValue();
      // Check each choice variable if all choices are present.
      for (Entry<ChoiceVariable, Set<LabeledValue>> entry : choices.entrySet())
      {
        // check if for this choice variable all options are present
        if (entry.getKey().getOptions() == entry.getValue().size())
        {
          // if yes, find the max value
          
          TimeDistance max = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
          TimeDistance maxCont = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
          boolean allContingent = false;
          LabeledTimePoint trigger = null;
          for (LabeledValue l : entry.getValue())
          {
            TimeDistance value = l.getValue();
            int compare = value.compareTo(max);
            if (compare > 0)
            {
              max = value;
            }  
            
          }

          // propose to add the max value as new value for the context
          markForAdd.add(new DefaultLabeledValue(context, max));
                    
          // remove any value that is equal (or greater) to the max value as it is subsumed by the new value added
          // for the context
          for (LabeledValue l : entry.getValue())
          {
            TimeDistance value = l.getValue();
            int compare = value.compareTo(max);
            if (compare >= 0)
            {
              markForRemoval.add(l.getLabel());
            }
          }
          
        }
      }
    }
    
    for (Label lbl : markForRemoval)
    {
      changes |= labels.remove(lbl) != null;
    }
    
    for (LabeledValue lbl : markForAdd)
    {
      changes |= addOrReplaceIfNecessary(labels, lbl);
    }
    
    
    
    assert labels.size() > 0 : "MarkForAdd: "+markForAdd+ " | MarkForRemoval: "+markForRemoval;
    return changes;
  }

  
  /**
   * <p>Redundant label elimination rule 1.</p>
   *  
   * <p>This is an extension of the RLE1 of Combi et al. (2013) considering >= instead of ==.
   * See Conrad et al. 2011 for details on when a labeled value is dominated by another labeled value and thus redundant.</p>
   * 
   * Complexity of this is roughly O(n^2) in the number of entries of <code>labels<code>
   * 
   * @param labels
   * @return
   */
  protected boolean redundantLabelEliminationRule1(Map<Label, LabeledValue> labels)
  {
    
    boolean changes = false; 
    
    // Redundant Label Elimination Rule 1
    for (Iterator<LabeledValue> iter1 = labels.values().iterator(); iter1.hasNext(); )
    {
      LabeledValue v1 = iter1.next();
      
      for (Iterator<LabeledValue> iter2 = labels.values().iterator(); iter2.hasNext(); )
      {
        LabeledValue v2 = iter2.next();
        if (v1 == v2)
          continue;
                 
        if (v2.dominates(v1))
        {
          assert debug(Level.FINEST, "\tRLE1: Removing label %s because it subsumes %s", v1, v2);
        
          iter1.remove();
          changes = true;
          break;
        }
        
      }
    }
    
    assert labels.size() > 0;
  
    return changes;
  }
  
  protected Set<ChoiceVariable> conflictLiterals(Label label1, Label label2)
  {
    Proposition prop1 = (Proposition)label1;
    Proposition prop2 = (Proposition)label2;
    
    Set<ObservationBinding> observations1 = prop1.getObservations();
    Set<ObservationBinding> observations2 = prop2.getObservations();
    
    Set<ChoiceVariable> conflicts = new HashSet<ChoiceVariable>();

    for (ObservationBinding l1 : observations1)
    {
      for (ObservationBinding l2 : observations2)
      {
        // Same Variable but different value -> conflict
        if (l1.getVariable().equals(l2.getVariable()))
        {
          if (l1.getValue() != l2.getValue())
          {
            conflicts.add(l1.getVariable());
          }
          break;
        }
      }
    }

    
    return conflicts;
  }
  
  protected ChoiceVariable singlediff(Label label1, Label label2)
  {
    Proposition prop1 = (Proposition)label1;
    Proposition prop2 = (Proposition)label2;
    
    // Cannot be singlediff if different length
    Set<ObservationBinding> observations1 = prop1.getObservations();
    Set<ObservationBinding> observations2 = prop2.getObservations();
    if (observations1.size() != observations2.size())
      return null;
    
    Set<ChoiceVariable> conflicts = conflictLiterals(label1, label2);
    
    if (conflicts.size() == 1)
      return conflicts.iterator().next();
    
    return null;
  }

  protected void addOrReplaceIfNecessary(LabeledTimePoint Q, LabeledTimePoint T,
      Map<Label, LabeledValue> labels, LabeledValue label)
  {
    if (Q.getLabel().satisfiable(label.getLabel()) && T.getLabel().satisfiable(label.getLabel()))
    {
//      label.addOrReplaceIfNecessary(labels);
      addOrReplaceIfNecessary(labels, label);
    }
    else
      debug(Level.SEVERE, "\tDropping %s because the Label is inconsistent with the nodes!", label);
  }

  /**
   * Calculates the intersection of <code>op1</code> and <code>op2</code>
   * 
   * @param op1 first operand, may be null. The LabeledValueSet is considered to be minimal regarding the contained labels.
   * @param op2 second operand, may not be null.
   * @return the intersection of <code>op1</code> and <code>op2</code>,
   *         <code>op2</code> in case <code>op1<code> is <code>null</code>.
   */
  public Map<Label, LabeledValue> intersect(LabeledValueSet op1, Map<Label, LabeledValue> op2)
  {
    if (op1 == null)
      return op2;
    
    ArgChecks.checkForNull(op2, "op2");
    
    Map<Label, LabeledValue> labels = DefaultLabeledValueSet.toMap(op1);
    
    Collection<LabeledValue> lvs_op2 = op2.values();

    
    
    for (LabeledValue l_op2 : lvs_op2)
    { 
//      DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_op2);
      Label label = l_op2.getLabel();
      
      LabeledValue existingValue = labels.get(label);
      if (existingValue != null)
      {
        LabeledValue newValue = merge(l_op2, existingValue);

        if (existingValue != newValue) 
        {            
          labels.put(label, newValue);
        }
      } 
      else
      {
        // This sort of makes RLE1 dispensable
        boolean foundGeneralization = false;
        
        for (Iterator<LabeledValue> iterator = labels.values().iterator(); iterator.hasNext();)
        {
          LabeledValue lbl = iterator.next();
          if (lbl.dominates(l_op2))
          {
            foundGeneralization = true;
            break;
          } 
          else if (l_op2.dominates(lbl))
          {
            iterator.remove();
          }
        }
        
        if (!foundGeneralization)
        {
          labels.put(label, l_op2);
        }
      }
    }
    
    
    assert debug(Level.FINER, "\tIntersect LabeledValueSet\t (%s) \\cap (%s) = (%s)", op1, op2, labels.values());
    return labels;
  }

  public LabeledValue merge(
      LabeledValue newValue, LabeledValue oldValue)
  {
    // Replace if necessary
    int compare = oldValue.getValue().compareTo(newValue.getValue());
    if (compare > 0) // old > new
    {
      return newValue;
    }
    else
    {
      return oldValue;
    }
  }
  
  public boolean addOrReplaceIfNecessary(Map<Label, LabeledValue> labels, LabeledValue newValue)
  {
    boolean changes = false;
    Label label = newValue.getLabel();
    LabeledValue oldValue = labels.get(label);
    if (oldValue != null)
    {
      LabeledValue value = merge(newValue, oldValue);
      
      if (value != null && value != oldValue)
      {
        labels.put(label, value);
        changes = true;
      }
    }
    else
    {
      labels.put(label, newValue);
      changes = true;
    }
    return changes;
  }

  protected LabeledEdge createEdge(LabeledTimePoint from, LabeledTimePoint to, LabeledValueSet labeledValueSet,
      boolean controledge, boolean implicit, FlowDirection flowDirection)
  {
    // IMPROVE This should be done by some factory...    
    return new DefaultLabeledEdge(from, to, labeledValueSet, controledge, implicit, flowDirection);
  }

  /**
   * This considers the triangle Q -> S -> T, where newLabelSet belongs to Q -> T
   * @param cstnu
   * @param Q
   * @param S
   * @param T
   * @param newLabel
   * @return
   * @throws TCNInconsistencyException
   */
  protected void applyPropositionModificationRules(CSTN cstnu, LabeledTimePoint Q,
      LabeledTimePoint S, LabeledTimePoint T, Map<Label, LabeledValue> labeledValuesQT) throws TCNInconsistencyException
  {
    // CSTNU Label Modification Rules, cf. Combi13
    propositionModificationRuleR0(Q, T, labeledValuesQT);
    propositionModificationRuleR1(cstnu, Q, S, labeledValuesQT);
    propositionModificationRuleR2(cstnu, Q, T, labeledValuesQT);
    propositionModificationRuleR3(cstnu, S, T, Q, labeledValuesQT);
  }

  /**
   * <pre>P? - <\alpha\beta, -w*> -> X <- <\beta\gamma p, -x, R, -v> -> Y</pre>
   * We know x <= v.
   * 
   * 0 <= w* --> if \alpha\beta is true, X may be executed at most w before P.
   * 
   * Cases:<ol>
   * <li>x <= v <= w* : *ignore
   * <li>x <= w* <= v : 
   * <li>w* <= x <= v : 
   * </ol>
   * 
   * @param cstnu
   * @param P
   * @param X
   * @param labeledValuesYX
   * @return
   */
  public void propositionModificationRuleR3(CSTN cstnu, LabeledTimePoint P, LabeledTimePoint X,
      LabeledTimePoint Y, Map<Label, LabeledValue> labeledValuesYX)
  {
    assert P != null;
    assert X != null;
    
    // P is no observation node, hence this rule cannot be applied. 
    if (!(P instanceof DefaultObservationTimePoint))
      return;
    
    DefaultObservationTimePoint oP = (DefaultObservationTimePoint) P;
    
    // We do not need to check all possible propositions because this will be called for each possible
    // change of a constraint involving an Observation node, hence we can do it one by one...
    
    // Get the edge from P to X
    LabeledEdge PX = cstnu.getEdge(P, X);
    
    // If no such constraint exists we are fine...
    if (PX == null)
      return;
    
    // AL: This rule only makes sense if Y may happen before P 
    LabeledEdge YP = cstnu.getEdge(Y, P);
    if (YP != null)
    {
      boolean after = true;
      for (LabeledValue l_YP : YP.getValue().getLabeledValues())
      {
        if (l_YP.getValue().compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0)
          after = false;
      }
      if (after)
      {
        assert debug(Level.FINER, "Skipping RuleR3 for (%s->%s) because Y may only happen after P.", Y, X);
        return;
      }
    }
    
    List<LabeledValue> newValues = new ArrayList<LabeledValue>();
    
    //Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    for (Iterator<LabeledValue> iterator = labeledValuesYX.values().iterator(); iterator.hasNext();)
    {
      LabeledValue l_YX = iterator.next();
      // If its a lowercase constraint the rule does not apply
      if (l_YX instanceof ContingentLabeledValue && ConstraintCase.LOWERCASE.equals(((ContingentLabeledValue)l_YX).getCase()))
      {
        //DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_YX);
        continue;
      }
      
      // Check if the Label contains p at all...
      Label betaGammaP = l_YX.getLabel();
      Label betaGamma = betaGammaP.removeVariable(oP.getObservation());
      // No Label has been removed, i.e., the label does not contain P, so the rule does not apply to this constraint
      if (betaGamma.equals(betaGammaP))
      {
        //DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_YX);
        continue;
      }
      
  
      
      // Check if there is a suitable partner for this constraint having \alpha\beta as label and having
      // a value >= the value of this constraint
      Collection<LabeledValue> lvs_PX = PX.getValue().getLabeledValues();
      boolean foundSuitablePartner = false;
      for (LabeledValue l_PX : lvs_PX)
      {
        // If this constraint is contingent the rule does not apply
        if (l_PX instanceof ContingentLabeledValue)
          continue;
        
        TimeDistance w = l_PX.getValue().neg();
        // check if 0 <= w; i.e., 0 >= -w; otherwise this rule cannot be applied
        if (DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(w) >= 0) // 0 <= w // AL currently using 0 < w instead of 0 <= w
          continue;
        
        Label alphaBeta = l_PX.getLabel();
        // Make sure the resulting label will be satisfiable.
        if (!satisfiable(betaGamma, alphaBeta))
          continue;
        
        // If AlphaBeta contains p this rule cannot be applied. In this case Rule R0 applies...
        // Should never happen as Rule R0 is applied beforehand
        if (alphaBeta.containsVariable(oP.getObservation()))
          continue;
        
        
        TimeDistance v = l_YX.getValue().neg();
        // Check if v <= w; or x <= w; otherwise this rule cannot be applied.
        if (v.compareTo(w) >= 0) // v <= w // AL currently using v < w instead of v <= w
        {
          if (l_YX instanceof ContingentLabeledValue) // FIXME check sequence of check based on dependencies ie x < v
          {
            v = ((ContingentLabeledValue) l_YX).getContingentValue().neg();
            if (v.compareTo(w) >= 0) // AL currently using v < w instead of v <= w
              continue;
          }
          else
            continue;
        }

        
        assert debug(Level.FINE, "Applying RuleR3 to PX: (%1$s->%2$s) = %3$s and YX: (%4$s->%2$s) = %5$s", P, X, l_PX, Y, l_YX);
        
        // 0 <= w <--> 0 >= -w
        assert DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(l_PX.getValue()) >= 0;
            
         // v <= w*
        assert w.compareTo(l_YX.getValue().neg()) >= 0
            || l_YX instanceof ContingentLabeledValue 
            && w.compareTo(((ContingentLabeledValue)l_YX).getContingentValue().neg()) >= 0;   
        // All set lets go
                   
        // AL: This part of the rule only makes sense if Y does not depend on the outcome of P
        // Should not happen because then Y must be after P which should be excluded by one of the statements above,
        // but it seems this cannot always be guaranteed (Particularly, if we allow for duration 0 or distance 0).
        if (!Y.getLabel().containsVariable(oP.getObservation()))
        {
          // We found a suitable partner for this constraint so we can apply the rule....
          foundSuitablePartner = true; // TODO can i remove the value from the map right away?
           // Calculate \Alpha\Beta\Gamma
          Label alphaBetaGamma = mergeLabels(betaGamma, alphaBeta);
          
          // Update the existing label to \Alpha\Beta\Gamma (without p)
          LabeledValue newLabeledValue = updateLabel(l_YX, alphaBetaGamma);
          assert debug(Level.FINER, "\tR3:\t\t Replacing %s by %s", l_YX, newLabeledValue);
          newValues.add(newLabeledValue);
        }
        else
        {
          assert debug(Level.WARNING, String.format("WARNING: cannot apply R3 because of a label mismatch: %s vs. %s", PX, labeledValuesYX.values()));
        }
      
        // Calculate alpha
        Label alpha = alphaBeta.removeLabel(betaGamma);
        // If alpha is empty, i.e. \Beta\Gamma subsumes \Alpha\Beta the second half of
        // the rule cannot be applied. In particular there is no \not(\Alpha)
        if (!alpha.isEmptyLabel())
        {
          // We found a suitable partner for this constraint so we can apply the rule....
          foundSuitablePartner = true;
          // Calculate the negation of alpha
          // This is possibly a set of label as alpha may consist of several propositional letters
          // and as we know a negation of a conjunction is a disjunction (or a set of conjunctions)....
          Label[] notAlpha = alpha.negate();
          
          // Add an updated constraint for each of the negations...
          for (Label nA : notAlpha)
          {
            // Calculate \not(Alpha)BetaGammaP
            Label notAlphaBetaGammaS = mergeLabels(nA, betaGammaP);
            // Update the existing label to \not(Alpha)BetaGammaP
            LabeledValue newLabeledValue = updateLabel(l_YX, notAlphaBetaGammaS);
            newValues.add(newLabeledValue);
            
            assert debug(Level.FINER, "\tR3 (cont.):\t Replacing %s by %s", l_YX, newLabeledValue);
          }
        }
      }
      
      if (foundSuitablePartner)
      {
        iterator.remove();
      }
    }
    
    for (LabeledValue newValue : newValues)
    {
      addOrReplaceIfNecessary(labeledValuesYX, newValue);
    }
    
  }

  /**
   * <pre>P? - <\alpha\beta, -w> -> X - <\beta\gamma p, x, R, v> -> P?</pre>
   * We know x <= v
   * 
   * 0 <= w* --> if \alpha\beta is true, X may be executed at most w before P.
   * 
   * Cases<ol>
   * <li>v >= w* >= x :
   * <li>v >= x >= w* :
   * <li>w* >= v >= x :
   * </ol>
   * 
   * @param cstnu
   * @param X
   * @param P
   * @param labeledValuesQT
   * @return
   */
  public void propositionModificationRuleR2(CSTN  cstnu, LabeledTimePoint X, LabeledTimePoint P, Map<Label, LabeledValue> labeledValuesQT)
  {
    assert X != null;
    assert P != null;
    
    // P is no observation node, hence this rule cannot be applied.
    if (!(P instanceof DefaultObservationTimePoint))
      return;
    
    DefaultObservationTimePoint oP = (DefaultObservationTimePoint) P;
    
    // We do not need to check all possible proposition because this will be called for each possible
    // change of a constraint involving an Observation node, hence we can do it one by one...
    
    // Get the edge from P to X
    LabeledEdge PX = cstnu.getEdge(P, X);
    // If no such constraint exists we are fine...
    if (PX == null)
      return;
  
//    Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    List<LabeledValue> newValues = new ArrayList<LabeledValue>();
    for (Iterator<LabeledValue> iterator = labeledValuesQT.values().iterator(); iterator.hasNext();)
    {
      LabeledValue l_XP = iterator.next();
      if (l_XP instanceof ContingentLabeledValue && ConstraintCase.LOWERCASE.equals(((ContingentLabeledValue)l_XP).getCase()))
      {
        //DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_XP);
        continue;
      }
      
      // Check if the Label contains p at all...
      Label betaGammaP = l_XP.getLabel();
      Label betaGamma = betaGammaP.removeVariable(oP.getObservation());
      // No Label has been removed, i.e., the label does not contain P, so the rule does not apply to this constraint
      if (betaGamma.equals(betaGammaP))
      {
        //DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_XP);
        continue;
      }
  
      // Check if there is a suitable partner for this constraint having \beta\gammaP as label and having
      // a value >= the value of this constraint
      Collection<LabeledValue> lvs_PX = PX.getValue().getLabeledValues();
      boolean foundSuitablePartner = false;
      for (LabeledValue l_PX : lvs_PX)
      {
        // If this constraint is contingent the rule does not apply
        if (l_PX instanceof ContingentLabeledValue)
          continue;
        
        TimeDistance w = l_PX.getValue().neg();
        // check if 0 <= w; i.e., 0 >= -w; otherwise this rule cannot be applied
        if (DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(w) > 0) 
          continue;
        
        
        Label alphaBeta = l_PX.getLabel();
        // Make sure the resulting label will be satisfiable.
        if (!satisfiable(betaGamma, alphaBeta))
          continue;
        
        // If AlphaBeta contains p this rule cannot be applied. In this case Rule R0 applies...
        // Should never happen as Rule R0 is applied beforehand
        if (alphaBeta.containsVariable(oP.getObservation()))
          continue;
        
        TimeDistance v = l_XP.getValue();
        
        // Check if v >= w; or x >= w; otherwise this rule cannot be applied.
        if (v.compareTo(w) < 0) 
        {
          if (l_XP instanceof ContingentLabeledValue)
          {
            v = ((ContingentLabeledValue) l_XP).getContingentValue();
            if (v.compareTo(w) < 0)
              continue;
          }
          else
            continue;
        }
        

        
        // 0 <= w <--> 0 >= -w
        assert DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(l_PX.getValue()) >= 0;
            
         // v >= w*
        assert w.compareTo(l_XP.getValue()) <= 0
            || l_XP instanceof ContingentLabeledValue 
            && w.compareTo(((ContingentLabeledValue)l_XP).getContingentValue()) <= 0;   
        // All set lets go
        
        // AL: This part of the rule only makes sense if X does not depend on the outcome of P
        // Should not happen because then X must be after P which should be excluded by one of the statements above,
        // but it seems this cannot always be guaranteed.
        if (!X.getLabel().containsVariable(oP.getObservation()))
        {
          // We found a suitable partner for this constraint so we can apply the rule....
          foundSuitablePartner = true;
  
          // Calculate \Alpha\Beta\Gamma
          Label alphaBetaGamma = mergeLabels(betaGamma, alphaBeta);
          
          // Update the existing label to \Alpha\Beta\Gamma (without p)
          LabeledValue newLabeledValue = updateLabel(l_XP, alphaBetaGamma);
          
          debug(Level.FINER, "\tR2:\t\t Replacing %s by %s", l_XP, newLabeledValue);
          newValues.add(newLabeledValue);
        }
        else
        {
          assert debug(Level.SEVERE, "ERROR: cannot apply R2 because of a label mismatch");
        }
        
        // Calculate alpha
        Label alpha = alphaBeta.removeLabel(betaGamma);
        // If alpha is empty, i.e. \Beta\Gamma subsumes \Alpha\Beta the second half of
        // the rule cannot be applied. In particular there is no \not(\Alpha)
        if (!alpha.isEmptyLabel())
        {
        
          // We found a suitable partner for this constraint so we can apply the rule....
          foundSuitablePartner = true;
          
          // Calculate the negation of alpha
          // This is possibly a set of label as alpha may consist of several propositional letters
          // and as we know a negation of a conjunction is a disjunction (or a set of conjunctions)....
          Label[] notAlpha = alpha.negate();
          
          // Add an updated constraint for each of the negations...
          for (Label nA : notAlpha)
          {
            //Calculate \not(Alpha)BetaGammaP
            Label notAlphaBetaGammaP = mergeLabels(nA, betaGammaP);
            // Update the existing label to \not(Alpha)BetaGammaP
            LabeledValue newLabeledValue = updateLabel(l_XP, notAlphaBetaGammaP);
//            DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, newLabeledValue);
            newValues.add(newLabeledValue);
            
            debug(Level.FINER, "R2 (cont.): Replacing %s by %s", l_XP, newLabeledValue);
          }
        }
      }
      
      // The constraint is untouched by the label...
      // Hence we keep the original one...
      if (foundSuitablePartner)
      {
        iterator.remove();
      }
    }
    
    for (LabeledValue newValue : newValues)
    {
      addOrReplaceIfNecessary(labeledValuesQT, newValue);
    }
  }

  /**
   * <pre>P? - <\alpha\beta, -w*> -> X - <\beta\gamma p, x, R, v> -> Y</pre>
   * We know x <= v
   * 
   * 0 <= w* ==> if \alpha\beta is true, X may be executed at most w before P.
   * 
   * 
   * Cases<ol>
   * <li>x <= v <= w* : Y may be executed before P, thus in cases where alpha is true we cannot rely on p
   * <li>x <= w* <= v : Y may be executed before or after P, thus in cases where alpha is true we cannot rely on p
   * <li>w* <= x <= v : Y will be executed after P thus we know the value of P when we need to execute Y; ignore
   * </ol>
   * 
   * In case 1. and 2. 
   * <code>X - <\beta\gamma p, x, R, v> -> Y</code>
   * becomes
   * <code>X - <\alpha\beta\gamma, x, R, v> -> Y</code>
   * <code>X - <\not(\alpha)\beta\gamma p, x, R, v> -> Y</code>
   * 
   * @param cstnu
   * @param X
   * @param P
   * @param labeledValuesQT
   * @return
   */
  public void propositionModificationRuleR1(CSTN  cstnu, LabeledTimePoint X, LabeledTimePoint P, Map<Label, LabeledValue> labeledValuesQT)
  {
    assert X != null;
    assert P != null;
    
    // P is no observation node, hence this rule cannot be applied. 
    if (!(P instanceof DefaultObservationTimePoint))
      return;
    
    DefaultObservationTimePoint oP = (DefaultObservationTimePoint) P;
    
    // We do not need to check all possible proposition because this will be called for each possible
    // change of a constraint involving an Observation node, hence we can do it one by one...
    
    // Get the edge from P to X
    LabeledEdge PX = cstnu.getEdge(P, X);
    
    // If no such constraint exists we are fine...
    if (PX == null)
      return;
    
//    Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    List<LabeledValue> newValues = new ArrayList<LabeledValue>();
    
    for (Iterator<LabeledValue> iterator = labeledValuesQT.values().iterator(); iterator.hasNext();)
    {
      LabeledValue l_XY = iterator.next();
      // If its a lowercase constraint the rule does not apply
      if (l_XY instanceof ContingentLabeledValue && ConstraintCase.LOWERCASE.equals(((ContingentLabeledValue)l_XY).getCase()))
      {
//        DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_XY);
        continue;
      }
  
      // Check if the Label contains p at all...
      Label betaGammaP = l_XY.getLabel();
      Label betaGamma = betaGammaP.removeVariable(oP.getObservation());
      // No Label has been removed, i.e., the label does not contain P, so the rule does not apply to this constraint
      if (betaGamma.equals(betaGammaP))
      {
//        DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_XY);
        continue;
      }
      
      // Check if there is a suitable partner for this constraint having \alpha\beta as label and having
      // a value >= the value of this constraint
      Collection<LabeledValue> lvs_PX = PX.getValue().getLabeledValues();
      boolean foundSuitablePartner = false;
      for (LabeledValue l_PX : lvs_PX)
      {
        // If this constraint is contingent the rule does not apply
        if (l_PX instanceof ContingentLabeledValue)
          continue;
        
        TimeDistance w = l_PX.getValue().neg();
        // check if 0 <= w; i.e., 0 >= -w; otherwise this rule cannot be applied
        if (DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(w) >= 0) 
          continue;
        
        Label alphaBeta = l_PX.getLabel();
        // Make sure the resulting label will be satisfiable.
        if (!satisfiable(betaGamma, alphaBeta))
          continue;
        
        // If AlphaBeta contains p this rule cannot be applied. In this case Rule R0 applies...
        // Should never happen as Rule R0 is applied beforehand
        if (alphaBeta.containsVariable(oP.getObservation()))
          continue;
  
        TimeDistance v = l_XY.getValue();
        // Check if v <= w; or x <= w; otherwise this rule cannot be applied.
        if (v.compareTo(w) > 0) // v <= w
        {
          if (l_XY instanceof ContingentLabeledValue)
          {
            v = ((ContingentLabeledValue) l_XY).getContingentValue();
            if (v.compareTo(w) > 0)
              continue;
          }
          else
            continue;
        }
        
        // We found a suitable partner for this constraint so we can apply the rule....
        foundSuitablePartner = true; // TODO can we remove the value right away?
        
        // 0 <= w <--> 0 >= -w
        assert DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(l_PX.getValue()) >= 0;
            
         // v <= w*
        assert w.compareTo(l_XY.getValue()) >= 0
            || l_XY instanceof ContingentLabeledValue 
            && w.compareTo(((ContingentLabeledValue)l_XY).getContingentValue()) >= 0;   
        // All set lets go
        
        
        // Calculate \Alpha\Beta\Gamma
        Label alphaBetaGamma = mergeLabels(betaGamma, alphaBeta);
        LabeledValue newLabeledValue;
  
        // Update the existing label to \Alpha\Beta\Gamma (without p)
        newLabeledValue = updateLabel(l_XY, alphaBetaGamma);
        
        debug(Level.FINER, "\tR1:\t\t Replacing %s by %s", l_XY, newLabeledValue);
//        DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, newLabeledValue);
        newValues.add(newLabeledValue);
        
        // Calculate alpha
        Label alpha = alphaBeta.removeLabel(betaGamma);
        // If alpha is empty, i.e. \Beta\Gamma subsumes \Alpha\Beta the second half of
        // the rule cannot be applied. In particular there is no \not(\Alpha)
        if (alpha.isEmptyLabel())
          continue;
        
        // Calculate the negation of alpha
        // This is possibly a set of label as alpha may consist of several propositional letters
        // and as we know a negation of a conjunction is a disjunction (or a set of conjunctions)....
        Label[] notAlpha = alpha.negate();
  
        // Add an updated constraint for each of the negations...
        for (Label nA : notAlpha)
        {
          // Calculate \not(Alpha)BetaGammaP
          Label notAlphaBetaGammaS = mergeLabels(nA, betaGammaP);
          // Update the existing label to \not(Alpha)BetaGammaP
          newLabeledValue = updateLabel(l_XY, notAlphaBetaGammaS); 
//          DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, newLabeledValue);
          newValues.add(newLabeledValue);
          
          debug(Level.FINER, "R1 (cont.): Replacing %s by %s", l_XY, newLabeledValue);
        }
      }
      
      // The constraint is untouched by the label...
      // Hence we keep the original one...
//      if (!foundSuitablePartner)
//      {
//        DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_XY);
//        continue;
//      }
      if (foundSuitablePartner)
      {
        iterator.remove();
      }
    }
    
    for (LabeledValue newValue : newValues)
    {
      addOrReplaceIfNecessary(labeledValuesQT, newValue);
    }
  }

  protected LabeledValue updateLabel(LabeledValue labeledValue, Label label)
  {
    LabeledValue newLabeledValue = new DefaultLabeledValue(label, labeledValue.getValue());
    
    return newLabeledValue;
  }

  /**
   * <pre>P? -- <\alpha p, -x, R, -w> --> X</pre>
   * We know <code>x <= w</code>.
   * 
   * 
   * 
   * Three cases:<ol>
   * <li> 0 <= x <= w : X may be executed before P; we cannot rely on p in the constraint; remove p
   * <li> x <= 0 <= w : X may be executed before or after P; we cannot rely on p in the constraint; remove p
   * <li> x <= w <= 0 : X can only be executed after P; no issue. </li>
   * </ol>
   * 
   * In case 1. and 2. 
   * <code>P? -- <\alpha p, -x, R, -w> --> X</code>
   * becomes
   * <code>P? -- <\alpha, -x, R, -w> --> X</code>
   * 
   * 
   * @param P
   * @param labeledValuesQT
   * @return
   */
  public void propositionModificationRuleR0(LabeledTimePoint P, LabeledTimePoint X, Map<Label, LabeledValue> labeledValuesQT)
  {
    assert P != null;
    
    // P is no observation node, hence this rule cannot be applied. 
    if (!(P instanceof DefaultObservationTimePoint))
      return;
    
    DefaultObservationTimePoint oP = (DefaultObservationTimePoint) P;
  
    //Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    List<LabeledValue> newValues = new ArrayList<LabeledValue>();
    
    for (Iterator<LabeledValue> iterator = labeledValuesQT.values().iterator(); iterator.hasNext();)
    {
      LabeledValue l_PX = iterator.next();
      // If either the non-contingent or the contingent value is > 0 the rule cannot be applied
      // -w <= 0 <==> nw > 0
      TimeDistance n_w = l_PX.getValue();
      boolean applicable = DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(n_w) >= 0;
      if (!applicable && l_PX instanceof ContingentLabeledValue)
      {
        if (ConstraintCase.LOWERCASE.equals(((ContingentLabeledValue)l_PX).getCase()))
        {
          applicable = false;
        }
        else
        {
          n_w = ((ContingentLabeledValue) l_PX).getContingentValue();
          applicable = DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(n_w) >= 0;
        }
      }
      
      
      if (!applicable)
      {
//        DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_PX);
        continue;
      }
      
      Label label = l_PX.getLabel().removeVariable(oP.getObservation());
  
      // No Label has been removed
      if (label.equals(l_PX.getLabel()))
      {
//        DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, l_PX);
        continue;
      }
      
      // Either not contingent or uppercase
      assert !(l_PX instanceof ContingentLabeledValue) 
        || ConstraintCase.UPPERCASE.equals(((ContingentLabeledValue)l_PX).getCase());
      
      // Either the non-contingent or the contingent value is > 0
      assert DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(l_PX.getValue()) >= 0 
        || ((l_PX instanceof ContingentLabeledValue) 
          && DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(((ContingentLabeledValue)l_PX).getContingentValue()) >= 0);
      
      // P has been part of the original label
      //assert
      
      // Update the constraint with the new proposition
      LabeledValue labeledValue = updateLabel(l_PX, label);
      // Add the constraint to the set of constraints
      iterator.remove();
      newValues.add(labeledValue);
      
      debug(Level.FINER, "\tR0:\t\t Replacing %s by %s (in connection with %s)", l_PX, labeledValue, oP);
    }
    
    for (LabeledValue newValue : newValues)
    {
      addOrReplaceIfNecessary(labeledValuesQT, newValue);
    }
  }

  public LabeledValue observationCaseReduction(LabeledValue l_XY, LabeledValue l_YX, LabeledTimePoint P)
  {
    assert P != null;
    assert l_XY != null;
    assert l_YX != null;
    
    if (!(P instanceof DefaultObservationTimePoint))
      return null;
    
    if (l_XY instanceof ContingentLabeledValue)
      return null;
    
    if (l_YX instanceof ContingentLabeledValue)
      return null;
    
    if (satisfiable(l_XY.getLabel(), l_YX.getLabel()))
      return null;
    
    Set<ChoiceVariable> conflictLiterals = conflictLiterals(l_XY.getLabel(), l_YX.getLabel());
    
    
    if (conflictLiterals == null || conflictLiterals.size() != 1) // More than one or no conflicting literal
      return null;
    
    ChoiceVariable conflictLiteral = conflictLiterals.iterator().next();
    
    if (!conflictLiteral.getDecisionTimePoint().equals(P))
      return null;
    
    TimeDistance u = l_XY.getValue();
    TimeDistance v = l_YX.getValue().neg();
    
    if (u.compareTo(v) >= 0 || DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(u) > 0)
      return null;
    
    assert P instanceof DefaultObservationTimePoint;
    assert !(l_XY instanceof ContingentLabeledValue);
    assert !(l_YX instanceof ContingentLabeledValue);
    
    // "p = P?"
    assert conflictLiteral.getDecisionTimePoint().equals(P);
    
    
    // 0 <= u < v
    assert DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(u) <= 0;
    assert u.compareTo(v) < 0;
    
    Label r_XY = ((Proposition)l_XY.getLabel()).removeVariable(conflictLiteral);
    Label r_YX = ((Proposition)l_YX.getLabel()).removeVariable(conflictLiteral);
    
    assert satisfiable(r_XY, r_YX);
    
    Label newLabel = mergeLabels(r_XY, r_YX);
        
    LabeledValue labeledValue = new DefaultLabeledValue(newLabel, DefaultTimeDistance.ZERO_TIME_DISTANCE);
    
//    System.out.println(String.format("ObservationCase:\t %s & %s -> %s", l_XY, l_YX, labeledValue));
    debug(Level.FINER, "\tObservationCase:\t %s & %s -> %s", l_XY, l_YX, labeledValue);
    
    return labeledValue;
  }
  


  /**
   * 
   * Note that this will modify input variable <code>nodes</code>!!!
   * 
   * @param tcn
   * @param nodes
   * @param rigidComponents
   * @return
   * @throws TCNInconsistencyException 
   * @throws InterruptedException 
   * @deprecated Not yet working...
   */
  @Deprecated
  protected Map<LabeledTimePoint, Set<LabeledTimePoint>> collapseRigidComponents(ChangeableCSTN tcn,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet, 
      Set<LabeledTimePoint> nodes,
      Collection<Set<LabeledTimePoint>> rigidComponents) throws TCNInconsistencyException, InterruptedException
  {
    // First determine the representative of each rigid component
    Map<LabeledTimePoint, Set<LabeledTimePoint>> collapsedComponents = new HashMap<LabeledTimePoint, Set<LabeledTimePoint>>();
    for (Set<LabeledTimePoint> rigidComponent : rigidComponents)
    {
      // Tightly couple the rigid component in the tcn, i.e., calculate the minimal network of the respective
      // fragment of the tcn. Afterwards there should be an edge between each pair of nodes in the component (not
      // considering nodes belonging to different branches).
      tightlyCoupleComponent(tcn, rigidComponent);

      // Now determine the representative for that component
      // Make a copy of the rigid component because we are going to split rigid components
      // containing multiple observation nodes
      Set<LabeledTimePoint> collapsedComponent = new HashSet<LabeledTimePoint>(rigidComponent);
      LabeledTimePoint representative;
      List<LabeledTimePoint> observationNodes = getObservationNodes(collapsedComponent);
      
      if (observationNodes.size() == 0)
      {
        // No observation nodes so pick a random node as representative, i.e., the first in the set
        representative = collapsedComponent.iterator().next();
      }
      else if (observationNodes.size() == 1)
      {
        // One observation node, use it as representative, particularly do not reduce observation nodes as
        // they are required for the algorithms to work.
        representative = observationNodes.get(0);
      }
      else 
      {
        //Now it's getting complicated, we got multiple observation nodes in one component.
        // Hence we have to split the component in order to not reduce any observation node, as
        // they are required for the algorithms to work.
        
        // Pick the first observation node as representative and exclude the rest from the rigid component
        representative = observationNodes.remove(0);
        // !!!NOTE!!! it might be necessary to split the rigid component into multiple one (one for each observation node)
        // as the contained nodes may depend on the respective observation node, i.e., they may not have a connection
        // to the other nodes in the rigid component
        collapsedComponent.removeAll(observationNodes);
      }
      
      // Remove the representative from the collapsed component.
      collapsedComponent.remove(representative);
      // add the collapsed component to the set of collapsed components
      collapsedComponents.put(representative, collapsedComponent);
      // Remove any collapsed node from the set of nodes.
      nodes.removeAll(collapsedComponent);
    }
    
    // Now collapse each rigid component using it's representative and the set of uncollapsed nodes
    for (Entry<LabeledTimePoint, Set<LabeledTimePoint>> collapsedComponent : collapsedComponents.entrySet())
    {
      collapseComponent(tcn, changeSet, nodes, collapsedComponent.getKey(), collapsedComponent.getValue());
    }
    
    return collapsedComponents;
  }
  
  /**
   * Tightly couples the nodes in the rigid component by calculating the minimal
   * network for the respective fragment (i.e., component) of the tcn.
   * 
   * @param tcn
   * @param rigidComponent
   * @throws TCNInconsistencyException
   * @throws InterruptedException 
   * @deprecated not yet working
   */
  @Deprecated
  protected void tightlyCoupleComponent(ChangeableCSTN tcn, Set<LabeledTimePoint> rigidComponent) throws TCNInconsistencyException, InterruptedException
  {
    // This only makes sense if the rigid component has more than two nodes. Otherwise it is already tightly coupled.
    if (rigidComponent.size() > 2)
    {
      solveComponent(new CSTNUReductionProjection(tcn, rigidComponent), null);
    }
  }


  /**
   * Collapse a rigid component. Particularly, each edge related to a node in
   * <code>componentToCollapse</code> is redirected to
   * <code>representative</code>.
   * 
   * @param tcn the TCN.
   * @param nodes The set of uncollapsed nodes.
   * @param representative The representative used for the collapsed component.
   * @param componentToCollapse The component to be collapsed.
   * @throws TCNInconsistencyException in case collapsing reveals that the tcn
   *           (and in particular the component to collapse) is inconsistent.
   * @deprecated not yet working
   */
  @Deprecated
  protected void collapseComponent(ChangeableCSTN tcn, ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet, Set<LabeledTimePoint> nodes, LabeledTimePoint representative, Set<LabeledTimePoint> componentToCollapse)
      throws TCNInconsistencyException
 {
    for (LabeledTimePoint node : nodes) // Only consider uncollapsed nodes
    {
      if (node == representative)
        continue;
      
      for (LabeledTimePoint nodeToReduce : componentToCollapse)
      {
        assert (nodeToReduce != representative);
        
        // First reduce any edge point to nodeToReduce
        LabeledEdge edgeToReduce = tcn.getEdge(node, nodeToReduce);
        if (edgeToReduce != null)
        {           
          // This edge is internal to the rigid component, i.e., it's value is fixed
          LabeledEdge rigidEdge = tcn.getEdge(nodeToReduce, representative); 
          
          if (rigidEdge != null) 
            processPath(tcn, changeSet, new Path<LabeledTimePoint>(node, representative), nodeToReduce);        
          }
        
        // Second reduce any edge starting at nodeToReduce
        edgeToReduce = tcn.getEdge(nodeToReduce, node);
        if (edgeToReduce != null)
        {
          // This edge is internal to the rigid component, i.e., it's value is fixed
          LabeledEdge rigidEdge = tcn.getEdge(representative, nodeToReduce); 
          
          if (rigidEdge != null)
            processPath(tcn, changeSet, new Path<LabeledTimePoint>(representative, node), nodeToReduce);
        }
      }
      
    }
    
  }
  
  /**
   * Process one single path 
   * @param model
   * @param changeSet
   * @param nodes
   * @param e_ij
   * @param e_jk
   * @return
   * @throws TCNInconsistencyException
   */
//  protected abstract boolean processPath(ChangeableCSTN model,
//      ChangeSet<LabeledEdge, LabeledValueSet, Tuple<LabeledEdge, LabeledEdge>> changeSet,
//      final LabeledEdge e_ij, final LabeledEdge e_jk) throws TCNInconsistencyException;
  
  protected abstract boolean processPath(ChangeableCSTN model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet, 
      Path<LabeledTimePoint> p,
      final LabeledTimePoint j) throws TCNInconsistencyException;

  /**
   * Uncollapses any previously collapsed component. In particular, edges are
   * created for any previously reduced node based on the edges pointing to the
   * representative of the respective component.
   * 
   * @param tcn the tcn
   * @param nodes The set of uncollapsed nodes.
   * @param collapsedComponents the set of collapsed components and their representatives
   * @deprecated not yet working
   */
  @Deprecated
  protected void uncollapseRigidComponents(ChangeableCSTN tcn,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet,
      Collection<LabeledTimePoint> nodes, Map<LabeledTimePoint, Set<LabeledTimePoint>> collapsedComponents)
  {
    for (Entry<LabeledTimePoint, Set<LabeledTimePoint>> collapsedComponent : collapsedComponents.entrySet())
    {
      try
      {
        uncollapseComponent(tcn, changeSet, nodes, collapsedComponent.getKey(), collapsedComponent.getValue());
      }
      catch (TCNInconsistencyException e)
      {
        // Should never happen
        throw new RuntimeException(e);
      }
    }
  }

  /**
   * 
   * @param tcn the tcn
   * @param nodes The set of uncollapsed nodes.
   * @param representative the representative of the collapsed component
   * @param collapsedNodes the reduced nodes of the collapsed component
   * @param changeSet 
   * @throws TCNInconsistencyException
   * @deprecated not yet working 
   */
  @Deprecated
  protected void uncollapseComponent(ChangeableCSTN tcn,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet,
      Collection<LabeledTimePoint> nodes, LabeledTimePoint representative, Set<LabeledTimePoint> collapsedNodes)
      throws TCNInconsistencyException
  {
    for (LabeledTimePoint n : nodes)
    {
      LabeledEdge repEdge1 = tcn.getEdge(n, representative);
      LabeledEdge repEdge2 = tcn.getEdge(representative, n);
      
      if (repEdge1 != null || repEdge2 != null)
      {
        for (LabeledTimePoint reducedNode : collapsedNodes)
        {
          if (repEdge1 != null)
          {
            LabeledEdge rigidEdge = tcn.getEdge(representative, reducedNode);
            if (rigidEdge != null)
              processPath(tcn, changeSet, new Path<LabeledTimePoint>(n, reducedNode), representative);
          }
          if (repEdge2 != null)
          {
            LabeledEdge rigidEdge = tcn.getEdge(reducedNode, representative);
            if (rigidEdge != null)
              processPath(tcn, changeSet, new Path<LabeledTimePoint>(reducedNode, n), representative);    
          }
        }
      }
    }    
  }
  
  
  /**
   * Returns a list of observation nodes contained in the given collection of
   * nodes.
   * 
   * @param nodes the nodes to be considered
   * @return a set of nodes. Each node in the set is guaranteed to be an {@link DefaultObservationTimePoint} or a subclass thereof.
   */
  protected List<LabeledTimePoint> getObservationNodes(Collection<LabeledTimePoint> nodes)
  {
    List<LabeledTimePoint> observationNodes = new ArrayList<LabeledTimePoint>();
    for (LabeledTimePoint node : nodes)
    {
      if (node instanceof DefaultObservationTimePoint)
      {
        observationNodes.add(node);
      }
    }

    return observationNodes;
  }
  
  @Override
  protected boolean checkConsistency(LabeledValueSet value1, LabeledValueSet value2)
  {
    for (LabeledValue labeledValue1 : value1.getLabeledValues())
    {
      
      for (LabeledValue labeledValue2 : value2.getLabeledValues())
      {
        if (satisfiable(labeledValue1.getLabel(), labeledValue2.getLabel()))
        {
          if (!isConsistent(labeledValue1, labeledValue2))
          {
            return false;
          }
        }
      }
    }
    return true;
  }

//  protected abstract Map<Label, LabeledValue> applyEdgeGenerationRules(ChangeableCSTN model, final LabeledEdge e_ij, final LabeledEdge e_jk);

  
  protected abstract void solveComponent(ChangeableCSTN model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException;
  
}
