package de.aristaflow.adept2.model.timemodel.tcn.cstnImpl;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.STNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInstanceTimeModel;

public class CSTNInstanceTimeModel extends TCNInstanceTimeModel<LabeledTimePoint, STNEdge> implements
    InstanceTimeModel, UpdateableInstanceTimeModel
{

  private ChangeableCSTN cstn;

  public CSTNInstanceTimeModel(UUID instanceID,
      UUID templateID, ChangeableCSTN cstn, Collection<NodeMapping<LabeledTimePoint>> nodeMapping,
      LabeledTimePoint creationTimePoint, LabeledTimePoint startTimePoint, LabeledTimePoint endTimePoint,
      Map<Integer, List<URI>> updateTriggers)
  {
    super(instanceID, templateID, new CSTN2STNProjection(cstn), nodeMapping, 
          creationTimePoint, startTimePoint, endTimePoint, updateTriggers);
    this.cstn = cstn;
  }
 
  public void setInstanceCreationTime(long timestamp)
  {
    LabeledTimePoint node = getProcessCreationTimePoint();
    setToTimeStamp(node, timestamp);
    
    handleTriggerEvent(node);
  }

  public void setInstanceStartTime(long timestamp)
  {
//    NodeMapping<LabeledTimePoint> startNodeMapping = stn.getNodeMapping(getBaseTemplate().getStartNode().getID(), iterationChain);
    
    LabeledTimePoint processStart = getProcessStartTimePoint();

    setToTimeStamp(processStart, timestamp);    
    
    handleTriggerEvent(processStart);
  }

  private void setToTimeStamp(LabeledTimePoint event, long timestamp)
  {
    
    LabeledTimePoint timeBaseEvent = getTimeBaseEvent();

    TimeDistance timeDistance = timeModelOperations.millisecondsToTimeDistance(timestamp);
    
    setToDistance(timeBaseEvent, event, timeDistance);
  }

  private void setToDistance(LabeledTimePoint from, LabeledTimePoint to,
      TimeDistance timeDistance)
  {
    LabeledEdge maxEdge = new DefaultLabeledEdge(from, to, new DefaultLabeledValueSet(new DefaultLabeledValue(from.getLabel(), timeDistance)), false, false, FlowDirection.FORWARD);
    cstn.updateConstraint(maxEdge);
    
    LabeledEdge minEdge = new DefaultLabeledEdge(to, from, new DefaultLabeledValueSet(new DefaultLabeledValue(from.getLabel(), timeDistance.neg())), false, false, FlowDirection.BACKWARD);
    cstn.updateConstraint(minEdge);
  }

  public void setInstanceFinishTime(long timestamp)
  {
//    NodeMapping<LabeledTimePoint> startNodeMapping = stn.getNodeMapping(getBaseTemplate().getEndNode().getID(), iterationChain);
    
    LabeledTimePoint processEnd = getProcessEndTimePoint();

    setToTimeStamp(processEnd, timestamp);   
    
    handleTriggerEvent(processEnd);
  }

  public void setNodeActivationTime(int nodeID, int iteration, long timestamp)
  {
    // Nothing to do
    
  }

  public void setActivityInstanceStartTime(int nodeID, int iteration,
      long timestamp)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getStartNode();

    setToTimeStamp(node, timestamp);
    
    handleTriggerEvent(node);
  }

  public void setActivityInstanceFinishTime(int nodeID, int iteration,
      long timestamp)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getEndNode();

    setToTimeStamp(node, timestamp);   
    
    Interval<CalendarElement> timePoint = getTimePoint(nodeMapping.getStartNode());
    
    long startTime = toTimeStamp(timePoint.getMinValue());
    
    long duration = timestamp - startTime;
    
    setToDistance(nodeMapping.getStartNode(), nodeMapping.getEndNode(), 
      timeModelOperations.millisecondsToTimeDistance(duration));
    
    
    // Any intermediate event is also finished by now, so handle triggers
    for (LabeledTimePoint n : nodeMapping.getIntermediateEvents())
    {
      handleTriggerEvent(n);
    }
    // Remove any contingent link which have the respective node as trigger event
    handleTriggerEvent(node);
    
    setToTimeStamp(node, timestamp);
  }

  private long toTimeStamp(CalendarElement value)
  {
    return value.getTimeInMillis();
  }

  private void handleTriggerEvent(LabeledTimePoint node)
  {
    Collection<LabeledEdge> edges = cstn.getEdges();
    for (LabeledEdge edge : edges)
    {
      boolean needsUpdate = false;
      STNEdge triggerEdge = null;
      List<LabeledValue> newValues = new ArrayList<LabeledValue>();
      for (LabeledValue labeledValue : edge.getValue().getLabeledValues())
      {
        LabeledValue value = labeledValue;
        newValues.add(labeledValue);
        
      }
      
      if (needsUpdate)
      {
        LabeledEdge newEdge = new DefaultLabeledEdge(edge.getSourceNode(), edge.getTargetNode(),
          new DefaultLabeledValueSet(newValues), edge.isControlEdge(), false, edge.getFlowDirection());
        cstn.updateConstraint(newEdge);
      }
      
    }
  }


  public void setActivityInstanceProgress(int nodeID, int iteration,
      float progress)
  {
    // Not implemented, future use
    
  }

  public void createNewIteration(List<Node> nodesInLoop, int iteration)
  {
    // TODO create new Iteration
    
  }

  public CSTN getBaseTimeModel()
  {
    return cstn;
  }
  
  @Override
  protected TimeDistance getValue(STNEdge edge)
  {
    if (edge != null)
    {
      return edge.getValue().getValue();
    }
    return null;
  }
  
  @Override
  public void setEarliestStartDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getStartNode();
    setEarliestDate(node, date);
    
//    TCNTestTools.printTCN(cstn, this.getClass(), ""+nodeID+"."+iteration+".start");
  }

  private void setEarliestDate(LabeledTimePoint node, CalendarElement date)
  {
    LabeledTimePoint timeBaseEvent = getTimeBaseEvent();
    
    TimeDistance timeDistance = timeModelOperations.millisecondsToTimeDistance(date.getTimeInMillis());
    
    
    
//    LabeledEdge min = cstnu.getEdge(node, timeBaseEvent);

    DefaultLabeledValue minValue = new DefaultLabeledValue(timeBaseEvent.getLabel(), timeDistance.neg());
    LabeledEdge minEdge = new DefaultLabeledEdge(node, timeBaseEvent, new DefaultLabeledValueSet(minValue), false, false, FlowDirection.BACKWARD);
    cstn.updateConstraint(minEdge);
  }

  @Override
  public void setEarliestCompletionDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getEndNode();
    setEarliestDate(node, date);
    
//    TCNTestTools.printTCN(cstn, this.getClass(), ""+nodeID+"."+iteration+".completion");
  }

  @Override
  public void setLatestStartDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getStartNode();

    setLatestDate(node, date);
        
    
//    TCNTestTools.printTCN(cstn, this.getClass(), ""+nodeID+"."+iteration+".start");
  }

  private void setLatestDate(LabeledTimePoint node, CalendarElement date)
  {
    long timestamp = date.getTimeInMillis();
    LabeledTimePoint timeBaseEvent = getTimeBaseEvent();
    
    TimeDistance timeDistance = timeModelOperations.millisecondsToTimeDistance(timestamp);
    
    
//    LabeledEdge max = cstnu.getEdge(timeBaseEvent, node);
    
    DefaultLabeledValue maxValue = new DefaultLabeledValue(timeBaseEvent.getLabel(), timeDistance);
    LabeledEdge maxEdge = new DefaultLabeledEdge(timeBaseEvent, node, new DefaultLabeledValueSet(maxValue), false, false, FlowDirection.FORWARD);
    cstn.updateConstraint(maxEdge);
  }

  @Override
  public void setLatestCompletionDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getEndNode();

    setLatestDate(node, date);
        
    
//    TCNTestTools.printTCN(cstn, this.getClass(), ""+nodeID+"."+iteration+".completion");
  }




}
