package de.aristaflow.adept2.model.timemodel.tcn.cstnImpl;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Path;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tools.DuplicateFreeQueue;

/**
 * A simple basic PC2 Solver considering labels (Lanz08 S. 85). </p>
 * 
 * Algorithm:
 * 
 * <pre>
 * Q = {(i, k, j) | (i &lt; j) and (k != i,j)}
 * while !isEmpty(Q) do
 *   select and delete a path (i, k, j) from Q
 *   if T_ij != T_ij &amp; (T_ik + T_kj)
 *     T_ij = T_ij &amp; (T_ik + T_kj)
 *     if isEmpty(T_ij) then throw InconsistencyException
 *     Q = Q || {(i, j, k), (k, i, j) | 1 &lt;= k &lt;= n, i != k != j}
 *   end
 * end
 * </pre>
 * 
 * 
 */
public class LabelBasedPC2Solver extends AbstractCSTNSolver
{

  public LabelBasedPC2Solver()
  {
    super(Level.ALL);
  }
  
  public LabelBasedPC2Solver(Level debugLevel)
  {
    super(debugLevel);
  }
  
  public <T extends ChangeableTCN<LabeledTimePoint, LabeledEdge>> T solve(T tcn)
      throws TCNInconsistencyException, InterruptedException
  {
    return solve(tcn, null);
  }
  

  @SuppressWarnings("unchecked")
  public <T extends ChangeableTCN<LabeledTimePoint, LabeledEdge>> T solve(T tcn,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet)
      throws TCNInconsistencyException, InterruptedException
  {
    if (tcn instanceof ChangeableCSTN)
    {
      return (T)solve((ChangeableCSTN)tcn, changeSet);
    }

    throw new IllegalArgumentException(""+tcn.getClass());
  }
  
  /**
   * A simple basic PC2 Solver. </p>
   * 
   * Algorithm:
   * 
   * <pre>
   * Q = {(i, k, j) | (i &lt; j) and (k != i,j)}
   * while !isEmpty(Q) do
   *   select and delete a path (i, k, j) from Q
   *   if T_ij != T_ik + T_kj
   *     T_ij = T_ij &amp; T_ik + T_kj
   *     if isEmpty(T_ij) then throw InconsistencyException
   *     Q = Q || {(i, j, k), (k, i, j) | 1 &lt;= k &lt;= n, i != k != j}
   *   end
   * end
   * </pre>
   * 
   * @return
   * @throws InterruptedException 
   * 
   */
  public CSTN solve(
      ChangeableCSTN model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException
  {   
    logger.fine("Starting PC2");
    
    logger.fine("Initializing Paths");

    solveComponent(model, changeSet);

    model.setConsistent();
    
    return model;
  }

  @Override
  protected void solveComponent(ChangeableCSTN model, 
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException
  {
    Collection<LabeledTimePoint> nodes = model.getTimePoints();
    
    Collection<LabeledEdge> changed = null;//model.getChanges(); 
    
    int nodeCount = nodes.size();
    final Queue<Path<LabeledTimePoint>> queue = new DuplicateFreeQueue<Path<LabeledTimePoint>>(nodeCount*nodeCount*2);
    if (changed == null)
    {
      getInitialPaths(nodes, queue);
    }
    else
    {
      for (LabeledEdge e : changed)
      {
        //addRelatedPaths(nodes, queue, e.getSourceNode(), e.getTargetNode());
        queue.add(new Path<LabeledTimePoint>(e.getSourceNode(), e.getTargetNode()));
        
      }
    }

    while (!queue.isEmpty())
    {
      
      if (Thread.currentThread().isInterrupted())
        throw new InterruptedException();
      
      final Path<LabeledTimePoint> triangle = queue.remove();

      LabeledTimePoint i = triangle.getFirst();
      LabeledTimePoint j = triangle.getSecond();
//      LabeledTimePoint k = triangle.getThird();
      
      for (LabeledTimePoint k : nodes)
      {
        if (i.equals(k) || j.equals(k))
          continue;
        
        if (processPath(model, changeSet, new Path<LabeledTimePoint>(i, k), j))
        {
          queue.add(new Path<LabeledTimePoint>(i, k));
//          addRelatedPaths(nodes, queue, i, k);
        }
        
        if (processPath(model, changeSet, new Path<LabeledTimePoint>(k, j), i))
        {
          queue.add(new Path<LabeledTimePoint>(k, j));
//          addRelatedPaths(nodes, queue, i, k);
        }
      }
    }
  }

  @Override
  protected boolean processPath(ChangeableCSTN model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet, Path<LabeledTimePoint> p,
      LabeledTimePoint via) throws TCNInconsistencyException
  {

    
    LabeledTimePoint from = p.getFrom();
    LabeledTimePoint to = p.getTo();
    final LabeledEdge e_ij = model.getEdge(from, via);
    
    final LabeledEdge e_jk = model.getEdge(via, to);
    
    
    final LabeledEdge e_ik = model.getEdge(from, to);
    
    assert debug(Level.FINE, "Evaluating (%s, %s) = (%s) \\cap (%s) + (%s)", from, to, e_ik, e_ij, e_jk);
    Map<Label, LabeledValue> newValues = applyEdgeGenerationRules(model, from, via , to);
    
    if (newValues != null)
    {
      applySimplificationRules(newValues);
    
      LabeledValueSet ls_ik = null;
      if (e_ik != null)
      {
        ls_ik = e_ik.getValue();
      }
    
      newValues = intersect(ls_ik, newValues);
      applyPropositionModificationRules(model, from, via, to, newValues);
      assert debug(Level.FINE, "   Label Removal Rules resulted in %s", newValues.values());
      applySimplificationRules(newValues);
      
      LabeledValueSet newValueSet = new DefaultLabeledValueSet(newValues.values());
    
      if (!newValueSet.equals(ls_ik)) // Simplification
                                                             // destroys !=
      {
        LabeledEdge labeledEdge = createEdge(from, to, newValueSet, e_ik != null && e_ik.isControlEdge(),
                                   e_ik == null || e_ik.isImplicit(), e_ik != null ? e_ik.getFlowDirection()
                                                                                  : FlowDirection.UNKOWN);
    
        assert debug(Level.FINE, "Updating (%s) to (%s)", e_ik, labeledEdge);
    
        logChange(changeSet, p, via, labeledEdge, e_ik != null ? e_ik.getValue() : null);
    
        checkConsistency(model, labeledEdge, new Triangle<LabeledTimePoint>(p, via));
        
        try
        {
          model.updateConstraint(labeledEdge);
        }
        catch (IllegalArgumentException e)
        {
          debug(Level.SEVERE, "Trying to update contingent constraint %s -> %s: %s -> %s", from, to, e_ik, newValueSet);
          throw new TCNInconsistencyException("Trying to update contingent constraint!",  p, via, model, e);
        }
    
        debug(Level.FINER, "Adding related Paths for %s -> %s", from, to);
    
        return true;
      }
      debug(Level.FINE, " Ignoring\t %s to %s (because no changes)", e_ik, newValueSet);
    }
    
    return false;
  }

  
 

  
  protected Map<Label, LabeledValue> applyEdgeGenerationRules(ChangeableCSTN model, LabeledTimePoint i, LabeledTimePoint j, LabeledTimePoint k)
  {    
    
    final LabeledEdge e_ij = model.getEdge(i, j);
    
    if (e_ij == null)
      return null;
    
    final LabeledEdge e_jk = model.getEdge(j, k);
    final LabeledEdge e_ji = model.getEdge(j, i);

    
    Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    
    if (e_jk != null)
    {
      for (LabeledValue cl_ij : e_ij.getValue().getLabeledValues())
      {
      
 
        for (LabeledValue cl_jk : e_jk.getValue().getLabeledValues())
        {
          if (satisfiable(cl_ij.getLabel(), cl_jk.getLabel()))
          {
            LabeledValue labeledValue = add(cl_ij, cl_jk);
            
            addOrReplaceIfNecessary(i, k, labels, labeledValue);
          }
        }
      }
    }
    
    if (e_ji != null)
    {
      for (LabeledValue l_ij : e_ij.getValue().getLabeledValues())
      {
        for (LabeledValue l_ji : e_ji.getValue().getLabeledValues())
        {
          LabeledValue observationCase = observationCaseReduction(l_ji, l_ij, k);
          
          if (observationCase != null)
          {
            addOrReplaceIfNecessary(i, k, labels, observationCase);
          }
        }
      }
    }
    
    Map<Label, LabeledValue> sum = null;
    if (labels.size() > 0)
    {
      sum = labels;      
    }
    

    return sum;
  }

  protected LabeledValue add(LabeledValue sum1,
      LabeledValue sum2)
  {
    TimeDistance sum = sum1.getValue().add(sum2.getValue());
    Label label = mergeLabels(sum1.getLabel(), sum2.getLabel());
    LabeledValue labeledValue = new DefaultLabeledValue(label, sum);
    
    assert debug(Level.FINEST, "\tAdding LabeledValue\t (%s) + (%s) = (%s)", sum1, sum2, labeledValue);
    
    return labeledValue;
  }




}
