package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;

public class RTEDExecutionStrategy
{
  public static class ExecutionDecision
  {

    private TimeDistance goTime;
    private List<LabeledTimePoint> list;

    public ExecutionDecision(TimeDistance goTime, List<LabeledTimePoint> goList)
    {
      this.goTime = goTime;
      this.list = goList;
    }

    public TimeDistance getGoTime()
    {
      return goTime;
    }

    public List<LabeledTimePoint> getTimepoints()
    {
      return list;
    }

    @Override
    public String toString()
    {
      return "ExecutionDecision [goTime=" + goTime + ", list=" + list + "]";
    }
    
    

  }

  protected final Logger logger = Logger.getLogger(this.getClass().getName());
  
  protected ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> stnu;

  protected Map<LabeledTimePoint, TimeDistance> partialSchedule = new HashMap<LabeledTimePoint, TimeDistance>();

  //we denote by $\mu(\psi) = \max\{\psi(t)\mid t\in Dom(\psi)\}$ the maximum execution time of timepoints appearing in $\psi$
  protected TimeDistance maximumExecutionTime;

  // we denote by U(y) = fx j x 62 Dom(y)g the set of unexecuted timepoints
  protected HashSet<LabeledTimePoint> unexecutedTimePoints;
  
  // by Ux(y)  U(y) the set of noncontingent unexecuted timepoints
  protected HashSet<LabeledTimePoint> nonContUnexecutedTimePoints;
  
  // by Uc(y) U(y)   the set of contingent unexecuted timepoints
  protected HashSet<LabeledTimePoint> contUnexecutedTimePoints;
  
  // by   Ua(y) Uc(y) the set of contingent activated unexecuted timepoints
  protected HashSet<LabeledTimePoint> contActivatedUnexecutedTimePoints;

  protected TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> solver;
  
  public RTEDExecutionStrategy(TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> solver,  ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> stnu) throws TCNInconsistencyException, InterruptedException
  {
    this.solver = solver;
    this.stnu = stnu;
    
    
    partialSchedule.put(this.stnu.getTimeBaseEvent(), DefaultTimeDistance.ZERO_TIME_DISTANCE);    
    
    maximumExecutionTime = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    
    unexecutedTimePoints = new HashSet<LabeledTimePoint>(stnu.getVertices());
    unexecutedTimePoints.remove(this.stnu.getTimeBaseEvent());
    
    nonContUnexecutedTimePoints = new HashSet<LabeledTimePoint>(unexecutedTimePoints);
    
    contUnexecutedTimePoints = new HashSet<LabeledTimePoint>();
    contActivatedUnexecutedTimePoints = new HashSet<LabeledTimePoint>();
    
    for (LabeledEdge edge : stnu.getEdges())
    {
      if (edge.isContingentDuration())
      {
        for (LabeledValue v : edge.getValue().getLabeledValues())
        {
          if (v instanceof ContingentLabeledValue)
          {
            switch (((ContingentLabeledValue) v).getCase())
            {
              case UPPERCASE:
                
                nonContUnexecutedTimePoints.remove(edge.getSourceNode());
                contUnexecutedTimePoints.add(edge.getSourceNode());
                
                if (edge.getTargetNode().equals(this.stnu.getTimeBaseEvent()))
                  contActivatedUnexecutedTimePoints.add(edge.getSourceNode());
                break;
              case LOWERCASE:
                
                nonContUnexecutedTimePoints.remove(edge.getTargetNode());
                contUnexecutedTimePoints.add(edge.getTargetNode());
                
                if (edge.getSourceNode().equals(this.stnu.getTimeBaseEvent()))
                  activate(edge);
                break;
              default:
                throw new RuntimeException("Missing case: "+((ContingentLabeledValue) v).getCase());
            }

          }
        }
      }
    }
    
    updateNow(DefaultTimeDistance.ZERO_TIME_DISTANCE);
    if (!stnu.isControllable())
      solver.solve(stnu);
    
  }
  


  private void updateNow(TimeDistance now)
  {
      LabeledTimePoint Z = stnu.getTimeBaseEvent();

     for (LabeledTimePoint tp : nonContUnexecutedTimePoints)
     {
       LabeledEdge edge = stnu.getEdge(tp, Z);
       LabeledValueSet newValues;
       if (edge != null)
       {
         LabeledValueSet values = edge.getValue();
         
        newValues = values.addOrReplaceIfNecessary(new DefaultLabeledValue(Proposition.EMPTY_PROPOSITION, now.neg()));
       }
       else
       {
         newValues = new DefaultLabeledValueSet(new DefaultLabeledValue(Proposition.EMPTY_PROPOSITION, now.neg()));
       }
       
       LabeledEdge newEdge = new DefaultLabeledEdge(tp, Z, newValues, edge != null ? edge.isControlEdge() : false, 
                                                                                   edge != null ? edge.isImplicit() : true, 
                                                                                   edge != null ? edge.getFlowDirection() : FlowDirection.BACKWARD);
       stnu.updateConstraint(newEdge);
     }
  }



  public ExecutionDecision getExecutionDecision()
  {
    List<LabeledTimePoint> goList = new ArrayList<LabeledTimePoint>();
    TimeDistance goTime =  DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    
//    \eIf(\tcp*[f]{WAIT RTED!}){$(U^x(\psi)=\emptyset)$}{
    if (nonContUnexecutedTimePoints.isEmpty()) {
//      $\delta(\psi):= (\tau^x,\chi)$, where $\tau^x := \infty$ and $\chi:=\emptyset$\;
      goList.clear();
      goTime = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    }
//  }(\tcp*[f]{$(\tau, \chi)$ RTED!}){
    else
    {
//    \ForEach{$(x\in U^x(\psi))$}{
TPLoop:      for (LabeledTimePoint tp : nonContUnexecutedTimePoints)
      {
//        $[m(x),M(x)]$ = current time-window for $x$\;
        LabeledEdge maxEdge = stnu.getEdge(stnu.getTimeBaseEvent(), tp);
        LabeledEdge minEdge = stnu.getEdge(tp, stnu.getTimeBaseEvent());
        
        
        TimeDistance lowerBound = minEdge != null ? getValue(minEdge).neg() : DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
        TimeDistance upperBound = maxEdge != null ? getValue(maxEdge) : DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
        
//        $W(x):=-\infty$\;
        TimeDistance waits = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
        
//        \ForEach{$((A_i,C_i) \mid C_i\in U^a(\psi) \wedge \requirement{x}{C_i{:}{-w_i}}{A_i})$}{
        for (LabeledEdge e : stnu.getOutgoingEdges(tp))
        {
   
          
          for (LabeledValue value : e.getValue().getLabeledValues())
          {
            //\requirement{x}{C_i{:}{-w_i}}{A_i})
             if (value instanceof ContingentLabeledValue)
             {
               ContingentLabeledValue contValue = (ContingentLabeledValue) value;
               
               if (!ConstraintCase.UPPERCASE.equals(contValue.getCase()))
                 continue;
               
               //C_i\in U^a(\psi)
               if (contActivatedUnexecutedTimePoints.contains(contValue.getTrigger()))
               {
//               $W(x):= \max\{W(x), \psi(A_i) + w_i\}$\;
                 TimeDistance wait_i = contValue.getContingentValue().neg();
                 TimeDistance psiA_i = partialSchedule.get(e.getTargetNode());
                 waits = max(waits, psiA_i.add(wait_i));
               }
               // Activation time point of the wait has not yet been executed. Is this possible?
               else if (!partialSchedule.containsKey(e.getTargetNode()))
               {
                 // It is not possible to properly determine the wait for this time point yet. We need to
                 // wait till the activation time point of the wait constraint has been executed...
                 logger.warning("Activation time point of wait has not yet been executed!");
                 //waits = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
                 continue TPLoop;
                 //break;
               }
             }
             
             // If we have an outgoing edge with a negative value to an unexecuted timepoint we first need to wait for that timepoint to occur
             // This should actually only be a problem for contingent timepoints
             if (value.getValue().compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) < 0 && unexecutedTimePoints.contains(e.getTargetNode()))
             {
               continue TPLoop;
             }
          }
//        }
        }
        
        
//        $floor(x):= \max\{m(x), W(x)\}$\;
//        $go(x) := \min\{floor(x),M(x)\}$\;
        TimeDistance go = max(lowerBound, waits);
        
        //assert go.compareTo(upperBound) <= 0 : "Go time is "+go+" but upper bound is "+upperBound;
        
        if (goTime.compareTo(go) > 0)
        {
          goTime = go;
          goList.clear();
          goList.add(tp);
          
        }
        else if (goTime.compareTo(go) == 0)
        {
          goList.add(tp);
        }
        
//      }
      }
//      $\delta(\psi) := (\tau^x,\chi)$, where $\tau^x := \min\{go(x) \mid x \in U^x(\psi)\}$ and $\chi:=\{x\in U^x(\psi)\mid \tau^x =go(x)\}$\;
//    }
    }

//  \KwRet{$\delta(\psi);$}
    return new ExecutionDecision(goTime, goList);
  }

  
  public Collection<LabeledTimePoint> getActivatedContingents()
  {
    return contActivatedUnexecutedTimePoints;
  }
  
  public void execute(Collection<LabeledTimePoint> timepoints, TimeDistance now) throws TCNInconsistencyException, InterruptedException
  {
    LabeledValue maxValue = new DefaultLabeledValue(stnu.getTimeBaseEvent().getLabel(), now);
    LabeledValue minValue = new DefaultLabeledValue(stnu.getTimeBaseEvent().getLabel(), now.neg());
    
    
    for (LabeledTimePoint tp : timepoints)
    {
      // Add Timepoint to partial schedule
      partialSchedule.put(tp, now);
      // Remove from list of unexecuted time points
      unexecutedTimePoints.remove(tp);
      contUnexecutedTimePoints.remove(tp);
      nonContUnexecutedTimePoints.remove(tp);
      
      // remove from list of activated contingent time points (if contingent time point)
      if (contActivatedUnexecutedTimePoints.remove(tp))
      {
     // if removed this has been a contingent time point. we need to remove all the waits...
        executeContingentTimepoint(tp);
      }
      
      Collection<LabeledEdge> outgoingEdges = stnu.getOutgoingEdges(tp);
      for (LabeledEdge edge : outgoingEdges)
      {
        if (edge.isContingentDuration())
        {
          activate(edge);
        }
      }
      
      stnu.updateConstraint(new DefaultLabeledEdge(stnu.getTimeBaseEvent(), tp, new DefaultLabeledValueSet(maxValue)));
      stnu.updateConstraint(new DefaultLabeledEdge(tp, stnu.getTimeBaseEvent(), new DefaultLabeledValueSet(minValue)));
      
    }
    
    updateNow(now);
    
    solver.solve(stnu);
  }

  protected void activate(LabeledEdge edge)
  {
    contActivatedUnexecutedTimePoints.add(edge.getTargetNode());
  }

  private void executeContingentTimepoint(LabeledTimePoint contingentTP)
  {
    
    Collection<LabeledEdge> incomingEdges = stnu.getIncomingEdges(contingentTP);
    for (LabeledEdge edge : incomingEdges)
    {
      if (edge.isContingentDuration())
      {
        ContingentLabeledValue contValue = (ContingentLabeledValue) edge.getValue().getValues()[0];
        
        if (contValue.getTrigger().equals(contingentTP))
        {
          LabeledTimePoint activationTP = edge.getSourceNode();
          
          // Remove the lower case.
          LabeledEdge lowerCaseEdge = stnu.getEdge(activationTP, contingentTP);
          LabeledValue value = getConstraintValue(lowerCaseEdge);
          stnu.updateContingentConstraint(new DefaultLabeledEdge(activationTP, contingentTP, 
                                                                 new DefaultLabeledValueSet(new DefaultLabeledValue(value.getLabel(), value.getValue())),
                                                                 lowerCaseEdge.isControlEdge(), lowerCaseEdge.isImplicit(), lowerCaseEdge.getFlowDirection()));
          
          handleWaits(contingentTP, activationTP);
        }
      }
    }
  }

  private void handleWaits(LabeledTimePoint contingentTP, LabeledTimePoint activationTP)
  {
    Collection<LabeledEdge> incomingEdges = stnu.getIncomingEdges(activationTP);
    
    for (LabeledEdge e : incomingEdges)
    {
      if (e.isContingent())
      {
        List<LabeledValue> values = new ArrayList<LabeledValue>();
        
        for (LabeledValue value : e.getValue().getLabeledValues())
        {
        
          if (value instanceof ContingentLabeledValue)
          {
            ContingentLabeledValue cValue = (ContingentLabeledValue) getConstraintValue(e);
            if (cValue.getTrigger().equals(contingentTP))
            {
              values.add(new DefaultLabeledValue(cValue.getLabel(), cValue.getValue()));
            }
            else
            {
              values.add(value);
            }
          }
          else
          {
            values.add(value);
          }
        }
        
        stnu.updateContingentConstraint(new DefaultLabeledEdge(e.getSourceNode(), e.getTargetNode(), 
                                                               new DefaultLabeledValueSet(values), 
                                                               e.isControlEdge(), e.isImplicit(), e.getFlowDirection()));
      }
    }
  }
  
  protected TimeDistance max(TimeDistance v1, TimeDistance v2)
  {
    if (v1.compareTo(v2) < 0)
       return v2;
    else
      return v1;
  }

  protected TimeDistance getValue(LabeledEdge minEdge)
  {
    return getConstraintValue(minEdge).getValue();
  }

  protected LabeledValue getConstraintValue(LabeledEdge minEdge)
  {
    return minEdge.getValue().getLabeledValues().iterator().next();
  }
}
