package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;
import de.aristaflow.adept2.util.ArgChecks;

public abstract class AbstractTCN<N extends TCNTimePoint, V extends ConstraintValue, E extends TCNEdge<N, V>> implements
    TCN<N, E>, ChangeableTCN<N, E> 
{

  public static final class DirectedGraphImplementation<N extends TCNTimePoint, E extends TCNEdge<N, ?>> implements
      DirectedGraph<N, E>
  {
    private NodeMapping<N> nodeMapping;
  
    public DirectedGraphImplementation(
        NodeMapping<N> nodeMapping)
    {
      this.nodeMapping = nodeMapping;
    }
  
    @Override
    public Collection<N> getVertices()
    {
      return new ArrayList<N>();
    }
  
    @Override
    public Collection<E> getEdges()
    {
      return new ArrayList<E>();
    }
  
    @Override
    public E getEdge(N from, N to)
    {
      return null;
    }
  
    @Override
    public void visit(
        GraphVisitor<N, E> visitor)
    {
      visitor.discoverGraph(this);
      for (N e : this.nodeMapping.getNodes())
      {
        visitor.discoverVertex(e);
        visitor.finishVertex(e);
      }
      visitor.finishGraph(this);
    }
    
    @Override
    public String toString()
    {
      return this.nodeMapping.toString();
    }
  }

  @SuppressWarnings("deprecation")
  public static final LabeledTimePoint TIMEBASE_NODE = new DefaultLabeledNode("Origin", new NodeInstance(Integer.MIN_VALUE, new IterationChain(0), Integer.MIN_VALUE), Proposition.EMPTY_PROPOSITION)
  {

    /**
     * 
     */
    private static final long serialVersionUID = 1892338209580518661L;

    @Override
    public void addSupportedPlugin(String extensionPointID, String pluginID, PluginData data)
    {
      // Not supported
    }

    @Override
    public void setUserAttributeValue(String attributeName, String attributeValue)
    {
      // Not supported
    }

    @Override
    public void setUserAttributes(Map<String, String> userAttributes)
    {
   // Not supported
    }
    
    
  };
  
//  protected final UUID templateID;
  
  private Map<N, Integer> nodes;
  private int maxNodeID;
  private E[][] edges;

  
  private boolean consistent = false;


  private Set<E> changeSet;
  
  public AbstractTCN(Collection<N> nodes,
      Collection<E> edges)
  {

//    this.templateID = templateID != null ? templateID : new UUID(0, 0);
    this.maxNodeID = 0;
    this.nodes = new IdentityHashMap<N, Integer>();
    this.nodes.put(getTimeBaseEvent(), Integer.valueOf(maxNodeID++));
    for (N n : nodes)
    {
      if (!this.nodes.containsKey(n))
        this.nodes.put(n, Integer.valueOf(maxNodeID++));
    }
    
    ensureCapacity(maxNodeID);

    for (E edge : edges)
    {
      int source = getNodeKey(edge.getSourceNode());
      int target = getNodeKey(edge.getTargetNode());

      if (internalGetConstraint(source, target) != null)
        throw new IllegalStateException("Duplicate Edge: "+edge);
      
      //this.edges[source][target] = edge;
      internalSetConstraint(source, target, edge);
    }
    
    init();
    
    assert checkConsistency();
  }
  
  public abstract void init();

  protected boolean checkConsistency()
  {
    for (int i = 0; i < this.edges.length; i++)
    {
      for (int j = 0; j < i; j++)
      {
        if (!checkConsistency(this.edges[i][j], this.edges[j][i]))
          return false;
      }
    }
    
    for (N node : nodes.keySet())
    {
      if (!doCheckConsistency(node))
        return false;
    }
    

    
    return true;
  }

  protected abstract boolean doCheckConsistency(N node);
  
  protected boolean checkConsistency(E tcnEdge1, E tcnEdge2)
  {
    if (tcnEdge1 != null && tcnEdge2 != null)
    {
      return doCheckConsistency(tcnEdge1, tcnEdge2);
    }  
    return true;
  }

  protected abstract boolean doCheckConsistency(E edge1, E edge);
  
  @SuppressWarnings("unchecked")
  private void ensureCapacity(int size)
  {
    if (edges == null)
    {
      int newSize = size;//(int)(size*1.1);
      this.edges = (E[][])new TCNEdge[newSize][newSize];
    }
    else
    {
      // IMPROVE consider to perform a reorganization (i.e., removing or reusing unused rows and columns) (see removeTCNNode(N)); 
      if (this.edges.length <= size)
      {
        int newSize = (int)(size*1.1);
        E[][] newArray = (E[][])new TCNEdge[newSize][newSize];
        
        for (int i = 0; i < this.edges.length; i++)
        {
          System.arraycopy(this.edges[i], 0, newArray[i], 0, this.edges.length);
        }
        this.edges = newArray;
      }
    }
  }

  protected int getNodeKey(N node)
  {
    Integer key = this.nodes.get(node);
    
    if (key == null)
      throw new IllegalArgumentException("Node does not exist: "+node);
    
    assert key != null;
    
    return key.intValue();
  }

 
  @Override
  public E updateConstraint(E edge)
  {
    int source = getNodeKey(edge.getSourceNode());
    int target = getNodeKey(edge.getTargetNode());

    internalSetConstraint(source, target, edge);

    return edge;
  }

  protected void internalSetConstraint(int source, int target, E edge)
  {
//    ArgChecks.checkForNull(edge, "edge");
    this.edges[source][target] = edge;
    logChange(edge);
    
    modified();
    
    assert checkConsistency(this.edges[source][target], this.edges[target][source]);
  }
  
  public void removeEdge(E edge)
  {
    int source = getNodeKey(edge.getSourceNode());
    int target = getNodeKey(edge.getTargetNode());
    
    internalSetConstraint(source, target, null);
  }

  private void logChange(E edge)
  {
    if (this.changeSet != null)
    {
      if (edge != null)
        this.changeSet.add(edge); 
    }
  }
  
  public Collection<E> getChanges()
  {
    if (changeSet != null)
      return new ArrayList<E>(changeSet);
    return null;
  }

  protected void modified()
  {
    this.consistent = false;
  }

  @Override
  public Collection<N> getTimePoints()
  {
    return Collections.unmodifiableSet(this.nodes.keySet());
  }

  @Override
  public Collection<E> getEdges()
  {
    List<E> eg = new ArrayList<E>();
    for (int i = 0; i < maxNodeID; i++)
    {
      for (int j = 0; j < maxNodeID; j++)
      {
        E constraint = internalGetConstraint(i,j);
        if (constraint != null)
          eg.add(constraint);
      }
    }
    
    return Collections.unmodifiableList(eg);
  }

  @Override
  public E getEdge(N from, N to)
  {
    int fromID  = getNodeKey(from);
    int toID = getNodeKey(to);
    
    return internalGetConstraint(fromID, toID);
  }
  
  public Collection<E> getOutgoingEdges(N from)
  {
    int fromID  = getNodeKey(from);
    List<E> edges = new ArrayList<E>(maxNodeID);
    
    for (int i = 0; i< maxNodeID; i++)
    {
      E e = internalGetConstraint(fromID, i);
      if (e != null)
      {
        edges.add(e);
      }
    }
    return edges;
  }
  
  public Collection<E> getIncomingEdges(N to)
  {
    int toID  = getNodeKey(to);
    List<E> edges = new ArrayList<E>(maxNodeID);
    
    for (int i = 0; i< maxNodeID; i++)
    {
      E e = internalGetConstraint(i, toID);
      if (e != null)
      {
        edges.add(e);
      }
    }
    return edges;
  }

  protected E internalGetConstraint(int fromID, int toID)
  {
    return edges[fromID][toID];
  }

  @Override
  public void visit(GraphVisitor<N, E> visitor)
  {
    visitor.discoverGraph(this);
  
  
//    for (NodeMapping<N> n : nodeMappings)
//    {
//      
//      DirectedGraphImplementation<N, E> subGraph = new DirectedGraphImplementation<N, E>(n);
//      subGraph.visit(visitor);
//    }
    
    for (N e : getTimePoints())
    {
      visitor.discoverVertex(e);
      visitor.finishVertex(e);
    }
    
    for (N ei : getTimePoints())
    {
      for (N ej : getTimePoints())
      {
        E e = getEdge(ei, ej);
        if (e != null)
        {
          visitor.discoverEdge(e.getSourceNode(), 
                               e.getTargetNode(), 
                               e);
          visitor.finishEdge(e.getSourceNode(), 
                             e.getTargetNode(), 
                             e);
        }
      }
    }
  
    visitor.finishGraph(this);
  }

  
  
  public boolean contains(LabeledTimePoint timePoint)
  {
    return this.nodes.containsKey(timePoint);
  }

  
  /**
   * 
   * @param node
   * @throws IllegalArgumentException if the node is still part of any NodeMapping
   */
  public void removeTimePoint(N node) throws IllegalStateException
  {
//    for (NodeMapping<N> mapping : nodeMappings)
//    {
//      if (mapping.getNodes().contains(node))
//        throw new IllegalArgumentException("Node is still part of "+mapping+"! Please instead remove the mapping.");
//    }
    
    int nodeKey = getNodeKey(node);
    
    this.nodes.remove(node);
    
    for (int i = 0; i < edges.length; i++)
    {
      this.edges[nodeKey][i] = null;
      this.edges[i][nodeKey] = null;
    }
  }
  
  @Override
  public boolean isConsistent()
  {
    return this.consistent;
  }
  
  public void setConsistent()
  {
    this.consistent = true;
    if (this.changeSet == null)
      this.changeSet = new HashSet<E>();
    else
      this.changeSet.clear();
  }

//  public N getProcessCreationTimePoint()
//  {
//    return processCreationNode;
//  }
//
//  public N getProcessStartTimePoint()
//  {
//    return processStartNode;
//  }
//
//  public N getProcessEndTimePoint()
//  {
//    return processEndNode;
//  }


  
  public N insertTimePoint(N node)
  {
    if (this.nodes.containsKey(node))
      return node;
    
    int nodeID = Integer.MIN_VALUE;
    // Try to find an unused spot
    if (maxNodeID > this.nodes.size())
    {
      for (int i = 0; i < maxNodeID; i++)
      {
        if (!nodes.containsValue(Integer.valueOf(i)))
        {
          nodeID = i;
          break;
        }
      }
      
      if (nodeID == Integer.MIN_VALUE)
      {
        nodeID = maxNodeID++;
      }
      else
      {
        // Clear the row to be reused
        Arrays.fill(this.edges[nodeID], null);
        for (int i = 0; i < this.edges.length; i++)
        {
          this.edges[i][nodeID] = null;
        }
      }
    }
    else
    {
      nodeID = maxNodeID++;
    }
    
    this.nodes.put(node, Integer.valueOf(nodeID));
    modified();
    
    ensureCapacity(maxNodeID);

    return node; 
  }
  

  
  public Collection<N> getVertices()
  {
    return getTimePoints();
  }
  
  @Override
  public abstract TCN<N, E> clone();


  
}