package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Arrays;
import java.util.Map;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.processmodel.Edge;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.JoinNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.SplitNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTransformation;

public abstract class AbstractTemplateTransformation<E extends TCNEdge<LabeledTimePoint, ?>> extends
    TCNTransformation<LabeledTimePoint, E>
{

  public static final String CONF_XORSPLIT_MAX_DURATION = CONF_PREFIX + "XORSplitMaxDuration";
  
  public static TimeDistance DEFAULT_XORSPLIT_MAX_DURATION = new DefaultTimeDistance(5, 0, 0, 0, 0);
  
  protected  TimeDistance XORSPLIT_MAX_DURATION;

  public AbstractTemplateTransformation(Configuration configuration, TemporalTemplate template)
  {
    super(configuration, template);
  }

  public AbstractTemplateTransformation(Configuration configuration, TemporalTemplate template, int lookAhead)
  {
    super(configuration, template, lookAhead);
  }
  
  @Override
  protected void readConfiguration()
  {
    super.readConfiguration();
    XORSPLIT_MAX_DURATION = longToTimeDistance(configuration.getLong(CONF_XORSPLIT_MAX_DURATION,
      DEFAULT_XORSPLIT_MAX_DURATION.toMilliseconds()));
  }

  @Override
  protected LabeledTimePoint transformInstanceCreationEvent(IterationChain iterationChain, Label branch)
  {
    LabeledTimePoint instanceCreationEvent = new DefaultLabeledNode(new NodeInstance(-1, iterationChain, -1), "Instance", branch);

    // Create the instance Creation event
    instanceCreationEvent = addNode(instanceCreationEvent);
    return instanceCreationEvent;
  }



  @Override
  protected E[] transformSyncEdge(LabeledTimePoint from, LabeledTimePoint to, Label branches)
  {
    return createTCNEdge(from, to, branches, CONTROL_FLOW_MIN_TIME_LAG, CONTROL_FLOW_MAX_TIME_LAG, true);
  }

  @Override
  protected NodeMapping<LabeledTimePoint> transformProcessEndNode(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_ENDFLOW)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int topologicalID = template.getNodeTopologicalID(nodeID);

    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);

    addNode(A_S);
    addNode(A_E);

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

    return addNodeMapping(new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E));
  }

  protected LabeledTimePoint createTCNNode(NodeInstance nodeInstance, String eventID, Label branches)
  {
    return new DefaultLabeledNode(nodeInstance, eventID, branches);
  }

  @Override
  protected NodeMapping<LabeledTimePoint> transformProcessStartNode(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_STARTFLOW)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int topologicalID = template.getNodeTopologicalID(nodeID);

    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
    
    LabeledTimePoint A_S;
    if (NO_DISTINCT_TIMEBASE_EVENT)
    {
      A_S = getTimeBaseNode();
    }
    else
    {
      A_S = addNode(createTCNNode(nodeInstance, "S", branches));
    }   
    
    LabeledTimePoint A_E = addNode(createTCNNode(nodeInstance, "E", branches));

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

    return addNodeMapping(new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E));
  }

  @Override
  protected Label mergeANDBranches(Label baseBranch, Label[] newBranches)
  {
    // Will be calculate by CSTNU algorithm
    return baseBranch;
  }

  @Override
  protected Label mergeBranches(LabeledTimePoint node1, LabeledTimePoint node2)
  {
    Label label = node1.getLabel().merge(node2.getLabel());

    return label;
  }

  @Override
  protected SplitNodeMapping<LabeledTimePoint> transformXORSplitNode(int nodeID, IterationChain iterationChain,
      Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_XOR_SPLIT)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int[] succs = template.getSuccByEdgeType(nodeID, EdgeType.ET_CONTROL);

    // Map<Long, Label> decisionLabels = new HashMap<Long, Label>();
    Proposition parentProposition = (Proposition) branches;

    int topologicalID = template.getNodeTopologicalID(nodeID);

    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);

    // Label observationLabel = null;

    long[] edgeCodes = new long[succs.length];
    for (int i = 0; i < succs.length; i++)
    {
      int succ = succs[i];
      long edgeCode = getEdgeCode(nodeID, succ, EdgeType.ET_CONTROL);
      edgeCodes[i] = edgeCode;

      // Observation literal = Observation.createObservation(nodeInstance,
      // edgeCode, i+1, succs.length);
      // Proposition label = new Proposition(literal);
      // if (observationLabel == null)
      // observationLabel = label;
      //
      // decisionLabels.put(Long.valueOf(edgeCode), label);

    }

    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    DefaultObservationTimePoint A_E = createObservationNode(nodeInstance, "E", branches, edgeCodes);

    Map<Long, ObservationBinding> decisionLabels = A_E.getEdgeCodeMapping();

    addNode(A_S);
    addNode(A_E);

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, XORSPLIT_MAX_DURATION,
      true);

    LabeledTimePoint[] nodes = new LabeledTimePoint[succs.length];

    for (int i = 0; i < succs.length; i++)
    {
      int succ = succs[i];
      long edgeCode = getEdgeCode(nodeID, succ, EdgeType.ET_CONTROL);
      ObservationBinding observation = decisionLabels.get(Long.valueOf(edgeCode));

      Label xorBranch = parentProposition.addVariableBinding(observation);

      LabeledTimePoint s = createTCNNode(nodeInstance, "" + edgeCode, xorBranch);
      addNode(s);

      createTCNEdge(A_E, s, xorBranch, STRUCTURE_NODE_MIN_DURATION, XORSPLIT_MAX_DURATION, true);
      createTCNEdge(A_S, s, xorBranch, STRUCTURE_NODE_MIN_DURATION, XORSPLIT_MAX_DURATION, true);
      nodes[i] = s;
    }

    // Arrays.fill(nodes, A_E);

    return addNodeMapping(new SplitNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E, nodes));
  }

  protected DefaultObservationTimePoint createObservationNode(NodeInstance nodeInstance, String eventID, Label branches,
      long[] edgeCodes)
  {
    return new DefaultObservationTimePoint(nodeInstance, eventID, branches, edgeCodes);
  }

  @Override
  protected JoinNodeMapping<LabeledTimePoint> transformXORJoinNode(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_XOR_JOIN)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int topologicalID = template.getNodeTopologicalID(nodeID);
    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);

    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);

    addNode(A_S);
    addNode(A_E);

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

//    int[] preds = template.getPredByEdgeType(nodeID, EdgeType.ET_CONTROL);

//    LabeledTimePoint[] nodes = new LabeledTimePoint[preds.length];
//    Arrays.fill(nodes, A_S);

    return addNodeMapping(new JoinNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, null, A_S, null, A_E));
  }



  @Override
  protected SplitNodeMapping<LabeledTimePoint> transformANDSplit(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_AND_SPLIT)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int topologicalID = template.getNodeTopologicalID(nodeID);
    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);

    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);

    addNode(A_S);
    addNode(A_E);

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

//    int[] succs = template.getSuccByEdgeType(nodeID, EdgeType.ET_CONTROL);
//
//    LabeledTimePoint[] nodes = new LabeledTimePoint[succs.length];
//    Arrays.fill(nodes, A_E);
    return addNodeMapping(new SplitNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E, null));
  }

  @Override
  protected Label mergeXORBranches(Label branch, Label[] newBranchPaths)
  {
    return branch;
  }

  private long getEdgeCode(int startNodeID, int endNodeID, EdgeType edgeType)
  {
    Edge edge = template.getEdge(startNodeID, endNodeID, edgeType);

    return edge.getEdgeCode();
  }

  @Override
  protected JoinNodeMapping<LabeledTimePoint> transformLoopEndNode(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_ENDLOOP)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int topologicalID = template.getNodeTopologicalID(nodeID);
    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);

    addNode(A_S);
    addNode(A_E);

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

//    LabeledTimePoint[] nodes = new LabeledTimePoint[2];
//    Arrays.fill(nodes, A_S);

    return addNodeMapping(new JoinNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, null, A_S, null, A_E));
  }

  @Override
  protected SplitNodeMapping<LabeledTimePoint> transformLoopStartNode(int nodeID, IterationChain iterationChain,
      Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_STARTLOOP)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int topologicalID = template.getNodeTopologicalID(nodeID);

    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);

    // Map<Long, Label> decisionLabels = new HashMap<Long, Label>();

    int[] succsRepeat = template.getPredByEdgeType(nodeID, EdgeType.ET_LOOP);
    int[] succsExit = template.getSuccByEdgeType(nodeID, EdgeType.ET_CONTROL);

    long edgeCodeRepeat = getEdgeCode(succsRepeat[0], nodeID, EdgeType.ET_LOOP);
    long edgeCodeExit = getEdgeCode(nodeID, succsExit[0], EdgeType.ET_CONTROL);

    // Proposition observationLabel = new
    // Proposition(Observation.createObservation(nodeInstance, 0, REPEAT_BRANCH
    // +1, 2));
    // decisionLabels.put(Long.valueOf(edgeCodeRepeat), observationLabel);
    // decisionLabels.put(Long.valueOf(edgeCodeExit), new
    // Proposition(Observation.createObservation(nodeInstance, 1, EXIT_BRANCH
    // +1, 2)));

    long[] edgeCodes = new long[] { edgeCodeRepeat, edgeCodeExit };

    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    DefaultObservationTimePoint A_E = createObservationNode(nodeInstance, "E", branches, edgeCodes);

    Map<Long, ObservationBinding> decisionLabels = A_E.getEdgeCodeMapping();

    addNode(A_S);
    addNode(A_E);

    createTCNEdge(A_S, A_E, branches, DefaultTimeDistance.ZERO_TIME_DISTANCE, DefaultTimeDistance.ZERO_TIME_DISTANCE,
      true);

    Proposition parentProposition = (Proposition) branches;

    // FIXME todo fix loop structure (XOR at start vs XOR at end)

    Label repeatBranch = parentProposition.addVariableBinding(decisionLabels.get(Long.valueOf(edgeCodeRepeat)));

    Label exitBranch = parentProposition.addVariableBinding(decisionLabels.get(Long.valueOf(edgeCodeExit)));

    LabeledTimePoint[] nodes = new LabeledTimePoint[2];
    nodes[REPEAT_BRANCH] = createTCNNode(nodeInstance, "R", repeatBranch);
    addNode(nodes[REPEAT_BRANCH]);
    createTCNEdge(A_E, nodes[REPEAT_BRANCH], repeatBranch, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION,
      true);

    nodes[EXIT_BRANCH] = createTCNNode(nodeInstance, "L", exitBranch);
    addNode(nodes[EXIT_BRANCH]);

    createTCNEdge(A_E, nodes[EXIT_BRANCH], exitBranch, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

    return addNodeMapping(new SplitNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E, nodes));
  }

  @Override
  protected LabeledTimePoint getTimeBaseNode()
  {
    return AbstractTCN.TIMEBASE_NODE;
  }

}