package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.util.NullArgumentException;

/**
 * This object is immutable!
 * 
 * @author Andreas
 */
public class DefaultLabeledEdge implements TCNEdge<LabeledTimePoint, LabeledValueSet>, Cloneable, LabeledEdge
{
  private final LabeledTimePoint from;

  private final LabeledTimePoint to;

  private final LabeledValueSet labeledValueSet;

  private final boolean implicit;

  private final boolean controledge;

  private final FlowDirection flowDirection;

  
  public DefaultLabeledEdge(LabeledTimePoint from, LabeledTimePoint to,
      LabeledValueSet labelSet)
  {
    this(from, to, labelSet, false, false, FlowDirection.UNKOWN);
  }
  
  public DefaultLabeledEdge(LabeledTimePoint from, LabeledTimePoint to,
      LabeledValueSet labeledValueSet, boolean controledge, boolean implicit, FlowDirection flowDirection)
  {
    if (from == null || to == null || labeledValueSet == null)
      throw new NullArgumentException();
    
    assert labeledValueSet.satisfiable(from.getLabel()) : "Illegal Label combination! Edge value " + labeledValueSet + " vs. from node label set "+from.getLabel() + " @ "+from + " --> "+to;
    assert labeledValueSet.satisfiable(to.getLabel()) : "Illegal Label combination! Edge value " + labeledValueSet + " vs. to node label set "+to.getLabel() + " @ "+from + " --> "+to;
    
    assert labeledValueSet.subsumes(from.getLabel()) : "Illegal Label combination! Edge label " + labeledValueSet + " must subsume from label "+from.getLabel() + " @ "+from + " --> "+to;
    assert labeledValueSet.subsumes(to.getLabel()): "Illegal Label combination! Edge label " + labeledValueSet + " must subsume to label "+to.getLabel() + " @ "+from + " --> "+to;

    this.from = from;
    this.to = to;
    this.labeledValueSet = labeledValueSet;
    this.controledge = controledge;
    this.implicit = implicit;    
    this.flowDirection = flowDirection;
    
    assert checkContingentDuration();
  }


  private boolean checkContingentDuration()
  {
    for (LabeledValue l : labeledValueSet.getLabeledValues())
    {
      if (l instanceof ContingentLabeledValue)
      {
        ContingentLabeledValue cl = (ContingentLabeledValue) l;
        if ((cl.getTrigger() == this.getSourceNode()) || (cl.getTrigger() == this.getTargetNode()))
        {
          if (labeledValueSet.getLabeledValues().size() != 1)
            throw new IllegalArgumentException(String.valueOf(this));
        }
      }
    }
    return true;
  }

  public LabeledTimePoint getSourceNode()
  {
    return from;
  }


  public LabeledTimePoint getTargetNode()
  {
    return to;
  }

  
  @Override
  public String toString()
  {
    return from+" -["+labeledValueSet.toString()+"]-> "+to;
  }


  public boolean isContingentDuration()
  {
    for (LabeledValue l : labeledValueSet.getLabeledValues())
    {
      if (l instanceof ContingentLabeledValue)
      {
        ContingentLabeledValue cl = (ContingentLabeledValue) l;
        if ((cl.getTrigger() == this.getSourceNode()) || (cl.getTrigger() == this.getTargetNode()))
        {
          assert labeledValueSet.getLabeledValues().size() == 1 : String.valueOf(this);
          return true;
        }
        return false;
      }
    }
    return false;
  }
  
  public boolean isContingent()
  {
    for (LabeledValue l : labeledValueSet.getLabeledValues())
    {
      if (l instanceof ContingentLabeledValue)
      {
        return true;
      }
    }
    return false;
  }


  public LabeledValueSet getValue()
  {
    return this.labeledValueSet;
  }

  @Override
  public boolean isImplicit()
  {
    return implicit;
  }
  
  @Override
  public boolean isControlEdge()
  {
    return this.controledge;
  }

  @Override
  public int hashCode()
  {
    return this.from.hashCode() ^ this.to.hashCode();
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;
    
    if (obj instanceof LabeledEdge)
    {
      LabeledEdge other = (LabeledEdge) obj;
      boolean result = this.from.equals(other.getSourceNode());
        result &= this.to.equals(other.getTargetNode());
        result &= this.labeledValueSet.equals(other.getValue());
      return result;
    }
    
    return super.equals(obj);
  }

  @Override
  public LabeledEdge clone()
  {
    // LabeledNodes are immutable
    return new DefaultLabeledEdge(from, to, ((DefaultLabeledValueSet)labeledValueSet).clone(), controledge, implicit, this.flowDirection);
  }

  @Override
  public FlowDirection getFlowDirection()
  {
    return this.flowDirection;
  }

  

}
