package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.util.ArgChecks;

/**
 * Note: Object is immutable!
 * 
 * @immutable
 * 
 * @author lanz
 *
 */
public class DefaultLabeledValueSet implements LabeledValueSet
{

  private final Map<Label, LabeledValue> values;
  private final int cachedHashCode;
  
  public DefaultLabeledValueSet(LabeledValue... values)
  {
    this(Arrays.asList(values));
  }
  
  public DefaultLabeledValueSet(Collection<LabeledValue> values)
  {
    ArgChecks.checkForEmpty(values, "values");
    ArgChecks.checkForNulls(values, "values");
    
    Map<Label, LabeledValue> lbls = new HashMap<Label, LabeledValue>();
    
    for (LabeledValue labeledValue : values)
    {
      assert labeledValue.getLabel().satisfiable(labeledValue.getLabel()); // Make sure the label is satisfiable
      LabeledValue old = lbls.put(labeledValue.getLabel(), labeledValue);
      assert old == null;
    }
    
    this.values = Collections.unmodifiableMap(lbls); // Make map immutable
    this.cachedHashCode = this.values.hashCode();
  }
  
  private DefaultLabeledValueSet(Map<Label, LabeledValue> values)
  {
    ArgChecks.checkForEmpty(values.values(), "values");
    ArgChecks.checkForNulls(values.values(), "values");
    
    Map<Label, LabeledValue> lbls = new HashMap<Label, LabeledValue>(values);
    
    for (LabeledValue labeledValue : values.values())
    {
      assert labeledValue.getLabel().satisfiable(labeledValue.getLabel()); // Make sure the label is satisfiable
    }
    
    this.values = Collections.unmodifiableMap(lbls); // Make map immutable
    this.cachedHashCode = this.values.hashCode();
  }
  
  @Override
  public String toString()
  {
    StringBuilder builder = new StringBuilder();
    String sep = "";
    for (LabeledValue b : this.values.values())
    {
      builder.append(sep);
      builder.append(b.toString());
      sep = ";";
    }
    return builder.toString();
  }
  
  public Label[] getLabels()
  {    
    return this.values.keySet().toArray(new Label[this.values.size()]);
  }
  
  public boolean satisfiable(Label label)
  {
    for (Label b : getLabels())
    { 
      if (!b.satisfiable(label))
        return false;
    }
    
    return true;
  }

  public boolean subsumes(Label label)
  {
    for (Label l : getLabels())
    {
      if (!l.subsumes(label))
        return false;
    }
    return true;
  }

  public Collection<LabeledValue> getLabeledValues()
  {
    return values.values();
  }

  @Override
  public SimpleConstraintValue[] getValues()
  {
    return values.values().toArray(new SimpleConstraintValue[values.size()]);
  }
  
  @Override
  public int hashCode()
  {
    return cachedHashCode;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;
    
    if (obj instanceof DefaultLabeledValueSet)
    {
      DefaultLabeledValueSet other = (DefaultLabeledValueSet) obj;
      return this.values.equals(other.values);
    }
    
    return super.equals(obj);
  }
  
  @Override
  public LabeledValueSet clone()
  {
    // Labels are immutable, so a shallow copy is enough
    return new DefaultLabeledValueSet(this.values);
  }
  
  public LabeledValueSet merge(LabeledValueSet other)
  {
    Map<Label, LabeledValue> labelMapping = toMap(this);
    for (LabeledValue v : other.getLabeledValues())
    {
      addOrReplaceIfNecessary(labelMapping, v);
    }
    
    return new DefaultLabeledValueSet(labelMapping);
  }
  

  @Override
  public LabeledValueSet addOrReplaceIfNecessary(LabeledValue newLabel)
  {
    Map<Label, LabeledValue> map = toMap(this);
    addOrReplaceIfNecessary(map, newLabel);
    return new DefaultLabeledValueSet(map);
  }



  /*
   * Some static utility methods
   */
  
  public static Map<Label, LabeledValue> toMap(LabeledValueSet values)
  {
    Map<Label, LabeledValue> map = new HashMap<Label, LabeledValue>();
    if (values instanceof DefaultLabeledValueSet)
    {
      map.putAll(((DefaultLabeledValueSet) values).values);
    }
    else
    {
      for (LabeledValue v : values.getLabeledValues())
      {
        map.put(v.getLabel(), v);
      }
    }
    return map;
  }
  
  /**
   * 
   * @param labels
   * @param newValue
   * @return <code>true</code> if the map has changed because of the add operation
   */
  public static boolean addOrReplaceIfNecessary(Map<Label, LabeledValue> labels, LabeledValue newValue)
  {
    boolean changes = false;
    Label label = newValue.getLabel();
    LabeledValue oldValue = labels.get(label);
    if (oldValue != null)
    {
      LabeledValue value = merge(newValue, oldValue);
      
      if (value != null && value != oldValue)
      {
        labels.put(label, value);
        changes = true;
      }
    }
    else
    {
      labels.put(label, newValue);
      changes = true;
    }
    return changes;
  }

  public static LabeledValue merge(
      LabeledValue newValue, LabeledValue oldValue)
  {
    if (newValue instanceof ContingentLabeledValue)
    {
      if (oldValue instanceof ContingentLabeledValue)
      {
        // Both are contingent
        ContingentLabeledValue clNew = (ContingentLabeledValue) newValue;
        ContingentLabeledValue clOld = (ContingentLabeledValue) oldValue;
        
        
        int compareNonCont = clOld.getValue().compareTo(clNew.getValue());
        int compareCont = clOld.getContingentValue().compareTo(clNew.getContingentValue());
        
        // if one of the values has been decreased
        if (compareCont > 0 || compareNonCont > 0) //old > new
        {
          assert clNew.getCase() == clOld.getCase() : clOld + " vs. "+clNew;
          assert clNew.getTrigger() == clOld.getTrigger() : "Trying to merge contingent constraints with different triggers!";
          
          // Merge or replace
          if (compareCont >= 0 && compareNonCont >= 0) // old >= new
          {
            // Both have either been decreased or not changed so replace
            //labels.put(newValue.getLabel(), newValue);
            return newValue;
          }
          else // old < new
          {
            
            // Merge
            TimeDistance nonContValeu = compareNonCont > 0 ? clNew.getValue() : clOld.getValue();
            TimeDistance contValue = compareCont > 0 ? clNew.getContingentValue() : clOld.getContingentValue();
            //labels.put(clNew.getLabel(), createNewConstraint(clNew, nonContValeu, contValue));
            return updateLabeledValue(clNew, nonContValeu, contValue);
          }
        }
        else
        {
          return oldValue;
        }
      }
      else
      {
        // new is contingent old is non-contingent
        // Merge
        
        ContingentLabeledValue clNew = (ContingentLabeledValue) newValue;
                 
        TimeDistance oldTD = oldValue.getValue();
        TimeDistance newTD = clNew.getValue();
        int compareNonCont = oldTD.compareTo(newTD);
        
        // if one of the values has been decreased
        if (compareNonCont >= 0) // old > new
        {
          return newValue;
        }
        else
        {
          // Merge
          TimeDistance contValue = clNew.getContingentValue();
          TimeDistance nonContValue = oldValue.getValue();
          //labels.put(clNew.getLabel(), createNewConstraint(clNew, nonContValue, contValue));
          return updateLabeledValue(clNew, nonContValue, contValue);
        }
      }
    
    }
    else
    {
      if (oldValue instanceof ContingentLabeledValue)
      {
        // new is non-contingent old is contingent
        // Merge
        ContingentLabeledValue clOld = (ContingentLabeledValue) oldValue;
        
        int compareNonCont = clOld.getValue().compareTo(newValue.getValue());
        
        // if one of the values has been decreased
        if (compareNonCont > 0)
        {
          // Merge
          TimeDistance nonContValue = newValue.getValue();
          TimeDistance contValue = clOld.getContingentValue();
//          labels.put(clOld.getLabel(), createNewConstraint(clOld, nonContValue, contValue));
          return updateLabeledValue(clOld, nonContValue, contValue);
        }
        else
        {
          return oldValue;
        }
      }
      else
      {
        // Both are non-contingent
        // Replace if necessary
        int compare = oldValue.getValue().compareTo(newValue.getValue());
        if (compare > 0) // old > new
        {
//          labels.put(newValue.getLabel(), newValue);
          return newValue;
        }
        else
        {
          return oldValue;
        }
      }
    }
    //return oldValue;
  }

  /**
   * Updates the {@link ContingentLabeledValue} with the non-contingent and
   * contingent value. Returns a {@link LabeledValue} (i.e., non-contingent) or
   * {@link ContingentLabeledValue} depending on whether or not
   * <code>nonContValue &lt;= contValue</code>. This ensures that for a
   * {@link ContingentLabeledValue} the contingent values is always bigger than
   * the non-contingent value (i.e., the contingent value is more restrictive;
   * otherwise the contingent values is superfluous).
   * 
   * @param oldValue
   * @param nonContValue
   * @param contValue
   * @return a {@link LabeledValue} (i.e., non-contingent) or
   *         {@link ContingentLabeledValue} depending on whether or not
   *         <code>nonContValue &lt;= contValue</code>.
   */
  protected static LabeledValue updateLabeledValue(ContingentLabeledValue oldValue,
      TimeDistance nonContValue, TimeDistance contValue)
  {
    LabeledValue value;
    if (nonContValue.compareTo(contValue) <= 0) // Non-contingent value is more
                                               // restrictive than contingent
    {
      value = new DefaultLabeledValue(oldValue.getLabel(), nonContValue);
    }
    else
    {
      value = new DefaultContingentLabeledValue(oldValue.getLabel(), oldValue.getCase(),
                                    nonContValue, oldValue.getTrigger(), contValue);
    }
    return value;
  }

}
