package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.ObservationTimePoint;

/**
 * 
 * <p><b>Note:</b> This object is immutable!</p>
 * 
 * @author Andreas
 *
 */
public class DefaultObservationTimePoint extends DefaultLabeledNode implements ObservationTimePoint
{

  /**
   * 
   */
  private static final long serialVersionUID = -837847718619760659L;
  
  private final Map<Long, ObservationBinding> decisionLabels;
  private Observation observation;

  /**
   * An Observation with a label representing a specific event of a node instance
   * 
   * @param nodeInstance the node instance (i.e., node and iteration) this labeled tcn node belongs to
   * @param eventID the event of the <code>nodeInstance</code> this labeled tcn node describes
   * @param label the label representing the process instance types this labeled node belongs to
   * @param edgeCodes the edge codes of the process
   * @param observationIDs the observation ids corresponding to the given edge codes (i.e., <code>edgeCodes[i]</code> is the edge code for <code>observationIDs[i]</code>)
   */
  public DefaultObservationTimePoint(NodeInstance nodeInstance, String eventID, Label label, long[] edgeCodes, int[] observationIDs)
  {
    super(nodeInstance, eventID, label);
    if (edgeCodes.length != observationIDs.length)
      throw new IllegalArgumentException("edgeCodes must have same length than observationIDs");
    
    assert edgeCodes.length > 1;
    this.observation = new Observation(this, edgeCodes.length);
    
    this.decisionLabels = new HashMap<Long, ObservationBinding>();
    for (int i = 0; i < edgeCodes.length; i++)
    {
      long edgeCode = edgeCodes[i];
      this.decisionLabels.put(Long.valueOf(edgeCode), 
                              ObservationBinding.createObservationBinding(this.observation, edgeCode, observationIDs[i]));
    }
  }
  
  /**
   * An Observation with a label representing a specific event of a node instance
   * 
   * @param nodeInstance the node instance (i.e., node and iteration) this labeled tcn node belongs to
   * @param eventID the event of the <code>nodeInstance</code> this labeled tcn node describes
   * @param labels the set of labels representing the process instance types this labeled node belongs to
   * @param edgeCodes edge codes of the process
   */
  public DefaultObservationTimePoint(NodeInstance nodeInstance, String eventID, Label label, long[] edgeCodes)
  {
    super(nodeInstance, eventID, label);
    assert edgeCodes.length > 1;
    this.observation = new Observation(this, edgeCodes.length);
    
    this.decisionLabels = new HashMap<Long, ObservationBinding>();
    for (int i = 0; i < edgeCodes.length; i++)
    {
      long edgeCode = edgeCodes[i];
      this.decisionLabels.put(Long.valueOf(edgeCode), 
                              ObservationBinding.createObservationBinding(this.observation, edgeCode, i));
    }
  }


  @Override
  public String toString()
  {
    return "X "+super.toString();
  }

  /**
   * 
   * @return The decision labels of all branches resulting from this observation node
   */
  public ObservationBinding[] getChoiceVariableBindings()
  {
    return decisionLabels.values().toArray(new ObservationBinding[0]);
  }
  
  public Map<Long, ObservationBinding> getEdgeCodeMapping()
  {
    return Collections.unmodifiableMap(this.decisionLabels);
  }

  /**
   * 
   * @return the label observed by this observation node.
   */
  public Observation getObservation()
  {
    return this.observation;
  }
  
  /**
   * 
   * @return the label observed by this observation node.
   */
  public ObservationBinding getObservationForEdgeCode(long edgeCode)
  {
    return this.decisionLabels.get(Long.valueOf(edgeCode));
  }

  public ChoiceVariable getChoiceVariable()
  {
    return this.observation;
  }
  
//  public Label getObservationLabel()
//  {
//    return new Proposition(this.observation);
//  }
  
}
