package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.util.NullArgumentException;

/**
 * <p>
 * An <code>Observation</code> represents the variable observed by a particular
 * <code>ObservationNode</code> (given by {@link #observationNode}). That is, an
 * <code>Observation</code> encodes the possible values of the variable observed
 * by the respective Observation node.
 * </p>
 * 
 * 
 * see e.g., Conrad et al. (2011)
 * 
 * <pre>
 * Definition 3.5 (Choice Variables) Each choice is associated with a finite domain variable x_i. 
 * Each of these variables has a domain with size equal to the number of options
 * of that choice. An assignment is a selection of a single option for a choice, 
 * represented as an assignment to the choice's associated choice variable.
 * </pre>
 * 
 * <p><b>Note:</b> This object is immutable!</p>
 * @author lanz
 * @immutable
 * 
 */
public class Observation implements ChoiceVariable, Comparable<Observation>
{

  private final DefaultObservationTimePoint observationNode;

  private final int options;

  public Observation(DefaultObservationTimePoint observationNode, int options)
  {
    if (observationNode == null)
      throw new NullArgumentException("splitNodeIteration");

    this.observationNode = observationNode;
    this.options = options;

  }

  @Override
  public String toString()
  {
    int ofM = this.getOptions();

    StringBuilder s = new StringBuilder();
    s.append("(");
    s.append(observationNode);
    s.append(';');
    s.append('*');
    s.append(':');
    s.append(ofM);
    s.append(")");
    return s.toString();
  }

  @Override
  public int hashCode()
  {
    return (this.observationNode.hashCode() * 31);
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;

    if (obj instanceof Observation)
    {
      Observation other = (Observation) obj;
      return this.observationNode.equals(other.getDecisionTimePoint());
    }

    return super.equals(obj);
  }

  public DefaultObservationTimePoint getDecisionTimePoint()
  {
    return observationNode;
  }

  public int getOptions()
  {
    return options;
  }

  @Override
  public int compareTo(Observation o)
  {
    return this.observationNode.compareTo(o.observationNode);
  }

}