package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;


import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariableBinding;

/**
 * <p>An <code>ObservationBinding</code> represents an observation made by a particular <code>ObservationNode</code> (given 
 * by {@link #observation}). In particular, an <code>ObservationBinding</code> encodes the value bound to the variable
 * observed by the respective Observation node.</p>
 * 
 * <p>Note that the {@link #getValue()} may only be interpreted with respect to the respective {@link Observation}/{@link DefaultObservationTimePoint}. 
 * In particular, two {@link ObservationBinding}s may have the same {@link #getValue()} but belong to different {@link Observation}s 
 * and thus cannot be compared.</p>  
 * 
 * 
 * see e.g., Conrad et al. (2011)
 * 
 * <pre>
 * Definition 3.5 (Choice Variables) Each choice is associated with a finite domain variable x_i. 
 * Each of these variables has a domain with size equal to the number of options
 * of that choice. An assignment is a selection of a single option for a choice, 
 * represented as an assignment to the choice's associated choice variable.
 * </pre>
 * 
 * <p><b>Note:</b> This object is immutable!</p>
 * @author Andreas
 */
public class ObservationBinding implements ChoiceVariableBinding, Comparable<ObservationBinding> 
{
  
  private final long edgeCode;
  
  private final int value;

  private final Observation observation;
  
  /**
   * 
   * @param observationNode The node instance of the {@link DefaultObservationTimePoint} this observation belongs to.
   * @param edgeCode The edge code in the {@link Template} this observation belongs to. (This is only necessary for 
   *  linking the TCN with the process model).
   * @param value The branchID of the particular branch this {@link Observation} belongs to. This encodes the
   *  particular observation (see {@link Observation} for details of the encoding).
   */
  public ObservationBinding(Observation observation, long edgeCode, int value)
  {
    if (value < 0)
      throw new IllegalArgumentException("n");
    int ofM = observation.getOptions();
    if (value >= ofM)
      throw new IllegalArgumentException("n");
    
    this.observation = observation;
    this.edgeCode = edgeCode;
    this.value = value;
  }
  
  /**
   * Creates a new observation for a particular branch of an {@link DefaultObservationTimePoint}. Thereby <code>n</code> and 
   * <code>ofM</code> represent that this is the <code>n</code>-th branch out of <code>ofM</code> branches leaving
   * the observation node. This is used to determine the number of propositional letters required to encode the 
   * different branches and to determine the encoding of the respective branch.
   * 
   * @param observationNode The node instance of the {@link DefaultObservationTimePoint}
   * @param edgeCode The edge code in the {@link Template} this observation belongs to.
   * @param n The number of this branch with respective to the overall number of branches (<code>ofM</code>).
   * @param ofM The number of branches leaving this particular observation node.
   * @return The {@link Observation} encoding the particular branch of the {@link DefaultObservationTimePoint}.
   */
  public static ObservationBinding createObservationBinding(Observation observation, long edgeCode, int n)
  {
    
    if (n < 0)
      throw new IllegalArgumentException("n");
    int ofM = observation.getOptions();
    if (n >= ofM)
      throw new IllegalArgumentException("n");
    
    assert n >= 0;
    assert ofM > n;
    
//    int branchID = ofM << 16;
//    branchID += n-1;
    
    return new ObservationBinding(observation, edgeCode, n);
  }
  
  /**
   * 
   * @return The branchID encoding the particular branch of the {@link DefaultObservationTimePoint} (see {@link Observation} for 
   * details of the encoding).
   */
  public int getValue()
  {
    return value;
  }

  @Override
  public int hashCode()
  {
    return (this.observation.hashCode() * 31) ^ this.value;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;
    
    if (obj instanceof ObservationBinding)
    {
      ObservationBinding other = (ObservationBinding) obj;
      return this.observation.equals(other.getVariable()) && this.value == other.value;
    }
    
    return super.equals(obj);
  }
  
  public Observation getVariable()
  {
    return observation;
  }

  public long getEdgeCode()
  {
    return edgeCode;
  }
  
  @Override
  public String toString()
  {
    int n = value;
    int ofM = observation.getOptions();

    
    StringBuilder s = new StringBuilder();
    s.append("(");
    s.append(observation.getDecisionTimePoint());
    s.append(';');
    s.append(n+1);
    s.append(':');
    s.append(ofM);
    s.append(")");
    return s.toString();
  }

  @Override
  public int compareTo(ObservationBinding other)
  {
    int compare = this.observation.compareTo(other.observation);
        
    if (compare != 0)
      return compare;
    
    int thisVal = this.value;
    int anotherVal = other.value;
    compare = thisVal < anotherVal ? -1 : (thisVal == anotherVal ? 0 : 1);
    return compare;
  }
}
