package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Collection;
import java.util.Queue;
import java.util.logging.Level;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Path;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tools.DuplicateFreeQueue;

/**
 * A simple basic PC2 Solver (Lanz08 S. 85). </p>
 * 
 * Algorithm:
 * 
 * <pre>
 * Q = {(i, k, j) | (i &lt; j) and (k != i,j)}
 * while !isEmpty(Q) do
 *   select and delete a path (i, k, j) from Q
 *   if T_ij != T_ij &amp; (T_ik + T_kj)
 *     T_ij = T_ij &amp; (T_ik + T_kj)
 *     if isEmpty(T_ij) then throw InconsistencyException
 *     Q = Q || {(i, j, k), (k, i, j) | 1 &lt;= k &lt;= n, i != k != j}
 *   end
 * end
 * </pre>
 * 
 * 
 */
public class PC2Solver<E extends STNEdge> extends AbstractSTNSolver<LabeledTimePoint, SimpleConstraintValue, E> implements TCNSolver<LabeledTimePoint, SimpleConstraintValue, E>
{
  
  public PC2Solver()
  {
    super();
  }
  
  protected PC2Solver(Level debugLevel)
  {
    super(debugLevel);
  }
  
  @SuppressWarnings("unchecked")
  public <T extends ChangeableTCN<LabeledTimePoint, E>> T solve(T tcn)
      throws TCNInconsistencyException, InterruptedException
  {
    if (tcn instanceof ChangeableSTN<?, ?>)
    {
      return (T)solve((ChangeableSTN<LabeledTimePoint, E>)tcn, null, null);
    }

    throw new IllegalArgumentException();
  }

  @SuppressWarnings("unchecked")
  public <T extends ChangeableTCN<LabeledTimePoint, E>> T solve(T tcn,
      ChangeSet<E, SimpleConstraintValue, Triangle<LabeledTimePoint>> changeSet)
      throws TCNInconsistencyException, InterruptedException
  {
    if (tcn instanceof ChangeableSTN<?, ?>)
    {
      return (T)solve((ChangeableSTN<LabeledTimePoint, E>)tcn, null, changeSet);
    }

    throw new IllegalArgumentException();
  }
  
  /**
   * A simple basic PC2 Solver. </p>
   * 
   * Algorithm:
   * 
   * <pre>
   * Q = {(i, k, j) | (i &lt; j) and (k != i,j)}
   * while !isEmpty(Q) do
   *   select and delete a path (i, k, j) from Q
   *   if T_ij != T_ik + T_kj
   *     T_ij = T_ij &amp; T_ik + T_kj
   *     if isEmpty(T_ij) then throw InconsistencyException
   *     Q = Q || {(i, j, k), (k, i, j) | 1 &lt;= k &lt;= n, i != k != j}
   *   end
   * end
   * </pre>
   * 
   * @return
   * @throws InterruptedException 
   * 
   */
  public STN<LabeledTimePoint, E> solve(
      ChangeableSTN<LabeledTimePoint, E> model, Collection<Path<LabeledTimePoint>> changed,
      ChangeSet<E, SimpleConstraintValue, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException
  {   
    assert debug(Level.FINE, "Starting PC2");
    Collection<LabeledTimePoint> timePoints = model.getTimePoints();
    int nodeCount = timePoints.size();
    final Queue<Path<LabeledTimePoint>> queue = new DuplicateFreeQueue<Path<LabeledTimePoint>>(nodeCount*nodeCount*2);
    assert debug(Level.FINE, "Initializing Paths");
    if (changed == null)
    {
      getInitialPaths(timePoints, queue);
    }
    else
    {
//      for (Path<LabeledTimePoint> e : changed)
//      {
//        queue.add(e);
//      }
      queue.addAll(changed);
    }

    while (!queue.isEmpty())
    {
      if (Thread.currentThread().isInterrupted())
        throw new InterruptedException();
      
      final Path<LabeledTimePoint> triangle = queue.remove();

      LabeledTimePoint i = triangle.getFirst();
      LabeledTimePoint j = triangle.getSecond();
//      LabeledTimePoint k = triangle.getThird();
      

      for (LabeledTimePoint k : timePoints)
      {
      
        if (i.equals(k) || j.equals(k))
          continue;
        
        if (processPath(model, changeSet, new Path<LabeledTimePoint>(i, k), j))
        {
          assert debug(Level.FINEST, "Adding related Paths for %s -> %s", i, k);
          queue.add(new Path<LabeledTimePoint>(i, k));
        }
        
        if (processPath(model, changeSet, new Path<LabeledTimePoint>(k, j), i))
        {
          assert debug(Level.FINEST, "Adding related Paths for %s -> %s", k, j);
          queue.add(new Path<LabeledTimePoint>(k, j));
        }
      }
    }

    
    model.setConsistent();
    return model;
  }


  private boolean processPath(ChangeableSTN<LabeledTimePoint, E> model,
      ChangeSet<E, SimpleConstraintValue, Triangle<LabeledTimePoint>> changeSet, Path<LabeledTimePoint> p, LabeledTimePoint via) throws TCNInconsistencyException
  {
    LabeledTimePoint from = p.getFrom();
    LabeledTimePoint to = p.getTo();
    
    final E e_ij = model.getEdge(from, via);
    
    final E e_jk = model.getEdge(via, to);
//    

    if (e_ij == null || e_jk == null)
      return false;
          
    assert model != null;
    assert e_ij != null;
    assert e_jk != null;
    
    assert e_ij.getTargetNode() == e_jk.getSourceNode();
    
    final E e_ik = model.getEdge(from, to);

    assert debug(Level.FINE, "Evaluating (%s, %s) = (%s) \\cap (%s) + (%s)", from, to, e_ik, e_ij, e_jk);
    
//    final E e_ik = model.getEdge(i, k);
    
    SimpleConstraintValue ls_ij = e_ij.getValue();
    SimpleConstraintValue ls_jk = e_jk.getValue();
    
    TimeDistance v_ij = ls_ij.getValue();
    TimeDistance v_jk = ls_jk.getValue();
    
    // If one or both are infinity the rest is useless as the sum will also be infinity which is always greater or equals than the current value
//    if (timeModelOperations.isInfinity(v_ij) != 0 || timeModelOperations.isInfinity(v_jk) != 0)
//      continue;
    
    TimeDistance sum;
    try
    {
      sum = v_ij.add(v_jk);
    }
    catch (IllegalArgumentException e1) // infinity + (- infinity)
    {
      throw new TCNInconsistencyException(e1.getMessage(), p, via, model);
    }
    
    assert debug(Level.FINER, "\tSum\t (%s) + (%s) = (%s)", v_ij, v_jk, sum);
    
    TimeDistance v_ik = null;
    if (e_ik != null)
    {
      v_ik = getTimeDistance(e_ik);
    }
    
    
    TimeDistance newValue;
    if (v_ik != null)
    {
      newValue = intersect(v_ik, sum);
    }
    else
    {
      newValue = sum;
    }
    
    assert debug(Level.FINER, "\tIntersect\t (%s) u (%s) = (%s)", v_ik, sum, newValue);

    

    if (v_ik == null || !v_ik.equals(newValue))
    {

      //Label branch = mergeLabels(cl_ij.getLabel(), cl_jk.getLabel());
      SimpleConstraintValue value = new DefaultConstraintValue( 
                                                 newValue);
//      LabeledValueSet labeledValueSet = new DefaultLabeledValueSet(label);        
      
      E stnEdge = createEdge(from, to, value, false, false, 
                                 //e_ik != null && e_ik.isControlEdge(), 
                                 //e_ik == null || e_ik.isImplicit(),
                                 e_ik != null ? e_ik.getFlowDirection() : FlowDirection.UNKOWN);
      
      assert debug(Level.FINE, "Updating (%s) to (%s)", e_ik, stnEdge);
      
      logChange(changeSet, p, via, stnEdge, e_ik != null ? e_ik.getValue() : null);
      
      checkConsistency(model, stnEdge, new Triangle<LabeledTimePoint>(p, via));
      
      try
      {
        model.updateConstraint(stnEdge);
      }
      catch (IllegalArgumentException e)
      {
        assert debug(Level.SEVERE, "Trying to update contingent constraint %s -> %s: %s -> %s",
                                         from, to, v_ik , newValue);
        throw new TCNInconsistencyException("Trying to update contingent constraint!", p, via, model, e);
      }
      catch (IllegalStateException e)
      {
        debug(Level.SEVERE, "Inconsistent: %s -> %s: %s", stnEdge.getSourceNode(), stnEdge.getTargetNode(), stnEdge);
        throw new TCNInconsistencyException(String.format("Inconsistent: %s -> %s: %s", stnEdge.getSourceNode(),
          stnEdge.getTargetNode(), stnEdge), new Path<LabeledTimePoint>(stnEdge.getSourceNode(),
          stnEdge.getTargetNode()), via, model, e);
      }

      return true;

    }
    debug(Level.FINE, " Ignoring\t %s to %s (because no changes)", e_ik, newValue);
    return false;
  }

  public TimeDistance intersect(TimeDistance v1, TimeDistance v2)
  {
    if (v1.compareTo(v2) <= 0)
      return v1;
    else
    return v2;
  }
  

  @SuppressWarnings("unchecked")
  protected E createEdge(LabeledTimePoint from, LabeledTimePoint to, SimpleConstraintValue value,
      boolean controledge, boolean implicit, FlowDirection flowDirection)
  {
    // IMPROVE This should be done by some factory...    
    return (E) new STNEdge(from, to, value, controledge, implicit, flowDirection);
  }

  @Override
  protected boolean checkConsistency(SimpleConstraintValue value1, SimpleConstraintValue value2)
  {
    return isConsistent(value1, value2);
  }


}
