package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.aristaflow.adept2.model.timemodel.tcn.Label;

/**
 * A {@link Proposition} consists of a (possibly empty) set of
 * {@link Observation}s from different {@link DefaultObservationTimePoint}s. Each
 * {@link Observation} encodes a particular Branch of the respective
 * {@link DefaultObservationTimePoint} (see {@link Observation}).
 * 
 * see e.g., Conrad et al. (2011)
 * 
 * <pre>
 * Definition 3.7 (Environment) An environment is a partial assignment of the choice 
 * variables in X, written e = {x_i = d_{ij}, ...}. An environment may have at most one assignment
 * to each variable to be consistent. A complete environment contains an assignment to every
 * choice variable in X. An empty environment provides no assignments and is written {}.
 * We denote the set of possible environments as E and the set of complete environments as
 * E_c. The length of an environment is the number of assigned variables, denoted |e|.
 * </pre>
 * 
 * <p><b>Note:</b> This object is immutable!</p>
 * 
 * @author Andreas
 * 
 */
public class Proposition implements Label
{
  
  public static final String EMPTY_LABEL = "\u22A1"; // 'o';
  
  /**
   * The empty proposition
   */
  public static final Proposition EMPTY_PROPOSITION = new Proposition();

  private final SortedSet<ObservationBinding> observations;

  private int cachedHashCode;
  
  private Proposition()
  {
    this(Collections.<ObservationBinding> emptyList());
  }
  
  public Proposition(ObservationBinding... literals)
  {
    this(Arrays.asList(literals));
  }
  
  public Proposition(Collection<ObservationBinding> literals)
  {
    this.observations = Collections.unmodifiableSortedSet(new TreeSet<ObservationBinding>(literals));
    this.cachedHashCode = this.observations.hashCode();
    // assert the proposition is satisfiable with itself
    assert satisfiable(this);
  }

  @Override
  public String toString()
  {
    
    StringBuilder s = new StringBuilder();
    if (observations.size() > 0)
    {
      for (ObservationBinding l : observations)
      {
        s.append(l);
        s.append(" ");
      }
    }
    else
    {
      s.append(EMPTY_LABEL);
    }
    
    return s.toString();
  }

  public Set<ObservationBinding> getObservations()
  {
   return this.observations;
  }

  @Override
  public int hashCode()
  {
    return cachedHashCode;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    
    if (obj instanceof Proposition)
    {
      Proposition other = (Proposition) obj;
      return this.observations.equals(other.observations);
    }
    
    return super.equals(obj);
  }
  
  /**
   * In any case <code>this</code> is true <code>label2</code> must also be
   * true.
   * 
   * Put otherwise: Any observation binding in <code>other</code> must also be present in
   * <code>this</code> with same sign
   * 
   *  
   * cf. Conrad et al. (2011) (Note that the definition there has inverse operators compared to ours here)
   * 
   * <pre>
   * Definition 4.5 (Subsumption of Environments) An environment e is subsumed by e' if for
   * every assignment x_i = d_{ij} \in e, the same assignment exists in e', denoted x_i = d_{ij} \in e'.
   * </pre>
   * 
   * @param label2
   * @return <code>true</code> if this subsumes <code>label2</code>, <code>false</code> otherwise
   */
  @Override
  public boolean subsumes(Label label2)
  {
    Proposition prop2 = (Proposition)label2;
    
    // Empty literal is subsumed by anything
    if (prop2.isEmptyLabel())
      return true;
    
    for (ObservationBinding l2 : prop2.observations)
    {
      if (!this.observations.contains(l2))
      {
        return false;
      }
    }
    return true;
  }
  
  /**
   * Uses the property that ObservationBinding in proposition are sorted in
   * ascending order to achieve a complexity of roughly <code>O(2N)</code>
   */
  public boolean satisfiable(Label other)
  {
    if (!(other instanceof Proposition))
      throw new IllegalArgumentException("first argument must be a proposition");
    
    Proposition p2 = (Proposition) other;
    
    SortedSet<ObservationBinding> literals1 = this.observations;
    SortedSet<ObservationBinding> literals2 = p2.observations;
    
    Iterator<ObservationBinding> iterator1 = literals1.iterator();
    Iterator<ObservationBinding> iterator2 = literals2.iterator();
    while (iterator1.hasNext() && iterator2.hasNext())
    {
      ObservationBinding l1 = iterator1.next();
      ObservationBinding l2 = iterator2.next();
      
      int compareTo = l1.getVariable().compareTo(l2.getVariable());
      
      // Find the matching literal (if any)
      // That is: same split node, same branchID, same iteration
      while (compareTo != 0)
      {
        if (compareTo > 0 && iterator2.hasNext()){
          l2 = iterator2.next();
        }
        else if (compareTo < 0 && iterator1.hasNext())
        {
          l1 = iterator1.next();
        }
        else if (!iterator1.hasNext() || !iterator2.hasNext())
        {
          return true;
        }
        compareTo = l1.getVariable().compareTo(l2.getVariable());
      }
      
      assert l1.getVariable().equals(l2.getVariable());
    
      // If both propositions contain the same literal it must have the same "value" for the proposition to be satisfiable
      if (l1.getValue() != l2.getValue())
        return false;
    }
    
    return true;
  }

  public Label merge(Label other)
  {
    if (!(other instanceof Proposition))
      throw new IllegalArgumentException("first argument must be a proposition");
    
    if (this.isEmptyLabel())
      return other;
    
    if (other.isEmptyLabel())
      return this;
    
    Proposition prop2 = (Proposition)other;
    
    List<ObservationBinding> lit = new ArrayList<ObservationBinding>(this.observations.size()+prop2.observations.size());
    lit.addAll(this.observations);
    
    ListIterator<ObservationBinding> iterator = lit.listIterator();
    
    for (ObservationBinding l2 : prop2.observations)
    {
//      boolean found = false;
      if (iterator.hasNext())
      {
        ObservationBinding lx = iterator.next();
        int compareTo = lx.getVariable().compareTo(l2.getVariable());
        
        while (compareTo < 0 && iterator.hasNext())
        {
          lx = iterator.next();
          compareTo = lx.getVariable().compareTo(l2.getVariable());
        }
        
        if (compareTo == 0)
        {
          if (lx.getValue() != l2.getValue())
            throw new IllegalArgumentException(String.format("Cannot merge %s and %s because they are not satisfiable", this, other));
          else
            continue;
        }
      }
      
      iterator.add(l2);
    }
    
    return new Proposition(lit);
  }

  public Label[] negate()
  {
    Set<Label> res = new HashSet<Label>();
    
    ObservationBinding[] literals = this.observations.toArray(new ObservationBinding[this.observations.size()]);
    if (literals.length == 0)
      throw new IllegalStateException("Empty Label cannot be negated");
    
    
    int[] gen = new int[literals.length];
    int[] genBase = new int[literals.length];
    for (int i = 0; i < gen.length; i++)
    { 
      ObservationBinding lg = literals[i];
      gen[i] = lg.getVariable().getOptions();
      genBase[i] = gen[i];
    }
    
    
    
    int pos = gen.length -1;
    
    while (pos >= 0)
    {
      //System.out.println(Arrays.toString(gen));
      boolean diff = false;
      ObservationBinding[] lt = new ObservationBinding[gen.length];
      
      for (int i = 0; i < lt.length; i++)
      {
        ObservationBinding baseLit = literals[i];
        int branchID = baseLit.getValue();
        int one = branchID+1;
        
        if (one != gen[i])
          diff = true;
        lt[i] = ObservationBinding.createObservationBinding(baseLit.getVariable(), Long.MIN_VALUE, gen[i]-1);
      }
      
      if (diff)
        res.add(new Proposition(lt));
      
      gen[pos]--;
      
      if (gen[pos] <= 0)
      {
        while (--pos >= 0 && --gen[pos] <= 0){
          // Not content required
        } 
        
        if (pos < 0)
        {
          break;
        }
        System.arraycopy(genBase, pos+1, gen, pos+1, gen.length-1-pos);
        pos = gen.length -1;
      }
      
      
    }
    
    return res.toArray(new Label[res.size()]);
  }


  public Label removeVariable(ChoiceVariable other)
  {
    if (!(other instanceof Observation))
      throw new IllegalArgumentException("first argument must be an Observation "+other.getClass());
    
    SortedSet<ObservationBinding> reducedLiterals = new TreeSet<ObservationBinding>();
    
    for (ObservationBinding l : this.observations)
    {
      if (other != l.getVariable())
      {
        reducedLiterals.add(l);
      }
    }
    
    if (reducedLiterals.size() != this.observations.size())
      return new Proposition(reducedLiterals);
    
    // no changes
    return this;
  }
  
  public Label removeLabel(Label other)
  {
    if (!(other instanceof Proposition))
      throw new IllegalArgumentException("first argument must be an Observation");
    
    assert this.satisfiable(other);
    
    Proposition proposition = (Proposition) other;
    SortedSet<ObservationBinding> reduceLiterals = new TreeSet<ObservationBinding>(this.observations);
    
    if (reduceLiterals.removeAll(proposition.observations))
      return new Proposition(reduceLiterals);
    
    // no changes
    return this;
  }

  public boolean isEmptyLabel()
  {
    return this.observations.size() == 0;
  }

  public boolean containsVariable(ChoiceVariable other)
  {
    for (ObservationBinding l : this.observations)
    {
      if (other == l.getVariable())
      {
        return true;
      }
    }

    
    return false;
  }

  public Label addVariableBinding(ChoiceVariableBinding choiceVariable)
  {   
    ObservationBinding observation = (ObservationBinding) choiceVariable;
    
    for (ObservationBinding lx : this.observations)
    {
      if (lx.getVariable().equals(observation.getVariable()))
      {
        if (lx.getValue() != observation.getValue())
          throw new IllegalArgumentException(String.format("Cannot append %s and %s because result is not satisfiable", this, observation));
        else
          return this;
      }
    }
    
    // Not found. Otherwise we would not be here
    SortedSet<ObservationBinding> lit = new TreeSet<ObservationBinding>();
    lit.addAll(this.observations);
    lit.add(observation);
    return new Proposition(lit);

  }

  @Override
  public ChoiceVariableBinding getBinding(ChoiceVariable variable)
  {
    for (ChoiceVariableBinding binding : this.observations)
    {
      if (binding.getVariable().equals(variable))
        return binding;
    }

    return null;
  }
}
