package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.GuardedInterval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultGuardedInterval;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.CompoundConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.STNU;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.LabelBasedPC2Solver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUDCSolver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;

/**
 * Class containing different algorithms for handling TCNs
 * 
 * @author lanz
 * 
 */
public class TCNAlgorithms
{
  
  public static TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> getSolver(
    final TCN<LabeledTimePoint, LabeledEdge> cstnu)
  {
//    final TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> solver;
//    if (cstnu instanceof CSTNPSUC)
//    {
//      solver = new CSTNPSUCDCSolver();
//    }
//    else if (cstnu instanceof CSTNPSU)
//    {
//      solver = new CSTNPSUDCSolver();
//    }
//    else if (cstnu instanceof CSTNU)
//    {
//      solver = new CSTNUDCSolver();
//    }
//    else if (cstnu instanceof CSTN)
//    {
//      solver = new LabelBasedPC2Solver();
//    }
//    else
//    {
//      throw new IllegalStateException();
//    }
//    return solver;
    return getSolver(cstnu.getClass());
  }
  
  public static TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> getSolver(
      @SuppressWarnings("rawtypes") Class<? extends TCN> clazz)
  {
    final TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> solver;
    if (CSTNU.class.isAssignableFrom(clazz))
    {
      solver = new CSTNUDCSolver();
    }
    else if (CSTN.class.isAssignableFrom(clazz))
    {
      solver = new LabelBasedPC2Solver();
    }
    else
    {
      throw new IllegalStateException();
    }
    return solver;
  }

  /**
   * Finds any rigid component within the TCN. A rigid component is a set of
   * nodes within the tcn where the distance between the nodes is fixed (i.e.,
   * lower distance == upper distance).
   * 
   * <p>
   * Algorithm has a complexity of approximately <code>O(n^2)</code>, where
   * <code>n</code> is the number of nodes in the tcn.
   * </p>
   * 
   * <p>
   * Note this implementation currently only support {@link SimpleConstraintValue} or {@link LabeledValueSet} as edge
   * constraints.</p>
   * 
   * @param tcn The tcn to search for rigid components
   * @return the rigid components in the tcn as collection of sets, i.e., each set is one rigid component
   * @throws IllegalArgumentException if <code>V</code> does not either implement {@link SimpleConstraintValue} 
   *        or {@link LabeledValueSet}
   */
  public static <N extends TCNTimePoint, V extends ConstraintValue, E extends TCNEdge<N, V>> Collection<Set<N>> findRigidComponents(
      TCN<N, E> tcn)
  {
    Map<N, Set<N>> localRigidComponents = new HashMap<N, Set<N>>();

    List<N> nodes = new ArrayList<N>(tcn.getTimePoints());
    // It's sufficient to check each pair once. So use an iterator removing each element
    // Process the list backwards because removing is cheaper this way
    for (ListIterator<N> iter = nodes.listIterator(nodes.size()); iter.hasPrevious(); ) 
    {
      N n1 = iter.previous();
      iter.remove();
      for (N n2 : nodes)
      {
        //if (n1 != n2)
        E edge1 = tcn.getEdge(n1, n2);
        if (edge1 == null)
          continue;

        E edge2 = tcn.getEdge(n2, n1);
        if (edge2 == null)
          continue;

        if (isRigidComponent(edge1, edge2))
        {
          // Rigid component
          Set<N> set1 = localRigidComponents.get(n1);
          Set<N> set2 = localRigidComponents.get(n2);
          Set<N> set;
          if (set1 == null && set2 == null)
          {
            // None of the nodes is part of a rigid component yet
            set = new HashSet<N>();
            localRigidComponents.put(n1, set);
            localRigidComponents.put(n2, set);
          }
          else if (set1 != null && set2 == null)
          {
            // node1 is already part of a rigid component, add node2 to the component
            set = set1;
            localRigidComponents.put(n2, set);
          }
          else if (set1 == null && set2 != null)
          {
            // node2 is already part of a rigid component, add node1 to the component
            set = set2;
            localRigidComponents.put(n1, set);
          }
          else // set1 != null && set2 != null
          {
            // both are already part of a rigid component so merge the two components
            set = new HashSet<N>(set1);
            set.addAll(set2);
            
            for (N node : set)
            {
              localRigidComponents.put(node, set);
            }
          }
            
          set.add(n1);
          set.add(n2);
        }
        
      }
    }

    return new HashSet<Set<N>>(localRigidComponents.values());
  }

  /**
   * This is just a switch for different kinds implementations of
   * {@link TCNEdge#getValue()}. Currently {@link SimpleConstraintValue} and
   * {@link LabeledValueSet} are supported.
   * 
   * @param edge1
   * @param edge2
   * @throws IllegalArgumentException if <code>V</code>  does not either implement {@link SimpleConstraintValue} 
   *        or {@link LabeledValueSet}
   */
  protected static <V extends ConstraintValue, E extends TCNEdge<?, V>> boolean isRigidComponent(E edge1, E edge2)
  {
    if (edge1.getValue() instanceof SimpleConstraintValue)
    {
      assert edge2.getValue() instanceof SimpleConstraintValue;
      return isRigidComponent((SimpleConstraintValue)edge1.getValue(), (SimpleConstraintValue)edge2.getValue());
    }
    else if (edge1 instanceof LabeledEdge)
    {
      return isRigidComponent((LabeledEdge)edge1, (LabeledEdge)edge2);
    }
    throw new IllegalArgumentException();
  }
  
  protected static boolean isRigidComponent(SimpleConstraintValue edge1, SimpleConstraintValue edge2)
  {
    TimeDistance sum = edge1.getValue().add(edge2.getValue());
    return sum.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) == 0;
  }
  
  /**
   * A pair of {@link LabeledValueSet} is considered to form a rigid component
   * if it is a rigid component for at least one of the two nodes. Particularly it needs to be a rigid component
   * in each projection the node is in.
   * 
   * @param edge1
   * @param edge2
   * @return
   */
  protected static boolean isRigidComponent(LabeledEdge edge1, LabeledEdge edge2)
  {
    LabeledTimePoint node1 = edge1.getSourceNode();
    LabeledTimePoint node2 = edge2.getSourceNode();
    boolean[] rigidNode = new boolean[]{true, true};
    boolean rigid = false;
    for (LabeledValue value1 : edge1.getValue().getLabeledValues())
    {
      for (LabeledValue value2 : edge2.getValue().getLabeledValues())
      {
        if (value1.getLabel().satisfiable(value2.getLabel()))
        {
          if (!isRigidComponent(value1, value2)) // Do this check first because it's "cheap" and will filter probably 90% of the cases
            return false;
          else
          {
            rigid = true; // Could be rigid, lets see if the nodes fit...
            
            // LOW: This is too strict. It would be sufficient if the union of all the labels of the edges subsume the nodes label. But this is darn complicated
            // The label of an edge needs to be either empty (i.e., apply in any
            // case) or the edge needs to apply in any case the node applies
            // (i.e., the label of the node needs to subsume the label of the
            // edge)
            if (!(value1.getLabel().isEmptyLabel() || node1.getLabel().subsumes(value1.getLabel()))
                || !(value2.getLabel().isEmptyLabel() || node1.getLabel().subsumes(value2.getLabel())))
              rigidNode[0] = false;

            if (!(value1.getLabel().isEmptyLabel() || node2.getLabel().subsumes(value1.getLabel()))
                || !(value2.getLabel().isEmptyLabel() || node2.getLabel().subsumes(value2.getLabel())))
              rigidNode[1] = false;

            if (!(rigidNode[0] || rigidNode[1])) // If it's not rigid towards at least one of the nodes it's not rigid at all
              return false;
            
          }
            
        }
      }
    }
    
    return rigid && (rigidNode[0] || rigidNode[1]); // Is it rigid at all? And if yes, is it rigid towards at least one of the nodes?
  }

  /**
   * Finds the single time point in the TCN with no incoming edge with value > 0
   * @param tcn
   * @return the single time point in <code>tcn</code> with no incoming edge with value > 0
   * @throws IllegalArgumentException in case there are multiple time points with no incoming edge with value > 0
   */
  public static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> N 
    getSourceTimePoint(TCN<N, E> tcn) throws IllegalArgumentException
  {
    N source = null;
    for (N node : tcn.getTimePoints())
    {
      boolean hasPredecessor = false;
      
      for (N node2 : tcn.getTimePoints())
      {
        E outgoing = tcn.getEdge(node, node2);
        E incoming = tcn.getEdge(node2, node);
        
        if (incoming != null)
        {
          TimeDistance inDist = getTimeDistance(incoming, false);
          if (inDist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) > 0)
          {
            hasPredecessor = true;
            break;
          }
        }
        
        if (outgoing != null)
        {
          TimeDistance outDist = getTimeDistance(outgoing, true);
          if (outDist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) <= 0)
          {
            hasPredecessor = true;
            break;
          }
        }
        
      }
      
//      for (E edge : tcn.getIncomingEdges(node))
//      {
//        TimeDistance timeDistance = getTimeDistance(edge, false);
//        if (timeDistance.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) > 0)
//        {
//          hasPredecessor = true;
//        }
//      }
//      
//      for (E edge : tcn.getOutgoingEdges(node))
//      {
//        TimeDistance timeDistance = getTimeDistance(edge, true);
//        if (timeDistance.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) <= 0)
//        {
//          hasPredecessor = true;
//        }
//      }
      
      if (!hasPredecessor){
        if (source == null)
        {
          source = node;
        }
        else
        {
          throw new IllegalArgumentException(String.format("More than one source node: '%s' and '%s'", source, node));
        }
      }
    }
    
    return source;
  }
  
  /**
   * Finds the single time point in the TCN with no outgoing edge with value >= 0
   * @param tcn
   * @return the single time point in <code>tcn</code> with no outgoing edge with value >= 0
   * @throws IllegalArgumentException in case there are multiple time points with no outgoing edge with value >= 0
   */
  public static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> N 
    getTargetTimePoint(TCN<N, E> tcn) throws IllegalArgumentException
  {
    N source = null;
    for (N node : tcn.getTimePoints())
    {
      boolean hasSuccessor = false;
//      for (E edge : tcn.getOutgoingEdges(node))
//      {
//        TimeDistance timeDistance = getTimeDistance(edge, true);
//        if (timeDistance.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) > 0)
//        {
//          hasSuccessor = true;
//        }
//      }
      
      for (N node2 : tcn.getTimePoints())
      {
        E outgoing = tcn.getEdge(node, node2);
        E incoming = tcn.getEdge(node2, node);
        
        if (incoming != null)
        {
          TimeDistance inDist = getTimeDistance(incoming, false);
          if (inDist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) <= 0)
          {
            hasSuccessor = true;
            break;
          }
        }
        
        if (outgoing != null)
        {
          TimeDistance outDist = getTimeDistance(outgoing, true);
          if (outDist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) > 0)
          {
            hasSuccessor = true;
            break;
          }
        }
      }
      
      if (!hasSuccessor){
        if (source == null)
        {
          source = node;
        }
        else
        {
          throw new IllegalArgumentException(String.format("More than one target node: '%s' and '%s'", source, node));
        }
      }
    }
    
    return source;
  }

  public static <N extends TCNTimePoint>
    GuardedInterval<TimeDistance> getRestrictableDuration(TCNTemplateTimeModel<N, ?> timeModel)
  {
    @SuppressWarnings("unchecked")
    TCN<N, TCNEdge<N, CompoundConstraintValue>> tcn = (TCN<N, TCNEdge<N, CompoundConstraintValue>>) timeModel.getBaseTimeModel();
    if (!tcn.isConsistent())
      throw new IllegalArgumentException("tcn must be consistent!");

    N sourceNode = timeModel.getProcessStartTimePoint();
    N targetNode = timeModel.getProcessEndTimePoint();

    if (sourceNode == null || targetNode == null)
      throw new IllegalArgumentException("Process creation time point and process end time point of tcn must be set!");

    return getRestrictableDuration(tcn, sourceNode, targetNode);
  }
  
  public static <N extends TCNTimePoint>
  GuardedInterval<TimeDistance> getRestrictableDuration(TCNInstanceTimeModel<N, ?> timeModel)
{
  @SuppressWarnings("unchecked")
  TCN<N, TCNEdge<N, CompoundConstraintValue>> tcn = (TCN<N, TCNEdge<N, CompoundConstraintValue>>) timeModel.getBaseTimeModel();
  if (!tcn.isConsistent())
    throw new IllegalArgumentException("tcn must be consistent!");

  N sourceNode = timeModel.getProcessStartTimePoint();
  N targetNode = timeModel.getProcessEndTimePoint();

  if (sourceNode == null || targetNode == null)
    throw new IllegalArgumentException("Process creation time point and process end time point of tcn must be set!");

  return getRestrictableDuration(tcn, sourceNode, targetNode);
}

  /**
   * 
   * @param tcn
   * @param sourceNode source time-point of the tcn
   * @param targetNode target time-point of the tcn
   * @return
   */
  @SuppressWarnings("unchecked")
  public static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
    GuardedInterval<TimeDistance> getRestrictableDuration(
      TCN<N, E> tcn, N sourceNode, N targetNode)
  {
    if (sourceNode == null || targetNode == null)
      throw new IllegalArgumentException("source and target node of the tcn must be given!");
    
    if (!tcn.isConsistent())
      throw new IllegalArgumentException("TCN not minimal");
    
    E flexMinDurEdge = tcn.getEdge(targetNode, sourceNode);
    
    // A minimum edge always exits
    if (flexMinDurEdge == null)
      throw new IllegalArgumentException("TCN not minimal");
    
    TimeDistance flexMinDur = getTimeDistance(flexMinDurEdge, true).neg();
    
    
    E flexMaxDurEdge = tcn.getEdge(sourceNode, targetNode);
    
    TimeDistance flexMaxDur;
    if (flexMaxDurEdge == null) // If the network is unbounded (i.e., has no maximum duration this edge may be null
    {
      flexMaxDur = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    }
    else
    {
      flexMaxDur = getTimeDistance(flexMaxDurEdge, true);
    }
    
    TimeDistance guardedMinDur;
    TimeDistance guardedMaxDur;
    TimeDistance contingency;
    
    if (tcn instanceof STNU)
    {
      guardedMinDur = getMinGuardDistance((STNU<N, V, E>)tcn, sourceNode, targetNode); 
      
      guardedMaxDur = getMaxGuardDistance(tcn, sourceNode, targetNode);
      
      if (guardedMinDur == null)
        guardedMinDur = flexMaxDur;
      
      if (guardedMaxDur == null)
        guardedMaxDur = flexMinDur;
      
      
      contingency = getContingency(tcn, sourceNode, targetNode);
      
      if (contingency == null)
        contingency = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    
    }
    else
    {
        guardedMinDur = flexMaxDur;
      
        guardedMaxDur = flexMinDur;
      
        contingency = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    }
      
    assert flexMinDur.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0;
    
    assert guardedMinDur.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0;
    assert guardedMinDur.compareTo(flexMinDur) >= 0;
    
    assert guardedMaxDur.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0;
    assert guardedMaxDur.compareTo(flexMinDur) >= 0;
    
    assert flexMaxDur.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0;
    assert flexMaxDur.compareTo(guardedMaxDur) >= 0;
    
    
    return new DefaultGuardedInterval<TimeDistance>(flexMinDur, guardedMinDur, guardedMaxDur, flexMaxDur, contingency);
  }
  
  public static < V extends CompoundConstraintValue, E extends TCNEdge<?, V>> TimeDistance 
    getTimeDistance(E edge, boolean maxValue)
  {
    int sgn = maxValue ? 1 : -1;
    
    TimeDistance timeDistance = null;

    SimpleConstraintValue[] values = edge.getValue().getValues();
    
    for (SimpleConstraintValue constraintValue : values)
    {
      if (timeDistance == null)
      {
        timeDistance = constraintValue.getValue();
      }
      else
      {
        if (constraintValue.getValue().compareTo(timeDistance) >= sgn)
        {
          timeDistance = constraintValue.getValue();
        }
      }
      
    }
    
    return timeDistance;
  }
  
  public static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
  TimeDistance getMinGuardDistance(STNU<N, V, E> tcn, N baseNode, N targetNode)
  {
    TimeDistance result;
//    if (baseNode == tcn.getTimeBaseEvent())
//    {
//    
    
      Map<TCNTimePoint, TimeDistance> cycleDetection = new HashMap<TCNTimePoint, TimeDistance>();
      result = internalGetMinGuardIncoming(tcn, baseNode, targetNode, cycleDetection);
      System.out.println("minGuard("+baseNode+", "+targetNode+") = "+result.toSeconds());
//    BellmanFord<N> bellmanFord = new BellmanFord<N>();
//    SemiReduciblePathsGraph<N> graph = new SemiReduciblePathsGraph<N>((TCN<N, ? extends LabeledEdge>) tcn);
//    Map<N, Operand> run = bellmanFord.run(graph, baseNode);
//    result = run.get(targetNode);
      
//      cycleDetection.clear();
//      TimeDistance outgoing = internalGetMinGuardOutgoing(tcn, baseNode, targetNode, cycleDetection);
//      
//      result = incoming != null ? outgoing != null ? (incoming.compareTo(outgoing) <= 0 ? incoming : outgoing) : incoming : outgoing;
//    }
//    else
//    {
//      Map<TCNTimePoint, TimeDistance> cycleDetection = new HashMap<TCNTimePoint, TimeDistance>();
//      TimeDistance minG = internalGetMinGuardDistance(tcn, tcn.getTimeBaseEvent(), targetNode, cycleDetection);
//      
//      cycleDetection.clear();
//      TimeDistance maxG = internalGetMaxGuardDistance(tcn, tcn.getTimeBaseEvent(), baseNode, cycleDetection);
//      
//      result = minG.sub(maxG);
//    }
   

    return result;
  }
  
//  protected static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
//  TimeDistance internalGetMinGuardOutgoing(TCN<N, E> tcn, N baseNode, N targetNode, Map<TCNTimePoint, TimeDistance> cycleDetection)
//  {
//    if (cycleDetection.containsKey(targetNode))
//    {
//      return cycleDetection.get(targetNode);
//    }
//    
//    if (baseNode == targetNode) // This is the Base case...
//      return DefaultTimeDistance.ZERO_TIME_DISTANCE;
//      
//    
//    // Add the current target node to the list of visited nodes to avoid loops
//    cycleDetection.put(targetNode, null);
//    
//    TimeDistance distance = null;
//        
//  
//    Collection<E> outgoingEdges = tcn.getOutgoingEdges(targetNode);
//    for (E e : outgoingEdges)
//    {
//      TimeDistance dist = getTimeDistance(e, false);
//      if (dist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) <= 0)
//      {
//        TimeDistance contDistance = getContingentMaxValue(e);
//        if (contDistance != null)
//        {
//          dist = contDistance;
//        }
//        
//        TimeDistance timePoint = internalGetMinGuardOutgoing(tcn, baseNode, e.getTargetNode(), cycleDetection);
//        
//        if (timePoint == null || timePoint.isInfinity())
//          continue;
//        
//        TimeDistance td = dist.add(timePoint);
//        
//        distance = distance != null && distance.compareTo(td) <= 0 ? distance : td;
//      }
//    }
//    
//    cycleDetection.put(targetNode, distance);
//
//    return distance;
//  }
  
  protected static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
  TimeDistance internalGetMinGuardIncoming(TCN<N, E> tcn, N baseNode, N targetNode, Map<TCNTimePoint, TimeDistance> cycleDetection)
  {
    if (cycleDetection.containsKey(targetNode))
    {
      return cycleDetection.get(targetNode);
    }
    
    if (baseNode == targetNode) // This is the Base case...
      return DefaultTimeDistance.ZERO_TIME_DISTANCE;
     
    TimeDistance dist;
    TimeDistance contDistance;
    // Add the current target node to the list of visited nodes to avoid loops (using the direct distance between base and target)
//    E edge = tcn.getEdge(baseNode, targetNode);
//    
//    TimeDistance dist = getTimeDistance(edge, false);
//    
//    TimeDistance contDistance = getContingentValue(edge);
//  // If there is a contingent minimum value it will always be smaller than
//  // the non-contingent one, so we can ignore the non-contingent one
//    if (contDistance != null) 
//    {
//                              
//      dist = contDistance;
//    }
    
    cycleDetection.put(targetNode, null);
    
//    System.out.println("minGuard("+baseNode+", "+targetNode+") init to "+dist.toSeconds());
    
    TimeDistance distance = null;
        
    Collection<E> incomingEdges = tcn.getIncomingEdges(targetNode);
    
    for (E e : incomingEdges)
    {
      dist = getTimeDistance(e, false);
      contDistance = getContingentValue(e);
      //if (dist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0 || contDistance != null)
      {
        
      // If there is a contingent minimum value it will always be smaller than
      // the non-contingent one, so we can ignore the non-contingent one
        if (contDistance != null) 
        {
                                  
          dist = contDistance;
        }

        TimeDistance timePoint = internalGetMinGuardIncoming(tcn, baseNode, e.getSourceNode(), cycleDetection);
        
        if (timePoint == null || timePoint.isInfinity())
          continue;
        
        TimeDistance td = dist.add(timePoint);
        
        if (distance == null || distance.compareTo(td) > 0)
        {
          distance = td;
          //System.out.println("minGuard("+baseNode+", "+targetNode+") = minGuard("+baseNode+", "+e.getSourceNode()+") + "+e+" = "+timePoint.toSeconds()+" + "+dist.toSeconds()+" = "+distance);
        }
      }
    }
    
    cycleDetection.put(targetNode, distance);
    
    //System.out.println("minGuard("+baseNode+", "+targetNode+") = "+distance);

    return distance;
  }
  
  public static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
  TimeDistance getMaxGuardDistance(TCN<N, E> tcn, N baseNode, N targetNode)
  {
  
    TimeDistance result;
//    if (baseNode == tcn.getTimeBaseEvent())
//    {
    
      Map<TCNTimePoint, TimeDistance> cycleDetection = new HashMap<TCNTimePoint, TimeDistance>();
      result = internalGetMaxGuardDistance(tcn, baseNode, targetNode, cycleDetection);
//    }
//    else
//    {
//      Map<TCNTimePoint, TimeDistance> cycleDetection = new HashMap<TCNTimePoint, TimeDistance>();
//      TimeDistance maxG = internalGetMaxGuardDistance(tcn, tcn.getTimeBaseEvent(), targetNode, cycleDetection);
//      
//      cycleDetection.clear();
//      TimeDistance minG = internalGetMinGuardIncoming(tcn, tcn.getTimeBaseEvent(), baseNode, cycleDetection);
//      
//      result = maxG.sub(minG);
//    }
    
    return result;
  }
  
  protected static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
  TimeDistance internalGetMaxGuardDistance(TCN<N, E> tcn, N baseNode, N targetNode, Map<TCNTimePoint, TimeDistance> cycleDetection)
  {
    if (cycleDetection.containsKey(targetNode))
    {
      return cycleDetection.get(targetNode);
    }
    
    if (baseNode == targetNode) // This is the Base case...
      return DefaultTimeDistance.ZERO_TIME_DISTANCE;
    
    // Add the current target node to the list of visited nodes to avoid loops
    cycleDetection.put(targetNode, null);
    
    TimeDistance distance = null;
 
    
    Collection<E> outgoingEdges = tcn.getOutgoingEdges(targetNode);
    
    for (E e : outgoingEdges)
    {
      TimeDistance dist = getTimeDistance(e, false).neg();
      
//      if (dist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0)
//      {
        TimeDistance contDistance = getContingentValue(e);
        if (contDistance != null)
        {
          dist = contDistance.neg();
        }
        
        TimeDistance timePoint = internalGetMaxGuardDistance(tcn, baseNode, e.getTargetNode(), cycleDetection);
        
        if (timePoint == null || timePoint.isInfinity())
          continue;
        
        TimeDistance td = dist.add(timePoint);
        
        distance = distance != null && distance.compareTo(td) >= 0 ? distance : td;
       
//      }
    }
    
    cycleDetection.put(targetNode, distance);
    
    return distance;
  }
  
  public static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
  TimeDistance getContingency(TCN<N, E> tcn, N baseNode, N targetNode)
  {
    Map<TCNTimePoint, TimeDistance> cycleDetection = new HashMap<TCNTimePoint, TimeDistance>();
    TimeDistance result = internalGetContingency(tcn, baseNode, targetNode, cycleDetection);

    if (result == null)
      return null;
    
    return result.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) > 0 ? result : DefaultTimeDistance.ZERO_TIME_DISTANCE;
  }
  
  protected static <N extends TCNTimePoint, V extends CompoundConstraintValue, E extends TCNEdge<N, V>> 
  TimeDistance internalGetContingency(TCN<N, E> tcn, N baseNode, N targetNode, Map<TCNTimePoint, TimeDistance> cycleDetection)
  {
    if (cycleDetection.containsKey(targetNode))
    {
      return cycleDetection.get(targetNode);
    }
    
    if (baseNode == targetNode) // This is the Base case...
      return DefaultTimeDistance.ZERO_TIME_DISTANCE;
      
    
    // Add the current target node to the list of visited nodes to avoid loops
    cycleDetection.put(targetNode, null);
    
    TimeDistance contingency = null;
        
    Collection<E> incomingEdges = tcn.getIncomingEdges(targetNode);
    
    for (E e : incomingEdges)
    {
      TimeDistance distUpper = getTimeDistance(e, false);
      
      if (distUpper.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0)
      {
        E ec = tcn.getEdge(e.getTargetNode(), e.getSourceNode());
        TimeDistance distLower = null;
        if (ec != null)
        {
          distLower = getTimeDistance(ec, true);
        }
        
        TimeDistance cont;
        if (distLower == null || distLower.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0)
        {
          cont = distUpper;
        }
        else
        {
          cont = distUpper.add(distLower);
        }
        
        TimeDistance recursion = internalGetContingency(tcn, baseNode, e.getSourceNode(), cycleDetection);
        
        if (recursion == null)
          continue;
        
        TimeDistance td = cont.neg().add(recursion);
        contingency = contingency != null && contingency.compareTo(td) >= 0 ? contingency : td;
        
        
        TimeDistance contMinDistance = getContingentMinValue(e);
        TimeDistance contMaxDistance = null;
        if (ec != null)
        {
          contMaxDistance = getContingentMaxValue(ec);
        }
        if (contMinDistance != null) 
        {
          if (contMaxDistance == null) // May have been removed due to label removal rule...
          {
            contMaxDistance = distLower;
          }
          
          cont = contMinDistance.sub(contMaxDistance).neg();
          
          if (recursion.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) < 0)
          {
            td = cont.add(recursion);
          }
          else
          {
            td = cont;
          }

          
          contingency = contingency != null && contingency.compareTo(td) >= 0 ? contingency : td;
          
        }
      }
    }
    
    cycleDetection.put(targetNode, contingency);

    return contingency;
  }
  
  public static <V extends CompoundConstraintValue, E extends TCNEdge<?, V>> TimeDistance getContingentMaxValue(E edge)
  {
    TimeDistance contMaxValue = null;
    for (SimpleConstraintValue v : edge.getValue().getValues())
    {
      if (v instanceof ContingentLabeledValue)
      {
        ContingentLabeledValue contingentLabeledValue = (ContingentLabeledValue) v;
        if (ConstraintCase.UPPERCASE.equals(contingentLabeledValue.getCase()))
        {
          TimeDistance contingentValue = contingentLabeledValue.getContingentValue().neg();
          if (contMaxValue == null || contMaxValue.compareTo(contingentValue) < 0)
          {
            contMaxValue = contingentValue;
          }
        }
      }
    }

    return contMaxValue;
  }
  
  public static <V extends CompoundConstraintValue, E extends TCNEdge<?, V>> TimeDistance getContingentMinValue(E edge)
  {
    TimeDistance contMinValue = null;
    for (SimpleConstraintValue v : edge.getValue().getValues())
    {
      if (v instanceof ContingentLabeledValue)
      {
        ContingentLabeledValue contingentLabeledValue = (ContingentLabeledValue) v;
        if (ConstraintCase.LOWERCASE.equals(contingentLabeledValue.getCase()))
        {
          TimeDistance contingentValue = contingentLabeledValue.getContingentValue();
          if (contMinValue == null || contMinValue.compareTo(contingentValue) > 0)
          {
            contMinValue = contingentValue;
          }
        }
      }
    }

    return contMinValue;
  }
  
  public static <V extends CompoundConstraintValue, E extends TCNEdge<?, V>> TimeDistance getContingentValue(E edge)
  {
    TimeDistance contValue = null;
    for (SimpleConstraintValue v : edge.getValue().getValues())
    {
      if (v instanceof ContingentLabeledValue)
      {
        ContingentLabeledValue contingentLabeledValue = (ContingentLabeledValue) v;
        TimeDistance contingentValue = contingentLabeledValue.getContingentValue();
        if (contValue == null || contValue.compareTo(contingentValue) > 0)
        {
          contValue = contingentValue;
        }
        
      }
    }

    return contValue;
  }

//  public static STN<LabeledTimePoint, LabeledEdge> 
//    toSTN(STNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcn)
//  {
//    ChangeableSTN<LabeledTimePoint, LabeledEdge> stn;
//    
//    if (tcn instanceof CSTN)
//    {
//      stn = new DefaultCSTN(tcn.getTimePoints(), new ArrayList<LabeledEdge>(), ((CSTN) tcn).getObservations());
//    }
//    else
//    {   
//      stn = new DefaultCSTN(tcn.getTimePoints(), new ArrayList<LabeledEdge>(), new ArrayList<ChoiceVariable>());
//    }
//    
//    for (LabeledEdge edge : tcn.getEdges())
//    {
//      if (edge.isContingentDuration())
//      {
//        List<LabeledValue> values = new ArrayList<LabeledValue>();
//        for (LabeledValue v : edge.getValue().getLabeledValues())
//        {
//          values.add(new DefaultLabeledValue(v.getLabel(), v.getValue()));
//        }
//        LabeledValueSet labelSet = new DefaultLabeledValueSet(values);
//        DefaultLabeledEdge e = new DefaultLabeledEdge(edge.getSourceNode(), edge.getTargetNode(), labelSet);
//        stn.updateConstraint(e);
//      }
//      else
//      {
//        List<LabeledValue> values = new ArrayList<LabeledValue>();
//        for (LabeledValue v : edge.getValue().getLabeledValues())
//        {
//          values.add(new DefaultLabeledValue(v.getLabel(), v.getValue()));
//        }
//        LabeledValueSet labelSet = new DefaultLabeledValueSet(values);
//        DefaultLabeledEdge e = new DefaultLabeledEdge(edge.getSourceNode(), edge.getTargetNode(), labelSet);
//        stn.updateConstraint(e);
//      }
//        
//    }
//    
//    return stn;
//  }
  
  
}
