package de.aristaflow.adept2.model.timemodel.tcn.stnuImpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Queue;
import java.util.logging.Level;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Path;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.STNU;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractSTNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.PC2Solver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.STNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tools.DuplicateFreeQueue;
import de.aristaflow.adept2.util.Tuple;

public class STNUDCSolver extends AbstractSTNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> implements
    TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge>
{

  private int updateCount;
  private STNUAllMaxProjection allMax;
  private Collection<Path<LabeledTimePoint>> changedEdges;
  
  
  public STNUDCSolver()
  {
    super();
  }
  
  public STNUDCSolver(Level debugLevel)
  {
    super(debugLevel);
  }
  
  @Override
  public <T extends ChangeableTCN<LabeledTimePoint, LabeledEdge>> T solve(T tcn) throws TCNInconsistencyException, InterruptedException
  {
    return solve(tcn, null);
  }
  
  @SuppressWarnings("unchecked")
  public <T extends ChangeableTCN<LabeledTimePoint, LabeledEdge>> T solve(
      T tcn,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet)
      throws TCNInconsistencyException, InterruptedException
  {
    if (tcn instanceof ChangeableSTNU)
    {
      return (T)solve((ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge>)tcn, changeSet);
    }
    else
    {
      throw new IllegalArgumentException();
    }
  }
  
  /**
   * Morris et al.
   * 
   * <p><code>
   * Boolean procedure determineDC()<br/>
   *  loop from 1 to Cutoff Bound do<br/>
   *    if AllMax projection inconsistent<br/>
   *      return false;<br/>
   *    Perform needed No-Case Reductions;<br/>
   *    Perform needed Upper-Case Reductions;<br/>
   *    Perform any Cross-Case Reductions;<br/>
   *    Perform any Lower-Case Reductions;<br/>
   *    Perform any Label Removal Reductions;<br/>
   *    if no reductions were found above<br/>
   *      return true;
   * </code></p>
   * 
   * @param cstnu
   * @throws TCNInconsistencyException 
   * @throws InterruptedException 
   */
  public ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> solve(ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException
  {   
    ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu = model;//model.clone();
    
    if (cstnu.isControllable())
      return cstnu;

    // Determine the CutOff Bound
    int nodeCount = cstnu.getTimePoints().size();
    int cutoff = nodeCount * cstnu.getEdges().size();
    
    // Create the solver used for calculating the minimal network and checking consistency of the allmax projection.
    PC2Solver<STNEdge> pc2solver = new PC2Solver<STNEdge>();
    
    // Create the AllMax Projection of the CSTNU. This will be reused in each iteration. cf. update[Contingent]Constraint()
    allMax = new STNUAllMaxProjection(cstnu);
    
    // Create the queue holding the edges to be updated
    Queue<Path<LabeledTimePoint>> queue = new DuplicateFreeQueue<Path<LabeledTimePoint>>(nodeCount*nodeCount*2);

    Collection<LabeledTimePoint> timePoints = model.getTimePoints();
    
    // Repeat until cutoff bound reached
    int iter;
    for (iter = 0; iter < cutoff; iter++)
    {
      if (Thread.currentThread().isInterrupted())
        throw new InterruptedException();
      assert debug(Level.FINE, "\nIteration %s", iter);
      updateCount = 0;
      
      // Check the allMax projection for consistency. Use the changed edges from
      // the last iteration as starting point! Only these have been changed
      // during the last run of DCSolver. Note that changedEdges may initially
      // be null (i.e., during the first run)
      pc2solver.solve(allMax, changedEdges, null);

      // Get the initial paths to checked
      
      getInitialPaths(timePoints, queue);

      // Initialize Data Structure holding all changes during this run.
      changedEdges = new ArrayList<Path<LabeledTimePoint>>();
      
      // Variable determining if there are any changes
      boolean changes = false;
      
      // Repeat for each Triple
      while (!queue.isEmpty())
      {
        if (Thread.currentThread().isInterrupted())
          throw new InterruptedException();
        final Path<LabeledTimePoint> path = queue.remove();
        
        
        LabeledTimePoint i = path.getFirst();
        LabeledTimePoint j = path.getSecond();
//        LabeledTimePoint k = triangle.getThird();
        
        for (LabeledTimePoint k : timePoints)
        {
          
          if (i.equals(k) || j.equals(k))
            continue;
          
          changes |= processPath(cstnu, changeSet, i, j, k);
          changes |= processPath(cstnu, changeSet, k, i, j);
        }
      
      }
      // If there are no more changes we are finished and the CSTNU is controllable
      if (!changes)
      {
        assert debug(Level.FINE, "Controllable!");
        break;
      }
      
      assert debug(Level.FINE, "%s Updates", updateCount );
      
        
    }
    
    // Check if we left the loop due to the Cutoff Bound 
    if (iter >= cutoff)
    {
      assert debug(Level.FINE, "Upps Cutoff bound reached");
      throw new TCNInconsistencyException("CutoffBound reached", cstnu);
    }
    
    cstnu.setControllable();
    return cstnu;
  }
  
  public boolean processPath(ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet, 
      LabeledTimePoint i,
      LabeledTimePoint j, LabeledTimePoint k) throws TCNInconsistencyException
  {

    
    //assert debug(Level.FINE, String.format("%s -> %s -> %s", i, j, k));
    
    // Get all three edges of the triple
    final LabeledEdge e_ij = model.getEdge(i, j);
    final LabeledEdge e_jk = model.getEdge(j, k);
    
           
    // Check if the triple has enough "real" edges, i.e., at least the two main edges must be non-null
    if (e_ij == null || e_jk == null)
    {
      return false;
    }

    boolean localChanges = false;
    assert model != null;
    assert e_ij != null;
    assert e_jk != null;
    
    assert e_ij.getTargetNode() == e_jk.getSourceNode();

    
    final LabeledEdge e_ik = model.getEdge(i, k);
    
    assert debug(Level.FINE, "Evaluating (%s, %s) = (%s) \\cap (%s) + (%s)", i, k, e_ik, e_ij, e_jk);
    
    LabeledValue value = null;
    if (e_ik != null)
    {
      value = e_ik.getValue().getLabeledValues().iterator().next();
    }
    // Apply Reduction Rules
    LabeledValue noCaseReduction = NoCaseReduction(model, e_ij, e_jk);
    if (noCaseReduction != null)
    {
      LabeledValue newValue = value != null ? merge(noCaseReduction, value) : noCaseReduction;
      assert debug(Level.FINEST, "\tAfter merge: %s", newValue);
      localChanges = newValue != value;
      value = newValue;
    }
    
    LabeledValue upperCaseReduction = UpperCaseReduction(model, e_ij,
                                                         e_jk);
    if (upperCaseReduction != null)
    {
      LabeledValue newValue = value != null ? merge(upperCaseReduction, value) : upperCaseReduction;
      assert debug(Level.FINEST, "\tAfter merge: %s", newValue);
      localChanges = newValue != value;
      value = newValue;
    }
    
    LabeledValue crossCaseReduction = CrossCaseReduction(model, e_ij,
                                                         e_jk);
    if (crossCaseReduction != null)
    {
      LabeledValue newValue = value != null ? merge(crossCaseReduction, value) : crossCaseReduction;
      assert debug(Level.FINEST, "\tAfter merge: %s", newValue);
      localChanges = newValue != value;
      value = newValue;
    }
    
    LabeledValue lowerCaseReduction = LowerCaseReduction(model, e_ij,
                                                         e_jk);
    if (lowerCaseReduction != null)
    {
      LabeledValue newValue = value != null ? merge(lowerCaseReduction, value) : lowerCaseReduction;
      assert debug(Level.FINEST, "\tAfter merge: %s", newValue);
      localChanges = newValue != value;
      value = newValue;
    }
    
    assert debug(Level.FINE, "   Generation Rules generated %s", value);
    assert debug(Level.FINE, "   Intersection resulted in %s", value);
    
    LabeledValue labelRemovalReduction = LabelRemovalReduction(model,
                                                               value,
                                                               e_ij.getSourceNode(),
                                                               e_jk.getTargetNode());
    if (labelRemovalReduction != null)
    {
      LabeledValue newValue = labelRemovalReduction;
      assert debug(Level.FINEST, "\tAfter merge: %s", newValue);
      localChanges = newValue != value;
      value = newValue;
    }
    
    assert debug(Level.FINE, "   Label Removal Rules resulted in %s", value);
    
    
    LabeledValueSet labeledValueSet = new DefaultLabeledValueSet(value);
    if (localChanges 
        && (e_ik == null || !labeledValueSet.equals(e_ik.getValue())))
    {
      assert debug(Level.FINEST, "Equals: %s", (e_ik != null ? Boolean.valueOf(labeledValueSet.equals(e_ik.getValue())) : "null"));
      LabeledEdge labeledEdge = new DefaultLabeledEdge(i, k, labeledValueSet, 
                                                e_ik != null && e_ik.isControlEdge(), 
                                                e_ik == null || e_ik.isImplicit(),
                                                e_ik != null ? e_ik.getFlowDirection() : FlowDirection.UNKOWN);
      assert debug(Level.FINE, "  Updating\t %s to %s (because of %s and %s)", e_ik, labeledEdge, e_ij, e_jk);
      // FIXME why contingent?
      updateContingentConstraint(changeSet, model, labeledEdge, 
                               new Triangle<LabeledTimePoint>(new Path<LabeledTimePoint>(i, k), j), "STNUDC");
      
      return true;
    }
    else
    {
      assert debug(Level.FINE, "  Ignoring\t %s to %s (because no changes)", e_ik, labeledValueSet);
    }
    return false;
  }
  
  protected LabeledValue merge(
      LabeledValue newValue, LabeledValue oldValue)
  {
    if (newValue instanceof ContingentLabeledValue)
    {
      if (oldValue instanceof ContingentLabeledValue)
      {
        // Both are contingent
        ContingentLabeledValue clNew = (ContingentLabeledValue) newValue;
        ContingentLabeledValue clOld = (ContingentLabeledValue) oldValue;
        
        
        int compareNonCont = clOld.getValue().compareTo(clNew.getValue());
        int compareCont = clOld.getContingentValue().compareTo(clNew.getContingentValue());
        
        // if one of the values has been decreased
        if (compareCont > 0 || compareNonCont > 0) //old > new
        {
          assert clNew.getCase() == clOld.getCase() : clOld + " vs. "+clNew;
          assert clNew.getTrigger() == clOld.getTrigger() : "Trying to merge contingent constraints with different triggers!";
          
          // Merge or replace
          if (compareCont >= 0 && compareNonCont >= 0) // old >= new
          {
            // Both have either been decreased or not changed so replace
            //labels.put(newValue.getLabel(), newValue);
            return newValue;
          }
          else // old < new
          {
            
            // Merge
            TimeDistance nonContValeu = compareNonCont > 0 ? clNew.getValue() : clOld.getValue();
            TimeDistance contValue = compareCont > 0 ? clNew.getContingentValue() : clOld.getContingentValue();
            //labels.put(clNew.getLabel(), createNewConstraint(clNew, nonContValeu, contValue));
            return updateLabeledValue(clNew, nonContValeu, contValue);
          }
        }
        else
        {
          return oldValue;
        }
      }
      else
      {
        // new is contingent old is non-contingent
        // Merge
        
        ContingentLabeledValue clNew = (ContingentLabeledValue) newValue;
                 
        TimeDistance oldTD = oldValue.getValue();
        TimeDistance newTD = clNew.getValue();
        int compareNonCont = oldTD.compareTo(newTD);
        
        // if one of the values has been decreased
        if (compareNonCont >= 0) // old > new
        {
          return newValue;
        }
        else
        {
          // Merge
          TimeDistance contValue = clNew.getContingentValue();
          TimeDistance nonContValue = oldValue.getValue();
          //labels.put(clNew.getLabel(), createNewConstraint(clNew, nonContValue, contValue));
          return updateLabeledValue(clNew, nonContValue, contValue);
        }
      }
    
    }
    else
    {
      if (oldValue instanceof ContingentLabeledValue)
      {
        // new is non-contingent old is contingent
        // Merge
        ContingentLabeledValue clOld = (ContingentLabeledValue) oldValue;
        
        int compareNonCont = clOld.getValue().compareTo(newValue.getValue());
        
        // if one of the values has been decreased
        if (compareNonCont > 0)
        {
          // Merge
          TimeDistance nonContValue = newValue.getValue();
          TimeDistance contValue = clOld.getContingentValue();
//          labels.put(clOld.getLabel(), createNewConstraint(clOld, nonContValue, contValue));
          return updateLabeledValue(clOld, nonContValue, contValue);
        }
        else
        {
          return oldValue;
        }
      }
      else
      {
        // Both are non-contingent
        // Replace if necessary
        int compare = oldValue.getValue().compareTo(newValue.getValue());
        if (compare > 0) // old > new
        {
//          labels.put(newValue.getLabel(), newValue);
          return newValue;
        }
        else
        {
          return oldValue;
        }
      }
    }
    //return oldValue;
  }
  
  protected LabeledValue updateLabeledValue(ContingentLabeledValue oldValue,
      TimeDistance nonContValue, TimeDistance contValue)
  {
    LabeledValue value;
    if (nonContValue.compareTo(contValue) <= 0) // Non-contingent value is more
                                               // restrictive than contingent
    {
      value = new DefaultLabeledValue(oldValue.getLabel(), nonContValue);
    }
    else
    {
      value = new DefaultContingentLabeledValue(oldValue.getLabel(), oldValue.getCase(),
                                    nonContValue, oldValue.getTrigger(), contValue);
    }
    return value;
  }
  
  protected ContingentLabeledValue getContingentDurationConstraint(STNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu,
      LabeledTimePoint from, LabeledTimePoint to)
  {
    LabeledEdge edge = cstnu.getEdge(from,
                                   to);
    
    return getContingentDurationConstraint(edge);
  }

  protected ContingentLabeledValue getContingentDurationConstraint(LabeledEdge edge)
  {
    ContingentLabeledValue res = null;
    
    if (edge != null)
    {
      assert edge.getValue().getLabeledValues().size() == 1;
      for (LabeledValue v : edge.getValue().getLabeledValues())
      {
        if (v instanceof ContingentLabeledValue)
        {
          if (((ContingentLabeledValue) v).getTrigger() == edge.getSourceNode() || ((ContingentLabeledValue) v).getTrigger() == edge.getTargetNode())
          {
            
          if (res != null)
            throw new IllegalStateException();
          res = (ContingentLabeledValue) v;
          }
        }
      }
      
      if (res == null)
        throw new IllegalStateException();
    }
    
    return res;
  }
  
  /**
   * 
   * B <-b:x- A <-B:z C adds A <-z- C if z>=-x
   * 
   * @param cstnu
   * @param CA
   * @param AB
   * @param changeSet 
   * @return
   * @throws TCNInconsistencyException 
   */
  private LabeledValue LabelRemovalReduction(STNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu, LabeledValue l_CA,
      LabeledTimePoint C, LabeledTimePoint A) throws TCNInconsistencyException
  {
    //assert AB.getValue().getLabeledValues().size() == 1;
    
    
    // C -B:z-> A
    if (!(l_CA instanceof ContingentLabeledValue && ConstraintCase.UPPERCASE.equals(((ContingentLabeledValue)l_CA).getCase())))
      return null;
    
    ContingentLabeledValue cl_CA = (ContingentLabeledValue) l_CA;
    
    // Never apply this rule if this is a contingent duration constraint!!!
    if (cl_CA.getTrigger() == C || cl_CA.getTrigger() == A) 
      return null;
    
    ContingentLabeledValue cl_AB = getContingentDurationConstraint(cstnu, A, cl_CA.getTrigger());
    
    
    // T -b:x-> R
    if (!(ConstraintCase.LOWERCASE.equals(cl_AB.getCase()))) // Is this even possible or should this be an assertion?
      return null;
         
    TimeDistance z = cl_CA.getContingentValue();
    TimeDistance x = cl_AB.getContingentValue();
    
    if (z.compareTo(x.neg()) < 0) // z >= -x
      return null;
    
    
    //B <-b:x- A A <-B:z C adds A <-z- C if z>=-x
    //B <-b:x- A
    assert cl_AB.getCase().equals(ConstraintCase.LOWERCASE);
    // A <-B:z C
    assert cl_CA.getCase().equals(ConstraintCase.UPPERCASE);
    // z >= -x
    //assert z >= -x;

    // Calculate the new Label (same trigger and same case as CA)
    // S -z-> T
    TimeDistance value = z;
    Label proposition = cl_CA.getLabel();
    // Label is downgraded to a non-contingent label so update is required
    LabeledValue newLabel = new DefaultLabeledValue(proposition, value);
    
    assert debug(Level.FINER, "\tLabelRemoval:\t Replacing %s by %s", cl_CA, newLabel);
    
    return newLabel;
  }

  /**
   * 
   * A <-x- C <-c:y- D adds A <-x+y- D if x<=0
   * 
   * @param cstnu
   * @param QS
   * @param ST
   * @param QT
   * @param changeSet 
   * @return
   * @throws TCNInconsistencyException 
   */
  private LabeledValue LowerCaseReduction(STNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu, LabeledEdge QS,
      LabeledEdge ST) throws TCNInconsistencyException
  {
    assert QS.getValue().getLabeledValues().size() == 1;
    assert ST.getValue().getLabeledValues().size() == 1;
    
    LabeledValue l_ST = ST.getValue().getLabeledValues().iterator().next();
    
    //S -x-> T
    if (l_ST instanceof ContingentLabeledValue)
      return null;
    
    LabeledValue l_QS = QS.getValue().getLabeledValues().iterator().next();

    // Q -c:y-> S
    if (!(l_QS instanceof ContingentLabeledValue && ConstraintCase.LOWERCASE.equals(((ContingentLabeledValue)l_QS).getCase())))
      return null;
         
    
    ContingentLabeledValue cl_QS = (ContingentLabeledValue) l_QS;
          
    TimeDistance v = l_ST.getValue();
    long cv = v.sign();
    if (!((cv < 0) || (cv == 0 && ST.getSourceNode() != ST.getTargetNode()))) // x <= 0
      return null;
          
          
    //A <-x- C <-c:y- D adds A <-x+y- D if x<=0
    //A <-x- C
    // C <-c:y- D
    assert ConstraintCase.LOWERCASE.equals(cl_QS.getCase());
    // x <= 0
    //assert cl_CA.getValue() <= 0;
    // "C = C"
    assert QS.getTargetNode() == ST.getSourceNode();
    // "C = c"
    assert cl_QS.getTrigger() == QS.getTargetNode();
    
    // Calculate the new Label
    // A <-x+y- D
    TimeDistance value = l_ST.getValue().add(cl_QS.getContingentValue());
    Label proposition = mergeLabels(l_ST.getLabel(), cl_QS.getLabel());
    //CSTNULabel newLabel = new CSTNULabel(proposition, null, null, value);
    
    LabeledValue newLabel = new DefaultLabeledValue(proposition, value);
    
    assert debug(Level.FINER, "\tLowerCase:\t %s + %s -> %s", cl_QS, l_ST, newLabel);
    
    return newLabel;
    
  }

  /**
   * 
   * A <-B:x- C <-c:y- D adds A<-B:(x+y)-D
   * 
   * @param cstnu
   * @param DC
   * @param CA
   * @param DA
   * @param changeSet 
   * @return
   * @throws TCNInconsistencyException 
   */
  private LabeledValue CrossCaseReduction(STNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu, LabeledEdge DC,
      LabeledEdge CA) throws TCNInconsistencyException
  {
    assert DC.getValue().getLabeledValues().size() == 1;
    assert CA.getValue().getLabeledValues().size() == 1;
    
    LabeledValue l_CA = CA.getValue().getLabeledValues().iterator().next();
    
    //A <-B:x- C
    if (!(l_CA instanceof ContingentLabeledValue && ConstraintCase.UPPERCASE.equals(((ContingentLabeledValue)l_CA).getCase())))
      return null;
        
    ContingentLabeledValue cl_CA = (ContingentLabeledValue) l_CA;

    LabeledValue l_DC = DC.getValue().getLabeledValues().iterator().next();
          
    // C <-c:y- D
    if (! (l_DC instanceof ContingentLabeledValue && ConstraintCase.LOWERCASE.equals(((ContingentLabeledValue)l_DC).getCase())) )
      return null;
    
    ContingentLabeledValue cl_DC = (ContingentLabeledValue) l_DC;
          
    // x<=0, B != C
    if (!(cl_CA.getContingentValue().sign() <= 0 && cl_CA.getTrigger() != CA.getSourceNode())) 
      return null;
    
    
    //A <-B:x- C C <-c:y- D adds A<-B:(x+y)-D If x <= 0, B != C
    //A <-B:x- C
    assert cl_CA.getCase().equals(ConstraintCase.UPPERCASE);
    // C <-c:y- D
    assert cl_DC.getCase().equals(ConstraintCase.LOWERCASE);
    // x <= 0
    //assert cl_CA.getValue() <= 0;
    // "C = C"
    assert DC.getTargetNode() == CA.getSourceNode();
    // B != C
    assert cl_CA.getTrigger() != CA.getSourceNode();
    // "C = c"
    assert cl_DC.getTrigger() == DC.getTargetNode();
          
    // Calculate the new Label (same trigger and same case as CA)
    // A<-B:(x+y)-D
    TimeDistance value = cl_CA.getContingentValue().add(cl_DC.getContingentValue());
    Label proposition = mergeLabels(cl_CA.getLabel(), cl_DC.getLabel());
    //CSTNULabel newLabel = new CSTNULabel(proposition, cl_CA.getTrigger(), cl_CA.getCase(), value);

    
    
    
    LabeledValue newLabel = createContingentConstraint(proposition, value, cl_CA.getTrigger());
    
    assert debug(Level.FINER, "\tCrossCase:\t %s + %s -> %s", cl_DC, cl_CA, newLabel);
    
    return newLabel;
  }

  /**
   * 
   * Q -u-> S -R:v-> T
   * ==> Q -R:u+v-> T
   * 
   * @param cstnu
   * @param QS
   * @param ST
   * @param QT
   * @param changeSet 
   * @return
   * @throws TCNInconsistencyException 
   */
  private LabeledValue UpperCaseReduction(STNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu, LabeledEdge QS,
      LabeledEdge ST) throws TCNInconsistencyException
  {
    assert QS.getValue().getLabeledValues().size() == 1;
    assert ST.getValue().getLabeledValues().size() == 1;

        
    LabeledValue l_QS = QS.getValue().getLabeledValues().iterator().next();

    // Q -u-> S
    if (l_QS instanceof ContingentLabeledValue)
      return null;    
    
    LabeledValue l_ST = ST.getValue().getLabeledValues().iterator().next();
    
    //S -R:v-> T
    if (!(l_ST instanceof ContingentLabeledValue && ConstraintCase.UPPERCASE.equals(((ContingentLabeledValue)l_ST).getCase())))
      return null;
    
    ContingentLabeledValue cl_ST = (ContingentLabeledValue) l_ST;
    
    // Q -u-> S -R:v-> T
    
    // Q -u-> S
    // S -R:v-> T
    assert cl_ST.getCase().equals(ConstraintCase.UPPERCASE);
    // "S = S"
    assert QS.getTargetNode() == ST.getSourceNode();   
    
    // Calculate the new Label (same trigger and same case as ST)
    // Q -R:u+v-> T
    TimeDistance value = l_QS.getValue().add(cl_ST.getContingentValue());
    Label proposition = mergeLabels(l_QS.getLabel(), cl_ST.getLabel());

    
    LabeledValue newLabel = createContingentConstraint(proposition, value, cl_ST.getTrigger());  
    
    assert debug(Level.FINER, "\tUpperCase:\t %s + %s -> %s", l_QS, cl_ST, newLabel);
    
    return newLabel;
  }


  protected LabeledValue createContingentConstraint(Label proposition, TimeDistance value,
      LabeledTimePoint trigger)
  {
    LabeledValue newLabel = new DefaultContingentLabeledValue(proposition, ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, value);

    return newLabel;
  }

  /**
   * 
   * A <-x- C <-y- D adds A <-x+y- D
   * 
   * @param cstnu
   * @param QS
   * @param ST
   * @param QT 
   * @param changeSet 
   * @return
   * @throws TCNInconsistencyException 
   */
  private LabeledValue NoCaseReduction(STNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu, LabeledEdge QS,
      LabeledEdge ST) throws TCNInconsistencyException
  {
    assert QS.getValue().getLabeledValues().size() == 1;
    assert ST.getValue().getLabeledValues().size() == 1;
    
    
    LabeledValue cl_ST = ST.getValue().getLabeledValues().iterator().next();
            
    LabeledValue cl_QS = QS.getValue().getLabeledValues().iterator().next();
        
    //A <-x- C C <-y- D adds A <-x+y- D
    //A <-x- C
//    assert !(cl_ST instanceof ContingentLabeledValue);
    // C <-y- D
//    assert !(cl_QS instanceof ContingentLabeledValue);
    // Not null
    assert cl_ST.getValue() != null;
    assert cl_QS.getValue() != null;
    // "C = C"
    assert QS.getTargetNode() == ST.getSourceNode();        
        
    TimeDistance value = cl_QS.getValue().add(cl_ST.getValue());
    Label proposition = mergeLabels(cl_QS.getLabel(), cl_ST.getLabel());
    //CSTNULabel newLabel = new CSTNULabel(proposition, null, null, value);
        
    LabeledValue newLabel = new DefaultLabeledValue(proposition, value);
    
    assert debug(Level.FINER, "\tNoCase:\t\t %s + %s -> %s", cl_QS, cl_ST, newLabel);
    
    return newLabel;
  }

  private void updateContingentConstraint(
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet,
      ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> cstnu, LabeledEdge edge,
      Triangle<LabeledTimePoint> causedBy, String reason)
      throws TCNInconsistencyException
  {
    LabeledEdge oldEdge = cstnu.getEdge(edge.getSourceNode(), edge.getTargetNode());
    if (changeSet != null)
    {
      LabeledValueSet oldValue = oldEdge != null ? oldEdge.getValue() : null;
      logChange(changeSet, causedBy.getFirst(), causedBy.getSecond(), edge, oldValue, reason);
    }
        
    checkConsistency(cstnu, edge, causedBy);

    cstnu.updateContingentConstraint(edge);
    changedEdges.add(new Path<LabeledTimePoint>(edge.getSourceNode(), edge.getTargetNode()));

    allMax.removeDelta(edge);
    updateCount++;
  }

  @Override
  public boolean checkConsistency(LabeledValueSet value1, LabeledValueSet value2)
  {
    for (LabeledValue labeledValue1 : value1.getLabeledValues())
    {
      for (LabeledValue labeledValue2 : value2.getLabeledValues())
      {
        if (!isConsistent(labeledValue1, labeledValue2))
        {
          return false;
        }
      }
    }
    return true;
  }

  


}
