package de.aristaflow.adept2.model.processmodel.simulator;

import java.util.HashMap;
import java.util.Map;
import java.util.PriorityQueue;

import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeState;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools;

public class TemplateSimulator
{

  public class FutureStateChange implements Comparable<FutureStateChange>
  {

    public int nodeID;
    public long startTime;
    public int iteration;
    public NodeState nextState;

    public FutureStateChange(int nodeID, int iteration, NodeState nextState, long timestamp)
    {
      this.nodeID = nodeID;
      this.iteration = iteration;
      this.nextState = nextState;
      this.startTime = timestamp;
    }

    public int compareTo(FutureStateChange o)
    {
      return (int)(this.startTime - o.startTime);
    }


  }

  protected TemporalTemplate template;
  private SimulationCallback callback;
  private SimulationStrategy simulationStrategy;
  private Map<Integer, FutureStateChange> nodeStates = new HashMap<Integer, FutureStateChange>();
  private Map<Integer, Integer> baseIteration = new HashMap<Integer, Integer>();

  public TemplateSimulator(TemporalTemplate template, SimulationCallback callback)
  {
    this.template = template;
    this.callback = callback;
    this.simulationStrategy = new DefaultSimulationStrategy(template, new UniformDistribution());
  }
  
  public void setSimulationStrategy(SimulationStrategy strategy)
  {
    this.simulationStrategy = strategy;
  }
  
  
  public void simulate()
  {
    int startNodeID = template.getStartNode().getID();
    
    long currentTime = this.simulationStrategy.getProcessInstanceCreationTime();
    
    callback.instanceCreated(currentTime);
    currentTime += 1000; // 1sec
    //callback.instanceStarted(currentTime); // This is the same as startNode switching to state running
    
    TemporalNode node = template.getNode(startNodeID);
        
    FutureStateChange f = new FutureStateChange(node.getID(), 0, NodeState.NS_RUNNING, currentTime);
    
    PriorityQueue<FutureStateChange> queue = new PriorityQueue<FutureStateChange>();
    queue.offer(f);
    
    while (!queue.isEmpty())
    {
      FutureStateChange stateChange = queue.poll();
      
      switch (stateChange.nextState)
      {
        case NS_ACTIVATED:
          currentTime = activateNode(stateChange, queue);
          break;
        case NS_RUNNING:
          currentTime = performNode(stateChange, queue);
          break;
        case NS_COMPLETED:
          currentTime = finishNode(stateChange, queue);
          break;
        case NS_SKIPPED:
          currentTime = skipNode(stateChange, queue);
          break;
        // Don't care...
        case NS_ENQUIRED:
        case NS_FAILED:
        case NS_NOT_ACTIVATED:
        case NS_SELECTED:
        case NS_SUSPENDED:
        default:
          break;
          
      }
      
      
    }
    
    callback.instanceFinished(currentTime);
  }


  private long skipNode(FutureStateChange stateChange,
      PriorityQueue<FutureStateChange> queue)
  {
    long currentTime = stateChange.startTime;
    System.out.println(String.format("Node %s skipped at %d", stateChange.nodeID, currentTime));
    callback.activitySkipped(stateChange.nodeID, stateChange.iteration, currentTime);

    nodeStates.put(stateChange.nodeID, stateChange);
    
    skipOrActivateSuccessors(stateChange, queue);
    
    return currentTime;
  }


  private long finishNode(FutureStateChange stateChange,
      PriorityQueue<FutureStateChange> queue)
  {
    long currentTime = stateChange.startTime;
    System.out.println(String.format("Node %s finished at %d", stateChange.nodeID, currentTime));
    callback.activityFinished(stateChange.nodeID, stateChange.iteration, currentTime);

    nodeStates.put(stateChange.nodeID, stateChange);
    
    activateSuccessors(stateChange, queue);
    
    return currentTime;
  }

  private long performNode(FutureStateChange stateChange,
      PriorityQueue<FutureStateChange> queue)
  {
    long currentTime = stateChange.startTime;
    System.out.println(String.format("Node %s started at %d", stateChange.nodeID, currentTime));
    callback.activityStarted(stateChange.nodeID, stateChange.iteration, currentTime);
    nodeStates.put(stateChange.nodeID, stateChange);
    
    TemporalNode node = template.getNode(stateChange.nodeID);
    long endTime = currentTime + simulationStrategy.determineNodeDuration(node);
    
    queue.offer(new FutureStateChange(stateChange.nodeID, stateChange.iteration, NodeState.NS_COMPLETED, endTime));
    
    return currentTime;
  }


  private long activateNode(FutureStateChange stateChange,
      PriorityQueue<FutureStateChange> queue)
  {
    long currentTime = stateChange.startTime;
    System.out.println(String.format("Node %s activated at %d", stateChange.nodeID, currentTime));
    callback.activityActivated(stateChange.nodeID, stateChange.iteration, currentTime);
    
    nodeStates.put(stateChange.nodeID, stateChange);
    
    TemporalNode node = template.getNode(stateChange.nodeID);

    long startTime = simulationStrategy.determineNodeStartTime(node, currentTime);
    
    queue.offer(new FutureStateChange(stateChange.nodeID, stateChange.iteration, NodeState.NS_RUNNING, startTime));  
    
    return currentTime;
  }


  private void activateSuccessors(FutureStateChange futureNode, PriorityQueue<FutureStateChange> queue)
  {
    
    NodeType nodeType = template.getNodeType(futureNode.nodeID);
    
    switch (nodeType)
    {
      case NT_XOR_SPLIT:
        {
          int[] succs = template.getSuccByEdgeType(futureNode.nodeID, EdgeType.ET_CONTROL);
          
          TemporalNode node = template.getNode(futureNode.nodeID);
          
          int sel = simulationStrategy.determineXORBranch(node);
          
          for (int nodeID : succs)
          {
            if (nodeID == sel)
            {
              if (isNodeReady(nodeID, futureNode.iteration))
              {
                 queue.offer(new FutureStateChange(nodeID, futureNode.iteration, NodeState.NS_ACTIVATED, futureNode.startTime));
              }
            }
            else
            {
              queue.offer(new FutureStateChange(nodeID, futureNode.iteration, NodeState.NS_SKIPPED, futureNode.startTime));
            }
          }
          
          int[] syncSucss = template.getSuccByEdgeType(futureNode.nodeID, EdgeType.ET_SYNC);
          for (int nodeID : syncSucss)
          {
            if (isNodeReady(nodeID, futureNode.iteration))
            {
               queue.offer(new FutureStateChange(nodeID, futureNode.iteration, NodeState.NS_ACTIVATED, futureNode.startTime));
            }
          }
        }
        break;
      
      case NT_ENDLOOP:
        TemporalNode node = template.getNode(futureNode.nodeID);
        boolean repeat = simulationStrategy.determineLoop(node);

        if (repeat)
        {
          int[] succs = template.getSuccByEdgeType(futureNode.nodeID, EdgeType.ET_LOOP);
          int nodeID = succs[0];// only one!
          
          if (!baseIteration.containsKey(futureNode.nodeID))
          {
            baseIteration.put(futureNode.nodeID, futureNode.iteration);
          }
          
          queue.offer(new FutureStateChange(nodeID, futureNode.iteration+1, NodeState.NS_ACTIVATED, futureNode.startTime));
        }
        else
        {
          int[] succs = template.getSuccByEdgeType(futureNode.nodeID, EdgeType.ET_CONTROL);
          int nodeID = succs[0];// only one!
          
          int iteration;
          Integer iter = baseIteration.remove(futureNode.nodeID);
          if (iter == null)
          {
            iteration = futureNode.iteration;
          }
          else
          {
            iteration = iter;
          }
          
          
          if (isNodeReady(nodeID, futureNode.iteration))
          {
            queue.offer(new FutureStateChange(nodeID, iteration, NodeState.NS_ACTIVATED, futureNode.startTime));
          }
        }
        
        break;
      
      case NT_NORMAL:
      case NT_STARTFLOW:
      case NT_STARTLOOP:
      case NT_ENDFLOW:
      case NT_AND_SPLIT:
      case NT_AND_JOIN:
      case NT_XOR_JOIN:
      default:
      {
        int[] possibleSuccs = template.getSuccByEdgeType(futureNode.nodeID, EdgeType.ET_CONTROL, EdgeType.ET_SYNC);
        for (int nodeID : possibleSuccs)
        {
          if (isNodeReady(nodeID, futureNode.iteration))
          {
             queue.offer(new FutureStateChange(nodeID, futureNode.iteration, NodeState.NS_ACTIVATED, futureNode.startTime));
          }
        }
      }
      break;
    }

    
  }
  
  private void skipOrActivateSuccessors(FutureStateChange stateChange,
      PriorityQueue<FutureStateChange> queue)
  {
    
    int[] succs = template.getSuccByEdgeType(stateChange.nodeID, EdgeType.ET_CONTROL);
    for (int nodeID : succs)
    {
      boolean nodeSkipped = true;
      int[] preds = template.getPredByEdgeType(stateChange.nodeID, EdgeType.ET_CONTROL);
      for (int predNodeID : preds)
      {
        FutureStateChange predStateChange = nodeStates.get(predNodeID);
        if (predStateChange == null)
        {
          nodeSkipped = false;
        }
        else
        {
          NodeState nodeState = predStateChange.nextState;
          if (!NodeState.NS_SKIPPED.equals(nodeState) || !(predStateChange.iteration == stateChange.iteration))
          {
            nodeSkipped = false;
          }
        }
      }
      
      // All controll-flow predecessors have been skipped, skip node as well
      if (nodeSkipped)
      {
        queue.offer(new FutureStateChange(nodeID, stateChange.iteration, NodeState.NS_SKIPPED, stateChange.startTime));
      }
      else
      {
        if (isNodeReady(nodeID, stateChange.iteration))
        {
           queue.offer(new FutureStateChange(nodeID, stateChange.iteration, NodeState.NS_ACTIVATED, stateChange.startTime));
        }
      }
    }
    
    int[] syncSucss = template.getSuccByEdgeType(stateChange.nodeID, EdgeType.ET_SYNC);
    for (int nodeID : syncSucss)
    {
      if (isNodeReady(nodeID, stateChange.iteration))
      {
         queue.offer(new FutureStateChange(nodeID, stateChange.iteration, NodeState.NS_ACTIVATED, stateChange.startTime));
      }
    }
  }


  private boolean isNodeReady(int nodeID, int iteration)
  {
    int[] preds = template.getPredByEdgeType(nodeID, EdgeType.ET_CONTROL, EdgeType.ET_SYNC);
    for (int predNodeID : preds)
    {
      FutureStateChange predStateChange = nodeStates.get(predNodeID);
      if (predStateChange == null)
        return false;
      
      NodeState nodeState = predStateChange.nextState;
      if (!(predStateChange.iteration >= iteration))
        return false;
      
      
      if (!(NodeState.NS_COMPLETED.equals(nodeState) || NodeState.NS_SKIPPED.equals(nodeState) || baseIteration.containsKey(predNodeID)))
      {
        return false;
      }
    }
    return true;
  }


//  private long performNode(int nodeID, int iteration, long currentTime)
//  {
//    callback.activityStarted(nodeID, iteration, currentTime);
//    
//    
//    TemporalNode node = template.getNode(nodeID);
//    long duration = simulationStrategy.determineNodeDuration(node);
//    
//    currentTime += duration;
//    
//    callback.activityFinished(nodeID, iteration, currentTime);
//    
//    return currentTime;
//  }





  protected long getStartTime()
  {
    return System.currentTimeMillis();
  }

}
