package de.aristaflow.adept2.model.processmodel.simulator;

import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.InconsistencyException;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTimeModelFactory;

public class TimeModelSimulationStrategy extends DefaultSimulationStrategy implements SimulationCallback
{

  private TimeModelSolver solver;
  private UpdateableInstanceTimeModel timeModel;
  private TimeModelOperations timeModelOperations;

  public TimeModelSimulationStrategy(TemporalTemplate template,
      Distribution distribution)
  {
    super(template, distribution);
    
    TimeModelFactory factory = new CSTNUTimeModelFactory();
    
    TemplateTimeModel templateTimeModel = factory.createTimeModelForTemplate(template);
    timeModelOperations = factory.getTimeModelOperations();
    timeModel = (UpdateableInstanceTimeModel) factory.createTimeModelForInstance(null, null, templateTimeModel);
    solver = factory.getTimeModelSolver(timeModel.getClass());
  }
  
  @Override
  public long determineNodeStartTime(TemporalNode node, long currentTime)
  {
    Interval<CalendarElement> startTimeFrame = timeModel.getActivityInstanceStartTimeframe(node.getID(), 0);
    
    long earliestStartTime = startTimeFrame.getMinValue().getTimeInMillis();
    long latestStartTime = startTimeFrame.getMaxValue().getTimeInMillis();
    
    assert (earliestStartTime >= currentTime) : "Earliest time is "+earliestStartTime+ " but current time is "+currentTime; // Should be ensured by time model
    assert currentTime <= latestStartTime;
    
    //long startTime = distribution.drawRandom(minStartTimeFrame, maxStartTimeFrame);
    long startTime = earliestStartTime;
    
    return startTime;
  }
  

  @Override
  public long determineNodeDuration(TemporalNode node)
  {
    Interval<TimeDistance> activityDuration = timeModel.getActivityInstanceDuration(node.getID(), 0);
    
    
    long duration;
    
    if (activityDuration != null)
    {
      TimeDistance minimumDuration = activityDuration.getMinValue();
      TimeDistance maximumDuration = activityDuration.getMaxValue();
      
      long minDuration = minimumDuration.toMilliseconds();
      long maxDuration = maximumDuration.toMilliseconds();
      
      assert minDuration > 0;
      
      
      duration = distribution.drawRandom(minDuration, maxDuration);
      
//      System.out.println("Duration of "+node.getID()+": "+duration+" in ["+minDuration+", "+maxDuration+"]");
    }
    else
    {
      duration = secondsToMillis(60);
    }
    
    
    return duration;  
  }

  public void instanceCreated(long timestamp)
  {
    timeModel.setInstanceCreationTime(timestamp);
    
//    CSTNUInstanceTimeModel tm = (CSTNUInstanceTimeModel) timeModel;
//    tm.getBaseTCN().visit(visitor.getInstance("InstanceCreated_STN_Pre"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("InstanceCreated_CSTNU_Pre"));
    
    updateTimeModel();
    
//    tm.getBaseTCN().visit(visitor.getInstance("InstanceCreated_STN_Post"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("InstanceCreated_CSTNU_Post"));
  }

  private void updateTimeModel()
  {
    try
    {
      solver.solve(timeModel);
    }
    catch (InconsistencyException e)
    {
      throw new IllegalStateException(e);
    }
    catch (InterruptedException e)
    {
      // This means we are shutting down so ignore
    }
  }

  public void instanceStarted(long timestamp)
  {
    timeModel.setInstanceStartTime(timestamp);
    
//    CSTNUInstanceTimeModel tm = (CSTNUInstanceTimeModel) timeModel;
//    tm.getBaseTCN().visit(visitor.getInstance("InstanceStarted_STN_Pre"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("InstanceStarted_CSTNU_Pre"));
    
    updateTimeModel();
    
//    tm.getBaseTCN().visit(visitor.getInstance("InstanceStarted_STN_Post"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("InstanceStarted_CSTNU_Post"));
  }

  public void activityStarted(int nodeID, int iteration, long timestamp)
  { 
//    checkConstitution(nodeID, iteration, true);
    
    timeModel.setActivityInstanceStartTime(nodeID, iteration, timestamp);
    
//    CSTNUInstanceTimeModel tm = (CSTNUInstanceTimeModel) timeModel;
//    tm.getBaseTCN().visit(visitor.getInstance("Activity"+nodeID+"Started_STN_Pre"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("Activity"+nodeID+"Started_CSTNU_Pre"));
    
    updateTimeModel();
    
//    tm.getBaseTCN().visit(visitor.getInstance("Activity"+nodeID+"Started_STN_Post"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("Activity"+nodeID+"Started_CSTNU_Post"));    
  }

//  private void checkConstitution(int nodeID, int iteration, boolean start)
//  {
//    CSTNUInstanceTimeModel tm = (CSTNUInstanceTimeModel) timeModel;
//    
//    TCN<LabeledNode,DefaultTCNEdge> baseTCN = tm.getBaseTCN();
//    TCN<LabeledNode,LabeledEdge> cstnu = tm.getCSTNUProjection();
//    
//    LabeledNode baseEvent = baseTCN.getTimeBaseEvent();
//    NodeMapping<LabeledNode> nodeMapping = baseTCN.getNodeMapping(nodeID, tm.iterationChain);
//    
//    LabeledNode node;
//    if (start)
//      node = nodeMapping.getStartNode();
//    else
//      node = nodeMapping.getEndNode();
//    
//    DefaultTCNEdge edge1;
//    LabeledEdge edge2;
////    edge1 = baseTCN.getEdge(baseEvent, node);
////    edge2 = cstnu.getEdge(baseEvent, node);
////    
////    if (edge1 != null && !edge1.equals(edge2))
////    {
////      System.out.println("Differing opinions: "+edge1+" vs. "+edge2);
////    }
//    
//    edge1 = baseTCN.getEdge(node, baseEvent);
//    edge2 = cstnu.getEdge(node, baseEvent);
//    
////    if (edge1 != null && !edge1.equals(edge2))
////    {
////      System.out.println("Differing opinions: "+edge1+" vs. "+edge2);
////    }
//  }

  public void activityFinished(int nodeID, int iteration, long timestamp)
  {
//    checkConstitution(nodeID, iteration, false);
    
    CSTNUInstanceTimeModel tm = (CSTNUInstanceTimeModel) timeModel;
    
//    System.out.println("=======Pre==========");
//    TCNTestTools.printCSV(tm.getCSTNU());
//    System.out.println("=======Pre==========\n");
//    
    timeModel.setActivityInstanceFinishTime(nodeID, iteration, timestamp);
    
    
    //TCNTestTools.printTCN(tm.getCSTNU(), this.getClass(), "Activity"+nodeID+"Finished_CSTNU_Pre");
//    System.out.println("=======Mid==========");
//    TCNTestTools.printCSV(tm.getCSTNU());
//    System.out.println("=======Mid==========\n");
//    
    updateTimeModel();
//  TCNTestTools.printTCN(tm.getCSTNU(), this.getClass(), "Activity"+nodeID+"Finished_CSTNU_Post");
//    System.out.println("=======Post==========");
//    TCNTestTools.printCSV(tm.getCSTNU());
//    System.out.println("=======Post==========\n");
//    
  }

  public void activityActivated(int nodeID, int iteration, long timestamp)
  {
    // Not required here
  }

  public void activitySkipped(int nodeID, int iteration, long timestamp)
  {
    // TODO Auto-generated method stub
    
  }

  public void instanceFinished(long timestamp)
  {
    timeModel.setInstanceFinishTime(timestamp);
    
//    CSTNUInstanceTimeModel tm = (CSTNUInstanceTimeModel) timeModel;
//    tm.getBaseTCN().visit(visitor.getInstance("InstanceFinished_STN_Pre"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("InstanceFinished_CSTNU_Pre"));
    
    updateTimeModel();
    
//    tm.getBaseTCN().visit(visitor.getInstance("InstanceFinished_STN_Post"));
//    tm.getCSTNUProjection().visit(visitor.getInstance("InstanceFinished_CSTNU_Post"));
  }

}
