package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import org.apache.commons.lang.StringEscapeUtils;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.DotTCNVisitor;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tools.Graph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;
import de.aristaflow.adept2.tools.dot.AbstractDotGraph;
import de.aristaflow.adept2.tools.dot.DotColor;
import de.aristaflow.adept2.tools.dot.DotEdge;
import de.aristaflow.adept2.tools.dot.DotEdge.EdgeStyle;
import de.aristaflow.adept2.tools.dot.DotNode;
import de.aristaflow.adept2.tools.dot.DotSubGraph;
import de.aristaflow.adept2.tools.dot.DotSubGraph.ElementStyle;
import de.aristaflow.adept2.tools.dot.GraphFormatter;
import de.aristaflow.adept2.tools.dot.LabelProvider;

public class CSTNUGraphFormatter implements
    GraphFormatter<LabeledTimePoint, LabeledEdge>,
    LabelProvider<LabeledTimePoint, LabeledEdge>
{

  boolean htmlLabels = true;

  private DotTCNVisitor<LabeledTimePoint, LabeledEdge> visitor;

  private boolean drawImplicitConstraints = false;

  public CSTNUGraphFormatter()
  {
    this(true, false);
  }

  public CSTNUGraphFormatter(boolean htmlLabels, boolean drawImplicitConstraints)
  {
    this.htmlLabels = htmlLabels;
    this.drawImplicitConstraints = drawImplicitConstraints;
  }

  public void formatVertex(LabeledTimePoint vertex, DotNode node)
  {
    node.setHTMLLabel(htmlLabels);
  }

  public void formatEdge(LabeledEdge e, DotEdge edge)
  {   
    if (e.isContingentDuration())
    {
      edge.setStyle(EdgeStyle.BOLD);
      edge.setColor(DotColor.BLUE);
      edge.setFontColor(DotColor.BLUE);
      edge.setWeight(50);
    }
    else
    {     
      if (e.isContingent())
      {
        edge.setStyle(EdgeStyle.DOTTED);
        edge.setColor(DotColor.GREEN);
        edge.setFontColor(DotColor.GREEN);
      }
    }
    
    if (!e.isControlEdge())
    {
      edge.setConstraint(false);
      edge.setStyle(EdgeStyle.DASHED);
    }
    else
    {
      switch (e.getFlowDirection())
      {
        case FORWARD:
          edge.setConstraint(true);
          //edge.setColor(DotColor.RED);
          edge.setWeight(40);
          break;
        case BACKWARD:
          edge.setConstraint(false);
          //edge.setColor(DotColor.GREEN);
          edge.setWeight(10);
          break;
        default:
          edge.setConstraint(false);
      }
    }

    if (e.isImplicit())
    {
      edge.setStyle(EdgeStyle.DOTTED);
      edge.setConstraint(false);
      edge.setWeight(1);
    }

    edge.setHTMLLabel(htmlLabels);
  }

  public boolean isVisible(LabeledEdge e)
  {
    if (drawImplicitConstraints)
      return true;

    if (!e.isImplicit())
      return true;

    if (e.isContingentDuration())
      return true;

    if (e.isContingent())
      return true;
//
//    if (e.getValue().getLabeledValues().length > 1)
//      return true;

    return false;
  }

  public void formatGraph(Graph<LabeledTimePoint, LabeledEdge> graph,
      AbstractDotGraph g)
  {
    if (g instanceof DotSubGraph)
    {
      ((DotSubGraph) g).setStyle(ElementStyle.DOTTED);
    }

  }

  public String getVertexLabel(LabeledTimePoint o)
  {

    String nodeName = o.toString();
    if (nodeName == null || nodeName.length() == 0)
      nodeName = "Node " + o.getNodeInstance().getNodeID();

    StringBuilder stringBuilder = new StringBuilder();

//    if (o instanceof ObservationNode)
//    {
//      stringBuilder.append("X ");
//    }
    stringBuilder.append(escape(nodeName));

    if (o.getLabel() != null)
    {
      stringBuilder.append("\\n");
      stringBuilder.append(toEscapedString(o.getLabel()));
    }

    TCN<LabeledTimePoint, LabeledEdge> graph = (TCN<LabeledTimePoint, LabeledEdge>) visitor
      .getBaseGraph();

    LabeledTimePoint baseEvent = graph.getTimeBaseEvent();

    LabeledEdge maxEdge = graph.getEdge(baseEvent, o);
    LabeledEdge minEdge = graph.getEdge(o, baseEvent);

    TimeDistance minValue = getValue(minEdge);
    TimeDistance maxValue = getValue(maxEdge);
    minValue = minValue != null ? minValue.neg() : null;

    stringBuilder.append("\\n[");
    stringBuilder.append(toEscapedString(minValue));
    stringBuilder.append("; ");
    stringBuilder.append(toEscapedString(maxValue));
    stringBuilder.append("]");
    String result = stringBuilder.toString();

    if (htmlLabels)
    {
      result = formatHTML(result);
    }
    return result;
  }

  protected TimeDistance getValue(TCNEdge<?, ?> edge)
  {
    if (edge instanceof LabeledEdge)
    {
      LabeledEdge e = (LabeledEdge) edge;
      LabeledValue label = e.getValue()
        .getLabeledValues().iterator().next();
      return label.getValue();
    }
    else if (edge instanceof STNEdge)
    {
      STNEdge e = (STNEdge) edge;
      return e.getValue().getValue();
    }
    return null;
  }

  private String formatHTML(String string)
  {
    String s = StringEscapeUtils.escapeHtml(string);

    s = s.replaceAll("_\\{([^\\{]+)\\}", "<SUB>$1</SUB>");
    s = s.replaceAll("\\^\\{([^\\{]+)\\}", "<SUP>$1</SUP>");
    s = s.replace("\\n", "<BR/>");
    return s;
  }

  public String internalGetEdgeLabel(LabeledEdge o)
  {
    StringBuilder builder = new StringBuilder();
    try
    {

      for (LabeledValue label : o.getValue().getLabeledValues())
      {
        

        if (!(label instanceof ContingentLabeledValue))
        {
          builder.append("<");
          builder.append(toEscapedString(label.getLabel()));
          builder.append(", ");
          builder.append(toEscapedString(label.getValue()));
          builder.append(", #, NaN>");
        }
        else
        {
          ContingentLabeledValue c = (ContingentLabeledValue)label;
          String caseLabel = ConstraintCase.LOWERCASE.equals(c.getCase()) ? "\\"
                                                                       + c
                                                                         .getTrigger()
                                                                         .toString()
                                                                         .toLowerCase()
                                                                       + "/"
                                                                    : "/"
                                                                      + c
                                                                        .getTrigger()
                                                                        .toString()
                                                                        .toUpperCase()
                                                                      + "\\";
          builder.append("<");
          builder.append(toEscapedString(c.getLabel()));
          builder.append(", ");
          builder.append(toEscapedString(label.getValue()));
          builder.append(", ");
          builder.append(toEscapedString(caseLabel));
          builder.append(", ");
          builder.append(toEscapedString(c.getContingentValue()));
          builder.append(">");
        }

        builder.append("\\n");
      }
    }
    catch (NullPointerException e)
    {
      System.out.println(o.toString());
      throw e;
    }

    String string = builder.toString();

    if (htmlLabels)
    {
      string = formatHTML(string);
    }
    // System.out.println(string);
    return string;
  }

  private String toEscapedString(Object object)
  {
    if (object == null)
      return "null";

    return escape(object.toString());
  }

  private String escape(String label)
  {
    String s = label;
    s = s.replaceAll("\\\\", "\\\\");
    return s;
  }

  public String getGraphLabel(Graph<LabeledTimePoint, LabeledEdge> g)
  {
    return g != null ? g.toString() : "";
  }

  public void setGraphVisitor(GraphVisitor<LabeledTimePoint, LabeledEdge> visitor)
  {
    this.visitor = (DotTCNVisitor<LabeledTimePoint, LabeledEdge>) visitor;

  }

  @Override
  public String getEdgeLabel(LabeledEdge o)
  {
//    return internalGetEdgeLabel(o);
    return null;
  }

  @Override
  public String getEdgeTailLabel(LabeledEdge e)
  {
    return internalGetEdgeLabel(e);
    //return null;
  }
}