package  de.aristaflow.adept2.tools.dot;

import org.w3c.dom.Element;

import de.aristaflow.adept2.util.NullArgumentException;

public class DotEdge extends DotElement {

	public enum EdgeStyle {
		DASHED("dashed"), DOTTED("dotted"), SOLID("solid"), INVISIBLE("invis"), BOLD(
				"bold");

		private final String style;

		private EdgeStyle(String style) {
			this.style = style;
		}

		public String getDotValue() {
			return style;
		}
	}
	
	public enum Direction {
	  FORWARD("forward"),
	  BACKWARD("back"),
	  BIDIRECTIONAL("both"),
	  NONE("none");
	  
	  private String style;

    private Direction(String style) {
	    this.style = style;
	  }
    
    public String getDotValue() {
      return style;
    }
	}

	private final DotNode from;
	private final DotNode to;
	private DotColor color;
	private EdgeStyle style;
	private boolean constraint = true;
  private Direction direction;
  private String tailLabel;
  private String headLabel;
  private int weight = -1;
  private boolean directed;

	public DotEdge(DotNode from, DotNode to) {
	  if (from == null || to == null)
	    throw new NullArgumentException();
	  
		this.from = from;
		this.to = to;
	}

	public DotEdge setStyle(EdgeStyle style) {
		this.style = style;
		return this;
	}
	
	public DotEdge setTailLabel(String tailLabel)
	{
	  this.tailLabel = tailLabel;
	  return this;
	}
	
	public DotEdge setHeadLabel(String headLabel)
  {
    this.headLabel = headLabel;
    return this;
  }

	public DotEdge setConstraint(boolean constraint) {
		this.constraint = constraint;
		return this;
	}
	
	public boolean isConstraint()
	{
	  return this.constraint;
	}
	
	public DotEdge setDirection(Direction direction) {
	  this.direction = direction;
	  return this;
	}
	

	@Override
	protected void createDot(StringBuilder builder) {
		builder.append(String.format("%s %s %s", from.getNodeName(), isDirected() ? "->" : "--",
				to.getNodeName()));
		String attrib = buildAttributes();

		if (attrib.length() > 0) {
			builder.append(" [");
			builder.append(attrib);
			builder.append("];");
		}
	}

	@Override
	protected String buildAttributes() {
		StringBuilder attrib = new StringBuilder();
		attrib.append(super.buildAttributes());
		
		if (tailLabel != null)
		{
		  addAttribute(attrib,String.format("taillabel=%s", escapeLabel(tailLabel)));
		}
		
		if (headLabel != null)
    {
      addAttribute(attrib,String.format("headLabel=%s", escapeLabel(headLabel)));
    }

		if (color != null) {
			addAttribute(attrib, String.format("color=%s", color.getDotValue()));
		}

		if (style != null) {
			addAttribute(attrib, String.format("style=%s", style.getDotValue()));
		}
		
		if (direction != null) {
		  addAttribute(attrib, String.format("dir=%s", direction.getDotValue()));
		}
		
    if (weight > 0) {
      addAttribute(attrib, String.format("weight=%s", Integer.valueOf(weight)));
    }
		
		addAttribute(attrib, String.format("constraint=%s", Boolean.valueOf(constraint)));

		return attrib.toString();
	}

  public void setWeight(int weight)
  {
    this.weight = weight; 
    
  }

  @Override
  protected void createGML(Element element)
  {
    Element edgeElement = element.getOwnerDocument().createElement("edge");
    element.appendChild(edgeElement);
    
    edgeElement.setAttribute("source", from.getNodeName());
    edgeElement.setAttribute("target", to.getNodeName());
    
    
    if (label != null)
    {
      Element dataElement = edgeElement.getOwnerDocument().createElement("data");
      edgeElement.appendChild(dataElement);
      dataElement.setAttribute("key", "edgelabel");
      dataElement.setTextContent(this.label);
    }
  }

  public boolean isDirected()
  {
    return directed;
  }

  public void setDirected(boolean directed)
  {
    this.directed = directed;
  }




}
