package  de.aristaflow.adept2.tools.dot;

import java.util.HashMap;

import org.w3c.dom.Element;

public class DotGraph extends AbstractDotGraph {

	public enum GraphType {
		/** Undirected Graph */
		GRAPH("graph"),

		/** Directed Graph */
		DIGRAPH("digraph");

		/** Type of the graph */
		private final String type;

		GraphType(String type) {
			this.type = type;
		}

		public String getDotValue() {
			return type;
		}
	}

	/**
	 * Dot Manual: Sometimes it is natural to make edges point from left to
	 * right instead of from top to bottom. If rankdir=LR in the top-level
	 * graph, the drawing is rotated in this way. TB (top to bottom) is the
	 * default. The mode rankdir=BT is useful for drawing upward-directed
	 * graphs. For completeness, one can also have rankdir=RL.
	 * 
	 * @author Blizzard
	 * 
	 */
	public enum RankDir {
		TopBottom("TB"), LeftRight("LR"), BottomTop("BT"), RightLeft("RL");

		private final String value;

		private RankDir(String value) {
			this.value = value;
		}

		public String getDotValue() {
			return value;
		}
	}

	private final GraphType type;
	
	protected RankDir rankDir;

  
	public DotGraph(GraphType type, String name) {
		super(name);
		this.type = type;

		this.subgraphs = new HashMap<String, DotSubGraph>();
	}

	public void setRankDir(RankDir rankDir) {
		this.rankDir = rankDir;
	}

	public String createDot() {
		StringBuilder builder = new StringBuilder();
		createDot(builder);
		return builder.toString();
	}

	@Override
	protected void createDot(StringBuilder builder) {
		builder.append(String.format("%s %s {\n", getType().getDotValue(), name));

		String attrib = buildAttributes();

		if (attrib.length() > 0) {
			builder.append("graph [");
			builder.append(attrib);
			builder.append("];\n");
		}

		for (DotElement node : nodes.values()) {
			node.createDot(builder);
			builder.append("\n");
		}
		
		for (DotElement node : subgraphs.values()) {
			node.createDot(builder);
			builder.append("\n");
		}

		for (DotElement edge : edges.values()) {
			edge.createDot(builder);
			builder.append("\n");
		}

		builder.append("}");
	}

	@Override
	protected String buildAttributes() {
		StringBuilder attrib = new StringBuilder();
		attrib.append(super.buildAttributes());

		if (rankDir != null) {
			addAttribute(attrib,
					String.format("rankdir = %s", rankDir.getDotValue()));
		}

		return attrib.toString();
	}

	@Override
	public String toString() {
		return createDot();
	}

	@Override
	public DotNode getNode(String nodeID) {
		DotNode dotNode = super.getNode(nodeID);
		if (dotNode == null)
		{
			for (DotSubGraph graph : subgraphs.values()){
				dotNode = graph.getNode(nodeID);
				if (dotNode != null)
					break;
			}
		}
		return dotNode;
	}

  @Override
  protected void createGML(Element element)
  {
    Element graphElement = element.getOwnerDocument().createElement("graph");
    element.appendChild(graphElement);
    graphElement.setAttribute("id", "G");
    graphElement.setAttribute("edgedefault", "directed");
    
    for (DotElement node : nodes.values()) {
      node.createGML(graphElement);
    }
    
    for (DotElement node : subgraphs.values()) {
      node.createGML(graphElement);
    }

    for (DotElement edge : edges.values()) {
      edge.createGML(graphElement);
    }
  }

  public GraphType getType()
  {
    return type;
  }

  @Override
  protected boolean isDirected()
  {
    return GraphType.DIGRAPH.equals(type);
  }


	
	

}
